!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! TAMES  Topological Analysis of Metabolic Structures                 !
! Copyright (C) 2017                                                   !
! Paulo A. N. Dias, paulo.dias29@gmail.com                             !
! Marco Seabra dos Reis, marco@eq.uc.pt                                !
! Armindo Salvador, armindo.salvador@gmail.com                         !
! Pedro Martins, pmartins@iscac.pt                                     !
!                                                                      !
! This file is part of TAMES.                                          !
!                                                                      !
! TAMES is free software: you can redistribute it and/or modify        !
! it under the terms of the GNU General Public License as published by !
! the Free Software Foundation, either version 3 of the License, or    !
! any later version.                                                   !
!                                                                      !
! You should have received a copy of the GNU General Public License    !
! along with TAMES.  If not, see <http://www.gnu.org/licenses/>.       !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                         !
! Program "Instances Counting":                           !
! - Obtaining the instances of structures within a given  !
!   topological scope in metabolic networks. NOTE: Some   !
!   of the instances might be repeated                    !
!                                                         !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      Program instancesCounting

! CONSTANTS
!      n1 : number of metabolites
!      n2 : number of reactions (reversibility not considered)
!      repeat : logical variable used to determine the number of reactions and of metabolites in the files
!      maxNumInst : expected maximum number of instances. This parameter is needed to create the allocatable arrays
!                   used to store the required information for each one of the instances
!      maxNumMet : maximum number of metabolites considered for the structures within the considered number of
!                  generations maxNumGen
!      maxNumReact : maximum number of reactions considered for the structures within the considered number of
!                    generations maxNumGen
!      maxNumGen : maximum number of generations considered to generate structures
!      numStructures : number of structures for the considered generations (maxNumGen)
!      numInst :  number of instance for the structures within the considered scope
!      isoStruRef : reference number of the structure in matrix RevStruct(i3,i2,i1) to which the instance in matrix
!                   supRevStInst(i2,i1,1) is a concretization.
!      InstTest : indicates if the instance tested was found before. If it is a new instance, InstTest = 0; if it is not,
!                 InstTest = 1
!      InstScope : indicates if the (possibly) newly found instance is within the considered topological scope. If it is
!                  within the scope, InstScope = 0; if it is not, InstScope = 1
!      TempCut : support variable used to switch off the part of the subroutine "isoInstances" used to determine if
!                the (possibly) new instance was found before - it allows only to obtain the metabolites and the
!                of the (possibly) new instance organized from the higher to the lower reference numbers. If the variable
!                TempCut = 0, the mentioned part of the subroutine is not switch off; if the variable TempCut = 1, the
!                considered part of the subroutine is switch off
!      refNumPermMet : reference number of the permutation for the metabolites that allowed to the determine the structure
!                      in matrix RevStruct(i3,i2,i1) of which the instance in matrix supRevStInst(i3,i2,1) is a realization,
!                      i.e., PermutNode( i3 , refNumPermMet , numNode(1) )
!      refNumPermIrrReact : reference number of the permutation for the irreversible reactions that allowed to the determine the
!                           structure in matrix RevStruct(i3,i2,i1) of which the instance in matrix supRevStInst(i3,i2,1) is a
!                           realization, i.e., PermutNode( i3 , refNumPermIrrReact , numNode(4) )
!      refNumPermRevReact : reference number of the permutation for the reversible reactions that allowed to the determine the
!                           structure in matrix RevStruct(i3,i2,i1) of which the instance in matrix supRevStInst(i3,i2,1) is a
!                           realization, i.e., PermutNode( i3 , refNumPermRevReact , numNode(4) )
!      refNumPermDirInvReact : reference number of the permutation for the direct and inverse reactions of a reversible
!                              reaction that allowed to the determine the structure in matrix RevStruct(i3,i2,i1) of
!                              which the instance in matrix supRevStInst(i3,i2,1) is a realization, i.e.,
!                              PermutDirInvReact( i3, refNumPermDirInvReact , numNode(3) )
!      supNumPermutNode : support constant used to obtain the vector numPermutNode(i1)
!      supNumPermDirInvReact : support constant used to obtain the vector numPermDirInvReact(i1)
!      supTotMetInst : support constant with the sum of the reference numbers of the metabolites of the (possibly) new instance
!      supTotReactInst : support constant with the sum of the reference numbers of the reactions of the (possibly) new instance
!
!
! VECTORS
!      mn(i) : list of names of all metabolites in file "List_Metabolites.txt"
!      rn(i) : list of names of all reactions in file "List_Reactions.txt"
!      RevReact(i1) : reversibility of the reaction i1: 0 irreversible, 1 reversible;
!      sup(i1) : support vector of size integer*2
!      nrefStruLow(i1) : lower reference number of a structure of the i1 generation
!      nrefStruHigh(i1) : higher reference number of a structure of the i1 generation
!      nMinInstPGen(i1) : minimum reference number of a instance which is part of the pseudo-generation i1. In this work,
!                         the instances are obtained by adding three possible elements to previously found instances:
!                         1) metabolite; 2) reaction; 3) metabolite + reaction
!      nMaxInstPGen(i1) : maximum reference number of a instance which is part of the pseudo-generation i1. In this work,
!                         the instances are obtained by adding three possible elements to previously found instances:
!                         1) metabolite; 2) reaction; 3) metabolite + reaction
!      minRefNInstStru(i1) : minimum reference number of an instance of structure i1
!      maxRefNInstStru(i1) : maximum reference number of an instance of structure i1
!      numPermutNode(i1) : number of permutations for the given set of nodes i1
!      numPermDirInvReact(i1) : number of permutations between the direct and inverse reaction for a set of reversible
!                               reactions i1
!      supMetTemp(i1) : support vector with all the metabolites of the (possibly) new instance still not ordered
!      supReactTemp(i1) : support vector with all the reactions of the (possibly) new instance still not ordered
!      supReactRevInstTemp(i1) : information regarding the reversibility of the reactions of the (possibly) new instance. If a
!                                reaction is reversible, the corresponding cell is filled with 1; it it is irreversible, the
!                                corresponding cell is filled with 0. The reactions may not be ordered according to its reversibility
!      supMet(i1) : support vector with all the metabolites of the (possibly) new instance ordered according to their degree
!                   (total and IN)
!      supReact(i1) : support vector with all the reactions of the (possibly) new instance ordered according to their degree
!                     (total and IN) and to their reversibility (irreversible reactions are stored first)
!      supMetTemp(i1) : support vector with all the metabolites of the (possibly) new instance (might not be ordered)
!      supReactTemp(i1) : support vector with all the reactions of the (possibly) new instance (might not be ordered)
!      supMetOrd(i1) : support vector with all the metabolites of the (possibly) new instance ordered from the higher to
!                      the lower reference number
!      supReactOrd(i1) : support vector with all the reactions of the (possibly) new instance ordered from the higher to
!                        the lower reference number
!      numNodes(i1) : number of nodes of the (possibly) new instance. (i1=1) n of metabolites; (i1=2) n of reactions;
!                     (i1=3) n of reversible reactions; (i1=4) n of irreversible reactions
!      supDataInst(i1) : support vector to store provisionally the relevant information for the (possibly) newly found
!                        instance: (i1=1) number of interconnections; (i1=2) number of connections metabolite/reaction
!                        (reversibility not considered)
!      supMetDegInst(i1) : gives the degree for each metabolite of a (possibly) new instance. i1=[(3 x (nmet - 1)) + 1],
!                          total n of connections of the metabolite; i1=[(3 x (nmet - 1)) + 2], n of connections
!                          comming to the metabolite (in); i1=[(3 x (nmet - 1)) + 3], n of connections going out of
!                          the metabolite (out)
!      supReactDegInst(i1) : gives the degree for each reaction of a possibly new instance. The irreversible reactions
!                            are stored first, and then, in case at least one exists, the reversible reactions are stored.
!                            i1=[(3 x (nmet - 1)) + 1], connections of the metabolite; i1=[(3 x (nmet - 1)) + 2],
!                            n of connections comming to the total n of metabolite (in); i1=[(3 x (nmet - 1)) + 3],
!                            n of connections going out of the metabolite (out)
!      supDifNodesInst(i1) : stores the information indicating if a (possibly) new instance has different nodes by taking
!                            into account their degree (total and IN). If all the nodes of a given type (metabolites,
!                            irreversible reactions and reversible reactions) are all different, the corresponding cell
!                            is filled with "1"; if there is at least two nodes of the same type with the same degree
!                            (total and IN), the corresponding cell is filled with "0". (i1=1) => metabolites;
!                            (i1=2) => irreversible reactions; (i1=3) => reversible reactions
!      InstStruct(i1) : structure of the instance i1
!      TotMetInst(i1) : sum of the reference numbers of the metabolites of instance i1
!      TotReactInst(i1) : sum of the reference numbers of the reactions of instance i1
!      OutRevStruct(i1) : output data obtained for the structure i1 with the developed code => frequency (number of instances)
!      const(i1) : support vector with the dimension integer*4
!
!
! MATRICES
!      a1(i2,i1) : products x reactions data matrix, i1=products and i2=reactions
!      a2(i2,i1) : reactions x reactants data matrix, i1=reactions and i2=reactants
!      RevStruct(i3,i2,i1) : matrix with the possible structures using a bipartite and directed graph representation
!                            (no restrictions). i1) number of the structure; i2) number of the metabolite;
!                            i3) [4*(i3-1) + 1] reactant of the direct reaction, [4*(i3-1) + 2] product of
!                            the direct reaction, [4*(i3-1) + 3] reactant of the inverse reaction, [4*(i3-1) + 4]
!                            product of the inverse reaction. If a metabolite is connected to a reaction, the
!                            corresponding cell is filled with 1; otherwise, is filled with 0
!      DataRevStruTemp(i2,i1) : i1) reference number of the structure. i2) addtional data for each structure: (i2=1) number
!                               of generation; (i2=2) number of metabolites; (i2=3) number of reactions; (i2=4) number of
!                               reversible reactions; (i2=5) number of Ancestors; (i2=6) number of Parents; (i2=7) number of
!                               Childs; (i2=8) number of Offspring; (i2=9) number of interconnections; (i2=10) maximum possible
!                               number of interconnections; (i2=11) number of connections metabolite/reaction (reversibility not
!                               considered); (i2=12) maximum possible number of connections metabolite/reaction (reversibility
!                               not considered)
!      DataRevStru(i2,i1) : i1) reference number of the structure. i2) addtional data for each structure: (i2=1) number
!                           of generation; (i2=2) number of metabolites; (i2=3) number of reactions; (i2=4) number of
!                           reversible reactions; (i2=5) number of interconnections; (i2=6) number of connections
!                           metabolite/reaction (reversibility not considered)
!      MetDegRevStru(i2,i1) : gives the degree for each metabolite of structure i1: i2=[(3 x (nmet - 1)) + 1], total n of
!                             connections of the metabolite; i2=[(3 x (nmet - 1)) + 2], n of connections comming to the
!                             metabolite (in); i2=[(3 x (nmet - 1)) + 3], n of connections going out of the metabolite (out)
!      ReactDegRevStru(i2,i1) : gives the degree for each reaction of structure i1. The irreversible reactions are stored
!                               first and then, in case at least one exists, the reversible reactions are stored.
!                               i2=[(3 x (nmet - 1)) + 1], connections of the reaction; i2=[(3 x (nmet - 1)) + 2],
!                               n of connections comming to the reaction (in); i2=[(3 x (nmet - 1)) + 3],
!                               n of connections going out of the reaction (out)
!      DifNodesStru(i2,i1) : stores the information indicating if structure i1 has different nodes by taking into account their
!                            degree (total and IN). If all the nodes of a given type (metabolites, irreversible reactions and
!                            reversible reactions) are all different, the corresponding cell is filled with "1"; if there is
!                            at least two nodes of the same type with the same degree (total and IN), the corresponding cell
!                            is filled with "0". (i2=1) => metabolites; (i2=2) => irreversible reactions; (i2=3) => reversible
!                            reactions
!      MetInst(i2,i1) :  reference numbers of the metabolites of the instance i1 organized to fit the pattern of the structure
!                        as stored in matrix RevStruct(i3,i2,i1)
!      ReactInst(i2,i1) :  reference numbers of the reactions of the instance i1 organized to fit the pattern of the structure
!                          as stored in matrix RevStruct(i3,i2,i1)
!      OrdMetInst(i2,i1) : metabolites of the instance i1 ordered from the higher to the lower reference numbers
!      OrdReactInst(i2,i1) : reactions of the instance i1 ordered from the higher to the lower reference numbers
!      supRevStInst(i3,i2,1) : support matrix to store a (possibly) new instance. The nodes are organized according to their
!                              degree (total and IN) The reactions are previously separated according to their reversibility:
!                              irreversible reactions appear first, followed by the reversible reactions. The pattern of this
!                              instance will be compared to the ones in matrix RevStruct(i3,i2,i1) to find what is its structure
!      maxInterRevStr(i2,i1) : maximum number of interconnections for a structure with i1 metabolites and i2 reactions within
!                              the considered scope
!      maxMRConRevStr(i2,i1) : maximum number of connections metabolite/reaction (reversibility not considered) for a structure
!                              with i1 metabolites and i2 reactions within the considered scope
!      MinGenStruInst(i2,i1) : minimum number of the generation in which one can find structures with the same number of
!                              interconnections i1 and with the same number of connections metabolite/reaction i2 of the
!                              (possibly) new instance
!      MaxGenStruInst(i2,i1) : maximum number of the generation in which one can find structures with the same number of
!                              interconnections i1 and with the same number of connections metabolite/reaction i2 of the
!                              (possibly) new instance
!      maxNumRevReactStr(i2,i1): maximum number of reversible reactions of a structure within the topological scope with
!                                i1 metabolites and i2 reactions
!      PermutNode(i3,i2,i1) : possible permutations for a given number i1 of nodes (metabolites or reactions), obtained
!                             initially by calling the subroutine allPermutations. i2) number of the permutation;
!                             i3) number of the nodes ordered by this permutation
!      supPermutNode(i2,i1) : support matrix used to obtain the matrix PermutNode(i3,i2,i1)
!      PermutDirInvReact(i3,i2,i1) : possible permutations between the direct and inverse single direction reactions for
!                                    a set of i1 reversible reactions. The matrix was obtained by calling the subroutine
!                                    allRevReactPerm. i2) number of the permutation; i3) ordering the direct and inverse single
!                                    direction reactions for all the reactions
!      supPermutDirInvReact(i2,i1) : support matrix used to obtain the matrix PermutDirInvReact(i3,i2,i1)

      implicit none

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      DECLARATION OF THE VARIABLES     !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!      Static arrays: declaration     ! -fmax-stack-var-size=20000
       ! Parameters
      integer, PARAMETER :: maxNumInst=55000000  ! 4 bits
       ! Integer: 1 bit
      integer*1 :: a1(1400,1400), a2(1400,1400), RevStruct(24,6,4000), supRevStInst(24,6,1)
      integer*1 :: PermutNode(6,1000,6), PermutDirInvReact(12,100,6), DifNodesStru(3,4000)
      integer*1 :: DataRevStru(6,4000), maxNumRevReactStr(8,8), maxInterRevStr(8,8)
      integer*1 :: maxMRConRevStr(8,8), MinGenStruInst(20,20), MaxGenStruInst(20,20)  ! HAD TO CHANGE THE MEMORY ALLOCATION
      integer*1 :: MetDegRevStru(18,4000), ReactDegRevStru(18,4000), RevReact(1400)   ! FOR THE VARIABLES MinGenStruInst(i1,i2)
      integer*1 :: supReactRevInstTemp(8), supMetDegInst(18), supReactDegInst(18)     ! AND MaxGenStruInst(i1,i2) FROM 15 TO 20
      integer*1 :: supDifNodesInst(3), numNodes(4), supDataInst(2), InstTest, InstScope
      integer*1 :: maxNumGen, maxNumMet, maxNumReact, numPseuGen, TempCut
       ! Integer: 2 bit
      integer*2 :: supMet(8), supReact(8), supMetTemp(8)
      integer*2 :: supReactTemp(8), supMetOrd(8), supReactOrd(8), nrefStruLow(15)   ! HAD TO CHANGE THE MEMORY ALLOCATION
      integer*2 :: nrefStruHigh(15), sup(5), numPermDirInvReact(8), supTotMetInst   ! FOR THE VARIABLES nrefStruLow(i1)
      integer*2 :: supTotReactInst, n1, n2, isoStruRef, refNumPermDirInvReact       ! nrefStruHigh(i1) from 7 to 15
       ! Integer: 4 bit
      integer :: OutRevStruct(4000), minRefNInstStru(4000), maxRefNInstStru(4000)
      integer :: nMinInstPGen(8), nMaxInstPGen(8), const(10), numPermutNode(8)
      integer :: numInst, numStructures, refNumPermMet, refNumPermIrrReact
      integer :: refNumPermRevReact
      integer :: i, ix
        ! Pseudo-structural elements
         ! i) Metabolite
      integer :: i1, j1, k1, iI_M1, iM1, iR1
         ! ii) Reaction
      integer :: i2, j2, k2, l2, iI_R2, iM2, iR2
         ! iii) Metabolite + Reaction
      integer :: i3, j3, k3, l3, iI_M3, iI_R3, iM3, IR3

!      Allocatable arrays: declaration
       ! Integer: 1 bit
      integer*1, DIMENSION(:,:), allocatable :: supPermutNode, supPermutDirInvReact
       ! Integer: 2 bits
      integer*2, DIMENSION(:,:), allocatable :: MetInst, ReactInst, in_nGen_nStru
      integer*2,  DIMENSION(:,:), allocatable :: OrdMetInst, OrdReactInst, DataRevStruTemp
      integer*2, DIMENSION(:), allocatable :: InstStruct, TotMetInst, TotReactInst
       ! Integer: 4 bits
      integer, allocatable :: supNumPermDirInvReact, supNumPermutNode
       ! Character
      character(len=220), DIMENSION(:), allocatable :: mn, rn
       ! Logical
      logical, allocatable :: repeat


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      INPUT FILES                      !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Reads the names of the reactions and of the metabolites and obtains the number of
!      metabolites and reactions
     !
!      Allocate the arrays mn(i1) and rn(i1), and the logical variable repeat
      allocate(mn(1400),rn(1400), repeat)
     !
     ! Metabolites
      !
      ! Logical variable. The loop is performed while the variable is "true"
      repeat = .true.
      ! Counts the number of metabolites
      n1 = 0
      open (1,file='List_Metabolites.txt')
      do while (repeat)
        ! Adds this new metabolite to the total
        n1 = n1 + 1
        ! Name of the metabolites
        read (1,1002) mn(n1)
        ! End the loop if the statement 'End' was found
        if( mn(n1) .eq. 'End' ) then
          ! Turn the logical variable into 'false' to end the loop
          repeat= .false.
          ! Subtract 1 to the total value of metabolites
          n1 = n1 - 1
        end if
      end do
      close (1)
      ! Deallocates the variable mn(i1)
      deallocate(mn)
     !
     ! Reactions
      !
      ! Logical variable. The cycle is performed while the variable is "true"
      repeat = .true.
      ! Counts the number of reactions (reversible and irreversible)
      n2 = 0
      open (1,file='List_Reactions.txt')
      do while (repeat)
        ! Adds this new reaction to the total
        n2 = n2 + 1
        ! Names of the unidirectional reactions
        read (1,1001) rn(n2)
        ! End the loop if the statement 'End' was found
        if( rn(n2) .eq. 'End' ) then
          ! Turn the logical variable into 'false' to end the loop
          repeat= .false.
          ! Subtract 1 to the total number of reactions
          n2 = n2 - 1
        end if
      end do
      close (1)
      ! Deallocates the variable rn(i1)
      deallocate(rn)
     !
     !
!      Reads the data from the files
    !
    ! Data matrix: products x reactions, i1=products and i2=reactions
      open (1,file='Products_X_Reactions.txt') !
      do i1=1,n1     ! Number of metabolites
        read (1,*) (a1(i2,i1), i2=1,n2) ! Number of reactions
      end do
      close (1)
    !
    ! Data matrix: reactions x reactants
      open (1,file='Reactions_X_Reactants.txt')
      do i1=1,n2  ! Number of reactions
        read (1,*) (a2(i2,i1), i2=1,n1)  ! Number of metabolites
      end do
      close (1)
     !
     !
    !  Reads the information regarding the reversibility of the reactions
      open (1,file='RevReact.txt') !
      do i1=1,n2
        read (1,*) RevReact(i1)
      end do
      close(1)


!      Reads the information regarding the possible structures till the number of generations
!      consideration (5 generations)
     !
    !  Read the file "in_nGen_nStru.txt"
      const(1) = 0
      ! Allocate the variable in_nGen_nStru(i2,i1)
      allocate(in_nGen_nStru(2,15))
      ! Logical variable. The cycle is performed while the variable is "true"
      repeat = .true.
      open (1,file='in_nGen_nStru.txt')
      do while (repeat)
        const(1) = const(1) + 1
        read(1,*) (in_nGen_nStru(i1,const(1)), i1=1,2)
        if( const(1) .eq. (in_nGen_nStru(1,1)+1) ) then
          repeat= .false.
        end if
      end do
      close(1)
      const(1) = 0
      maxNumGen = in_nGen_nStru(1,1)
      numStructures = in_nGen_nStru(2,1)
      do i=2,(maxNumGen+1)
        nrefStruLow(i-1) = in_nGen_nStru(1,i)
        nrefStruHigh(i-1) = in_nGen_nStru(2,i)
      end do
      ! Deallocate the variables in_nGen_nStru(i2,i1) and repeat
      deallocate(in_nGen_nStru, repeat)
     !
    !  Read the file 'in_nMet_nReact.txt'
      open (1,file='in_nMet_nReact.txt')
      do i1=1,2
        if( i1 .eq. 1 ) then
          read(1,*) maxNumMet
        else
          read(1,*) maxNumReact
        end if
      end do
      close(1)
     !
    !  Read the file 'Structures_Metabolite_A.txt', which stores part of the structures (no restrictions)
    !  corresponding to metabolite A (i2=1)
      const(1) = 1  ! Metabolite A: i2=1
      if( maxNumMet .ge. const(1) ) then
        open (1,file='Structures_Metabolite_A.txt')
        do i1=1,numStructures
          read(1,*) (RevStruct(i3,const(1),i1), i3=1,(4*maxNumReact))    ! Maximum number of generations = maximum number of reactions
        end do
        close(1)
      end if
     !
    !  Read the file 'Structures_Metabolite_B.txt', which stores part of the structures (no restrictions)
    !  corresponding to metabolite B (i2=2)
      const(1) = 2  ! Metabolite B: i2=2
      if( maxNumMet .ge. const(1) ) then
        open (1,file='Structures_Metabolite_B.txt')
        do i1=1,numStructures
          read(1,*) (RevStruct(i3,const(1),i1), i3=1,(4*maxNumReact))    ! Maximum number of generations = maximum number of reactions
        end do
        close(1)
      end if
     !
    !  Read the file 'Structures_Metabolite_C.txt', which stores part of the structures (no restrictions)
    !  corresponding to metabolite C (i2=3)
      const(1) = 3  ! Metabolite C: i2=3
      if( maxNumMet .ge. const(1) ) then
        open (1,file='Structures_Metabolite_C.txt')
        do i1=1,numStructures
          read(1,*) (RevStruct(i3,const(1),i1), i3=1,(4*maxNumReact))    ! Maximum number of generations = maximum number of reactions
        end do
        close(1)
      end if
     !
    !  Read the file 'Structures_Metabolite_D.txt', which stores part of the structures (no restrictions)
    !  corresponding to metabolite D (i2=4)
      const(1) = 4  ! Metabolite D: i2=4
      if( maxNumMet .ge. const(1) ) then
        open (1,file='Structures_Metabolite_D.txt')
        do i1=1,numStructures
          read(1,*) (RevStruct(i3,const(1),i1), i3=1,(4*maxNumReact))    ! Maximum number of generations = maximum number of reactions
        end do
        close(1)
      end if
     !
    !  Read the file 'Structures_Metabolite_E.txt', which stores part of the structures (no restrictions)
    !  corresponding to metabolite E (i2=5)
      const(1) = 5  ! Metabolite E: i2=5
      if( maxNumMet .ge. const(1) ) then
        open (1,file='Structures_Metabolite_E.txt')
        do i1=1,numStructures
          read(1,*) (RevStruct(i3,const(1),i1), i3=1,(4*maxNumReact))    ! Maximum number of generations = maximum number of reactions
        end do
        close(1)
      end if
     !
    !  Read the file 'Structures_Metabolite_F.txt', which stores part of the structures (no restrictions)
    !  corresponding to metabolite F (i2=6)
      const(1) = 6  ! Metabolite F: i2=6
      if( maxNumMet .ge. const(1) ) then
        open (1,file='Structures_Metabolite_F.txt')
        do i1=1,numStructures
          read(1,*) (RevStruct(i3,const(1),i1), i3=1,(4*maxNumReact))    ! Maximum number of generations = maximum number of reactions
        end do
        close(1)
      end if
      const(1) = 0
     !
    !  Read the 'Structures_General.txt', which stores the general information regarding the structures (no restrictions)
      ! Allocate the variable DataRevStruTemp(i2,i1)
      allocate(DataRevStruTemp(12,4000))
      open (1,file='Structures_General.txt')
      do i1=1,numStructures
        read(1,*) (DataRevStruTemp(i2,i1), i2=1,12)
      end do
      close(1)
      ! Stores the relevant information of matrix DataRevStruTemp(i2,i1) into matrix DataRevStru(i2,i1)
      do i1=1,numStructures
        ! N of the generation
        DataRevStru(1,i1) = DataRevStruTemp(1,i1)
        ! N of metabolites
        DataRevStru(2,i1) = DataRevStruTemp(2,i1)
        ! N of reactions
        DataRevStru(3,i1) = DataRevStruTemp(3,i1)
        ! N of reversible reactions
        DataRevStru(4,i1) = DataRevStruTemp(4,i1)
        ! N of interconnections
        DataRevStru(5,i1) = DataRevStruTemp(9,i1)
        ! N of different connections metabolite/reaction (reversibility not considered)
        DataRevStru(6,i1) = DataRevStruTemp(11,i1)
      end do
      ! Deallocate the variable DataRevStruTemp(i2,i1)
      deallocate(DataRevStruTemp)
     !
    !  Read the file 'MetDegreeRevStru.txt', which stores the information regarding the degrees (total, IN and OUT)
    !  of the metabolites of each one of the structures considered within the scope
      open (1,file='MetDegreeRevStru.txt')
      do i1=1,numStructures
        read(1,*) (MetDegRevStru(i2,i1), i2=1,(3*(maxNumMet)))
      end do
      close(1)
     !
    !  Read the file 'ReactDegreeRevStru.txt', which stores the information regarding the degrees (total, IN and OUT)
    !  of the metabolites of each one of the structures considered within the scope
      open (1,file='ReactDegreeRevStru.txt')
      do i1=1,numStructures
        read(1,*) (ReactDegRevStru(i2,i1), i2=1,(3*maxNumReact))
      end do
      close(1)
     !
    !  Read the file 'DifNodesStru.txt', which stores the information indicating if structure i1 has different nodes by
    !  taking into account their degree (total and IN)
      open (1,file='DifNodesStru.txt')
      do i1=1,numStructures
        read(1,*) (DifNodesStru(i2,i1), i2=1,3)
      end do
      close(1)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      SUPPORT MATRICES                 !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!      Obtains the maximum number of interconnections and of connections metabolite/reaction for the
!      i1 structures with DataRevStru(2,i1) metabolites and DataRevStru(3,i2) reactions within the
!      considered scope (patterns stored in matrix RevStruct(i3,i2,i1)). It also obtains the number
!      of the lower and of the higher generation in which the structure of an instance with a given
!      number of interconnections and of different connections metabolite/reaction may be found
      !
      ! Creates the matrices maxInterRevStr(i2,i1), maxMRConRevStr(i2,i1) and maxNumRevReactStr(i2,i1)
      do i1=1,maxNumMet  ! Maximum number of metabolites
        do i2=1,maxNumReact  ! Maximum number of reactions
          maxInterRevStr(i2,i1) = 0
          maxMRConRevStr(i2,i1) = 0
          maxNumRevReactStr(i2,i1) = 0
        end do
      end do
      !
      ! Creates the matrices MinGenStruInst(i2,i1) and MaxGenStruInst(i2,i1)
      do i1=1,(2*maxNumGen)
        do i2=1,(2*maxNumGen)
          MinGenStruInst(i2,i1) = 0
          MaxGenStruInst(i2,i1) = 0
        end do
      end do
      !
      ! Checks all the structures
      do i1=1,numStructures
        !
        ! Structure i1 has the higher number of interconnections of all the structures with the
        ! same number of metabolites and reactions
        if( DataRevStru(5,i1) .gt. maxInterRevStr(DataRevStru(3,i1),DataRevStru(2,i1)) ) then
          maxInterRevStr(DataRevStru(3,i1),DataRevStru(2,i1)) = DataRevStru(5,i1)
        end if
        !
        ! Structure i1 has the higher number of connections metabolite/reaction of all the structures
        ! with the same number of metabolites and reactions
        if( DataRevStru(6,i1) .gt. maxMRConRevStr(DataRevStru(3,i1),DataRevStru(2,i1)) ) then
          maxMRConRevStr(DataRevStru(3,i1),DataRevStru(2,i1)) = DataRevStru(6,i1)
        end if
        !
        ! Structure i1 has the higher number of reversible reactions of all the structures with the
        ! same number of metabolites and reactions
        if( DataRevStru(4,i1) .gt. maxNumRevReactStr(DataRevStru(3,i1),DataRevStru(2,i1)) ) then
          maxNumRevReactStr(DataRevStru(3,i1),DataRevStru(2,i1)) = DataRevStru(4,i1)
        end if
        !
        ! Determines the minimum number of the generation of a structure i1 with the number of interconnections
        ! DataRevStru(5,i1) and with the number of connections metabolite/reaction DataRevStru(6,i1)
        if( MinGenStruInst( DataRevStru(6,i1) , DataRevStru(5,i1) ) .eq. 0 ) then
          MinGenStruInst( DataRevStru(6,i1) , DataRevStru(5,i1) ) = DataRevStru(1,i1)
        else
          if( DataRevStru(1,i1) .lt. MinGenStruInst( DataRevStru(6,i1) , DataRevStru(5,i1) ) ) then
            MinGenStruInst( DataRevStru(6,i1) , DataRevStru(5,i1) ) = DataRevStru(1,i1)
          end if
        end if
        !
        ! Determines the maximum number of the generation of a structure i1 with the number of interconnections
        ! DataRevStru(5,i1) and with the number of connections metabolite/reaction DataRevStru(6,i1)
        if( MaxGenStruInst( DataRevStru(6,i1) , DataRevStru(5,i1) ) .eq. 0 ) then
          MaxGenStruInst( DataRevStru(6,i1) , DataRevStru(5,i1) ) = DataRevStru(1,i1)
        else
          if( DataRevStru(1,i1) .gt. MaxGenStruInst( DataRevStru(6,i1) , DataRevStru(5,i1) ) ) then
            MaxGenStruInst( DataRevStru(6,i1) , DataRevStru(5,i1) ) = DataRevStru(1,i1)
          end if
        end if
        !
        ! Creates the vectors OutRevStruct(i1), minRefNInstStru(i1) and maxRefNInstStru(i1)
        OutRevStruct(i1) = 0
        minRefNInstStru(i1) = 0
        maxRefNInstStru(i1) = 0
      end do


!      Obtains all the permutations for a given number of nodes and stores them into the matrix PermutNode(i3,i2,i1)
      ! Allocates the variables supPermutNode(i2,i1) and supNumPermutNode
      allocate(supPermutNode(6,1000),supNumPermutNode)
      do i1=1,maxNumMet
        sup(1) = i1
        call allPermutations(sup(1), supPermutNode, supNumPermutNode)
        numPermutNode(i1) = supNumPermutNode
        do i2=1,numPermutNode(i1)
          do i3=1,i1
            PermutNode(i3,i2,i1) = supPermutNode(i3,i2)
            supPermutNode(i3,i2) = 0
          end do
        end do
      end do
      ! Deallocates the variables supPermutNode(i2,i1) and supNumPermutNode
      deallocate(supPermutNode,supNumPermutNode)
     !
!      Obtains all the permutations between the direct and inverse single direction reactions for a given number of
!      reversible reactions
      ! Allocates the variables supPermutDirInvReact(i2,i1) and supNumPermDirInvReact
      allocate(supPermutDirInvReact(12,100),supNumPermDirInvReact)
      do i1=1,maxNumReact
         sup(1) = i1
         call allRevReactPerm(sup(1), supPermutDirInvReact, supNumPermDirInvReact)
         numPermDirInvReact(i1) = supNumPermDirInvReact
         do i2=1,numPermDirInvReact(i1)
           do i3=1,(2*i1)
             PermutDirInvReact(i3,i2,i1) = supPermutDirInvReact(i3,i2)
             supPermutDirInvReact(i3,i2) = 0
           end do
         end do
      end do
      ! Deallocates the variables supPermutDirInvReact(12,100) and supNumPermDirInvReact
      deallocate(supPermutDirInvReact,supNumPermDirInvReact)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      SEED INSTANCES                   !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!      Allocates the arrays MetInst(i2,i1), ReactInst(i2,i1), InstStruct(i1), TotMetInst(i1) and TotReactInst(i1)
      allocate(MetInst(maxNumMet,maxNumInst),ReactInst(maxNumReact,maxNumInst))
      allocate(InstStruct(maxNumInst),TotMetInst(maxNumInst),TotReactInst(maxNumInst))
      allocate(OrdMetInst(maxNumMet,maxNumInst),OrdReactInst(maxNumReact,maxNumInst))
      !
!      Determines all the instances with 1 reaction (reversible or irreversible) with 1 product and 1 reactant. This
!      instances will be used as our "seed-instances"
      const(2) = 0
      const(3) = 0
      const(4) = 0
      const(6) = 0
      !
      ! Creates the vectors nMinInstPGen(i1) and nMaxInstPGen(i1)
      do i1=1,8
        ! Will give the minimum reference number of a instance of a pseudo-generation i1
        nMinInstPGen(i1) = 0
        ! Will give the maximum reference number of a instance of a pseudo-generation i1
        nMaxInstPGen(i1) = 0
      end do
      !
      ! Minimum reference number of a instance of a pseudo-generation 1
      nMinInstPGen(1) = 1
      !
      ! Initializes the variable TempCut
      TempCut = 0
      !
      ! Creates the support vector supRevStInst(i3,i2,i1), used to provisionally store the
      ! pattern of the (possibly) new found structure
      do i1=1,maxNumMet  ! Number of metabolites
        do i2=1,(4*maxNumReact)  ! 4 x number of reactions
          supRevStInst(i2,i1,1) = 0
        end do
      end do
      !
      ! Metabolite i1
      do i1=1,(n1-1)  ! Looks only to the upper triangle cells of a matrix metabolite x metabolite. So, it will check
      ! only a pair of metabolites only once
        ! Metabolite i2
        do i2=i1+1,n1
          ! Reaction i3
          do 4 ix=1,n2
            !
            ! The reaction ix connects the metabolites i1 and i2 as a pair reactant/product or product/reactant?
            ! Metabolite i1 = product and metabolite i2 = reactant of reaction ix?
            if( ( ( a1(ix,i1) .eq. 1 ) .and. ( a2(i2,ix) .eq. 1 ) ) .or. &
            ! Metabolite i2 = product and metabolite i1 = reactant of reaction ix?
            ( ( a1(ix,i2) .eq. 1 ) .and. ( a2(i1,ix) .eq. 1 )) ) then
              !
              if(i1 .eq. i2) then
                const(6) = const(6) + 1
                goto 4
              end if
              ! A new instance was found: the pair i1/i2 is tested only once to each one of the reactions
              !
              ! Stores the number of metabolites
              numNodes(1) = 2
              ! Stores the number of reactions
              numNodes(2) = 1
              ! Stores the number of reversible and irreversible reactions and the information about the reversibility
              ! of the reactions
              if( RevReact( ix ) .eq. 1 ) then
                supReactRevInstTemp(1) = 1  ! Reaction supReactTemp(1) is reversible
                numNodes(3) = 1  ! N of reversible reactions
                numNodes(4) = 0  ! N of irreversible reactions
              else
                supReactRevInstTemp(1) = 0  ! Reaction supReactTemp(1) is irreversible
                numNodes(3) = 0  ! N of reversible reactions
                numNodes(4) = 1  ! N of irreversible reactions
              end if
              !
              !
              ! Stores provisionally the 2 metabolites
              supMetTemp(1) = i1
              supMetTemp(2) = i2
              !
              ! Stores provisionally the reaction ix
              supReactTemp(1) = ix
              !
              ! Obtains the sum of the reference numbers of the metatabolites and of the reactions of the (possibly) new instance
              ! Sum of the reference numbers of the metabolites
              supTotMetInst = 0
              supTotMetInst = i1 + i2  ! Sum of metabolite i1 + metabolite i2
              ! Sum of the reference numbers of the reactions
              supTotReactInst = 0
              supTotReactInst = ix  ! Reaction ix
              !
              ! Calls the subroutine organizingInst, which obtains the pattern of the (possibly) new instance organized
              ! according to the reversibility of the reactions and to the degree (total and IN) of the nodes (metabolites
              ! and reactions)
              call organizingInst(supMetTemp, supReactTemp, a1, a2, &
              maxInterRevStr, maxMRConRevStr, numNodes, supReactRevInstTemp, &
              supRevStInst, supMetDegInst, supReactDegInst, supDataInst, &
              supMet, supReact, InstScope, supDifNodesInst)
              ! Instance within the considered scope? The maximum number of interconnections and different connections
              ! metabolite/reaction for a given number of metabolites and reactions were considered
              if( InstScope .eq. 1 ) then
                ! Instance is outside the scope
                const(4) = const(4) + 1
                InstScope = 0
                goto 4
              end if
              ! Instance is within the scope
              !
              ! Determines the structure of the newly found instance
              ! Calls the subroutine isomorphismTest
              call isomorphismTest(supRevStInst, RevStruct, DataRevStru, &
              MetDegRevStru, ReactDegRevStru, numNodes, supDataInst, &
              supMetDegInst, supReactDegInst, PermutNode, numPermutNode, &
              PermutDirInvReact, numPermDirInvReact, &
              nrefStruLow(MinGenStruInst(supDataInst(2),supDataInst(1))), &
              nrefStruHigh(MaxGenStruInst(supDataInst(2),supDataInst(1))), &
              refNumPermMet, refNumPermIrrReact, refNumPermRevReact, &
              refNumPermDirInvReact, isoStruRef, DifNodesStru, supDifNodesInst)
              ! The instance found is within the considered scope? For that to happen, isoStruRef must be different
              ! from "0"
              if( isoStruRef .eq. 0 ) then
                !
                ! Adds 1 to the total number of instances not asigned to a specific structure (used to
                ! control the quality of the code)
                const(2) = const(2) + 1
                goto 4
                !
              end if
              !
              ! Call the subroutine isoInstances, to determine if this is a new instance
              call isoInstances(numNodes, supTotMetInst, supTotReactInst, supMet, &
              supReact, TotMetInst, TotReactInst, OrdMetInst, OrdReactInst, &
              InstStruct, minRefNInstStru, maxRefNInstStru, InstTest, supMetOrd, &
              supReactOrd, maxNumMet, maxNumReact, maxNumInst, isoStruRef, numInst, &
              TempCut)
              if( InstTest .eq. 1 ) then
                const(3) = const(3) + 1
                InstTest = 0
                goto 4
              end if
              !
              !
              ! Stores the information regarding the new instance numInst
              !
              ! Reference number of the instance
              numInst = numInst + 1
              ! Maximum reference number of a instance found for the 1st pseudo-generation
              nMaxInstPGen(1) = nMaxInstPGen(1) + 1
              ! Maximum reference number of a instance of structure isoStruRef
              maxRefNInstStru(isoStruRef) = numInst
              !
              ! Stores the number of the structure of the instance
              InstStruct(numInst) = isoStruRef
              !
              ! Stores the information regarding the metabolites and the reactions of instance numInst
              ! Calls the subroutine storeInst
              call storeInst(numInst, isoStruRef, supTotMetInst, &
              supTotReactInst, refNumPermMet, refNumPermIrrReact, &
              refNumPermRevReact, numNodes, supMet, supReact, &
              PermutNode, MetInst, ReactInst, TotMetInst, &
              TotReactInst, OutRevStruct, maxNumMet, maxNumReact, &
              maxNumInst, supMetOrd, supReactOrd, OrdMetInst, OrdReactInst)
              !
              ! Clears the variables isoStruRef, refNumPermMet, refNumPermIrrReact, refNumPermRevReact and refNumPermDirInvReact
              isoStruRef = 0
              refNumPermMet = 0
              refNumPermIrrReact = 0
              refNumPermRevReact = 0
              refNumPermDirInvReact = 0
              !
            end if
4         continue
        end do
      end do


!      Adds pseudo-structural elements to the instances obtained for the previous pseudo-generation to
!      obtain the instances of the next pseudo-generation. The pseudo-structural elements added are:
!      i) 1 metabolite; ii) 1 reaction; and iii) 1 metabolite + 1 reaction. The instances obtained for
!      each pseudo-generation can be part of different generations.
      !
      ! Exterior loop continues till the last pseudo-generation has been visited. In practise, the number
      ! of pseudo-generations is equal to the number of generations considered within the scope
      !
      ! To test the code, initially we will not consider the maxNumGen for the exterior loop. Instead,
      ! a temporary variable, numPseuGen, is used
      numPseuGen = max((maxNumMet-1),maxNumReact)
      !
      ! Support constant used to store the structure with the lower reference number that will be compared to
      ! the newly found instance. Initially, I will check all structures. So, nrefStruLow_ = const(1) = 1 and
      ! nrefStruHigh_ = numStructures
      const(1) = 1
      !
      do 104 i=2,max(2,numPseuGen)  ! The variable numPseuGen should be substituted by maxNumGen after the code been
      !                           validated
       ! Adds successively the 3 pseudo-structural elements
        write(*,*) 'Number of pseudo-generation'
        write(*,*) i-1
        write(*,*) 'Number of instances of this pseudo-generation'
        write(*,*) nMaxInstPGen(i-1) - nMinInstPGen(i-1) + 1
        write(*,*) 'Total number of instances'
        write(*,*) numInst
        write(*,*)
        !
        ! Is this the pseudo-generation numPseuGen? If it is, equals the variable TempCut to 1
        If( i .eq. numPseuGen ) TempCut = 1
        !
        ! Minimum reference number of the pseudo-generation i: equal to the maximum reference number + 1 of
        ! a instance of the previous pseudo-generation
        nMinInstPGen(i) = nMaxInstPGen(i-1) + 1
        !
        ! Maximum reference number of the pseudo-generation i: the value is initially equal to the maximum
        ! reference number of the previous pseudo-generation; every time a new instance is found, one adds 1
        ! to this variable
        nMaxInstPGen(i) = nMaxInstPGen(i-1)
      !
      ! i) adds a new metabolite to each one of the reactions of each one of the instances of the previous
      !    pseudo-generation
        do 101 i1=nMinInstPGen(i-1),nMaxInstPGen(i-1)
          !
          write(*,1004) numInst, i, 1, (i1 - nMinInstPGen(i-1) + 1), (nMaxInstPGen(i-1) - nMinInstPGen(i-1) + 1)
          !
          ! Is instance i1 a concretization of a structure of the generation maxNumGen?
          if( DataRevStru(1,InstStruct(i1)) .eq.  maxNumGen ) goto 101
          !
          ! Imposed restriction: only instances with maxNumMet or less metabolites
          ! Number of metabolites of the instances obtained by adding a metabolite to instance i1 = number of
          ! metabolites of instance i1 + 1 ( DataRevStru(2,InstStruct(i1)) + 1 )
          if( ( DataRevStru(2,InstStruct(i1)) + 1 ) .gt. maxNumMet ) goto 101
          !
          ! Imposed restriction: only instances wih maxNumReact or less reactions
          ! Number of reactions of the instances obtained by adding a metabolite to instance i1 = number of
          ! reactions of instance i1 (DataRevStru(3,InstStruct(i1)))
          if( DataRevStru(3,InstStruct(i1)) .gt. maxNumReact ) goto 101
          !
          ! Adds a metabolite to each one of the reactions of instance i1
          do j1=1,DataRevStru(3,InstStruct(i1))  ! Number of reactions of the instance i1
            !
            ! Checks if the metabolite k1 is connected to the reaction j1
            do 5 k1=1,n1  ! Number of metabolites of the metabolic network
              !
              ! Checks if the metabolite k1 is connected to the reaction j1 (as reactant or as a product)
              if( ( a1( ReactInst(j1,i1), k1 ) .ne. 0 ) .or. ( a2( k1, ReactInst(j1,i1) ) .ne. 0 ) ) then
                !
                ! Is k1 a metabolite of the instance i?
                do iI_M1=1,DataRevStru(2,InstStruct(i1))  ! Number of metabolites of instance i1
                  !
                  ! Metabolite k1 equal to metabolite iI_M1 of instance i1?
                  if( k1 .eq. MetInst(iI_M1,i1) ) goto 5
                  !
                end do
                !
                ! Metabolite k1 is not one of the metabolites of instance i1
                !
                ! Stores the number of metabolites and the number of reactions of the (possibly) new instance
                ! N of metabolites
                numNodes(1) = DataRevStru(2,InstStruct(i1)) + 1  ! N of metabolites of instance i1 + 1
                ! N of reactions
                numNodes(2) = DataRevStru(3,InstStruct(i1))  ! N of reactions of instance i1
                ! N of reversible reactions
                numNodes(3) = DataRevStru(4,InstStruct(i1))  ! N of reversible reactions of instance i1
                ! N of irreversible reactions
                numNodes(4) = numNodes(2) - numNodes(3)  ! N of reactions - n of reversible reactions
                !
                ! The number of reversible reactions of the new instance is higher than the maximum allowed within
                ! the scope for the same number of metabolites and reacions?
                if( numNodes(3) .gt. maxNumRevReactStr(numNodes(2),numNodes(1)) ) then
                  ! Instance is outside the scope
                  const(4) = const(4) + 1
                  goto 5
                end if
                !
                ! Stores temporarily the reference number of the metabolites of the (possibly) new instance
                ! Metabolites corresponding to instance i1
                do iM1=1,DataRevStru(2,InstStruct(i1))  ! Number of metabolites of instance i1
                  supMetTemp(iM1) = MetInst(iM1,i1)
                end do
                ! Metabolite k1
                supMetTemp(numNodes(1)) = k1
                !
                ! Stores temporarily the reference number of the reactions of the (possibly) new instance and
                ! its reversibility
                ! Reactions corresponding to instance i1
                do iR1=1,DataRevStru(3,InstStruct(i1))  ! Number of reactions of instance i1
                  supReactTemp(iR1) = ReactInst(iR1,i1)
                  if( RevReact( supReactTemp(iR1) ) .eq. 1 ) then
                    supReactRevInstTemp(iR1) = 1  ! Reaction supReactTemp(iR1) is reversible
                  else
                    supReactRevInstTemp(iR1) = 0  ! Reaction supReactTemp(iR1) is irreversible
                  end if
                end do
                !
                ! Obtains the sum of the reference numbers of the metatabolites and of the reactions of the (possibly) new instance
                ! Sum of the reference numbers of the metabolites
                supTotMetInst = 0
                supTotMetInst = TotMetInst(i1) + k1  ! Sum of instance i1 + metabolite k1
                ! Sum of the reference numbers of the reactions
                supTotReactInst = 0
                supTotReactInst = TotReactInst(i1)  ! Sum of instance i1
                !
                ! Calls the subroutine organizingInst, which obtains the pattern of the (possibly) new instance organized
                ! according to the reversibility of the reactions and to the degree (total and IN) of the nodes (metabolites
                ! and reactions)
                call organizingInst(supMetTemp, supReactTemp, a1, a2, &
                maxInterRevStr, maxMRConRevStr, numNodes, supReactRevInstTemp, &
                supRevStInst, supMetDegInst, supReactDegInst, supDataInst, &
                supMet, supReact, InstScope, supDifNodesInst)
                ! Instance within the considered scope? The maximum number of interconnections and different connections
                ! metabolite/reaction for a given number of metabolites and reactions were considered
                if( InstScope .eq. 1 ) then
                  ! Instance is outside the scope
                  const(4) = const(4) + 1
                  InstScope = 0
                  goto 5
                end if
                ! Instance is within the scope
                !
                ! Determines the structure of the (possibly) new instance
                ! Calls the subroutine isomorphismTest
                call isomorphismTest(supRevStInst, RevStruct, DataRevStru, MetDegRevStru, &
                ReactDegRevStru, numNodes, supDataInst, supMetDegInst, supReactDegInst, &
                PermutNode, numPermutNode, PermutDirInvReact, numPermDirInvReact, &
                nrefStruLow( MinGenStruInst( supDataInst(2) , supDataInst(1) ) ), &
                nrefStruHigh( MaxGenStruInst( supDataInst(2) , supDataInst(1) ) ), &
                refNumPermMet, refNumPermIrrReact, refNumPermRevReact, &
                refNumPermDirInvReact, isoStruRef, DifNodesStru, supDifNodesInst)
                ! The instance found is within the considered scope? For that to happen, isoStruRef must be different
                ! from "0"
                if( isoStruRef .eq. 0 ) then
                  !
                  ! Adds 1 to the total number of instances not asigned to a specific structure (used to
                  ! control the quality of the code)
                  const(2) = const(2) + 1
                  goto 5
                  !
                end if
                !
                ! Call the subroutine isoInstances, to determine if this is a new instance
                call isoInstances(numNodes, supTotMetInst, supTotReactInst, supMet, &
                supReact, TotMetInst, TotReactInst, OrdMetInst, OrdReactInst, &
                InstStruct, minRefNInstStru, maxRefNInstStru, InstTest, supMetOrd, &
                supReactOrd, maxNumMet, maxNumReact, maxNumInst, isoStruRef, numInst, &
                TempCut)
                if( InstTest .eq. 1 ) then
                  const(3) = const(3) + 1
                  InstTest = 0
                  goto 5
                end if
                !
                !
                ! Stores the information regarding the new instance numInst
                !
                ! Reference number of the instance
                numInst = numInst + 1
                ! Maximum reference number of a instance found for the ith pseudo-generation
                nMaxInstPGen(i) = nMaxInstPGen(i) + 1
                ! Maximum reference number of a instance of structure isoStruRef
                maxRefNInstStru(isoStruRef) = numInst
                !
                ! Stores the number of the structure of the instance
                InstStruct(numInst) = isoStruRef
                !
                ! Stores the information regarding the metabolites and the reactions of instance numInst
                ! Calls the subroutine storeInst
                call storeInst(numInst, isoStruRef, supTotMetInst, &
                supTotReactInst, refNumPermMet, refNumPermIrrReact, &
                refNumPermRevReact, numNodes, supMet, supReact, &
                PermutNode, MetInst, ReactInst, TotMetInst, &
                TotReactInst, OutRevStruct, maxNumMet, maxNumReact, &
                maxNumInst, supMetOrd, supReactOrd, OrdMetInst, OrdReactInst)
                !
                ! Clears the variables isoStruRef, refNumPermMet, refNumPermIrrReact, refNumPermRevReact and refNumPermDirInvReact
                isoStruRef = 0
                refNumPermMet = 0
                refNumPermIrrReact = 0
                refNumPermRevReact = 0
                refNumPermDirInvReact = 0
                !
              end if
5           continue
          end do
101     continue
      !
      ! ii) adds a new reaction to connect two metabolites of an instance of the previous
      !     pseudo-generation as pairs reactant/product
        do 102 i2=nMinInstPGen(i-1),nMaxInstPGen(i-1)
          !
          write(*,1004) numInst, i, 2, (i2 - nMinInstPGen(i-1) + 1), (nMaxInstPGen(i-1) - nMinInstPGen(i-1) + 1)
          !
          ! Is instance i2 a concretization of a structure of the generation maxNumGen?
          if( DataRevStru(1,InstStruct(i2)) .eq. maxNumGen ) goto 102
          !
          ! Imposed restriction: only instances with maxNumMet or less metabolites
          ! Number of metabolites of the instances of the pseudo-generation i obtained by adding a reaction to
          ! instance i2 = number of metabolites of instance i2 (DataRevStru(2,InstStruct(i2)))
          if( DataRevStru(2,InstStruct(i2)) .gt. maxNumMet ) goto 102
          !
          ! Imposed restriction: only instances with maxNumReact or less reactions
          ! Number of reactions of the instances of the pseudo-generation i obtained by adding a metabolite to
          ! instance i2 = number of reactions of instance i2 + 1 ( DataRevStru(3,InstStruct(i2)) + 1 )
          if( ( DataRevStru(3,InstStruct(i2)) + 1 ) .gt. maxNumReact ) goto 102
          !
          ! Adds a reaction to all the possible pairs of metabolites of the instance i2
          ! Loop to obtain the first metabolite
          do j2=1,( DataRevStru(2,InstStruct(i2)) - 1 )  ! Number of metabolites of the instance i2
            ! Loop to obtain the second metabolite
            do k2=(j2 + 1),DataRevStru(2,InstStruct(i2))  ! Number of metabolites   of the instance i2
              ! Loop to check all the n2 reactions
              do 10 l2=1,n2  ! Number of reactions  of the metabolic network
                !
                ! Checks if the reaction l2 connects the metabolites MetInst(j2,i2) and MetInst(k2,i2) as a pair reactant/product
                if( ( ( a2(MetInst(j2,i2),l2) .ne. 0 ) .and. ( a1(l2,MetInst(k2,i2)) .ne. 0 ) )  .or. &
                ( ( a2(MetInst(k2,i2),l2) .ne. 0 ) .and. ( a1(l2,MetInst(j2,i2)) .ne. 0 ) ) ) then
                  !
                  ! Checks if the reaction l2 is already part of the instance i2
                  do iI_R2=1,DataRevStru(3,InstStruct(i2))  ! Number of reactions of instance i2
                    if( l2 .eq. ReactInst(iI_R2,i2) ) goto 10
                  end do
                  !
                  ! Reaction l2 is not part of instance i2
                  !
                  ! Stores the number of metabolites and the number of reactions of the (possibly) new instance
                  ! N of metabolites
                  numNodes(1) = DataRevStru(2,InstStruct(i2))  ! N of metabolites of instance i2
                  ! N of reactions
                  numNodes(2) = DataRevStru(3,InstStruct(i2)) + 1  ! N of reactions of instance i2 + 1
                  ! Reversibility of reaction l2?
                  if( RevReact( l2 ) .eq. 1 ) then
                    ! Reaction l2 is reversible
                    !
                    ! N of reversible reactions
                    numNodes(3) = DataRevStru(4,InstStruct(i2)) + 1  ! N of reversible reactions of instance i2 + 1
                    ! N of irreversible reactions
                    numNodes(4) = numNodes(2) - numNodes(3)  ! N of reactions - n of reversible reactions
                  else
                    ! Reaction l2 is irreversible
                    !
                    ! N of reversible reactions
                    numNodes(3) = DataRevStru(4,InstStruct(i2))  ! N of reversible reactions of instance i2
                    ! N of irreversible reactions
                    numNodes(4) = numNodes(2) - numNodes(3)  ! N of reactions - n of reversible reactions
                  end if
                  !
                  ! The number of reversible reactions of the new instance is higher than the maximum allowed within
                  ! the scope for the   same number of metabolites and reacions?
                  if( numNodes(3) .gt. maxNumRevReactStr(numNodes(2),numNodes(1)) ) then
                    ! Instance is outside the scope
                    const(4) = const(4) + 1
                    goto 10
                  end if
                  !
                  ! Stores temporarily the reference number of the metabolites of the (possibly) new instance
                  ! Metabolites corresponding to instance i2
                  do iM2=1,DataRevStru(2,InstStruct(i2))  ! Number of metabolites of instance i2
                    supMetTemp(iM2) = MetInst(iM2,i2)
                  end do
                  !
                  ! Stores temporarily the reference number of the reactions of the (possibly) new instance and
                  ! its reversibility
                  ! Reactions corresponding to instance i2
                  do iR2=1,DataRevStru(3,InstStruct(i2))  ! Number of reactions of instance i2
                    supReactTemp(iR2) = ReactInst(iR2,i2)
                    if( RevReact( supReactTemp(iR2) ) .eq. 1 ) then
                      supReactRevInstTemp(iR2) = 1  ! Reaction supReactTemp(iR2) is reversible
                    else
                      supReactRevInstTemp(iR2) = 0  ! Reaction supReactTemp(iR2) is irreversible
                    end if
                  end do
                  ! Reaction l2
                  supReactTemp(numNodes(2)) = l2
                  if( RevReact( l2 ) .eq. 1 ) then
                    supReactRevInstTemp(numNodes(2)) = 1  ! Reaction supReactTemp(iR2) is reversible
                  else
                    supReactRevInstTemp(numNodes(2)) = 0  ! Reaction supReactTemp(iR2) is irreversible
                  end if
                  !
                  ! Obtains the sum of the reference numbers of the metatabolites and of the reactions of the (possibly) new instance
                  ! Sum of the reference numbers of the metabolites
                  supTotMetInst = 0
                  supTotMetInst = TotMetInst(i2)  ! Sum of instance i2
                  ! Sum of the reference numbers of the reactions
                  supTotReactInst = 0
                  supTotReactInst = TotReactInst(i2) + l2  ! Sum of instance i2 + reaction l2
                  !
                  ! Calls the subroutine organizingInst, which obtains the pattern of the (possibly) new instance organized
                  ! according to the reversibility of the reactions and to the degree (total and IN) of the nodes (metabolites
                  ! and reactions)
                  call organizingInst(supMetTemp, supReactTemp, a1, a2, &
                  maxInterRevStr, maxMRConRevStr, numNodes, supReactRevInstTemp, &
                  supRevStInst, supMetDegInst, supReactDegInst, supDataInst, &
                  supMet, supReact, InstScope, supDifNodesInst)
                  ! Instance within the considered scope? The maximum number of interconnections and different connections
                  ! metabolite/reaction for a given number of metabolites and reactions were considered
                  if( InstScope .eq. 1 ) then
                    ! Instance is outside the scope
                    const(4) = const(4) + 1
                    InstScope = 0
                    goto 10
                  end if
                  ! Instance is within the scope
                  !
                  ! Determines the structure of the (possibly) new instance
                  ! Calls the subroutine isomorphismTest
                  call isomorphismTest(supRevStInst, RevStruct, &
                  DataRevStru, MetDegRevStru, ReactDegRevStru, numNodes, supDataInst, &
                  supMetDegInst, supReactDegInst, PermutNode, numPermutNode, &
                  PermutDirInvReact, numPermDirInvReact, &
                  nrefStruLow( MinGenStruInst( supDataInst(2) , supDataInst(1) ) ), &
                  nrefStruHigh( MaxGenStruInst( supDataInst(2) , supDataInst(1) ) ), &
                  refNumPermMet, refNumPermIrrReact, refNumPermRevReact, &
                  refNumPermDirInvReact, isoStruRef, DifNodesStru, supDifNodesInst)
                  ! The instance found is within the considered scope? For that to happen, isoStruRef must be different
                  ! from "0"
                  if( isoStruRef .eq. 0 ) then
                    !
                    ! Adds 1 to the total number of instances not asigned to a specific structure (used to
                    ! control the quality of the code)
                    const(2) = const(2) + 1
                    goto 10
                    !
                  end if
                  !
                  ! Call the subroutine isoInstances, to determine if this is a new instance
                  call isoInstances(numNodes, supTotMetInst, supTotReactInst, supMet, &
                  supReact, TotMetInst, TotReactInst, OrdMetInst, OrdReactInst, &
                  InstStruct, minRefNInstStru, maxRefNInstStru, InstTest, supMetOrd, &
                  supReactOrd, maxNumMet, maxNumReact, maxNumInst, isoStruRef, numInst, &
                  TempCut)
                  if( InstTest .eq. 1 ) then
                    const(3) = const(3) + 1
                    InstTest = 0
                    goto 10
                  end if
                  !
                  !
                  ! Stores the information regarding the new instance numInst
                  !
                  ! Reference number of the instance
                  numInst = numInst + 1
                  ! Maximum reference number of a instance found for the ith pseudo-generation
                  nMaxInstPGen(i) = nMaxInstPGen(i) + 1
                  ! Maximum reference number of a instance of structure isoStruRef
                  maxRefNInstStru(isoStruRef) = numInst
                  !
                  ! Stores the number of the structure of the instance
                  InstStruct(numInst) = isoStruRef
                  !
                  ! Stores the information regarding the metabolites and the reactions of instance numInst
                  ! Calls the subroutine storeInst
                  call storeInst(numInst, isoStruRef, supTotMetInst, &
                  supTotReactInst, refNumPermMet, refNumPermIrrReact, &
                  refNumPermRevReact, numNodes, supMet, supReact, &
                  PermutNode, MetInst, ReactInst, TotMetInst, &
                  TotReactInst, OutRevStruct, maxNumMet, maxNumReact, &
                  maxNumInst, supMetOrd, supReactOrd, OrdMetInst, OrdReactInst)
                  !
                  ! Clears the variables isoStruRef, refNumPermMet, refNumPermIrrReact, refNumPermRevReact and refNumPermDirInvReact
                  isoStruRef = 0
                  refNumPermMet = 0
                  refNumPermIrrReact = 0
                  refNumPermRevReact = 0
                  refNumPermDirInvReact = 0
                  !
                end if
10            continue
            end do
          end do
102     continue
      !
      ! iii) adds a new reaction and a new metabolite. The new reaction should connect the new
      !      metabolite to a existent metabolite of an instance of the previous pseudo-generation
      !      to form a pair reactant/product
      !
        !
   !     if( i .eq. 3 ) goto 104
        do 103 i3=nMinInstPGen(i-1),nMaxInstPGen(i-1)
          !
          write(*,1004) numInst, i, 3, (i3 - nMinInstPGen(i-1) + 1), (nMaxInstPGen(i-1) - nMinInstPGen(i-1) + 1)
          !
          ! Is instance i3 a concretization of a structure of the generation maxNumGen?
          if( DataRevStru(1,InstStruct(i3)) .eq. maxNumGen ) goto 103
          !
          ! Imposed restriction: only instances with maxNumMet or less metabolites
          ! Number of metabolites of the instances of the pseudo-generation i obtained by adding a reaction to
          ! instance i3 = number of metabolites of instance i3 + 1 ( DataRevStru(2,InstStruct(i3)) + 1 )
          if( ( DataRevStru(2,InstStruct(i3)) + 1 ) .gt. maxNumMet ) goto 103
          !
          ! Imposed restriction: only instances with maxNumReact or less reactions
          ! Number of reactions of the instances of the pseudo-generation i obtained by adding a metabolite to
          ! instance i3 = number of reactions of instance i3 + 1 ( DataRevStru(3,InstStruct(i3)) + 1 )
          if( ( DataRevStru(3,InstStruct(i3)) + 1 ) .gt. maxNumReact ) goto 103
          !
          ! Adds a reaction to connect a new metabolite to one of the metabolites of instance i3 to form a pair
          ! reactant/product
          ! Loop to obtain the metabolite of instance i3
          do j3=1,DataRevStru(2,InstStruct(i3))  ! Number of metabolites of instance i3
            ! Loop to obtain the new metabolite from the metabolic network
            do k3=1,n1  ! Number of metabolites of the metabolic network
              ! Loop to obtain the new reaction from the metabolic network
              do 15 l3=1,n2  ! Number of reactions of the metabolic network
                !
                ! Checks if the new reaction l3 connects the new metabolites k3 to the metabolite MetInst(j3,i3) as a
                ! pair reactant/product
                if( ( ( a2(MetInst(j3,i3),l3) .ne. 0 ) .and. ( a1(l3,k3) .ne. 0 ) )  .or. &
                ( ( a2(k3,l3) .ne. 0 ) .and. ( a1(l3,MetInst(j3,i3)) .ne. 0 ) ) ) then
                  !
                  ! Checks if the metabolite k3 is already part of instance i3
                  do iI_M3=1,DataRevStru(2,InstStruct(i3))  ! Number of metabolites of instance i3
                    if( k3 .eq. MetInst(iI_M3,i3) ) goto 15
                  end do
                  ! Checks if the reaction l3 is already part of intance i3
                  do iI_R3=1,DataRevStru(3,InstStruct(i3))  ! Number of reactions of instance i3
                    if( l3 .eq. ReactInst(iI_R3,i3) ) goto 15
                  end do
                  !
                  ! Reaction l3 and metabolite k3 are not part of instance i3
                  !
                  ! Stores the number of metabolites and the number of reactions of the (possibly) new instance
                  ! N of metabolites
                  numNodes(1) = DataRevStru(2,InstStruct(i3)) + 1  ! N of metabolites of instance i3 + 1
                  ! N of reactions
                  numNodes(2) = DataRevStru(3,InstStruct(i3)) + 1  ! N of reactions of instance i3 + 1
                  ! Reversibility of reaction l3?
                  if( RevReact( l3 ) .eq. 1 ) then
                    ! Reaction l3 is reversible
                    !
                    ! N of reversible reactions
                    numNodes(3) = DataRevStru(4,InstStruct(i3)) + 1  ! N of reversible reactions of instance i3 + 1
                    ! N of irreversible reactions
                    numNodes(4) = numNodes(2) - numNodes(3)  ! N of reactions - n of reversible reactions
                  else
                    ! Reaction l3 is irreversible
                    !
                    ! N of reversible reactions
                    numNodes(3) = DataRevStru(4,InstStruct(i3))  ! N of reversible reactions of instance i3
                    ! N of irreversible reactions
                    numNodes(4) = numNodes(2) - numNodes(3)  ! N of reactions - n of reversible reactions
                  end if
                  !
                  ! The number of reversible reactions of the new instance is higher than the maximum allowed within
                  ! the scope for the   same number of metabolites and reacions?
                  if( numNodes(3) .gt. maxNumRevReactStr(numNodes(2),numNodes(1)) ) then
                    ! Instance is outside the scope
                    const(4) = const(4) + 1
                    goto 15
                  end if
                  !
                  ! Stores temporarily the reference number of the metabolites of the (possibly) new instance
                  ! Metabolites corresponding to instance i3
                  do iM3=1,DataRevStru(2,InstStruct(i3))  ! Number of metabolites of instance i3
                    supMetTemp(iM3) = MetInst(iM3,i3)
                  end do
                  ! Metabolite k3
                  supMetTemp(numNodes(1)) = k3
                  !
                  ! Stores temporarily the reference number of the reactions of the (possibly) new instance and
                  ! its reversibility
                  ! Reactions corresponding to instance i3
                  do iR3=1,DataRevStru(3,InstStruct(i3))  ! Number of reactions of instance i3
                    supReactTemp(iR3) = ReactInst(iR3,i3)
                    if( RevReact( supReactTemp(iR3) ) .eq. 1 ) then
                      supReactRevInstTemp(iR3) = 1  ! Reaction supReactTemp(iR3) is reversible
                    else
                      supReactRevInstTemp(iR3) = 0  ! Reaction supReactTemp(iR3) is irreversible
                    end if
                  end do
                  ! Reaction l3
                  supReactTemp(numNodes(2)) = l3
                  if( RevReact( l3 ) .eq. 1 ) then
                    supReactRevInstTemp(numNodes(2)) = 1  ! Reaction l3 is reversible
                  else
                    supReactRevInstTemp(numNodes(2)) = 0  ! Reaction l3 is irreversible
                  end if
                  !
                  ! Obtains the sum of the reference numbers of the metatabolites and of the reactions of the (possibly) new instance
                  ! Sum of the reference numbers of the metabolites
                  supTotMetInst = 0
                  supTotMetInst = TotMetInst(i3) + k3  ! Sum of instance i3 + metabolite k3
                  ! Sum of the reference numbers of the reactions
                  supTotReactInst = 0
                  supTotReactInst = TotReactInst(i3) + l3  ! Sum of instance i3 + reaction l3
                  !
                  ! Calls the subroutine organizingInst, which obtains the pattern of the (possibly) new instance organized
                  ! according to the reversibility of the reactions and to the degree (total and IN) of the nodes (metabolites
                  ! and reactions)
                  call organizingInst(supMetTemp, supReactTemp, a1, a2, &
                  maxInterRevStr, maxMRConRevStr, numNodes, supReactRevInstTemp, &
                  supRevStInst, supMetDegInst, supReactDegInst, supDataInst, &
                  supMet, supReact, InstScope, supDifNodesInst)
                  ! Instance within the considered scope? The maximum number of interconnections and different connections
                  ! metabolite/reaction for a given number of metabolites and reactions were considered
                  if( InstScope .eq. 1 ) then
                    ! Instance is outside the scope
                    const(4) = const(4) + 1
                    InstScope = 0
                    goto 15
                  end if
                  ! Instance is within the scope
                  !
                  ! Determines the structure of the (possibly) new instance
                  ! Calls the subroutine isomorphismTest
                  call isomorphismTest(supRevStInst, RevStruct, &
                  DataRevStru, MetDegRevStru, ReactDegRevStru, numNodes, &
                  supDataInst, supMetDegInst, supReactDegInst, PermutNode, &
                  numPermutNode, PermutDirInvReact, numPermDirInvReact, &
                  nrefStruLow( MinGenStruInst( supDataInst(2) , supDataInst(1) ) ), &
                  nrefStruHigh( MaxGenStruInst( supDataInst(2) , supDataInst(1) ) ), &
                  refNumPermMet, refNumPermIrrReact, refNumPermRevReact, &
                  refNumPermDirInvReact, isoStruRef, DifNodesStru, supDifNodesInst)
                  ! The instance found is within the considered scope? For that to happen, isoStruRef must be different
                  ! from "0"
                  if( isoStruRef .eq. 0 ) then
                    !
                    ! Adds 1 to the total number of instances not asigned to a specific structure (used to
                    ! control the quality of the code)
                    const(2) = const(2) + 1
                    goto 15
                    !
                  end if
                  !
                  ! Call the subroutine isoInstances, to determine if this is a new instance
                  call isoInstances(numNodes, supTotMetInst, supTotReactInst, supMet, &
                  supReact, TotMetInst, TotReactInst, OrdMetInst, OrdReactInst, &
                  InstStruct, minRefNInstStru, maxRefNInstStru, InstTest, supMetOrd, &
                  supReactOrd, maxNumMet, maxNumReact, maxNumInst, isoStruRef, numInst, &
                  TempCut)
                  if( InstTest .eq. 1 ) then
                    const(3) = const(3) + 1
                    InstTest = 0
                    goto 15
                  end if
                  !
                  !
                  ! Stores the information regarding the new instance numInst
                  !
                  ! Reference number of the instance
                  numInst = numInst + 1
                  ! Maximum reference number of a instance found for the ith pseudo-generation
                  nMaxInstPGen(i) = nMaxInstPGen(i) + 1
                  ! Maximum reference number of a instance of structure isoStruRef
                  maxRefNInstStru(isoStruRef) = numInst
                  !
                  ! Stores the number of the structure of the instance
                  InstStruct(numInst) = isoStruRef
                  !
                  ! Stores the information regarding the metabolites and the reactions of instance numInst
                  ! Calls the subroutine storeInst
                  call storeInst(numInst, isoStruRef, supTotMetInst, &
                  supTotReactInst, refNumPermMet, refNumPermIrrReact, &
                  refNumPermRevReact, numNodes, supMet, supReact, &
                  PermutNode, MetInst, ReactInst, TotMetInst, &
                  TotReactInst, OutRevStruct, maxNumMet, maxNumReact, &
                  maxNumInst, supMetOrd, supReactOrd, OrdMetInst, OrdReactInst)
                  !
                  ! Clears the variables isoStruRef, refNumPermMet, refNumPermIrrReact, refNumPermRevReact and refNumPermDirInvReact
                  isoStruRef = 0
                  refNumPermMet = 0
                  refNumPermIrrReact = 0
                  refNumPermRevReact = 0
                  refNumPermDirInvReact = 0
                  !
                end if
15            continue
            end do
          end do
103     continue
104   continue


      write(*,*)
      write(*,*) 'Number of pseudo-generation'
      write(*,*) i
      write(*,*) 'Number of instances of this pseudo-generation'
      write(*,*) nMaxInstPGen(i) - nMinInstPGen(i) + 1
      write(*,*) 'Total number of instances'
      write(*,*) numInst
      write(*,*) nMaxInstPGen(i)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      OUTPUT FILES                     !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Store the information regarding the metabolites of the instances found organized
!      according to the degree of the nodes. File:
!      'MetInst.txt'
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='MetInst.txt')
      do i1=1,numInst
        write(8,1003) (MetInst(i2,i1), i2=1,maxNumMet)
      end do
      close(8)
      ! Deallocate variable MetInst(i2,i1)
      deallocate(MetInst)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Store the information regarding the reactionss of the instances found organized
!      according to the degree of the nodes. File:
!      'ReactInst.txt'
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='ReactInst.txt')
      do i1=1,numInst
        write(8,1003) (ReactInst(i2,i1), i2=1,maxNumReact)
      end do
      close(8)
      ! Deallocate variable ReactInst(i2,i1)
      deallocate(ReactInst)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Stores the reference number of the structure of each one of the instances. File:
!      'InstStruct.txt'
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='InstStruct.txt')
      do i1=1,numInst
        write(8,1003) InstStruct(i1)
      end do
      close(8)
      ! Deallocate variable InstStruct(i1)
      deallocate(InstStruct)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Store the reference numbers of the metabolites of the instances found organized
!      from the higher to the lower number. File:
!      'MetInstOrdRefNum.txt'
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='MetInstOrdRefNum.txt')
      do i1=1,numInst
        write(8,1003) (OrdMetInst(i2,i1), i2=1,maxNumMet)
      end do
      close(8)
      ! Deallocate variable OrdMetInst(i2,i1)
      deallocate(OrdMetInst)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Store the reference numbers of the reactions of the instances found organized
!      from the higher to the lower number. File:
!      'ReactInstOrdRefNum.txt'
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='ReactInstOrdRefNum.txt')
      do i1=1,numInst
        write(8,1003) (OrdReactInst(i2,i1), i2=1,maxNumReact)
      end do
      close(8)
      ! Deallocate variable OrdReactInst(i2,i1)
      deallocate(OrdReactInst)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Store the sum of the reference numbers of the metabolites of each one of the
!      instances found. File:
!      'SumMetInst.txt'
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='SumMetInst.txt')
      do i1=1,numInst
        write(8,1003) TotMetInst(i1)
      end do
      close(8)
      ! Deallocate variable TotMetInst(i1)
      deallocate(TotMetInst)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Store the sum of the reference numbers of the reactions of each one of the
!      instances found. File:
!      'SumReactInst.txt'
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='SumReactInst.txt')
      do i1=1,numInst
        write(8,1003) TotReactInst(i1)
      end do
      close(8)
      ! Deallocate variable TotReactInst(i1)
      deallocate(TotReactInst)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Store the number of found instances of the structures. File:
!      'OutRevStruct.txt'
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='OutRevStruct.txt')
      do i1=1,numStructures
        write(8,1006) OutRevStruct(i1)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Store the number of instances found. File:
!      'numInstances.txt'
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='numInstances.txt')
      write(8,1007) numInst
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Formats
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
1001  format (a220)
1002  format (a100)
1003  format (8(i6, ','))
1004  format (i10, ',  ' i2, '.' i1, ',  ' i10, ' -' i10)
!1005  format (20(i5, ','))  ! HAD TO CHANGE FROM 10 TO 20...
1006  format (8(i10, ','))
1007  format (i10)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Program contains the following subroutines: i) storeInst, ii) organizingInst,
!      ii) allPermutations, iv) allRevReactPerm, v) isomorphismTest, vi) orderVector
!      and vii) isoInstances
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      contains


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!     Subroutine to store the new instance numInst into the corresponding variables
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine storeInst(numInst_, isoStruRe_, supTotMetInst_, supTotReactInst_, refNumPermMeta_, refNumPermIrrReacti_, &
      refNumPermRevReacti_, numN_, supMe_, supReac_, PermutNod_, MetIns_, ReactIns_, TotMetInst_, TotReactInst_, &
      OutRevStruct_, maxNumMeta_, maxNumReacti_, maxNumIns_, supMetOr_, supReactOr_, OrdMetIns_, OrdReactIns_)

! CONSTANTS
!      maxNumIns_ : expected maximum number of instances. This parameter is needed to create the allocatable arrays
!                   used to store the required information for each one of the instances
!      maxNumMeta_ : maximum number of metabolites considered for the structures within the considered number of
!                    generations maxNumGen
!      maxNumReacti_ : maximum number of reactions considered for the structures within the considered number of
!                      generations maxNumGen
!      numInst_ :  number of instance for the structures within the considered scope
!      isoStruRe_ : reference number of the structure in matrix RevStruct(i3,i2,i1) to which the instance in matrix
!                   supRevStInst(i2,i1,1) is a concretization.
!      refNumPermMeta_ : reference number of the permutation for the metabolites that allowed to the determine the structure
!                        in matrix RevStruct(i3,i2,i1) of which the instance in matrix supRevStInst(i3,i2,1) is a realization,
!                        i.e., PermutNode( i3 , refNumPermMet , numNode(1) )
!      refNumPermIrrReacti_ : reference number of the permutation for the irreversible reactions that allowed to the determine the
!                             structure in matrix RevStruct(i3,i2,i1) of which the instance in matrix supRevStInst(i3,i2,1) is a
!                             realization, i.e., PermutNode( i3 , refNumPermIrrReact , numNode(4) )
!      refNumPermRevReacti_ : reference number of the permutation for the reversible reactions that allowed to the determine the
!                           structure in matrix RevStruct(i3,i2,i1) of which the instance in matrix supRevStInst(i3,i2,1) is a
!                           realization, i.e., PermutNode( i3 , refNumPermRevReact , numNode(4) )
!      supTotMetInst_ : support constant with the sum of the reference numbers of the metabolites of the (possibly) new instance
!      supTotReactInst_ : support constant with the sum of the reference numbers of the reactions of the (possibly) new instance
!
!
! VECTORS
!      numN_(i1) : number of nodes of the (possibly) new instance. (i1=1) n of metabolites; (i1=2) n of reactions;
!                  (i1=3) n of reversible reactions; (i1=4) n of irreversible reactions
!      supMe_(i1) : support vector with all the metabolites of the (possibly) new instance ordered according to their degree
!                   (total and IN)
!      supReac_(i1) : support vector with all the reactions of the (possibly) new instance ordered according to their degree
!                     (total and IN) and to their reversibility (irreversible reactions are stored first)
!      supMetOr_(i1) : support vector with all the metabolites of the (possibly) new instance ordered from the higher to
!                      the lower reference number
!      supReactOr_(i1) : support vector with all the reactions of the (possibly) new instance ordered from the higher to
!                        the lower reference number
!      TotMetInst_(i1) : sum of the reference numbers of the metabolites of instance i1
!      TotReactInst_(i1) : sum of the reference numbers of the reactions of instance i1
!      OutRevStruct_(i1) : output data obtained for the structure i1 with the developed code => frequency (number of instances)
!
!
! MATRICES
!      PermutNod_(i3,i2,i1) : possible permutations for a given number i1 of nodes (metabolites or reactions), obtained
!                             initially by calling the subroutine allPermutations. i2) number of the permutation;
!                             i3) number of the nodes ordered by this permutation
!      MetIns_(i2,i1) :  reference numbers of the metabolites of the instance i1 organized to fit the pattern of the structure
!                        as stored in matrix RevStruct(i3,i2,i1)
!      ReactIns_(i2,i1) :  reference numbers of the reactions of the instance i1 organized to fit the pattern of the structure
!                          as stored in matrix RevStruct(i3,i2,i1)
!      OrdMetIns_(i2,i1) : metabolites of the instance i1 ordered from the higher to the lower reference numbers
!      OrdReactIns_(i2,i1) : reactions of the instance i1 ordered from the higher to the lower reference numbers

      implicit none

      ! IN variables
       ! Static variables
        ! Parameters
      integer, intent(in) :: maxNumIns_  ! 4 bits
        ! Integer: 1 bit
      integer*1, intent(in) :: PermutNod_(:,:,:), numN_(:)
      integer*1, intent(in) :: maxNumMeta_, maxNumReacti_
        ! Integer: 2 bits
      integer*2, intent(in) :: supMe_(:), supReac_(:), supMetOr_(:), supReactOr_(:)
      integer*2, intent(in) :: isoStruRe_, supTotMetInst_, supTotReactInst_
        ! Integer: 4 bits
      integer, intent(in) :: numInst_, refNumPermMeta_, refNumPermIrrReacti_
      integer, intent(in) :: refNumPermRevReacti_

      ! IN/OUT variables
      ! Static variables
      integer, intent(inout) :: OutRevStruct_(:)  ! 4 bits
       ! Allocatable variables
      integer*2, DIMENSION(maxNumMeta_,maxNumIns_), intent(inout) :: MetIns_, OrdMetIns_
      integer*2, DIMENSION(maxNumReacti_,maxNumIns_), intent(inout) :: ReactIns_, OrdReactIns_  ! 2 bits
      integer*2, DIMENSION(maxNumIns_), intent(inout) :: TotMetInst_, TotReactInst_  ! 2 bits

      ! General variables
      integer :: st1_  ! 4 bits

!      Stores the information regarding the new instance numInst_ into the corresponding variables:
      ! Sum of the reference numbers of the metabolites of instance numInst_
      TotMetInst_(numInst_) = supTotMetInst_
      !
      ! Sum of the reference numbers of the reactions of instance numInst_
      TotReactInst_(numInst_) = supTotReactInst_
      !
      ! Metabolites of the instance
      do st1_=1,numN_(1)  ! Number of metabolites of the instance numInst_
        !
        ! Stores the reference number of the metabolites according to the order present in the pattern
        ! of the corresponding structure
        MetIns_(st1_,numInst_) = supMe_( PermutNod_(st1_,refNumPermMeta_,numN_(1)) )
        !
        ! Stores the metabolites ordered from the higher to the lower reference number
        OrdMetIns_(st1_,numInst_) = supMetOr_(st1_)
        !
      end do
      !
      ! Irreversible reactions of the instance
      if( numN_(4) .ne. 0 ) then
        ! The instance has at least one irreversible reaction
        do st1_=1,numN_(4)  ! Number of irreversible reactions of the instance numInst_
          !
          ! Stores the reference number of the irreversible reactions according to the order present in the pattern
          ! of the corresponding structure
          ReactIns_(st1_,numInst_) = supReac_( PermutNod_(st1_,refNumPermIrrReacti_,numN_(4)) )
          !
          ! Stores the reactions fromthe higher to the lower number
          OrdReactIns_(st1_,numInst_) = supReactOr_(st1_)
          !
        end do
      end if
      !
      ! Reversible reactions of the instance
      if( numN_(3) .ne. 0 ) then
        ! The instance has at least one reversible reaction
        do st1_=1,numN_(3)  ! Number of reversible reactions of the instance numInst_
          !
          ! Stores the reference number of the reversible reactions according to the order present in the pattern
          ! of the corresponding structure
          ReactIns_( (numN_(4) + st1_) , numInst_ ) = supReac_( numN_(4) + PermutNod_(st1_,refNumPermRevReacti_,numN_(3)) )
          !
          ! Stores the reactions fromthe higher to the lower number
          OrdReactIns_( (numN_(4) + st1_) , numInst_) = supReactOr_( numN_(4) + st1_ )
          !
        end do
      end if
      !
      ! Adds this instance to structure isoStruRe_ (determination of the frequency)
      OutRevStruct_(isoStruRe_) = OutRevStruct_(isoStruRe_) + 1

      end subroutine storeInst


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!     Subroutine to obtain and organize the topological structures of the instances
!!     according to the degree of the nodes, metabolites and reactions. The reactions
!!     should be ordered first according to its reversibility: irreversible reactions
!!     are stored first, and then, if at least one exists, reversible reactions are
!!     stored.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine organizingInst(supMetTemp_, supReactTemp_, a1_, &
      a2_, maxInterRevStr_, maxMRConRevStr_, numNodes_, supReactRevInstTemp_, &
      supRevStInst_, supMetDegInst_, supReactDegInst_, supDataInstance_, &
      supMet_, supReact_, InstScope_, supDifNodesInst_)

! CONSTANTS
!      InstScope_ : indicates if the (possibly) newly found instance is within the considered topological scope. If it is
!                   within the scope, InstScope_ = 0; if it is not, InstScope_ = 1
!
!
! VECTORS
!      supMetTemp_(i1) : support vector with all the metabolites of the (possibly) new instance (might not be ordered)
!      supReactTemp_(i1) : support vector with all the reactions of the (possibly) new instance (might not be ordered)
!      supMetDegInstTemp_(i1) : gives the degree for each metabolite of a (possibly) new instance. i1=[(3 x (nmet - 1)) + 1],
!                                  total n of connections of the metabolite; i1=[(3 x (nmet - 1)) + 2], n of connections
!                                  comming to the metabolite (in); i1=[(3 x (nmet - 1)) + 3], n of connections going out of
!                                  the metabolite (out). The metabolites may not be organized according to their degree
!      supReactDegInstTemp_(i1) : gives the degree for each reaction of a possibly new instance. The irreversible reactions
!                                    are stored first, and then, in case at least one exists, the reversible reactions are stored.
!                                    i1=[(3 x (nmet - 1)) + 1], connections of the metabolite; i1=[(3 x (nmet - 1)) + 2],
!                                    n of connections comming to the total n ofmetabolite (in); i1=[(3 x (nmet - 1)) + 3],
!                                    n of connections going out of the metabolite (out). The reactions may not be organized
!                                    according to their degree
!      supMetDegInst_(i1) : gives the degree for each metabolite of a (possibly) new instance. i1=[(3 x (nmet - 1)) + 1],
!                           total n of connections of the metabolite; i1=[(3 x (nmet - 1)) + 2], n of connections
!                           comming to the metabolite (in); i1=[(3 x (nmet - 1)) + 3], n of connections going out of
!                           the metabolite (out)
!      supReactDegInst_(i1) : gives the degree for each reaction of a possibly new instance. The irreversible reactions
!                             are stored first, and then, in case at least one exists, the reversible reactions are stored.
!                             i1=[(3 x (nmet - 1)) + 1], connections of the metabolite; i1=[(3 x (nmet - 1)) + 2],
!                             n of connections comming to the total n of metabolite (in); i1=[(3 x (nmet - 1)) + 3],
!                             n of connections going out of the metabolite (out)
!      supDifNodesInst_(i1) : stores the information indicating if a (possibly) new instance has different nodes by taking
!                             into account their degree (total and IN). If all the nodes of a given type (metabolites,
!                             irreversible reactions and reversible reactions) are all different, the corresponding cell
!                             is filled with "1"; if there is at least two nodes of the same type with the same degree
!                             (total and IN), the corresponding cell is filled with "0". (i1=1) => metabolites;
!                             (i1=2) => irreversible reactions; (i1=3) => reversible reactions
!      supDataInstance_(i1) : support vector to store provisionally the relevant information for the (possibly) newly found
!                             instance: (i1=1) number of interconnections; (i1=2) number of connections metabolite/reaction
!                             (reversibility not considered)
!      numNodes_(i1) : number of nodes of the (possibly) new struture: (i1=1) number of metabolites; (i1=2) number of reactions;
!                      (i1=3) number of reversible reactions; (i1=4) number of irreversible reactions
!      supReactRevInstTemp_(i1) : information regarding the reversibility of the reactions of the (possibly) new instance. If a
!                                 reaction is reversible, the corresponding cell is filled with 1; it it is irreversible, the
!                                 corresponding cell is filled with 0. The reactions may not be ordered according to its reversibility
!      supMet_(i1) : support vector with all the metabolites of the (possibly) new instance ordered according to their degree
!                    (total and IN)
!      supReact_(i1) : support vector with all the reactions of the (possibly) new instance ordered according to their degree
!                      (total and IN) and to their reversibility (irreversible reactions are stored first)
!      supOrdMet_(i1) : stores the order of the metabolites of the (possibly) new instance according to their degree:
!                       (i1=[(3 x (nmet - 1)) + 1 ]), reference number of the metabolite; (i1=[(3 x (nmet - 1)) + 2 ]),
!                       total number of connections; (i1=[(3 x (nmet - 1)) + 3 ]), number of connections IN
!      supOrdReact_(i1) : stores the order of the reactions of the (possibly) new instance according to their degree. The
!                         reactions are stored according to their reversibility: irreversible reactions first, and
!                         reersible reactions after. (i1=[(3 x (nreact - 1)) + 1 ]), reference number of the reaction;
!                         (i1=[(3 x (nreact - 1)) + 2 ]), total number of connections; (i1=[(3 x (nreact - 1)) + 3 ]),
!                         number of connections IN
!      supMetDegInstTemp_b(i1) :    the same as supMetDegInst_(i1)
!      supReactDegInstTemp_b(i1) : the same as supReactDegInst_(i1)
!      supReactRevInstTemp_b(i1) : the same as supReactRevInstTemp_(i1)
!      supConst_(i1) : support vector used to organize the (possibly) new instance according to the degree of its nodes
!
!
! MATRICES
!      a1_(i2,i1) : products x reactions data matrix, i1=products and i2=reactions
!      a2_(i2,i1) : reactions x reactants data matrix, i1=reactions and i2=reactants
!      maxInterRevStr_(i2,i1) : maximum number of interconnections for a structure with i1 metabolites and i2 reactions within
!                               the considered scope
!      maxMRConRevStr_(i2,i1) : maximum number of connections metabolite/reaction (reversibility not considered) for a structure
!                               with i1 metabolites and i2 reactions within the considered scope
!      supRevStInstTemp_(i3,i2,1) : support matrix to store temporarily a new structure. This structure will be reorganized
!                                   into matrix supRevStInst_(i3,i2,1)
!      supRevStInst_(i3,i2,1) : support matrix to store the pattern of the (possibly) new structure stored in matrix
!                               supRevStInstTemp_(i3,i2,1) ordered according to degree of the nodes (metabolites and reactions).
!                               This representation of the structure will be compared to others already found to determine
!                               if it is a isomorphic representation of one of them

      implicit none

      ! IN variables
       ! Integer: 1 bit
      integer*1, intent(in) :: a1_(:,:), a2_(:,:), maxInterRevStr_(:,:)
      integer*1, intent(in) :: maxMRConRevStr_(:,:), supReactRevInstTemp_(:), numNodes_(:)
       ! Integer: 2 bits
      integer*2, intent(in) :: supMetTemp_(:), supReactTemp_(:)

      ! OUT variables
       ! Integer: 1 bit
      integer*1, intent(inout) :: supRevStInst_(:,:,:), supMetDegInst_(:), InstScope_
      integer*1, intent(inout) :: supReactDegInst_(:), supDataInstance_(:)
      integer*1, intent(inout) :: supDifNodesInst_(:)
       ! Integer: 2 bits
      integer*2, intent(inout) :: supMet_(:), supReact_(:)

      !Internal support variables
      integer*1 :: supRevStInstTemp_(32,8,1), supMetDegInstTemp_(24), supMetDegInstTemp_b(24)
      integer*1 :: supReactDegInstTemp_(24), supReactDegInstTemp_b(24), supOrdMet_(24)
      integer*1 :: supOrdReact_(24), supReactRevInstTemp_b(8)
      integer*2 :: supConst_(2)

      ! General variables
      integer :: iOrg_1, iOrg_2, iOrg_3

!     Initializes the support variables
      ! Initializes the variable InstScope_. Initially, we will consider the instance is within the scope
      InstScope_ = 0
      ! Initializes the support vectors supMetDegInstTemp_(i1), supReactDegInstTemp_(i1) and supDataInstance_(i1)
      do iOrg_1=1,(3*numNodes_(1))
        supMetDegInstTemp_(iOrg_1) = 0
      end do
      do iOrg_1=1,(3*numNodes_(2))
        supReactDegInstTemp_(iOrg_1) = 0
      end do
      supDataInstance_(1) = 0
      supDataInstance_(2) = 0

!      Obtains the structure and the relevant data regarding the instance
      do iOrg_1=1,numNodes_(1)  ! Number of metabolites of this (possibly) new instance
        do iOrg_2=1,numNodes_(2)  ! Number of reactions of this (possibly) new instance
          !
          ! Metabolite supMetTemp_(iOrg_1) reactant of reaction supReactTemp_(iOrg_2)?
          supRevStInstTemp_( ( (4*(iOrg_2 - 1)) + 1 ), iOrg_1, 1 ) = a2_( supMetTemp_(iOrg_1), supReactTemp_(iOrg_2) )
          !
          ! Metabolite supMetTemp_(iOrg_1) product of reaction supReactTemp_(iOrg_2)?
          supRevStInstTemp_( ( (4*(iOrg_2 - 1)) + 2 ), iOrg_1, 1 ) = a1_( supReactTemp_(iOrg_2), supMetTemp_(iOrg_1) )
          !
          ! Counts the total degree of the metabolite supMetTemp_(iOrg_1)
          supMetDegInstTemp_( ((3*(iOrg_1 - 1)) + 1) ) = supMetDegInstTemp_( ((3*(iOrg_1 - 1)) + 1) ) +  &
          a1_( supReactTemp_(iOrg_2), supMetTemp_(iOrg_1) ) + a2_( supMetTemp_(iOrg_1), supReactTemp_(iOrg_2) )
          ! Counts the degree IN (product) of the metabolite supMetTemp_(iOrg_1)
          supMetDegInstTemp_( ((3*(iOrg_1 - 1)) + 2) ) = supMetDegInstTemp_( ((3*(iOrg_1 - 1)) + 2) ) + &
          a1_( supReactTemp_(iOrg_2), supMetTemp_(iOrg_1) )
          ! Counts the degree OUT (reactant) of the metabolite supMetTemp_(iOrg_1)
          supMetDegInstTemp_( ((3*(iOrg_1 - 1)) + 3) ) = supMetDegInstTemp_( ((3*(iOrg_1 - 1)) + 3) ) + &
          a2_( supMetTemp_(iOrg_1), supReactTemp_(iOrg_2) )
          !
          ! Counts the total degree of the reaction supReactTemp_(iOrg_2)
          supReactDegInstTemp_( ((3*(iOrg_2 - 1)) + 1) ) = supReactDegInstTemp_( ((3*(iOrg_2 - 1)) + 1) ) + &
          a1_( supReactTemp_(iOrg_2), supMetTemp_(iOrg_1) ) + a2_( supMetTemp_(iOrg_1), supReactTemp_(iOrg_2) )
          ! Counts the degree IN (1 reactant) of the reaction supReactTemp_(iOrg_2)
          supReactDegInstTemp_( ((3*(iOrg_2 - 1)) + 2) ) = supReactDegInstTemp_( ((3*(iOrg_2 - 1)) + 2) ) + &
          a2_( supMetTemp_(iOrg_1), supReactTemp_(iOrg_2) )
          ! Counts the degree OUT (1 product) of the reaction supReactTemp_(iOrg_2)
          supReactDegInstTemp_( ((3*(iOrg_2 - 1)) + 3) ) = supReactDegInstTemp_( ((3*(iOrg_2 - 1)) + 3) ) + &
          a1_( supReactTemp_(iOrg_2), supMetTemp_(iOrg_1) )
          !
          ! Is this a new connection metabolite/reaction?
          if( ( a1_( supReactTemp_(iOrg_2), supMetTemp_(iOrg_1) ) &
          +  a2_( supMetTemp_(iOrg_1), supReactTemp_(iOrg_2) ) )  .ne. 0 )  then
            ! Adds new connection metabolite/reaction to this (possibly) new instance
            supDataInstance_(2) = supDataInstance_(2) + 1
          end if
          !
          ! Adds the interconnections to this (possibly) new instance
          supDataInstance_(1) = supDataInstance_(1) + a1_( supReactTemp_(iOrg_2), supMetTemp_(iOrg_1) ) + &
          a2_( supMetTemp_(iOrg_1), supReactTemp_(iOrg_2) )
          !
          ! The reaction supReactTemp_(iOrg_2) is reversible?
          if( supReactRevInstTemp_(iOrg_2) .eq. 1 ) then
            !
            ! We dont need to store the part of the pattern corresponding to the inverse reaction
            !
            ! Counts the total degree of the metabolite supMetTemp_(iOrg_1)
            supMetDegInstTemp_( ((3*(iOrg_1 - 1)) + 1) ) = supMetDegInstTemp_( ((3*(iOrg_1 - 1)) + 1) ) +  &
            a1_( supReactTemp_(iOrg_2), supMetTemp_(iOrg_1) ) + a2_( supMetTemp_(iOrg_1), supReactTemp_(iOrg_2) )
            ! Counts the degree IN (product of the inverse reaction) of the metabolite supMetTemp_(iOrg_1)
            supMetDegInstTemp_( ((3*(iOrg_1 - 1)) + 2) ) = supMetDegInstTemp_( ((3*(iOrg_1 - 1)) + 2) ) + &
            a2_( supMetTemp_(iOrg_1), supReactTemp_(iOrg_2) )
            ! Counts the degree OUT (reactant of the inverse reaction) of the metabolite supMetTemp_(iOrg_1)
            supMetDegInstTemp_( ((3*(iOrg_1 - 1)) + 3) ) = supMetDegInstTemp_( ((3*(iOrg_1 - 1)) + 3) ) + &
            a1_( supReactTemp_(iOrg_2), supMetTemp_(iOrg_1) )
            !
            ! Counts the total degree of the reaction supReactTemp_(iOrg_2)
            supReactDegInstTemp_( ((3*(iOrg_2 - 1)) + 1) ) = supReactDegInstTemp_( ((3*(iOrg_2 - 1)) + 1) ) + &
            a1_( supReactTemp_(iOrg_2), supMetTemp_(iOrg_1) ) + a2_( supMetTemp_(iOrg_1), supReactTemp_(iOrg_2) )
            ! Counts the degree IN (1 reactant for the inverse reaction) of the reaction supReactTemp_(iOrg_2)
            supReactDegInstTemp_( ((3*(iOrg_2 - 1)) + 2) ) = supReactDegInstTemp_( ((3*(iOrg_2 - 1)) + 2) ) + &
            a1_( supReactTemp_(iOrg_2), supMetTemp_(iOrg_1) )
            ! Counts the degree OUT (1 product for the inverse reaction) of the reaction supReactTemp_(iOrg_2)
            supReactDegInstTemp_( ((3*(iOrg_2 - 1)) + 3) ) = supReactDegInstTemp_( ((3*(iOrg_2 - 1)) + 3) ) + &
            a2_( supMetTemp_(iOrg_1), supReactTemp_(iOrg_2) )
            !
            ! Adds the interconnections to this (possibly) new instance
            supDataInstance_(1) = supDataInstance_(1) + a1_( supReactTemp_(iOrg_2), supMetTemp_(iOrg_1) ) + &
            a2_( supMetTemp_(iOrg_1), supReactTemp_(iOrg_2) )
            !
          else
            supRevStInstTemp_( ( (4*(iOrg_2 - 1)) + 3 ), iOrg_1, 1 ) = 0
            supRevStInstTemp_( ( (4*(iOrg_2 - 1)) + 4 ), iOrg_1, 1 ) = 0
          end if
          !
        end do
      end do

!      Tests if the instance is within the considered scope
      ! The number of interconnections is higher than the maximum value allowed for the considered scope?
      if( supDataInstance_(1) .gt. maxInterRevStr_(numNodes_(2),numNodes_(1)) ) then
        InstScope_ = 1  ! Instance is outside the scope
        goto 2510
      end if
      ! The number of connections metabolite/reaction is higher than the maximum value allowed for the
      ! considered scope?
      if( supDataInstance_(2) .gt. maxMRConRevStr_(numNodes_(2),numNodes_(1)) ) then
        InstScope_ = 1  ! Instance is outside the scope
        goto 2510
      end if

!      Organizes the (possibly) new structure according to the degree of the nodes
      ! Creates the support vectors supMetDegInstTemp_b(i1), supReactDegInstTemp_b(i1) and supReactRevInstTemp_b(i1)
      do iOrg_1=1,(3*numNodes_(1))  ! 3 x number of metabolites of the (possibly) new structure
        supMetDegInstTemp_b(iOrg_1) = supMetDegInstTemp_(iOrg_1)
      end do
      do iOrg_1=1,(3*numNodes_(2))  ! 3 x number of reactions of the (possibly) new structure
        supReactDegInstTemp_b(iOrg_1) = supReactDegInstTemp_(iOrg_1)
      end do
      do iOrg_1=1,numNodes_(2)  ! Number of reactions
        supReactRevInstTemp_b(iOrg_1) = supReactRevInstTemp_(iOrg_1)
      end do
      !
      ! Obtains the ordered list for the metabolites of the (possibly) new structure according to their degree
      do iOrg_1=1,numNodes_(1)  ! Number of metabolites of the (possibly) new structure
        supOrdMet_( (3*(iOrg_1 - 1)) + 1 ) = 0   ! "Reference" number of the metabolite at the ordered position iOrg_1
        supOrdMet_( (3*(iOrg_1 - 1)) + 2 ) = 0   ! Number of connections of the metabolite at the ordered position iOrg_1
        supOrdMet_( (3*(iOrg_1 - 1)) + 3 ) = 0   ! Number of connections IN of the metabolite at the ordered position iOrg_1
        supConst_(1) = 0  ! Support constant: stores the "reference" number of the metabolite for the new ordered position iOrg_1
        do iOrg_2=1,numNodes_(1)  ! Number of metabolites of the (possibly) new structure
          !
          ! Checks the metabolite with higher degree (total and IN)
          !     tests: i) number of connections;
          if( supMetDegInstTemp_b( (3*(iOrg_2 - 1)) + 1 ) .ge. supOrdMet_( (3*(iOrg_1 - 1)) + 2 ) )  then !.and. &
            ! Number of connections are the same?
            if( supMetDegInstTemp_b( (3*(iOrg_2 - 1)) + 1 ) .eq. supOrdMet_( (3*(iOrg_1 - 1)) + 2 ) ) then
              ! If the number of connections are the same, tests: ii) number of connections IN
              if( supMetDegInstTemp_b( (3*(iOrg_2 - 1)) + 2 ) .le. supOrdMet_( (3*(iOrg_1 - 1)) + 3 ) ) goto 2500
            end if
            !
            ! Support constant: stores the "reference" number of the metabolite for the new ordered position iOrg_1
            supConst_(1) = iOrg_2
            !
            ! Previous "reference" number of the metabolite to introduce in the ordered position iOrg_1
            supOrdMet_( (3*(iOrg_1 - 1)) + 1 ) = iOrg_2
            ! Total number of connections
            supOrdMet_( (3*(iOrg_1 - 1)) + 2 ) = supMetDegInstTemp_b( (3*(iOrg_2 - 1)) + 1 )
            ! Number of connections IN
            supOrdMet_( (3*(iOrg_1 - 1)) + 3 ) = supMetDegInstTemp_b( (3*(iOrg_2 - 1)) + 2 )
            !
          end if
          goto 2500
2500      continue
          !
          ! iOrg_2 equal to numNodes_(1) (number of metabolites)?
          if( iOrg_2 .eq. numNodes_(1) ) then
            ! Turns to zero the cells corresponding to the metabolite supConst_(1) to allow the next metabolite with
            ! higher degree to be found
            supMetDegInstTemp_b( (3*(supConst_(1) - 1)) + 1 ) = 0
            supMetDegInstTemp_b( (3*(supConst_(1) - 1)) + 2 ) = 0
            supMetDegInstTemp_b( (3*(supConst_(1) - 1)) + 3 ) = 0
          end if
        end do
      end do
      !
      ! Obtains the ordered list of the reations of the (possibly) new structure according to their reversibility and
      ! degree
      ! Creates the vector supOrdReact_(i1)
      do iOrg_1=1,(3*numNodes_(2))  ! 3 x number of reactions of the (possibly) new structure
        supOrdReact_(iOrg_1) = 0
      end do
      ! Does the loop till the maximum between the number of irreversible reactions (numNodes_(4)) and
      ! the number of reversible reactions (numNodes_(3))
      do iOrg_1=1,max(numNodes_(4),numNodes_(3))
        supConst_(1) = 0
        supConst_(2) = 0
        do iOrg_2=1,numNodes_(2)  ! Number of reactions of the (possibly) new structure
          !
          ! Checks the reversibility of the reaction
          !
          ! Reaction is irreversible
          if( supReactRevInstTemp_b(iOrg_2) .eq. 0 ) then
            !
            ! Checks the reaction with higher degree (total and IN)
            !       tests: i) number of connections;
            if( supReactDegInstTemp_b( (3*(iOrg_2 - 1)) + 1 ) .ge. supOrdReact_( (3*(iOrg_1 - 1)) + 2 ) ) then
              ! The number of connections is the same?
              if( supReactDegInstTemp_b( (3*(iOrg_2 - 1)) + 1 ) .eq. supOrdReact_( (3*(iOrg_1 - 1)) + 2 ) ) then
                ! If the number of connections are the same, tests: ii) number of connections IN
                if( supReactDegInstTemp_b( (3*(iOrg_2 - 1)) + 2 ) .le. supOrdReact_( (3*(iOrg_1 - 1)) + 3 ) ) goto 2505
              end if
              !
              ! Support constant: stores the "reference" number of the reaction for the new ordered position iOrg_1
              supConst_(1) = iOrg_2
              !
              ! Previous "reference" number of the reaction to introduce in the ordered position iOrg_1
              supOrdReact_( (3*(iOrg_1 - 1)) + 1 ) = iOrg_2
              ! Total number of connections
              supOrdReact_( (3*(iOrg_1 - 1)) + 2 ) = supReactDegInstTemp_b( (3*(iOrg_2 - 1)) + 1 )
              ! Number of connections IN
              supOrdReact_( (3*(iOrg_1 - 1)) + 3 ) = supReactDegInstTemp_b( (3*(iOrg_2 - 1)) + 2 )
              !
              goto 2505
              !
            end if
          end if
          !
          ! Reaction is reversible
          if( supReactRevInstTemp_b(iOrg_2) .eq. 1 ) then
            !
            ! Checks the reaction with higher degree (total and IN)
            !       tests: i) number of connections;
            if( supReactDegInstTemp_b( (3*(iOrg_2 - 1)) + 1 ) .ge. supOrdReact_( (3*(numNodes_(4) + iOrg_1 - 1)) + 2 ) ) then
              ! The number of connections is the same?
              if( supReactDegInstTemp_b( (3*(iOrg_2 - 1)) + 1 ) .eq. supOrdReact_( (3*(numNodes_(4) + iOrg_1 - 1)) + 2 ) ) then
                ! If the number of connections are the same, tests: ii) number of connections IN
                if( supReactDegInstTemp_b( (3*(iOrg_2 - 1)) + 2 ) .le. &
                supOrdReact_( (3*(numNodes_(4) + iOrg_1 - 1)) + 3 ) ) goto 2505
              end if
              !
              ! Support constant: stores the "reference" number of the reaction for the new ordered position (numNodes_(4) + iOrg_1)
              supConst_(2) = iOrg_2
              !
              ! Previous "reference" number of the reaction to introduce in the ordered position (numNodes_(4) + iOrg_1)
              supOrdReact_( (3*(numNodes_(4) + iOrg_1 - 1)) + 1 ) = iOrg_2
              ! Total number of connections
              supOrdReact_( (3*(numNodes_(4) + iOrg_1 - 1)) + 2 ) = supReactDegInstTemp_b( (3*(iOrg_2 - 1)) + 1 )
              ! Number of connections IN
              supOrdReact_( (3*(numNodes_(4) + iOrg_1 - 1)) + 3 ) = supReactDegInstTemp_b( (3*(iOrg_2 - 1)) + 2 )
              !
            end if
          end if
          !
          goto 2505
2505      continue
          ! iOrg_2 equal to numNodes_(2) (number of reactions)?
          if( iOrg_2 .eq. numNodes_(2) ) then
            !
            ! Irreversible reactions
            if( (numNodes_(4) .ne. 0) .and. (iOrg_1 .le. numNodes_(4)) ) then
              ! Turns to zero the cells corresponding to the irreversible reaction supConst_(1) to allow the next irreversible
              ! reaction with higher degree to be found
              supReactDegInstTemp_b( (3*(supConst_(1) - 1)) + 1 ) = 0
              supReactDegInstTemp_b( (3*(supConst_(1) - 1)) + 2 ) = 0
              supReactDegInstTemp_b( (3*(supConst_(1) - 1)) + 3 ) = 0
              supReactRevInstTemp_b(supConst_(1)) = 2
            end if
            !
            ! Reversible reactions
            if( (numNodes_(3) .ne. 0) .and. (iOrg_1 .le. numNodes_(3)) ) then
              ! Turns to zero the cells corresponding to the reversible reaction supConst_(2) to allow the next reversible
              ! reaction with higher degree to be found
              supReactDegInstTemp_b( (3*(supConst_(2) - 1)) + 1 ) = 0
              supReactDegInstTemp_b( (3*(supConst_(2) - 1)) + 2 ) = 0
              supReactDegInstTemp_b( (3*(supConst_(2) - 1)) + 3 ) = 0
              supReactRevInstTemp_b(supConst_(2)) = 2
            end if
          end if
        end do
      end do
      !
      !
      ! Organizes the pattern of the (possibly) new structure according to the new order of the metabolites and of
      ! the reactions
      do iOrg_1=1,numNodes_(1)  ! Number of metabolites of the (possibly) new structure
        do iOrg_2=1,numNodes_(2)  ! Number of reactions of the (possibly) new structure
          do iOrg_3=1,4  ! For each reaction, 2 unidirectional reactions with 2 cells for the reactant and product
            supRevStInst_( ((4*(iOrg_2 - 1)) + iOrg_3 ), iOrg_1, 1) = &
            supRevStInstTemp_(  ((4*( supOrdReact_((3*(iOrg_2 - 1)) + 1) - 1 )) + iOrg_3 ), supOrdMet_( (3*(iOrg_1 - 1)) + 1 ), 1)
          end do
        end do
      end do
      !
      ! Organizes the information regarding metabolites ant their degrees according to the new order
      do iOrg_1=1,numNodes_(1)  ! Number of metabolites of the (possibly) new structure
        supMet_(iOrg_1) = supMetTemp_( supOrdMet_((3*(iOrg_1 - 1)) + 1) )
        do iOrg_2=1,3
          supMetDegInst_( (3*(iOrg_1 - 1)) + iOrg_2 ) = &
          supMetDegInstTemp_( (3*( supOrdMet_((3*(iOrg_1 - 1)) + 1) - 1 )) + iOrg_2 )
        end do
      end do
      !
      ! Organizes the information regarding the degrees of the reactions according to the new order
      do iOrg_1=1,numNodes_(2)  ! Number of reactions of the (possibly) new structure
        supReact_(iOrg_1) = supReactTemp_( supOrdReact_( (3*(iOrg_1 - 1)) + 1) )
        do iOrg_2=1,3
          supReactDegInst_( (3*(iOrg_1 - 1)) + iOrg_2 ) = &
          supReactDegInstTemp_( (3*( supOrdReact_( (3*(iOrg_1 - 1)) + 1) - 1 )) + iOrg_2 )
        end do
      end do
      !
      ! Determines if the metabolites of this (possibly) new instance have all different degrees (total and IN)
      do iOrg_1=1,(numNodes_(1)-1)  ! Performs the loop from 1 to the number of metabolites - 1
        do iOrg_2=(iOrg_1+1),numNodes_(1)  ! Performs the loop from 1 to the number of metabolites
          !
          ! The metabolites at the positions iOrg_1 and iOrg_2 have the same degrees (total and IN)?
          if( ( supMetDegInst_( (3*(iOrg_1 -1)) + 1 ) .eq. &
          supMetDegInst_( (3*(iOrg_2 -1)) + 1 ) ) .and. &
          ( supMetDegInst_( (3*(iOrg_1 -1)) + 2 ) .eq. &
          supMetDegInst_( (3*(iOrg_2 -1)) + 2 ) ) ) then
            ! The metabolites have the same degree (total and IN)
            supDifNodesInst_(1) = 0
            goto 2515
          end if
          !
          ! All possible pairs of metabolites have been compared?
          if( (iOrg_1 .eq. (numNodes_(1)-1)) .and. (iOrg_2 .eq. numNodes_(1)) ) then
            ! All the metabolites have different degrees
            supDifNodesInst_(1) = 1
            goto 2515
          end if
        end do
      end do
2515  continue
      !
      ! Determines if the irreversible reactions of this (possibly) new instance have all different degrees (total and IN)
      ! Instance has >1 irreversible reactions?
      if( numNodes_(4) .gt. 1 ) then
        ! Instance has >1 irreversible reaction
        do iOrg_1=1,(numNodes_(4)-1)  ! Performs the loop from 1 to the number of irreversible reactions - 1
          do iOrg_2=(iOrg_1+1),numNodes_(4)  ! Performs the loop from iOrg_1+1 to the number of irreversible reactions
            !
            ! The irreversible reactions at the positions iOrg_1 and iOrg_2 have the same degrees (total and IN)?
            if( ( supReactDegInst_( (3*(iOrg_1 -1)) + 1 ) .eq. &
            supReactDegInst_( (3*(iOrg_2 -1)) + 1 ) ) .and. &
            ( supReactDegInst_( (3*(iOrg_1 -1)) + 2 ) .eq. &
            supReactDegInst_( (3*(iOrg_2 -1)) + 2 ) ) ) then
              ! The irreversible reactions have the same degree (total and IN)
              supDifNodesInst_(2) = 0
              goto 2520
            end if
            !
            ! All possible pairs of irreversible reactions have been compared?
            if( (iOrg_1 .eq. (numNodes_(4)-1)) .and. (iOrg_2 .eq. numNodes_(4)) ) then
              ! All the metabolites have different degrees
              supDifNodesInst_(2) = 1
              goto 2520
            end if
          end do
        end do
      else
        ! Instance has <2 irreversible reactions
        supDifNodesInst_(2) = 0
        goto 2520
      end if
2520  continue
      !
      ! Determines if the reversible reactions of this (possibly) new instance have all different degrees (total and IN)
      ! Instance has >1 reversible reactions?
      if( numNodes_(3) .gt. 1 ) then
        ! Instance has >1 reversible reaction
        do iOrg_1=1,(numNodes_(3)-1)  ! Performs the loop from 1 to the number of reversible reactions - 1
          do iOrg_2=(iOrg_1+1),numNodes_(3)  ! Performs the loop from iOrg_1+1 to the number of reversible reactions
            !
            ! The reversible reactions at the positions iOrg_1 and iOrg_2 have the same degrees (total and IN)?
            if( ( supReactDegInst_( (3*(numNodes_(4)+iOrg_1 -1)) + 1 ) .eq. &
            supReactDegInst_( (3*(numNodes_(4)+iOrg_2 -1)) + 1 ) ) .and. &
            ( supReactDegInst_( (3*(numNodes_(4)+iOrg_1 -1)) + 2 ) .eq. &
            supReactDegInst_( (3*(numNodes_(4)+iOrg_2 -1)) + 2 ) ) ) then
              ! The reversible reactions have the same degree (total and IN)
              supDifNodesInst_(3) = 0
              goto 2510
            end if
            !
            ! All possible pairs of irreversible reactions have been compared?
            if( (iOrg_1 .eq. (numNodes_(3)-1)) .and. (iOrg_2 .eq. numNodes_(3)) ) then
              ! All the metabolites have different degrees
              supDifNodesInst_(3) = 1
              goto 2510
            end if
          end do
        end do
      else
        ! Instance has <2 reversible reactions
        supDifNodesInst_(3) = 0
        goto 2510
      end if

2510  continue

      end subroutine organizingInst


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!     Subroutine to determine all the possible permutations of a increasing ordered
!!     set of numbers. The set is obtained from the number of the metabolites or
!!     reactions. Application of an algorithm to generate permutations in lexicographic
!!     order.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine allPermutations(inNumber_, Permut_, numPermut_)


! LOGICAL VARIABLES
!      repeat_ : loop continues till repeat is FALSE
!
!
! CONSTANTS
!      numPop_ : number of metabolites or reactions, needed to create the set of ordered
!                numbers. It is equal to inNumber_
!      inNumber_ : number of metabolites or reactions. It is the IN argument of the sub-routine
!
!
! VECTORS
!      supPer_(numPop_) : support vector super_(i1) needed to create the new permutation.
!                         i1) each cell corresponds to a number of the new permutation
!
!
! MATRICES
!      Permut_(numPop_,numPermut_) : matrix Permut_(i2,i1) with the permutations of a set of
!                                    numPop_ numbers. i1) number of the permutation; i2) column
!                                    for each numPop_ number.

      implicit none

      ! IN variables
       ! Integer: 2 bits
      integer*2, intent(in) :: inNumber_

      ! OUT variables
       ! Integer: 1 bit
      integer*1, intent(inout) :: Permut_(:,:)
       ! Integer: 4 bits
      integer, intent(inout) :: numPermut_

      !Internal support variables
       ! Integer: 2 bits
      integer*2 :: supPer_(8), support_(3), numPop_
       ! Logical
      logical :: repeat_

      ! General variables
      integer*2 :: i_1  ! 4 bits

!      Creating the permutations for a set of increasing ordered numbers. Application of an algorithm
!      to generate permutations in lexicographic order
      ! Generate the set of numbers from the total number of metabolites or reactions
      numPop_ = inNumber_
      do i_1=1,numPop_
        Permut_(i_1,1) = i_1
      end do
      ! Initializes the counting of the number of permutations
      numPermut_ = 1
      ! Logical variable
      repeat_ = .true.
      ! Finding new permutations occurs till "repeat_" is FALSE
      do while (repeat_)
        support_(1) = 0
        ! Determines the highest index k such that Permt_[k] < Permut[k + 1], and k + 1 =< numPop_
        do i_1=1,(numPop_-1)
          if( (Permut_(i_1,numPermut_) .lt. Permut_((i_1+1),numPermut_)) .and. ( i_1 .gt. support_(1) ) ) then
            support_(1) = i_1
          end if
        end do
        ! An index k was found such that Permt_[k] < Permut[k + 1] (support_(1) is not equal to 0)
        if( support_(1) .ne. 0 ) then
          support_(2) = 0
          ! Creates the support vector supPer_(i_1) to help obtaining the new permutation
          do i_1=1,numPop_
            ! The support vector stores the previous permutation
            supPer_(i_1) = Permut_(i_1,(numPermut_))
          end do
          ! Find the largest index l greater than k such that Permut_[k] < Permut_[l]
          do i_1=(support_(1)+1),numPop_
            if( (Permut_(i_1,numPermut_) .gt. Permut_(support_(1),numPermut_)) .and. (i_1 .gt. support_(2)) ) then
              support_(2) = i_1
            end if
          end do
          ! Swap the cells in position k and l: k=support_(1) and l=support_(2)
          supPer_(support_(1)) = Permut_(support_(2),(numPermut_))
          supPer_(support_(2)) = Permut_(support_(1),(numPermut_))
          ! Adds the new permutation to the total number of permutations
          numPermut_ = numPermut_ + 1
          ! Stores the new permutation in the matrix Permut_(j,i). Some changes still need to be done to this
          do i_1=1,support_(1)  ! new permutation
            Permut_(i_1,numPermut_) = supPer_(i_1)
          end do
          support_(3) = 0
          ! Reverse the sequence from Permut_[k + 1] up to and including the final element Permut_[numPop_]
          do i_1=(support_(1)+1),numPop_
            Permut_(i_1,numPermut_) = supPer_(numPop_-support_(3))
            support_(3) = support_(3) + 1
          end do
          ! Turns to zeros the support matrix supPer_(i_1) to be used in the next iteration
          do i_1=1,numPop_
            supPer_(i_1) = 0
          end do
        ! If no index was found were Permt_[k] < Permut[k + 1], there are no more permutations
        else
          ! Logical variable repeat_ is turn to FALSE to stop the loop
          repeat_ = .false.
        end if
      end do

      end subroutine allPermutations


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!     Subroutine to determine all the possible permutations considering the
!!     two single directions reactions of a reversible reaction. So, the possible
!!     permutations between the direct and the inverse reaction of a reversible
!!     reation are obtained.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine allRevReactPerm(numberIn_, RevPerm_, numRevPerm_)

! CONSTANTS
!      numberIn_ : number of reactions
!      nReact_ : number of reactions (equal to numberIn_)
!
!
! VECTORS
!      rdir(i1) : support vector that stores the ordenation of the direct reaction
!                 for all the reactions: 0 if it stays in its position, 2 if it
!                 moves to the position of the inverse reaction
!      rinv(i1) : support vector that stores the ordenation of the inverse reaction
!                 for all the i1 reactions: 2 if it stays in its position, 0 if it
!                 moves to the position of the direct reaction
!      suppor_(i1) : support vector used to count the consecutive number of times the
!                    positions for the direct and inverse reactions of a i1 reaction
!                    are keep the same in the permutations. After reaching a determined
!                    number, the positions of the direct and inverse reaction of a i1
!                    reaction are exchanged and count is turn to zero again
!
!
! MATRICES
!      RevPerm_(i2,i1) : possible i1 permutations between the direct and inverse single direction reactions for
!                        a set of numberIn_ reversible reactions. i2) ordering the direct and inverse single
!                        direction reactions for all the reactions

      implicit none

      ! IN variables
       ! Integer: 2 bits
      integer*2, intent(in) :: numberIn_

      ! OUT variables
       ! Integer: 1 bit
      integer*1, intent(inout) :: RevPerm_(:,:)
       ! Integer: 4 bits
      integer, intent(inout) :: numRevPerm_

      !Internal support variables
      integer*1 :: rdir_(8), rinv_(8)
      integer*2 :: nReact_, suppor_(8)  ! 2 bits

      ! General variables
      integer :: i1_, i2_  ! 4 bits


!      Obtains the permutations between the direct and the reverse reaction for all
!      the reactions
      ! number of reactions equal to numberIn_
      nReact_ = numberIn_
      ! Creates the support vectors rdir_(i1_), rinv_(i1_) and suppor(i1_)
      do i1_=1,nReact_
        rdir_(i1_) = 0  ! The direct reactions are initially at there natural position
        rinv_(i1_) = 2  ! The inverse reactions are initially at there natural position
        ! The support vector suppor_(i1_), used to count the number of consecutive
        suppor_(i1_) = 0  ! permutations without changing the position of the direct
      end do              ! and inverse reaction of a i1_ reaction
      !
      ! Number of permutations equal to 2^(nreact_)
      numRevPerm_ = (2**(nReact_))
      !
      do i1_=1,numRevPerm_
        do i2_=1,nreact_
          suppor_(i2_) = suppor_(i2_) + 1
          RevPerm_((2*(i2_ - 1) + 1),i1_) = rdir_(i2_)
          RevPerm_((2*(i2_ - 1) + 2),i1_) = rinv_(i2_)
          ! Test if the pair 0/2 or 2/0 for reaction i2_ was kept constant for the last
          ! 2^(total number of reactions - i2_) permutations
          if( ( i1_ .ne. numRevPerm_ ) .and. ( suppor_(i2_) .eq. (2**(nReact_ - i2_)) ) ) then
            ! If it was, change the order of the pair 0/2
            suppor_(i2_) = 0
            rdir_(i2_) = RevPerm_((2*(i2_ - 1) + 2),i1_)
            rinv_(i2_) = RevPerm_((2*(i2_ - 1) + 1),i1_)
          end if
        end do
      end do

      end subroutine allRevReactPerm


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!     Subroutine to determine the structure of a instance
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine isomorphismTest(supRevStIns_, RevStruct_, &
      DataRevStru_, MetDegRevStru_, ReactDegRevStru_, numNod_, supDataIns_, &
      supMetDegIns_, supReactDegIns_, PermutNode_, numPermutNode_, &
      PermutDirInvReact_, numPermDirInvReact_, nrefStruLow_, nrefStruHigh_, &
      refNumPermMet_, refNumPermIrrReact_, refNumPermRevReact_, &
      refNumPermDirInvReact_, isoStruRef_, DifNodesSt_, supDifNodesIn_)

! CONSTANTS
!      isoStruRef_ : reference number of the structure in matrix RevStruct_(i3,i2,i1) to which the instance in matrix
!                    supRevStIns_(i2,i1,1) is a concretization.
!      nrefStruLow_ : lower reference number of a structure of the considered generation stored in matrix RevStruct_(i3,i2,i1)
!      nrefStruHigh_ : higher reference number of a structure of the considered generation stored in matrix RevStruct_(i3,i2,i1)
!      refNumPermMet_ : reference number of the permutation for the metabolites that allowed to the determine the structure
!                       in matrix RevStruct(i3,i2,i1) of which the instance in matrix supRevStInst(i3,i2,1) is a realization,
!                       i.e., PermutNode( i3 , refNumPermMet , numNode(1) )
!      refNumPermIrrReact_ : reference number of the permutation for the irreversible reactions that allowed to the determine the
!                            structure in matrix RevStruct(i3,i2,i1) of which the instance in matrix supRevStInst(i3,i2,1) is a
!                            realization, i.e., PermutNode( i3 , refNumPermIrrReact , numNode(4) )
!      refNumPermRevReact_ : reference number of the permutation for the reversible reactions that allowed to the determine the
!                            structure in matrix RevStruct(i3,i2,i1) of which the instance in matrix supRevStInst(i3,i2,1) is a
!                            realization, i.e., PermutNode( i3 , refNumPermRevReact , numNode(4) )
!      refNumPermDirInvReact_ : reference number of the permutation for the direct and inverse reactions of a reversible
!                               reaction that allowed to the determine the structure in matrix RevStruct(i3,i2,i1) of
!                               which the instance in matrix supRevStInst(i3,i2,1) is a realization, i.e.,
!                               PermutDirInvReact( i3, refNumPermDirInvReact , numNode(3) )
!
!
! VECTORS
!      numNod_(i1) : number of nodes of the (possibly) new struture: (i1=1) number of metabolites; (i1=2) number of reactions;
!                    (i1=3) number of reversible reactions; (i1=4) number of irreversible reactions
!      supDataIns_(i1) : support vector to store provisionally the relevant information for the (possibly) newly found
!                        instance: (i1=1) number of interconnections; (i1=2) number of connections metabolite/reaction
!                        (reversibility not considered)
!      supMetDegIns_(i1) : gives the degree for each metabolite of a (possibly) new instance. i1=[(3 x (nmet - 1)) + 1],
!                              total n of connections of the metabolite; i1=[(3 x (nmet - 1)) + 2], n of connections
!                              comming to the metabolite (IN); i1=[(3 x (nmet - 1)) + 3], n of connections going out of
!                              the metabolite (OUT)
!      supReactDegIns_(i1) : gives the degree for each reaction of a possibly new instance. The irreversible reactions
!                               are stored first, and then, in case at least one exists, the reversible reactions are stored.
!                               i1=[(3 x (nmet - 1)) + 1], connections of the metabolite; i1=[(3 x (nmet - 1)) + 2],
!                               n of connections comming to the total n of metabolite (IN); i1=[(3 x (nmet - 1)) + 3],
!                               n of connections going out of the metabolite (OUT)
!      supDifNodesIn_(i1) : stores the information indicating if a (possibly) new instance has different nodes by taking
!                           into account their degree (total and IN). If all the nodes of a given type (metabolites,
!                           irreversible reactions and reversible reactions) are all different, the corresponding cell
!                           is filled with "1"; if there is at least two nodes of the same type with the same degree
!                           (total and IN), the corresponding cell is filled with "0". (i1=1) => metabolites;
!                           (i1=2) => irreversible reactions; (i1=3) => reversible reactions
!      numPermutNode_(i1) : number of permutations for the given set of nodes i1
!      numPermDirInvReact_(i1) : number of permutations between the direct and inverse reaction for a set of reversible
!                                reactions i1
!      supNumPermutNode_(i1) : support variable with the number of permutations for the three types of nodes considered (metabolites,
!                              irreversible reactions and reversible reactions) taking into account the possibility of all the nodes
!                              of a given type having different degrees (total and IN) - in this cases, the number of permutations
!                              is 1. (i1=1) => metabolites; (i1=2) => irreversible reactions; (i1=3) => reversible reactions
!
!
! MATRICES
!      supRevStIns_(i3,i2,1) : support matrix to store a (possibly) new instance. The nodes are organized according to their
!                              degree (total and IN) The reactions are previously separated according to their reversibility:
!                              irreversible reactions appear first, followed by the reversible reactions. The pattern of this
!                              instance will be compared to the ones in matrix RevStruct(i3,i2,i1) to find what is its structure
!      RevStruct_(i3,i2,i1) : matrix with the possible structures using a bipartite and directed
!                             graph representation. i1) number of the structure; i2) number of the metabolite;
!                             i3) [4*(i3-1) + 1] reactant of the direct reaction, [4*(i3-1) + 2] product of
!                             the direct reaction, [4*(i3-1) + 3] reactant of the inverse reaction, [4*(i3-1) + 4]
!                             product of the inverse reaction. If a metabolite is connected to a reaction, the
!                             corresponding cell is filled with 1; otherwise, is filled with 0
!      DataRevStru_(i2,i1) : i1) reference number of the structure. i2) aditional data for each structure: (i2=1) number
!                            of generation; (i2=2) number of metabolites; (i2=3) number of reactions; (i2=4) number of
!                            reversible reactions; (i2=5) number of interconnections; (i2=6) number of connections
!                            metabolite/reaction (reversibility not considered)
!      MetDegRevStru_(i2,i1) : gives the degree for each metabolite of structure i1: i2=[(3 x (nmet - 1)) + 1], total n of
!                              connections of the metabolite; i2=[(3 x (nmet - 1)) + 2], n of connections comming to the
!                              metabolite (IN); i2=[(3 x (nmet - 1)) + 3], n of connections going out of the metabolite (OUT)
!      ReactDegRevStru_(i2,i1) : gives the degree for each reaction of structure i1. The irreversible reactions are stored
!                                first and then, in case at least one exists, the reversible reactions are stored.
!                                i2=[(3 x (nmet - 1)) + 1], connections of the metabolite; i2=[(3 x (nmet - 1)) + 2],
!                                n of connections comming to the total n ofmetabolite (IN); i2=[(3 x (nmet - 1)) + 3],
!                                n of connections going out of the metabolite (OUT)
!      DifNodesSt_(i2,i1) : stores the information indicating if structure i1 has different nodes by taking into account their
!                           degree (total and IN). If all the nodes of a given type (metabolites, irreversible reactions and
!                           reversible reactions) are all different, the corresponding cell is filled with "1"; if there is
!                           at least two nodes of the same type with the same degree (total and IN), the corresponding cell
!                           is filled with "0". (i2=1) => metabolites; (i2=2) => irreversible reactions; (i2=3) => reversible
!                           reactions
!      PermutNode_(i3,i2,i1) : possible permutations for a given number i1 of nodes (metabolites or reactions), obtained
!                              initially by calling the subroutine allPermutations. i2) number of the permutation;
!                              i3) number of the nodes ordered by this permutation
!      PermutDirInvReact_(i3,i2,i1) : possible permutations between the direct and inverse single direction reactions for
!                                    a set of i1 reversible reactions. The matrix was obtained by calling the subroutine
!                                    allRevReactPerm. i2) number of the permutation; i3) ordering the direct and inverse single
!                                    direction reactions for all the reactions

      implicit none

      ! IN variables
       ! Integer: 1 bit
      integer*1, intent(in) :: supRevStIns_(:,:,:), RevStruct_(:,:,:)
      integer*1, intent(in) :: PermutNode_(:,:,:), PermutDirInvReact_(:,:,:)
      integer*1, intent(in) :: MetDegRevStru_(:,:), ReactDegRevStru_(:,:)
      integer*1, intent(in) :: DifNodesSt_(:,:), DataRevStru_(:,:)
      integer*1, intent(in) :: supDifNodesIn_(:), supMetDegIns_(:)
      integer*1, intent(in) :: supReactDegIns_(:), numNod_(:), supDataIns_(:)
       ! Integer: 2 bits_
      integer*2, intent(in) :: numPermDirInvReact_(:)
      integer*2, intent(in) :: nrefStruLow_, nrefStruHigh_
       ! Integer: 4 bits
      integer, intent(in) :: numPermutNode_(:)

      ! OUT variables
       ! Integer: 2 bits
      integer*2, intent(inout) :: isoStruRef_, refNumPermDirInvReact_
       ! Integer: 4 bits
      integer, intent(inout) :: refNumPermMet_, refNumPermIrrReact_, refNumPermRevReact_

      ! Internal support variables
       ! Integer: 4 bits
      integer :: supNumPermutNode_(3)

      ! General variables (4 bits)
      integer*2 :: iso1_, iso4_
      integer :: iso2_, iso3_, iso5_, com1_, com2_, com3_, com4_
      integer :: Nmet1_, Nmet2_, Nreact1_, Nreact2_, Nreact3_

 !     Initializes the variables iso3_ and iso5_
      iso3_ = 0
      iso5_ = 0

!      Creates the support variable supNumPermutNode_(
      ! Number of permutations for the metabolites
      if( supDifNodesIn_(1) .eq. 0 ) then
        ! The metabolites do not have all different degrees (total and IN)
        supNumPermutNode_(1) = numPermutNode_(numNod_(1))
      else
        ! The metabolites have all different degrees
        supNumPermutNode_(1) = 1
      end if
      !
      ! Number of permutations of the irreversible reactions
      if( supDifNodesIn_(2) .eq. 0 ) then
        ! The irreversible reactions do not have all different degrees (total and IN)
        supNumPermutNode_(2) = numPermutNode_(numNod_(4))
      else
        ! The irreversible reactions have all different degrees
        supNumPermutNode_(2) = 1
      end if
      !
      ! Number of permutations of the reversible reactions
      if( supDifNodesIn_(3) .eq. 0 ) then
        ! The reversible reactions do not have all different degrees (total and IN)
        supNumPermutNode_(3) = numPermutNode_(numNod_(3))
      else
        ! The reversible reactions have all different degrees
        supNumPermutNode_(3) = 1
      end if

!      Determine the structure iso1_ in matrix RevStruct_(i3,i2,i1) of the (possibly) new instance represented in
!      matrix supRevStIns_(i2,i1,1)
      ! Reference number of a structure in matrix RevStruct_(i3,i2,i1)
      do 3000 iso1_=nrefStruLow_,max(nrefStruLow_,nrefStruHigh_)
        !
        ! Checks if structure iso1_ and the (possibly) new instance both have the same values for the variables
        ! that store information indicating if the patterns have different nodes by taking into account their
        ! degree (total and IN)
        ! Metabolites
        if( supDifNodesIn_(1) .ne.  DifNodesSt_(1,iso1_) ) goto 3000
        ! Irreversible reactions
        if( supDifNodesIn_(2) .ne.  DifNodesSt_(2,iso1_) ) goto 3000
        ! Reversible reactions
        if( supDifNodesIn_(3) .ne.  DifNodesSt_(3,iso1_) ) goto 3000
        !
        ! The test should be done only for structures with the same structural characteristics of the (possibly) new
        ! number of metabolites, number of reactions, number of reversible reactions, number of interconnections,
        ! number of connections metabolite/reaction (reversibility not considered)
        if( ( numNod_(1) .eq. DataRevStru_(2,iso1_) ) .and. ( numNod_(2) .eq. DataRevStru_(3,iso1_) ) .and. &
        ( numNod_(3) .eq. DataRevStru_(4,iso1_) ) .and. ( supDataIns_(1) .eq. DataRevStru_(5,iso1_) ) .and. &
        ( supDataIns_(2) .eq. DataRevStru_(6,iso1_) ) ) then
          ! Tests if the nodes (metabolites and reactions) of the (possibly) new instance have the same degree (total and IN)
          ! of the nodes of structure iso1_
          ! The metabolites have the same degree?
          do Nmet1_=1,numNod_(1)  ! Number of metabolites of the (possibly) new instance
            if( MetDegRevStru_( ((3*(Nmet1_ - 1)) + 1) , iso1_) .ne. supMetDegIns_( ((3*(Nmet1_ - 1)) + 1) ) ) goto 3000
            if( MetDegRevStru_( ((3*(Nmet1_ - 1)) + 2) , iso1_) .ne. supMetDegIns_( ((3*(Nmet1_ - 1)) + 2) ) ) goto 3000  ! The (possibly)
          end do  ! new instance is not a concretization of structure iso1_
          ! The reactions have the same degree?
          do Nreact1_=1,numNod_(2)  ! Number of reactions of the (possibly) new instance
            if( ReactDegRevStru_( ((3*(Nreact1_ - 1)) + 1) , iso1_) .ne. supReactDegIns_( ((3*(Nreact1_ - 1)) + 1)) ) goto 3000
            if( ReactDegRevStru_( ((3*(Nreact1_ - 1)) + 2) , iso1_) .ne. supReactDegIns_( ((3*(Nreact1_ - 1)) + 2)) ) goto 3000  ! The
          end do  ! (possibly) new instance is not a concretization of structure iso1_
          !
          ! The (possibly) new instance might be a concretization of structure iso1_
          !
          ! Reference number of the permutation for the metabolites
          do 3005 iso2_=1,supNumPermutNode_(1)
            !
            ! Checks if it is the first iteration for the permutations of the metabolites
            if( iso2_ .ne. 1 ) then
              ! If it is not the first iteration, tests if the degree of metabolites are kept equal
              do Nmet2_=1,numNod_(1)  ! Number of metabolites of the (possibly) new instance
                if( MetDegRevStru_( ((3*(PermutNode_(Nmet2_,iso2_,numNod_(1)) - 1)) + 1) , iso1_) .ne. &
                supMetDegIns_( ((3*(Nmet2_ - 1)) + 1) ) ) goto 3005
                if( MetDegRevStru_( ((3*(PermutNode_(Nmet2_,iso2_,numNod_(1)) - 1)) + 2) , iso1_) .ne. &
                supMetDegIns_( ((3*(Nmet2_ - 1)) + 2) ) ) goto 3005  ! Checks the next permutation for the metabolites
              end do
            end if
            !
            ! The degree of the metabolites are kept with this permutation. One can continue to check if the (possibly) new
            ! instance is a concretization of structure iso1_
            !
            ! The (possibly) new instance has at least one irreversible reaction?
            if( numNod_(4) .ne. 0 ) then
              ! Checks the permutations for the irreversible reactions
              do 3010 iso5_=1,supNumPermutNode_(2)
                !
                ! Checks if it is the first permutation for the irreversible reactions
                if( iso5_ .ne. 1 ) then
                  ! If it is not the first iteration, tests if the degree of the reactions are kept equal
                  do Nreact2_=1,numNod_(4)  ! Number of irreversible reactions of the (possibly) new structure
                    if( ReactDegRevStru_( ((3*(PermutNode_(Nreact2_,iso5_,numNod_(4)) - 1)) + 1)  ,iso1_) .ne. &
                    supReactDegIns_( ((3*(Nreact2_ - 1)) + 1)) ) goto 3010
                    if( ReactDegRevStru_( ((3*(PermutNode_(Nreact2_,iso5_,numNod_(4)) - 1)) + 2)  ,iso1_) .ne. &
                    supReactDegIns_( ((3*(Nreact2_ - 1)) + 2)) ) goto 3010  ! Checks the next permutation for the irreversible
                  end do                                                         ! reactions
                end if
                !
                ! The part of the patterns with the connections metabolite/irreversible reactions of the (possibly) new
                ! instance and of structure iso1_) can be compared
                do com3_=1,numNod_(1)  ! Number of metabolites of the (possibly) new instance
                  do com4_=1,numNod_(4)  ! Number of irreversible reactions of the (possibly) new instance
                    ! We only need to compare the direct reactions since there is no inverse reaction
                    !
                    ! How to determine the position? ->
                    !     Matrix([(4x(nreaction - 1)) + direct/inverse + reactant/product],[nmetabolite],[nstructure])
                    !      where,
                    !           direct/inverse: "0", because we only need to compare the direct reactions
                    !           reactant/product: "1" for a reactant, and "2" for a product
                    !     For the matrix supRevStIns_(i3,i2,i1), the positions of the metabolites and of the reactions
                    !     are not changed.
                    !     However, for the matrix RevStruct_(i3,i2,i1) the positions of the metabolites and of the reactions
                    !     change according to the respective permutations:
                    !       n of reaction -> PermutNode_(com4_,iso5_,numNod_(4))
                    !       direct/inverse reation position ->
                    !        i3: 1st term + 2nd term + 3rd term
                    !          1st term: [4 x [PermutNode_(com4_,iso5_,numNod_(4)) - 1]]
                    !          2nd term: 0
                    !          3rd term: "1" for a reactant, and "2" for a product (the same value as the one considered to
                    !                    matrix supRevStIns_(i3,i2,i1))
                    !        i2: PermutNode_(com3_,iso2_,numNod_(1))
                    !        i1: iso1_
                    !
                    ! Compares the direct single-direction reaction (direct/inverse term equal to "0"):
                    ! Reactant position (reactant/product term equal to "1")
                    if( supRevStIns_( ((4*(com4_ - 1)) + 0 + 1) ,com3_ ,1) .ne. &
                    RevStruct_( ((4*( PermutNode_(com4_,iso5_,numNod_(4)) - 1 )) + 0 + 1 ) , &
                    PermutNode_(com3_,iso2_,numNod_(1)) , iso1_) ) goto 3010
                    !          |            1st term                    | |2nd||3rd |
                    ! Product position (reactant/product term equal to "2")
                    if( supRevStIns_( ((4*(com4_ - 1)) + 0 + 2) ,com3_ ,1) .ne. &
                    RevStruct_( ((4*( PermutNode_(com4_,iso5_,numNod_(4)) - 1 )) + 0 + 2 ) , &
                    PermutNode_(com3_,iso2_,numNod_(1)) , iso1_) ) goto 3010
                    !
                    ! Is this part of the pattern, i.e., the connections metabolites/irreversible reactions the same between the
                    ! (possibly) new instance and structure iso1_?
                    if( (com3_ .eq. numNod_(1)) .and. (com4_ .eq. numNod_(4)) ) then
                      ! We can now test the part of the pattern corresponding to the connections between metabolites and reversible
                      goto 3015  ! reactions
                    end if
                  end do
                end do
3010          continue
              ! This part of the code is not available to the cases were there is a correspondence between the connections
              ! metabolite/irreversible reaction of the (possibly) new instance and structure iso1_. So, it means a new
              ! permutation for the metabolites must be tested
              goto 3005
              !
            end if
            !
            ! The part of the pattern between the metabolites and the reversible reactions can now be tested
            goto 3015
3015        continue
            !
            ! The (possibly) new instance has at least one reversible reaction?
            if( numNod_(3) .ne. 0 ) then
              ! Reference number of the permutation for the reversible reactions
              do 3020 iso3_=1,supNumPermutNode_(3)
                ! Checks if it is the first permutation for the reversible reactions
                if( iso3_ .ne. 1 ) then
                  ! If it is not the first iteration, tests if the degree of the reactions are kept equal (only the total
                  ! degree is needed to test the reversible reactions of the two structures: for reversible reactions, the
                  ! IN degree is equal to the OUT degree
                  do Nreact3_=1,numNod_(3)  ! Number of reversible reactions of the (possibly) new structure
                    if( ReactDegRevStru_( ((3*(numNod_(4) + PermutNode_(Nreact3_,iso3_,numNod_(3)) - 1)) + 1)  ,iso1_) .ne. &
                    supReactDegIns_( ((3*(numNod_(4) + Nreact3_ - 1)) + 1)) ) goto 3020  ! Checks the next permutation for the reversible
                  end do                                                         ! reactions
                end if
                !
                ! The degree of the metabolites are kept with this permutation. One can continue to check if the (possibly) new
                ! instance is a concretization of structure iso1_
                !
                ! Reference number of the permutation for the direct and inverse single direction reactions for
                ! all the reversible reactions
                do 3025 iso4_=1,numPermDirInvReact_(numNod_(3))
                  !
                  ! Each possible permutation (perm. Met x perm. RevReact x perm. DirInvReact.) will be used to compare
                  ! the (possibly) new instance with structure iso1_
                  do com1_=1,numNod_(1)  ! Number of metabolites of the (possibly) new instance
                    do com2_=1,numNod_(3)  ! Number of reversible reactions of the (possibly) new instance
                      !
                      ! How to determine the position? ->
                      !     Matrix([(4x(nreaction - 1)) + direct/inverse + reactant/product],[nmetabolite],[nstructure])
                      !      where,
                      !           direct/inverse: "0" if it is the direct reaction, and "2" if it is the inverse reaction
                      !           reactant/product: "1" for a reactant, and "2" for a product
                      !     For the matrix supRevStIns_(i3,i2,i1), the positions of the metabolites and of the reactions
                      !     are not changed.
                      !     However, for the matrix RevStruct_(i3,i2,i1) the positions of the metabolites and of the reactions
                      !     change according to the respective permutations:
                      !       n of reaction -> PermutNode_(com2_,iso3_,numNod_(3))
                      !       direct/inverse reation position ->
                      !        i3: 1st term + 2nd term + 3rd term
                      !          1st term: [4 x [PermutNode_(com2_,iso3_,numNod_(3)) - 1]]
                      !          2nd term: PermutDirInvReact_([[2 x [n of reaction - 1]] + a],iso4_,numNod_(3))
                      !                    where "a" is according to the position considered to matrix supRevStIns_(i3,i2,i1):
                      !                    "1" for the direct reaction, and "2" for the inverse reaction
                      !          3rd term: "1" for a reactant, and "2" for a product (the same value as the one considered to
                      !                    matrix supRevStIns_(i3,i2,i1))
                      !        i2: PermutNode_(com1_,iso2_,numNod_(1))
                      !        i1: iso1_
                      !
                      ! Compares the direct single-direction reaction (direct/inverse term equal to "0"):
                      ! Reactant position (reactant/product term equal to "1")
                      if( supRevStIns_( ((4*(numNod_(4) + com2_ - 1)) + 0 + 1) ,com1_ ,1) .ne. &
                      RevStruct_( ((4*( numNod_(4) + PermutNode_(com2_,iso3_,numNod_(3)) - 1 )) + &
                      PermutDirInvReact_( ((2*( PermutNode_(com2_,iso3_,numNod_(3)) - 1)) + 1 ) , &
                      iso4_ , numNod_(3) ) + 1 ) , PermutNode_(com1_,iso2_,numNod_(1)) , iso1_) ) goto 3025
                      !          |                      1st term                              |   |                         2nd term                               ||3rd term|
                      ! Product position (reactant/product term equal to "2")
                      if( supRevStIns_( ((4*(numNod_(4) + com2_ - 1)) + 0 + 2) ,com1_ ,1) .ne. &
                      RevStruct_( ((4*( numNod_(4) + PermutNode_(com2_,iso3_,numNod_(3)) - 1 )) + &
                      PermutDirInvReact_( ((2*( PermutNode_(com2_,iso3_,numNod_(3)) - 1)) + 1 ) , &
                      iso4_ , numNod_(3) ) + 2 ) , PermutNode_(com1_,iso2_,numNod_(1)) , iso1_) ) goto 3025
                      !
                      ! We do not need to compare the inverse single-direction reaction
                      !
                      ! Is this (possibly) new instance a concretization of structure iso1_? For that to happen, all the cells
                      ! must be compared: all the metabolites (com1_ equal to number of metabolites) and all reactions
                      ! (com2_ equal to number of reversible reactions)
                      if( (com1_ .eq. numNod_(1)) .and. (com2_ .eq. numNod_(3)) )   then
                        !
                        ! Stores the reference number i1=iso1_ of the structure of matrix RevStruct_(i3,i2,i1), of which the
                        ! instance represented in matrix supRevStIns_(i2,i1,1) is one of its concretizations
                        isoStruRef_ = iso1_
                        !
                        ! Stores the reference number of the permutation of the metabolites that led to the tested instance
                        ! being classified as a realization of structure iso1_
                        refNumPermMet_ = iso2_
                        !
                        ! The instance has at least one irreversible reaction?
                        if( numNod_(4) .ne. 0 ) then
                          ! Stores the reference number of the permutation for the irreversible reactions that led to the
                          ! tested instance to be classified as a realization of structure iso1_
                          refNumPermIrrReact_ = iso5_
                        else
                          ! The instance does not have irreversible reactions
                          refNumPermIrrReact_ = 0
                        end if
                        !
                        ! Stores the reference number of the permutation for the reversible reactions that led to the
                        ! tested instance to be classified as a realization of structure iso1_
                        refNumPermRevReact_ = iso3_
                        !
                        ! Stores the reference number of the permutation for the direct and inverse reactions of a
                        ! reversible reaction that led to the tested instance to be classified as a realization of
                        ! structure iso1_
                        refNumPermDirInvReact_ = iso4_
                        !
                        goto 3030
                      end if
                    end do
                  end do
3025            continue
3020          continue
            else
              ! The structure has no reversible reactions
              !
              ! Stores the reference number i1=iso1_ of the structure of matrix RevStruct_(i3,i2,i1), of which the
              ! instance represented in matrix supRevStIns_(i2,i1,1) is one of its concretizations
              isoStruRef_ = iso1_
              !
              ! Stores the reference number of the permutation of the metabolites that led to the tested instance
              ! being classified as a realization of structure iso1_
              refNumPermMet_ = iso2_
              !
              ! Stores the reference number of the permutation for the irreversible reactions that led to the
              ! tested instance to be classified as a realization of structure iso1_
              refNumPermIrrReact_ = iso5_
              !
              ! The instance has no reversible reactions
              refNumPermRevReact_ = 0
              refNumPermDirInvReact_ = 0
              !
              goto 3030
              !
            end if
3005      continue
        end if
3000  continue
      !
      ! This part of the code is not available to the situations in which the (possibly) new instance represented in matrix
      ! supRevStIns_(i2,i1,1) is a concretization of structure i1=iso1_ stored in matrix RevStruct_(i3,i2,i1)
      !
      ! The instance found is not within the considered scope. So, isoStruRef_ equal to "0"
      isoStruRef_ = 0
      refNumPermMet_ = 0
      refNumPermIrrReact_ = 0
      refNumPermRevReact_ = 0
      refNumPermDirInvReact_ = 0
      goto 3030
3030  continue

      end subroutine isomorphismTest


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!     Subroutine to order a vector from the higher to the lower number
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine orderVector(Vector_, numElem_, VectorOrd_)

! CONSTANTS
!      numElem_ : number of elements of the vector
!      position_ :  support constant used to store the position with the node with higher reference number in the support
!                   vector supNodeOrd_(i1)
!
!
! VECTORS
!      Vector_(i1) : input vector with numElem_ cells
!      VectorOrd_(i1) : vector with the cells of Vector_(i1) ordered from the higher to the lower number
!      supVectorOrd_(i1) : support vector used to order the cells of Vector_(i1) into VectorOrd_(i1)

      implicit none

      ! IN variables
       ! Integer: 1 bit
      integer*1, intent(in) :: numElem_
       ! Integer: 2 bits
      integer*2, intent(in) :: Vector_(:)

      ! OUT variables
       ! Integer: 2 bits
      integer*2, intent(inout) :: VectorOrd_(:)

      !Internal support variables
       ! Integer: 1 bit
      integer*1 :: position_
       ! Integer: 2 bits
      integer*2 :: supVectorOrd_(8)

      ! General variables
      integer*1 :: iOrd1_, iOrd2_  ! 4 bits


!      Creates the ordered list for the vector from the higher to the lower number
      !
      do iOrd1_=1,numElem_  ! Number of elements of the vector
        supVectorOrd_(iOrd1_) = Vector_(iOrd1_)
      end do
      do iOrd1_=1,numElem_  ! Number of elements of the vector
        VectorOrd_(iOrd1_) = 0
        do iOrd2_=1,numElem_  ! Number of elements of the vector
          !
          ! A greater number was found than Vector_(iOrd1_)?
          if( ( supVectorOrd_(iOrd2_) .gt. VectorOrd_(iOrd1_) ) ) then
            ! Stores the new higher value at the cell VectorOrd_(iOrd1_)
            VectorOrd_(iOrd1_) = supVectorOrd_(iOrd2_)
            position_ = iOrd2_
          end if
          !
          ! Is iOrd2_ equal to numElem_?
          if( iOrd2_ .eq. numElem_ ) then
            ! Clears the cell corresponding to position_
            supVectorOrd_(position_) = 0
            position_ = 0
          end if
          !
        end do
      end do

      end subroutine orderVector


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!     Subroutine to determine if an apparently new instance had been found before.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine isoInstances(numNode_, supTotMetIns_, supTotReactIns_, supMet_, &
      supReact_, TotMetIns_, TotReactIns_, OrdMetInst_, OrdReactInst_, &
      InstStruct_, minRefNInstStru_, maxRefNInstStru_, InstTest_, supMetOrd_, &
      supReactOrd_, maxNumMet_, maxNumReact_, maxNumInst_, isoStruRefe_, &
      numIns_, TempCut_)

! CONSTANTS
!      isoStruRefe_ : reference number of the structure in matrix RevStruct(i3,i2,i1) to which the instance in matrix
!                     supRevStInst(i2,i1,1) is a concretization.
!      maxNumInst_ : expected maximum number of instances. This parameter is needed to create the allocatable arrays
!                    used to store the required information for each one of the instances
!      maxNumMet_ : maximum number of metabolites considered for the structures within the considered number of
!                   generations maxNumGen
!      maxNumReact_ : maximum number of reactions considered for the structures within the considered number of
!                     generations maxNumGen
!      minRefNInstStru_ : minimum reference number of an instance of structure isoStruRefe_
!      maxRefNInstStru_ : maximum reference number of an instance of structure isoStruRefe_
!      TempCut_ : support variable used to switch off the part of the subroutine "isoInstances" used to determine if
!                 the (possibly) new instance was found before - it allows only to obtain the metabolites and the
!                 of the (possibly) new instance organized from the higher to the lower reference numbers. If the variable
!                 TempCut_ = 0, the mentioned part of the subroutine is not switch off; if the variable TempCut_ = 1, the
!                 considered part of the subroutine is switch off
!      InstTest_ : indicates if the instance tested was found before. If it is a new instance, InstTest_ = 0; if it is not,
!                  InstTest_ = 1
!      supTotMetIns_ : support constant with the sum of the reference numbers of the metabolites of the (possibly) new instance
!      supTotReactIns_ : support constant with the sum of the reference numbers of the reactions of the (possibly) new instance
!
!
! VECTORS
!      numNode_(i1) : number of nodes of the (possibly) new instance. (i1=1) n of metabolites; (i1=2) n of reactions;
!                     (i1=3) n of reversible reactions; (i1=4) n of irreversible reactions
!      supMet_(i1) : support vector with all the metabolites of the (possibly) new instance
!      supReact_(i1) : support vector with all the reactions of the (possibly) new instance
!      InstStruct_(i1) : reference number of the structure of the instance i1
!      TotMetIns_(i1) : sum of the reference numbers of the metabolites of instance i1
!      TotReactIns_(i1) : sum of the reference numbers of the reactions of instance i1
!      supMetOrd_(i1) : support vector with all the metabolites of the (possibly) new instance ordered from the higher to
!                       the lower reference number
!      supReactOrd_(i1) : support vector with all the reactions of the (possibly) new instance ordered from the higher to
!                         the lower reference number
!      MetInstOrd_(i1) : support vector with all the metabolites of the instance inst_ ordered from the higher to the
!                        lower reference number
!      ReactInstOrd_(i1) : support vector with all the reactions of the instance inst_ ordered from the higher to the
!                          lower reference number
!      supNodeOrd_(i1) : support vector used to order the nodes (metabolites or reactions) of the considered instances
!                        from the higher to the lower reference number
!
!
! MATRICES
!      OrdMetInst_(i2,i1) : metabolites of the instance i1 ordered from the higher to the lower reference numbers
!      OrdReactInst_(i2,i1) : reactions of the instance i1 ordered from the higher to the lower reference numbers

      implicit none

      ! IN variables
       ! Static variables
        ! Parameters
      integer, intent(in) :: maxNumInst_  ! 4 bits
        ! Integer: 1 bit
      integer*1, intent(in) :: numNode_(:), maxNumMet_, maxNumReact_, TempCut_
        ! Integer: 2 bits
      integer*2, intent(in) :: supMet_(:), supReact_(:), supTotMetIns_
      integer*2, intent(in) :: supTotReactIns_, isoStruRefe_
        ! Integer: 4 bits
      integer*4, intent(in) :: maxRefNInstStru_(:), numIns_
       ! Allocatable variables: 2 bits
      integer*2, DIMENSION(maxNumMet_,maxNumInst_), intent(in) :: OrdMetInst_
      integer*2, DIMENSION(maxNumReact_,maxNumInst_), intent(in) :: OrdReactInst_
      integer*2, DIMENSION(maxNumInst_), intent(in) :: InstStruct_, TotMetIns_, TotReactIns_

      ! IN/OUT variables
        ! Integer: 4 bits
      integer*4, intent(inout) :: minRefNInstStru_(:)

      ! OUT variables
        ! Integer: 1 bit
      integer*1, intent(inout) :: InstTest_  ! 1 bit
        ! Integer: 2 bits
      integer*2, intent(inout) :: supMetOrd_(:), supReactOrd_(:)

      ! General variables
      integer :: inst_, met_, react_  ! 4 bits

!      Creates the ordered list for the metabolites and reactions of the newly found instance (from the higher to the lower
!      number of the specific type of node, metabolites or reactions)
      !
      ! Orders the metabolites
      ! Calls the subroutine orderVector
      call orderVector(supMet_, numNode_(1), supMetOrd_)
      !
      ! Orders of the reactions
      ! Calls the subroutine orderVector
      call orderVector(supReact_, numNode_(2), supReactOrd_)
      !
      ! Is this the first instance of structure isoStruRefe_ found?
      if( minRefNInstStru_(isoStruRefe_) .eq. 0 ) then
        ! It is a new instance
        InstTest_ = 0
        ! Stores the minimum reference number of a instance of structure isoStruRefe_
        minRefNInstStru_(isoStruRefe_) = numIns_ + 1
        goto 3040
      end if
      !
      ! Variable TempCut_ is equal to 1?
      if( TempCut_ .eq. 1 ) then
        ! The (possibly) new instance will not be compared with previously found instances. The
        ! variable InstTest_ is turn to 1 to allow the (possibly) new instance to be stored.
        ! It will be tested against the other instances of the same structure in another
        ! Fortran project
        InstTest_ = 0
        !
        goto 3040
      end if

!      Determines if the instance inst_ is the same that the supposedly new instance tested
      ! Test all the instances with reference numbers between the minimum and maximum found before
      ! for structure isoStruRefe_
      do 3035 inst_=minRefNInstStru_(isoStruRefe_),maxRefNInstStru_(isoStruRefe_)
        !
        ! Test if the instance inst_ has the same structure that the (possibly) new instance
        if( isoStruRefe_ .ne. InstStruct_(inst_) ) goto 3035
        !
        ! Test if the instances have an equal sum of the reference numbers of the metabolites
        if( TotMetIns_(inst_) .ne. supTotMetIns_ ) goto 3035
        !
        ! Test if the instances have an equal sum of the reference numbers of the reactions
        if( TotReactIns_(inst_) .ne. supTotReactIns_ ) goto 3035
        !
        ! The instances can be compared
        !
        ! The metabolites of instance inst_ are equal to the tested instance?
        do met_=1,numNode_(1)  ! Number of metabolites
          if( supMetOrd_(met_) .ne. OrdMetInst_(met_,inst_) ) goto 3035  ! The instances are not the same; tests the next
        end do                                                     ! instance inst_
        !
        ! The reactions of instance inst_ are equal to the tested instance?
        do react_=1,numNode_(2)  ! Number of reactions
          if( supReactOrd_(react_) .ne. OrdReactInst_(react_,inst_) ) goto 3035  ! The instances are not the same; tests
        end do                                                                   ! the next instance inst_
        !
        ! The instances are the same!
        InstTest_ = 1
        goto 3040
        !
3035  continue
      !
      ! This part of the code is available only to new instances
      InstTest_ = 0
      goto 3040
      !
3040  continue

      end subroutine isoInstances


      end program instancesCounting

