!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! TAMES  Topological Analysis of Metabolic Structures                 !
! Copyright (C) 2017                                                   !
! Paulo A. N. Dias, paulo.dias29@gmail.com                             !
! Marco Seabra dos Reis, marco@eq.uc.pt                                !
! Armindo Salvador, armindo.salvador@gmail.com                         !
! Pedro Martins, pmartins@iscac.pt                                     !
!                                                                      !
! This file is part of TAMES.                                          !
!                                                                      !
! TAMES is free software: you can redistribute it and/or modify        !
! it under the terms of the GNU General Public License as published by !
! the Free Software Foundation, either version 3 of the License, or    !
! any later version.                                                   !
!                                                                      !
! You should have received a copy of the GNU General Public License    !
! along with TAMES.  If not, see <http://www.gnu.org/licenses/>.       !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                         !
! Program "Organizing Instances":                         !
! - Determines the actual number of instances obtained by !
!   the program "Instances Counting". Organizes the       !
!   instances according to their structure.               !
!                                                         !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      Program OrgInst

! CONSTANTS
!      numInst : number of instances for the considered scope. NOTE: includes repeated instances
!      numInstReal : real number of instances for the considered scope (does not include repeated instances)
!      repeat : logical variable used to read the file "in_nGen_nStru.txt"
!      maxNumGen : maximum number of generations considered to generate structures
!      maxNumMet : maximum number of metabolites considered for the structures within the considered number of
!                  generations maxNumGen
!      maxNumReact : maximum number of reactions considered for the structures within the considered number of
!                    generations maxNumGen
!      numStructures : number of structures for the considered generations (maxNumGen)
!
!
! VECTORS
!      InstStruct(i1) : structure of the instance i1. NOTE: includes repeated instances
!      OutRevStruct(i1) : number of instances of structure i1. NOTE: includes repeated instances
!      OutRevStructReal(i1) : real number of instances of structure i1 (does not include repeated instances)
!      nrefInstLow(i1) : lower reference number of a instance of structure i1 in organized matrices
!                        for the instances - OrgMetInst(i2,i1), OrgReactInst(i2,i1), MetInstOrdRefNum(i2,i1)
!                        and ReactInstOrdRefNum(i2,i1) - by structure. NOTE: includes repeated instances
!      nrefInstHigh(i1) : higher reference number of a instance of structure i1 in organized matrices
!                         for the instances - OrgMetInst(i2,i1), OrgReactInst(i2,i1), MetInstOrdRefNum(i2,i1)
!                         and ReactInstOrdRefNum(i2,i1) - by structure. NOTE: includes repeated instances
!      nrefInstLowReal(i1) : lower reference number of a instance of structure i1 in organized matrices
!                            for the instances - OrgMetInstReal(i2,i1), OrgReactInstReal(i2,i1) - by
!                            structure (does not include repeated instances)
!      nrefInstHighReal(i1) : higher reference number of a instance of structure i1 in organized matrices
!                             for the instances - OrgMetInstReal(i2,i1), OrgReactInstReal(i2,i1) - by
!                             structure (does not include repeated instances)
!      SumMetInst(i1) : sum of the reference numbers of the metabolites of instance i1. NOTE: some of the
!                       instances may be repeated
!      SumReactInst(i1) : sum of the reference numbers of the reactions of instance i1. NOTE: some of the
!                         instances may be repeated
!      RepeatInst(i1) : indicates if instance i1 is the same that a lower reference number instance or if it
!                       is the first time this particular instance was found. If it is a repeated instance fills
!                       the corresponding cell with "1"; if it is the first this instance was found, fills the
!                       correponding cell with "0"
!      supOrgInst(i1) : support vector used to organized the instances by their respective structure
!      const(i1) : support vector of size integer*4
!      supOutRevStruct(i1) : support vector with the number of instances of the structures. Used to obtain
!                            a frequency ranking for the structures. NOTE: includes repeated instances
!      supRankStru(i1) : rank of the number of instances of the structures. NOTE: includes repeated instances
!
!
! MATRICES
!      DataRevStruTemp(i2,i1) : i1) reference number of the structure. i2) addtional data for each structure: (i2=1) number
!                               of generation; (i2=2) number of metabolites; (i2=3) number of reactions; (i2=4) number of
!                               reversible reactions; (i2=5) number of Ancestors; (i2=6) number of Parents; (i2=7) number of
!                               Childs; (i2=8) number of Offspring; (i2=9) number of interconnections; (i2=10) maximum possible
!                               number of interconnections; (i2=11) number of connections metabolite/reaction (reversibility not
!                               considered); (i2=12) maximum possible number of connections metabolite/reaction (reversibility
!                               not considered)
!      NumNodes(i2,i1) : i1) reference number of the structure. i2) addtional data for each structure: (i2=1) number of
!                            metabolites; (i2=2) number of reactions
!      OrgMetInst(i2,i1) :  reference numbers of the metabolites of the instance i1 ordered according
!                           to the degree of the nodes. The instances are organized according to their
!                           structure: instances of structures of lower reference numbers are stored
!                           first. NOTE: some of the instances may be repeated
!      OrgReactInst(i2,i1) : reference numbers of the reactions of the instance i1 ordered according
!                            to the degree of the nodes. The instances are organized according to their
!                            structure: instances of structures of lower reference numbers are stored
!                            first. NOTE: some of the instances may be repeated
!      MetInstOrdRefNum(i2,i1) : metabolites of the instance i1 ordered from the higher to the lower
!                                reference numbers. The instances are organized according to their
!                                structure: instances of structures of lower reference numbers are
!                                stored first. NOTE: some of the instances may be repeated
!      ReactInstOrdRefNum(i2,i1) : reactions of the instance i1 ordered from the higher to the lower
!                                  reference numbers. The instances are organized according to their
!                                  structure: instances of structures of lower reference numbers are
!                                  stored first. NOTE: some of the instances may be repeated

      implicit none

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      DECLARATION OF THE VARIABLES     !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!      Allocatable arrays: declaration
       ! Integer: 1 bit
      integer*1, DIMENSION(:), allocatable :: RepeatInst
       ! Integer: 2 bits
      integer*2, DIMENSION(:,:), allocatable :: OrgMetInst, OrgReactInst, in_nGen_nStru
      integer*2, DIMENSION(:,:), allocatable :: MetInstOrdRefNum, ReactInstOrdRefNum
      integer*2, DIMENSION(:,:), allocatable :: OrgMetInstReal, OrgReactInstReal
      integer*2, DIMENSION(:,:), allocatable :: DataRevStruTemp
      integer*2, DIMENSION(:), allocatable :: InstStruct, SumMetInst, SumReactInst

!      Static arrays: declaration
      ! Integer: 1 bit
      integer*1 :: NumNodes(2,4000), maxNumGen, maxNumMet, maxNumReact
      ! Integer: 2 bits
      integer*2 :: supRankStru(4000), numStructures
      ! Integer: 4 bits
      integer*4 :: OutRevStruct(4000), nrefInstLow(4000), nrefInstHigh(4000)
      integer*4 :: OutRevStructReal(4000), nrefInstLowReal(4000), nrefInstHighReal(4000)
      integer*4 :: supOrgInst(4000), supOutRevStruct(4000), const(10)
      integer*4 :: numInst, numInstReal, ID, omp_get_thread_num
      integer*4 :: i, i1, i2, iMet, iReact
      ! Logical
      logical :: repeat

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      INPUT FILES                      !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Reads the information regarding the possible structures till the number of generations
!      considered (5 generations)
     !
    !  Read the file "in_nGen_nStru.txt"
      ! Allocates the variable in_nGen_nStru(i2,i1)
      allocate(in_nGen_nStru(2,15))
      !
      const(1) = 0
      ! Logical variable. The cycle is performed while the variable is "true"
      repeat = .true.
      open (1,file='in_nGen_nStru.txt')
      do while (repeat)
        const(1) = const(1) + 1
        read(1,*) (in_nGen_nStru(i1,const(1)), i1=1,2)
        if( const(1) .eq. (in_nGen_nStru(1,1)+1) ) then
          repeat= .false.
        end if
      end do
      close(1)
      const(1) = 0
      maxNumGen = in_nGen_nStru(1,1)
      numStructures = in_nGen_nStru(2,1)
      ! Deallocates the variable in_nGen_nStru(i2,i1)
      deallocate(in_nGen_nStru)
     !
    !  Read the file "in_nMet_nReact.txt", which stores  the maximum number of metabolites and of reactions considered
    !  for the structures within the assumed number of generations
      open (1,file='in_nMet_nReact.txt')
      do i1=1,2
        if( i1 .eq. 1 ) then
          read(1,*) maxNumMet
        else
          read(1,*) maxNumReact
        end if
      end do
      close(1)
     !
    !  Read the 'Structures_General.txt', which stores the general information regarding the structures (no restrictions)
      ! Allocate the variable DataRevStruTemp(i2,i1)
      allocate(DataRevStruTemp(12,4000))
      open (1,file='Structures_General.txt')
      do i1=1,numStructures
        read(1,*) (DataRevStruTemp(i2,i1), i2=1,12)
      end do
      close(1)
      ! Stores the relevant information of matrix DataRevStruTemp(i2,i1) into matrix NumNodes(i2,i1)
      do i1=1,numStructures
        ! N of metabolites
        NumNodes(1,i1) = DataRevStruTemp(2,i1)
        ! N of reactions
        NumNodes(2,i1) = DataRevStruTemp(3,i1)
      end do
      ! Deallocate the variable DataRevStruTemp(i2,i1)
      deallocate(DataRevStruTemp)

!      Reads the information regarding the instances of the structures of the considered scope found
!      in the metabolic network
      !
    !
    ! Read the file 'OutRevStruct.txt', which stores the naked frequency of the structures in the metabolic network
    ! and the number of different metabolites and reactions that appear in at least one instance of the corresponding
    ! structure
      open(1,file='OutRevStruct.txt')
      do i1=1,numStructures
        read(1,*) OutRevStruct(i1)
      end do
      close(1)
     !
    ! Reads the file 'numInstances.txt', which stores the number of instances found in the metabolic network
      open(1,file='numInstances.txt')
      read(1,*) numInst
      close(1)
     !
      ! Allocate the arrays OrgMetInst(i2,i1), OrgReactInst(i2,i1), MetInstOrdRefNum(i2,i1), ReactInstOrdRefNum(i2,i1)
      ! and InstStruct(i1)
      allocate(OrgMetInst(maxNumMet,numInst),OrgReactInst(maxNumReact,numInst))
      allocate(MetInstOrdRefNum(maxNumMet,numInst),ReactInstOrdRefNum(maxNumReact,numInst))
      allocate(InstStruct(numInst), SumMetInst(numInst), SumReactInst(numInst))
     !
    ! Reads the file 'InstStruct.txt', that stores the number of the structure of each instance
      open(1,file='InstStruct.txt')
      do i1=1,numInst
        read(1,*) InstStruct(i1)
      end do
      close(1)
      !
      ! Creates the vectors nrefInstLow(i1), nrefInstHigh(i1) and supOrgInst(i1)
      do i1=1,numStructures
        nrefInstLow(i1) = 0
        nrefInstHigh(i1) = 0
        supOrgInst(i1) = 0
      end do
      !
      ! Obtains the minimum and maximum reference numbers of the instances of each structure
      ! in organized matrices OrgMetInst(i2,i1) and OrgReactInst(i2,i1): stores first instances
      ! of structures of lower reference numbers
      const(1) = 0
      do i1=1,numStructures
        ! Structure i1 has at least one instance?
        if( OutRevStruct(i1) .ne. 0 ) then
          nrefInstLow(i1) = const(1) + 1
          nrefInstHigh(i1) = const(1) + OutRevStruct(i1)
          const(1) = nrefInstHigh(i1)
        end if
      end do
     !
    ! Reads the file 'MetInst.txt', that stores the reference numbers of the metabolites of
    ! each instance. The instances are automatically organized into matrix OrgMetInst(i2,i1)
    ! according to their structure
      open(1,file='MetInst.txt')
      do i1=1,numInst
        read(1,*) (OrgMetInst(i2, (nrefInstLow(InstStruct(i1)) + supOrgInst(InstStruct(i1))) ), &
        i2=1,maxNumMet)
        supOrgInst(InstStruct(i1)) = supOrgInst(InstStruct(i1)) + 1
      end do
      close(1)
      !
      ! Turns to zero the support vector supOrgInst(i1)
      do i1=1,numStructures
        supOrgInst(i1) = 0
      end do
    ! Reads the file 'ReactInst.txt', that stores the reference numbers of the reactions of each
    ! instance ordered according to the degree of the nodes. The instances are automatically
    ! organized into matrix OrgReactInst(i2,i1) according to their structure
      open(1,file='ReactInst.txt')
      do i1=1,numInst
        read(1,*) (OrgReactInst(i2, (nrefInstLow(InstStruct(i1)) + supOrgInst(InstStruct(i1))) ), &
        i2=1,maxNumReact)
        supOrgInst(InstStruct(i1)) = supOrgInst(InstStruct(i1)) + 1
      end do
      close(1)
      !
      ! Turns to zero the support vector supOrgInst(i1)
      do i1=1,numStructures
        supOrgInst(i1) = 0
      end do
    ! Reads the file 'MetInstOrdRefNum.txt', that stores the reference numbers of the metabolites of
    ! each instance ordered from the higher to the lower number. The instances are automatically
    ! organized into matrix MetInstOrdRefNum(i2,i1) according to their structure
      open(1,file='MetInstOrdRefNum.txt')
      do i1=1,numInst
        read(1,*) (MetInstOrdRefNum(i2, (nrefInstLow(InstStruct(i1)) + supOrgInst(InstStruct(i1))) ), &
        i2=1,maxNumMet)
        supOrgInst(InstStruct(i1)) = supOrgInst(InstStruct(i1)) + 1
      end do
      close(1)
      !
      ! Turns to zero the support vector supOrgInst(i1)
      do i1=1,numStructures
        supOrgInst(i1) = 0
      end do
    ! Reads the file 'ReactInstOrdRefNum.txt', that stores the reference numbers of the reactions of
    ! each instance ordered from the higher to the lower number. The instances are automatically
    ! organized into matrix ReactInstOrdRefNum(i2,i1) according to their structure
      open(1,file='ReactInstOrdRefNum.txt')
      do i1=1,numInst
        read(1,*) (ReactInstOrdRefNum(i2, (nrefInstLow(InstStruct(i1)) + supOrgInst(InstStruct(i1))) ), &
        i2=1,maxNumReact)
        supOrgInst(InstStruct(i1)) = supOrgInst(InstStruct(i1)) + 1
      end do
      close(1)
      !
      ! Turns to zero the support vector supOrgInst(i1)
      do i1=1,numStructures
        supOrgInst(i1) = 0
      end do
    ! Reads the file 'SumMetInst.txt', that stores the sum of the reference numbers of the metabolites
    ! of each instance. The instances are automatically organized into vector SumMetInst(i1) according
    ! to their structure
      open(1,file='SumMetInst.txt')
      do i1=1,numInst
        read(1,*) SumMetInst(nrefInstLow(InstStruct(i1)) + supOrgInst(InstStruct(i1)))
        supOrgInst(InstStruct(i1)) = supOrgInst(InstStruct(i1)) + 1
      end do
      close(1)
      !
      ! Turns to zero the support vector supOrgInst(i1)
      do i1=1,numStructures
        supOrgInst(i1) = 0
      end do
    ! Reads the file 'SumReactInst.txt', that stores the sum of the reference numbers of the reactions
    ! of each instance. The instances are automatically organized into vector SumReactInst(i1) according
    ! to their structure
      open(1,file='SumReactInst.txt')
      do i1=1,numInst
        read(1,*) SumReactInst(nrefInstLow(InstStruct(i1)) + supOrgInst(InstStruct(i1)))
        supOrgInst(InstStruct(i1)) = supOrgInst(InstStruct(i1)) + 1
      end do
      close(1)
      ! Deallocates the vector InstStruct(i1)
      deallocate(InstStruct)
      write(*,*) 'Stop reading the files'
      write(*,*)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      DETERMINE THE REPEATED INSTANCES !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Determines the instances that are equivalent to instances with lower reference numbers
      ! Allocates the variable RepeatInst(i1)
      allocate(RepeatInst(numInst))
      ! Creates the variable RepeatInst(i1)
      do i1=1,numInst
        RepeatInst(i1) = 0
      end do
      ! Creates the variable OutRevStructReal(i1)
      do i1=1,numStructures
        OutRevStructReal(i1) = 0
      end do
      !
      write(*,*) 'Obtains the ranking of frequencies for the structures'
      write(*,*)
      ! Determines a ranking of frequency for the structures. NOTE: some of the instances may
      ! be repeated
      ! Obtains the variable supOutRevStruct(i1)
      do i1=1,numStructures
        supOutRevStruct(i1) = OutRevStruct(i1)
      end do
      ! Obtains the rank
      const(3) = 0
      do 50 i1=1,numStructures
        const(1) = 0
        const(2) = 0
        do i2=1,numStructures
          ! Structure i2 has a higher frequency than the frequency const(2) of the previous
          ! found structure const(1)?
          if( supOutRevStruct(i2) .gt. const(2) )then
            const(2) = supOutRevStruct(i2)
            const(1) = i2
          end if
          !
          ! Tested all the numStructures structures?
          if( i2 .eq. numStructures) then
            ! A structure was found with at least 1 instance?
            if( const(2) .ne. 0 ) then
              ! Stores the structure const(1) with the rank i1
              supRankStru(i1) = const(1)
              !
              ! Turns to "0" the frequency of structure const(1), to allow the determine the next
              ! structure with higher ranking
              supOutRevStruct(const(1)) = 0
              !
              ! Search the structure with the next higher frequency
              goto 50
            else
              ! The structures that are still not ranked do not have instances
              !
              ! Stores the rank that should have been found in this iteration
              const(3) = i1
              !
              ! Leaves the loop
              goto 55
            end if
          end if
        end do
50    continue
      ! This part of the code is only available in the case all the structures have at least
      ! 1 instance. Goes to next part of the code
      goto 55
55    continue
      ! Is there at least 1 structure with no instances?
      if( const(3) .ne. 0 ) then
        ! Finds the rank of the structures with no instances
        do i1=1,numStructures
          ! Structure is has no instances?
          if( OutRevStruct(i1) .eq. 0 ) then
            supRankStru(const(3)) = i1
            const(3) = const(3) + 1
          end if
        end do
      end if
      !
      write(*,*) 'Starts finding the real number of instances'
      write(*,*)
      !
      ID = 0
!      PARALLELIZATION USING OPENMP
      ! Parallelizes the loop
      ! Option: define the number of threads used?
  !    CALL OMP_SET_NUM_THREADS(2)
      !$omp parallel do default(shared) private(i,i1,iMet,iReact,ID) &
      !$omp schedule(dynamic)
      ! Compares the instances of each structure, starting with the structures with higher
      ! number of instances. NOTE: some of the instances may be repeated
      do i=1,numStructures
        !
        ! Structure supRankStru(i) has at least one instance?
        if( OutRevStruct(supRankStru(i)) .ne. 0 ) then
          ! Structure supRankStru(i) has more than 1 instance?
          if( OutRevStruct(supRankStru(i)) .gt. 1 ) then
            ! Check all the instances of structure supRankStru(i)
            do i1=nrefInstLow(supRankStru(i)),(nrefInstHigh(supRankStru(i))-1)
              ! Instance i1 is the first time it appears?
              if( RepeatInst(i1) .eq. 0 ) then
                !
                ! Adds to the sum of the real number of instances of structure supRankStru(i)
                OutRevStructReal(supRankStru(i)) = OutRevStructReal(supRankStru(i)) + 1
                !
                do 100 i2=(i1+1),nrefInstHigh(supRankStru(i))
                  !
                  ! Checks if instance i2 was still not found has being repeated
                  if( RepeatInst(i2) .ne. 0 ) goto 100
                  !
                  ! Test if the instances have an equal sum of the reference numbers of the metabolites
                  if( SumMetInst(i1) .ne. SumMetInst(i2) ) goto 100
                  !
                  ! Test if the instances have an equal sum of the reference numbers of the reactions
                  if( SumReactInst(i1) .ne. SumReactInst(i2) ) goto 100
                  !
                  ! The instances can be compared
                  !
                  ! The metabolites of instance i2 are equal to the ones of instance i1?
                  do iMet=1,NumNodes(1,supRankStru(i))  ! Number of metabolites
                    if( MetInstOrdRefNum(iMet,i1) .ne. MetInstOrdRefNum(iMet,i2) ) goto 100  ! The instances are
                  end do                                              ! not the same; tests the next instance i2
                  !
                  ! The reactions of instance i2 are equal to the ones of instance i1?
                  do iReact=1,NumNodes(2,supRankStru(i))  ! Number of reactions
                    if( ReactInstOrdRefNum(iReact,i1) .ne. ReactInstOrdRefNum(iReact,i2) ) goto 100  ! The instances
                  end do                                              ! are not the same; tests the next instance i2
                  !
                  ! Instance i2 is equal to instance i1
                  RepeatInst(i2) = 1
                  !
100             continue
              end if
            end do
            !
            ! Instance nrefInstHigh(supRankStru(i)) is the first time it appears?
            if( RepeatInst(nrefInstHigh(supRankStru(i))) .eq. 0 ) then
              !
              ! Adds to the sum of the real number of instances of structure supRankStru(i)
              OutRevStructReal(supRankStru(i)) = OutRevStructReal(supRankStru(i)) + 1
              !
            end if
          else
            ! Structure supRankStru(i) has only 1 instance
            !
            ! Adds to the sum of the real number of instances of structure supRankStru(i)
            OutRevStructReal(supRankStru(i)) = OutRevStructReal(supRankStru(i)) + 1
            !
          end if
        end if
        ! Stores the ID number of the thread used
        ID = omp_get_thread_num()
        write(*,1003) supRankStru(i), i, OutRevStruct(supRankStru(i)), OutRevStructReal(supRankStru(i)), ID
      end do
!     END PARALLELIZATION
      !
      ! Determining the real number of instances, numInstReal
      numInstReal = 0
      do i1=1,numStructures
        numInstReal = numInstReal + OutRevStructReal(i1)
      end do
      write(*,*)
      write(*,*) 'Found the real number of instances:'
      write(*,*) numInstReal
      write(*,*)
      !
      ! Deallocate variables MetInstOrdRefNum(i2,i1), ReactInstOrdRefNum(i2,i1), SumMetInst(i1)
      ! and SumReactInst(i1)
      deallocate(MetInstOrdRefNum,ReactInstOrdRefNum,SumMetInst,SumReactInst)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      ORGANIZE THE INSTANCES           !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Obtains the minimum and maximum reference numbers of the instances of each structure
!      in organized matrices OrgMetInstReal(i2,i1) and OrgReactInstReal(i2,i1): stores first instances
!      of structures of lower reference numbers
      write(*,*) 'Starts organizing the instances'
      write(*,*)
      const(1) = 0
      do i1=1,numStructures
        ! Structure i1 has at least one instance?
        if( OutRevStructReal(i1) .ne. 0 ) then
          nrefInstLowReal(i1) = const(1) + 1
          nrefInstHighReal(i1) = const(1) + OutRevStructReal(i1)
          const(1) = nrefInstHighReal(i1)
        end if
      end do


!      Organizes the instances into matrices OrgMetInstReal(i2,i1) and OrgReactInstReal(i2,i1)
!      according to their structure: stores first instances of structures of lower reference numbers
      ! Allocates variables OrgMetInstReal(i2,i1) and OrgReactInstReal(i2,i1)
      allocate(OrgMetInstReal(maxNumMet,numInstReal),OrgReactInstReal(maxNumReact,numInstReal))
      ! Turns to zero the support vector supOrgInst(i1)
      do i1=1,numStructures
        supOrgInst(i1) = 0
      end do
      do i=1,numStructures
        ! Structure i has at least 1 instance?
        if( OutRevStructReal(i) .ne. 0) then
          do i1=nrefInstLow(i),nrefInstHigh(i)
            ! Is instance i1 not a repeated instance of a lower reference number?
            if( RepeatInst(i1) .eq. 0 ) then
              !
              ! Stores the metabolites of instance i1
              do iMet=1,maxNumMet
                OrgMetInstReal(iMet, (nrefInstLowReal(i) + supOrgInst(i)) ) = &
                OrgMetInst(iMet,i1)
              end do
              !
              ! Stores the reactions of instance i1
              do iReact=1,maxNumReact
                OrgReactInstReal(iReact, (nrefInstLowReal(i) + supOrgInst(i)) ) = &
                OrgReactInst(iReact,i1)
              end do
              supOrgInst(i) = supOrgInst(i) + 1
              !
            end if
          end do
        end if
      end do
      ! Deallocates the variables OrgMetInst(i2,i1), OrgReactInst(i2,i1) and RepeatInst(i1)
      deallocate(OrgMetInst,OrgReactInst,RepeatInst)
      write(*,*) 'Instances organized by structure'


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Stores the real number of instances of the structures. File:
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='OutRevStructReal.txt')
      do i1=1,numStructures
        write(8,1001) OutRevStructReal(i1)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Stores the lower and higher reference numbers of instances of a given structure.
!      The instances are stored according to their structure: first the instances of
!      structures of lower reference numbers. File:
!      "numRefInstLowHigh.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='numRefInstLowHigh.txt')
      do i1=1,numStructures
        ! Structure i1 has at least one instance?
        if( OutRevStructReal(i1) .ne. 0 )then
          write(8,1001) nrefInstLowReal(i1), nrefInstHighReal(i1)
        else
          write(8,1001) 0, 0
        end if
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Stores the real number of instances of the considered scope. File:
!      "numInstancesReal.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='numInstancesReal.txt')
      write(8,1001) numInstReal
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Stores the reference numbers of the metabolites of each instance. The instances
!      are organized according to their structure: instances of structures of lower
!      reference numbers are stored first. File:
!      "OrgMetInst.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='OrgMetInst.txt')
      do i1=1,numInstReal
        write(8,1002) (OrgMetInstReal(i2,i1), i2=1,maxNumMet)
      end do
      close(8)
      ! Deallocate the matrix OrgMetInstReal(i2,i1)
      deallocate(OrgMetInstReal)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Stores the reference numbers of the reactions of each instance. The instances
!      are organized according to their structure: instances of structures of lower
!      reference numbers are stored first. File:
!      "OrgReactInst.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='OrgReactInst.txt')
      do i1=1,numInstReal
        write(8,1002) (OrgReactInstReal(i2,i1), i2=1,maxNumReact)
      end do
      close(8)
      ! Deallocate the matrix OrgReactInstReal(i2,i1)
      deallocate(OrgReactInstReal)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Formats
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
1001  format (2(i10, ','))
1002  format (20(i5, ','))
1003  format ('Nstru=', i5, ';  Nrank=', i5, ';  Ninst++=', i10 ';  Ninst=', i10, ';  Nthread=', i5)

      end program OrgInst

