!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! TAMES  Topological Analysis of Metabolic Structures                 !
! Copyright (C) 2017                                                   !
! Paulo A. N. Dias, paulo.dias29@gmail.com                             !
! Marco Seabra dos Reis, marco@eq.uc.pt                                !
! Armindo Salvador, armindo.salvador@gmail.com                         !
! Pedro Martins, pmartins@iscac.pt                                     !
!                                                                      !
! This file is part of TAMES.                                          !
!                                                                      !
! TAMES is free software: you can redistribute it and/or modify        !
! it under the terms of the GNU General Public License as published by !
! the Free Software Foundation, either version 3 of the License, or    !
! any later version.                                                   !
!                                                                      !
! You should have received a copy of the GNU General Public License    !
! along with TAMES.  If not, see <http://www.gnu.org/licenses/>.       !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                         !
! Program "Preferential Paths":                           !
! - Testing the existence of preferential paths within    !
!   the network of directly related structures. The       !
!   developed algorithm should answer to the following    !
!   question: frequent Parents lead to frequent Childs?   !
!                                                         !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      Program PrefPaths

! CONSTANTS
!      repeat : logical variable used to determine the number of reactions and of metabolites in the files
!      maxNumGen : maximum number of generations considered to generate structures
!      numStructures : number of structures for the considered generations (maxNumGen)
!      freqType : column number of matrix ResultsRevStruct(i2,i1) where the type of frequency ("naked",
!                 "embedded", ...) that one intend to use to obtain the preferential paths is stored
!
!
! VECTORS
!      nrefStruLow(i1) : lower reference number of a structure of the i1 generation
!      nrefStruHigh(i1) : higher reference number of a structure of the i1 generation
!      const(i1) : support vector with the dimension integer*4
!
!
! MATRICES
!      DataRevStru(i2,i1) : i1) reference number of the structure. i2) addtional data for each structure: (i2=1) number
!                           of generation; (i2=2) number of metabolites; (i2=3) number of reactions; (i2=4) number of
!                           reversible reactions; (i2=5) number of Ancestors; (i2=6) number of Parents; (i2=7) number of
!                           Childs; (i2=8) number of Offspring; (i2=9) number of interconnections; (i2=10) maximum possible
!                           number of interconnections; (i2=11) number of connections metabolite/reaction (reversibility not
!                           considered); (i2=12) maximum possible number of connections metabolite/reaction (reversibility
!                           not considered)
!      FamRevStru(i2,i1) : matrix with the family relationships Parent/Child. i1) Parent structure; i2) Child structure.
!                          If two structures are not directly related, 0 is introduced in the respective cell; if they are
!                          directly related, a number corresponding to the strutural component in which they differ is
!                          introduced in the respective cell: 1 => new reactant/product to a irreversible reaction (internal
!                          connection), 2 => new reactant/product to a reversible reaction (internal connection),
!                          3 => new reaction, 4 => new reversible reaction, 5 => new metabolite (reactant or product) of a
!                          irreversible reaction, 6 => new metabolite (reactant or product) of a reversible reaction,
!                          7 => new reaction and metabolite
!      ResultsRevStruct(i2,i1) : output data obtained for the structure i1 with the developed code. (i2=1) "naked" frequency (number of
!                                instances); (i2=2) "embedded" frequency; (i2=3) n of stand-alone instances of type i - 0 metabolites
!                                and 0 reactions shared with other instances of the same structure; (i2=4) n of stand-alone instances
!                                of type ii - 1 metabolite and 0 reactions shared with other instances of the same structure; (i2=5) total
!                                n stand-alone instances - type i + ii; (i2=6) n of different metabolites; (i2=7) n of different reactions
!      RankStru(i2,i1) : stores the rank of the prevalence of structure i1: (i2=1) absolute rank; (i2=2) generation rank
!      PrefPath(i2,i1) : preferential path starting from structure i1. General information of structure i1 is stored in the first columns:
!                        (i2=1) n of generation of structure i1; (i2=2) "naked" frequency"; (i2=3) overall ranking; (i2=4) generation
!                        ranking. The following columns stores information of the most frequent descendents of structure i1. So, for a
!                        descendent X of structure i1, we define the following variables needed to determine the position in matrix:
!                        a = 4 + [(n generation structure X - n generation structure i1 - 1) x 5]. For a structure X, the following
!                        information is stored: (i2 = a + 1) n of generation; (i2 = a + 2) n of structure; (i2 = a + 3) "naked"
!                        frequency; (i2 = a + 4) overall rank; (i2 = a + 5) generation rank

      implicit none

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      DECLARATION OF THE VARIABLES     !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!      Static arrays: declaration
      ! Integer: 1 bit
      integer*1 :: FamRevStru(4000,4000), maxNumGen, freqType
      ! Integer: 2 bits
      integer*2 :: DataRevStru(12,4000), in_nGen_nStru(2,15), RankStru(2,4000)
      integer*2 :: nrefStruLow(15), nrefStruHigh(15)
      ! Integer: 4 bits
      integer*4 :: ResultsRevStruct(7,4000), PrefPath(74,4000), const(10)
      integer*4 :: numStructures
      integer*4 :: i1, i2
      ! Logical
      logical :: repeat


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      INPUT FILES                      !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Reads the information regarding the possible structures till the number of generations
!      consideration (5 generations)
     !
    !  Read the file "in_nGen_nStru.txt"
      const(1) = 0
      ! Logical variable. The cycle is performed while the variable is "true"
      repeat = .true.
      open (1,file='in_nGen_nStru.txt')
      do while (repeat)
        const(1) = const(1) + 1
        read(1,*) (in_nGen_nStru(i1,const(1)), i1=1,2)
        if( const(1) .eq. (in_nGen_nStru(1,1)+1) ) then
          repeat= .false.
        end if
      end do
      close(1)
      const(1) = 0
      maxNumGen = in_nGen_nStru(1,1)
      numStructures = in_nGen_nStru(2,1)
      do i1=2,(maxNumGen+1)
        nrefStruLow(i1-1) = in_nGen_nStru(1,i1)
        nrefStruHigh(i1-1) = in_nGen_nStru(2,i1)
      end do
     !
    !  Read the file 'Structures_General.txt', which stores the general information
    !  regarding the structures (no restrictions)
      open (1,file='Structures_General.txt')
      do i1=1,numStructures
        read(1,*) (DataRevStru(i2,i1), i2=1,12)
      end do
      close(1)
     !
    !  Read the file 'Family_Structures.txt' which stores the family relationships Parent-Child between
    ! the structures
      open (1,file='Family_Structures.txt')
      do i1=1,numStructures
        read(1,*) (FamRevStru(i2,i1), i2=1,numStructures)
      end do
      close(1)


!      Reads the information regarding the structures in the metabolic network: "naked" frequency,
!      "embedded" frequency, ...
      open (1,file='ResultsRevStruct.txt')
      do i1=1,numStructures
        read(1,*) (ResultsRevStruct(i2,i1), i2=1,7)
      end do
      close(1)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      STRUCTURES RANKING               !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Determines the ranking of the prevalent for the structures
      ! Type of frequency used to obtain the rankning
      freqType = 1  ! "Naked" frequency
      !
      ! Calls the subroutine rankOvGen
      call rankOvGen(ResultsRevStruct, nrefStruLow, nrefStruHigh,&
      maxNumGen, numStructures, freqType, RankStru)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      PREFERENTIAL PATHS               !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Determines the preferential paths by starting from a given
!      structure
      ! Type of frequency used to obtain the preferential paths: "Naked" frequency
      ! (freqType = 1)
      !
      ! Calls the subroutine preferentialPaths
      call preferentialPaths(maxNumGen, numStructures, DataRevStru, ResultsRevStruct, &
      nrefStruLow, nrefStruHigh, FamRevStru, RankStru, freqType, PrefPath)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      OUTPUT FILES                     !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the prevalence rank (absolute and per generation) of the structure to file:
!      "RankStru.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='RankStru.txt')
      do i1=1,numStructures
        write(8,1001) (RankStru(i2,i1), i2=1,2)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information of the most frequent path by starting from a given
!      structure to file:
!      'PrefPath.txt'
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='PrefPath.txt')
      do i1=1,numStructures
        write(8,1001) (PrefPath(i2,i1), i2=1,(4 + ((maxNumGen - 1)*5)))
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Formats
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
1001  format (100(i9, ','))


      contains


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!     Subroutine to obtain the rank of prevalence ("naked" or " embedded" frequency,
!!     for example) of the structures
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine rankOvGen(ResultsRevStruct_, nrefStruLow_, nrefStruHigh_,&
      maxNumGen_, numStructures_, freqType_, RankStru_)

! CONSTANTS
!      maxNumGen_ : maximum number of generations considered to generate structures
!      numStructures_ : number of structures for the considered generations (maxNumGen_)
!      freqType_ : column number of matrix ResultsRevStruct_(i2,i1) where the type of frequency ("naked", "embedded", ...)
!                 that one intend to use to obtain the preferential paths is stored
!
!
! VECTORS
!      nrefStruLow_(i1) : lower reference number of a structure of the i1 generation
!      nrefStruHigh_(i1) : higher reference number of a structure of the i1 generation
!      supRank_(i1) : support vector used to determine the ranking of prevalence of the structures: (i1=1) n of the structure;
!                     (i1=2) frequency ("naked", "embedded", ...) of the structure
!
!
! MATRICES
!      ResultsRevStruct_(i2,i1) : output data obtained for the structure i1 with the developed code. (i2=1) "naked" frequency (number of
!                            instances); (i2=2) "embedded" frequency; (i2=3) n of stand-alone instances of type i - 0 metabolites
!                            and 0 reactions shared with other instances of the same structure; (i2=4) n of stand-alone instances
!                            of type ii - 1 metabolite and 0 reactions shared with other instances of the same structure; (i2=5) total
!                            n stand-alone instances - type i + ii; (i2=6) n of different metabolites; (i2=7) n of different reactions
!      RankStru_(i2,i1) : stores the rank of the prevalence of structure i1: (i2=1) absolute rank; (i2=2) generation rank

      implicit none

      ! IN variables
       ! Integer: 1 bit
      integer*1, intent(in) :: maxNumGen_, freqType_
      integer*2, intent(in) :: nrefStruLow_(:), nrefStruHigh_(:)
       ! Integer: 4 bits
      integer*4, intent(in) :: ResultsRevStruct_(:,:), numStructures_

      ! OUT variables
       ! Integer: 2 bits
      integer*2, intent(inout) :: RankStru_(:,:)

      !Internal support variables
      integer*4 :: supRank_(2)

      ! General variables
      integer*4 :: iR1_, iR2_, iR3_

!      Determines the ranking of the prevalent for the structures
      ! Creates the matrix RankStru_(i2,i1)
      do iR1_=1,numStructures_
        do iR2_=1,2
          RankStru_(iR2_,iR1_) = 0
        end do
      end do
      !
      ! Determines the overall ranking of prevalence
      do iR1_=1,numStructures_
        !
        ! Restores supRank_(i1) to "0"
        supRank_(1) = 0
        supRank_(2) = 0
        !
        ! Finds the structure iR2_ with the rank of prevalence iR1_
        do iR2_=1,numStructures_
          ! Checks if the structure is still not ranked
          if( RankStru_(1,iR2_) .eq. 0 ) then
            ! Structure iR2_ has a frequency higher than the value stored in supRank_(2)?
            if( (supRank_(1) .eq. 0) .or. &
            (ResultsRevStruct_(freqType_,iR2_) .gt. supRank_(2)) ) then
              ! Stores the n of structure iR2_ and its frequency
              supRank_(1) = iR2_
              supRank_(2) = ResultsRevStruct_(freqType_,iR2_)
            end if
          end if
        end do
        !
        ! Stores the ranking i for structure supRank_(1)
        RankStru_(1,supRank_(1)) = iR1_
        !
      end do
     !
     !
      ! Rank of prevalence per generation
      ! Performs the loop for all the considered generations
      do iR1_=1,maxNumGen_
        !
        ! Performs the loop for the structures of generation iR1_
        do iR2_=1,(nrefStruHigh_(iR1_) - nrefStruLow_(iR1_) + 1)
          !
          ! Restores supRank_(i1) to "0"
          supRank_(1) = 0
          supRank_(2) = 0
          !
          ! Finds the structure iR3_ with the rank iR2_
          do iR3_=nrefStruLow_(iR1_),nrefStruHigh_(iR1_)
            ! Checks if the structure iR3_ is still not ranked
            if( RankStru_(2,iR3_) .eq. 0 ) then
              ! Structure iR3_ has a frequency higher than the value stored in supRank_(2)?
              if( (supRank_(1) .eq. 0) .or. &
              (ResultsRevStruct_(freqType_,iR3_) .gt. supRank_(2)) ) then
                ! Stores the n of structure iR3_ and its frequency
                supRank_(1) = iR3_
                supRank_(2) = ResultsRevStruct_(freqType_,iR3_)
              end if
            end if
          end do
          !
          ! Stores the ranking iR1_ for structure supRank_(1)
          RankStru_(2,supRank_(1)) = iR2_
          !
        end do
      end do

      end subroutine rankOvGen


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!     Subroutine to obtain the preferential paths in the network of directly
!!     related structures by starting at a given structure
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine preferentialPaths(maxNumGe_, numStructure_, DataRevStru_, ResultsRevStruc_, &
      nrefStruLower_, nrefStruHigher_, FamRevStru_, RankStr_, freqTyp_, PrefPath_)

! CONSTANTS
!      maxNumGe_ : maximum number of generations considered to generate structures
!      numStructure_ : number of structures for the considered generations (maxNumGe_)
!      freqTyp_ : column number of matrix ResultsRevStruc_(i2,i1) where the type of frequency ("naked", "embedded", ...)
!                 that one intend to use to obtain the preferential paths is stored
!
!
! VECTORS
!      nrefStruLower_(i1) : lower reference number of a structure of the i1 generation
!      nrefStruHigher_(i1) : higher reference number of a structure of the i1 generation
!      supp_(i1) : support vector used to determine the preferential paths
!
!
! MATRICES
!      DataRevStru_(i2,i1) : i1) reference number of the structure. i2) addtional data for each structure: (i2=1) number
!                            of generation; (i2=2) number of metabolites; (i2=3) number of reactions; (i2=4) number of
!                            reversible reactions; (i2=5) number of Ancestors; (i2=6) number of Parents; (i2=7) number of
!                            Childs; (i2=8) number of Offspring; (i2=9) number of interconnections; (i2=10) maximum possible
!                            number of interconnections; (i2=11) number of connections metabolite/reaction (reversibility not
!                            considered); (i2=12) maximum possible number of connections metabolite/reaction (reversibility
!                            not considered)
!      ResultsRevStruc_(i2,i1) : output data obtained for the structure i1 with the developed code. (i2=1) "naked" frequency (number of
!                                instances); (i2=2) "embedded" frequency; (i2=3) n of stand-alone instances of type i - 0 metabolites
!                                and 0 reactions shared with other instances of the same structure; (i2=4) n of stand-alone instances
!                                of type ii - 1 metabolite and 0 reactions shared with other instances of the same structure; (i2=5) total
!                                n stand-alone instances - type i + ii; (i2=6) n of different metabolites; (i2=7) n of different reactions
!      RankStr_(i2,i1) : stores the rank of the prevalence of structure i1: (i2=1) absolute rank; (i2=2) generation rank
!      FamRevStru_(i2,i1) : matrix with the family relationships Parent/Child. i1) Parent structure; i2) Child structure.
!                           If two structures are not directly related, 0 is introduced in the respective cell; if they are
!                           directly related, a number corresponding to the strutural component in which they differ is
!                           introduced in the respective cell: 1 => new reactant/product to a irreversible reaction (internal
!                           connection), 2 => new reactant/product to a reversible reaction (internal connection),
!                           3 => new reaction, 4 => new reversible reaction, 5 => new metabolite (reactant or product) of a
!                           irreversible reaction, 6 => new metabolite (reactant or product) of a reversible reaction,
!                           7 => new reaction and metabolite
!      PrefPath_(i2,i1) : preferential path starting from structure i1. General information of structure i1 is stored in the first columns:
!                         (i2=1) n of generation of structure i1; (i2=2) "naked" frequency"; (i2=3) overall ranking; (i2=4) generation
!                         ranking. The following columns stores information of the most frequent descendents of structure i1. So, for a
!                         descendent X of structure i1, we define the following variables needed to determine the position in matrix:
!                         a = 4 + [(n generation structure X - n generation structure i1 - 1) x 5]. For a structure X, the following
!                         information is stored: (i2 = a + 1) n of generation; (i2 = a + 2) n of structure; (i2 = a + 3) "naked"
!                         frequency; (i2 = a + 4) overall rank; (i2 = a + 5) generation rank

      implicit none

      ! IN variables
       ! Integer: 1 bit
      integer*1, intent(in) :: FamRevStru_(:,:), maxNumGe_, freqTyp_
       ! Integer: 2 bits
      integer*2, intent(in) :: DataRevStru_(:,:), RankStr_(:,:)
      integer*2, intent(in) :: nrefStruLower_(:), nrefStruHigher_(:)
       ! Integer: 4 bits
      integer*4, intent(in) :: ResultsRevStruc_(:,:), numStructure_

      ! OUT variables
       ! Integer: 4 bits
      integer*4, intent(inout) :: PrefPath_(:,:)

      !Internal support variables
       ! Integer: 4 bits
      integer*4 :: supp_(2)

      ! General variables
      integer*4 :: iP1_, iP2_, iP3_, iP4_

!      Determines the preferential paths by starting from a given structure
      ! Stores relevant information concerning structure iP1_
      do iP1_=1,numStructure_
        ! N of generation
        PrefPath_(1,iP1_) = DataRevStru_(1,iP1_)
        ! "Naked" frequency => freqTyp_ = 1
        PrefPath_(2,iP1_) = ResultsRevStruc_(freqTyp_,iP1_)
        ! Overall rank
        PrefPath_(3,iP1_) = RankStr_(1,iP1_)
        ! Generation rank
        PrefPath_(4,iP1_) = RankStr_(2,iP1_)
      end do
      !
      ! Determines the preferential paths by starting with the structures of
      ! the penultimate generation
      do iP1_=1,(maxNumGe_-1)
        do iP2_=nrefStruLower_(maxNumGe_-iP1_),nrefStruHigher_(maxNumGe_-iP1_)
          !
          ! Turns to "0" the support constant supp_(j), used to store:
          supp_(1) = 0  ! Generation rank of the Child stucture
          supp_(2) = 0  ! Number of the Child Structure
          !
          ! Checks the Childs of structure iP2_
          do iP3_=nrefStruLower_(maxNumGe_-iP1_+1),nrefStruHigher_(maxNumGe_-iP1_+1)
            ! Structure iP3_ is a Child of structure iP2_?
            if(FamRevStru_(iP3_,iP2_) .ne. 0) then
              ! The generation rank of structure iP3_ is higher than the value
              ! stored in supp_(1)?
              if( (supp_(1) .eq. 0) .or. &
              (RankStr_(2,iP3_) .lt. supp_(1)) ) then
                supp_(1) = RankStr_(2,iP3_)
                supp_(2) = iP3_
              end if
            end if
          end do
          !
          ! Stores the relevant information of the most prevalent Child of structure iP2_
          ! N of generation
          PrefPath_(5,iP2_) = DataRevStru_(1,supp_(2))
          ! N of the Child structure supp_(2)
          PrefPath_(6,iP2_) = supp_(2)
          ! "Naked" frequency => freqTyp_ = 1
          PrefPath_(7,iP2_) = ResultsRevStruc_(freqTyp_,supp_(2))
          ! Overall rank
          PrefPath_(8,iP2_) = RankStr_(1,supp_(2))
          ! Generation rank
          PrefPath_(9,iP2_) = RankStr_(2,supp_(2))
          !
          ! Stores the most relevant information of the most prevalent path of the
          ! Child structure supp_(2)
          do iP4_=1,((maxNumGe_-(maxNumGe_-iP1_+1))*5)
            PrefPath_((9+iP4_),iP2_) = PrefPath_((4+iP4_),supp_(2))
          end do
        end do
      end do

      end subroutine preferentialPaths

      end program PrefPaths
