!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! TAMES  Topological Analysis of Metabolic Structures                 !
! Copyright (C) 2017                                                   !
! Paulo A. N. Dias, paulo.dias29@gmail.com                             !
! Marco Seabra dos Reis, marco@eq.uc.pt                                !
! Armindo Salvador, armindo.salvador@gmail.com                         !
! Pedro Martins, pmartins@iscac.pt                                     !
!                                                                      !
! This file is part of TAMES.                                          !
!                                                                      !
! TAMES is free software: you can redistribute it and/or modify        !
! it under the terms of the GNU General Public License as published by !
! the Free Software Foundation, either version 3 of the License, or    !
! any later version.                                                   !
!                                                                      !
! You should have received a copy of the GNU General Public License    !
! along with TAMES.  If not, see <http://www.gnu.org/licenses/>.       !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                         !
! Program "Children Information":                         !
! - Determining the entropy of Shannon for the network of !
!   directly related structures, in which the information !
!   is given by the ratio, RfN, of the naked frequency of !
!   the children of a given Parent structure,             !
!     RfN = fN/SumfNch, where fN is the naked frequency   !
!           of a given Child, and SumfNch is the sum of   !
!           the naked frequencies of all the children of  !
!           a given Parent structure                      !
!                                                         !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

program ShannonEntropy

! CONSTANTS
!      maxNumGen : maximum number of generations considered to generate structures
!      numStructures : number of structures for the considered generations (maxNumGen)
!      repeat : logical variable used to obtain the number of generations and of structures from the files
!
!
! VECTORS
!      nrefStruLow(i1) : lower reference number of a structure of the i1 generation
!      nrefStruHigh(i1) : higher reference number of a structure of the i1 generation
!      fN(i1) : naked frequency (number of instances) of the i1 structure
!      const(i1) : support vector of size integer*4
!
!
! MATRICES
!      DataRevStru(i2,i1) : i1) reference number of the structure. i2) addtional data for each structure: (i2=1) number
!                           of generation; (i2=2) number of metabolites; (i2=3) number of reactions; (i2=4) number of
!                           reversible reactions; (i2=5) number of Ancestors; (i2=6) number of Parents; (i2=7) number of
!                           Childs; (i2=8) number of Offspring; (i2=9) number of interconnections; (i2=10) maximum possible
!                           number of interconnections; (i2=11) number of connections metabolite/reaction (reversibility not
!                           considered); (i2=12) maximum possible number of connections metabolite/reaction (reversibility
!                           not considered)
!      FamRevStru(i2,i1) : matrix with the family relationships Parent/Child. i1) Parent structure; i2) Child structure.
!                          If two structures are not directly related, 0 is introduced in the respective cell; if they are
!                          directly related, a number corresponding to the strutural component in which they differ is
!                          introduced in the respective cell: 1 => new reactant/product to a irreversible reaction (internal
!                          connection), 2 => new reactant/product to a reversible reaction (internal connection),
!                          3 => new reaction, 4 => new reversible reaction, 5 => new metabolite (reactant or product) of a
!                          irreversible reaction, 6 => new metabolite (reactant or product) of a reversible reaction,
!                          7 => new reaction and metabolite
!      in_nGen_nStru(i2,i1) : stores temporarily information regarding the number of generations and
!                             of strutures. In the first line, (i1=1), stores: (i2=1, i1=1) number of
!                             generations; (i2=2, i1=1) number of structures. From the second line, (i1=2),
!                             to the line maxNumGen + 1, stores the minimum and the maximum reference number
!                             of a structure that is part of generation 1 to generation maxNumGen: (i2=1),
!                             minimum reference number; (i2=2), maximum reference number
!      ResultsRevStruct(i2,i1) : output data obtained for the structure i1 with the developed code. (i2=1) "naked" frequency (number of
!                                instances); (i2=2) "embedded" frequency; (i2=3) n of stand-alone instances of type i - 0 metabolites
!                                and 0 reactions shared with other instances of the same structure; (i2=4) n of stand-alone instances
!                                of type ii - 1 metabolite and 0 reactions shared with other instances of the same structure; (i2=5) total
!                                n stand-alone instances - type i + ii; (i2=6) n of different metabolites; (i2=7) n of different reactions
!      Nch(i2,i1) : number of children of structure i1 obtained by adding one of the five structural element (the number or numbers
!                   corresponding to the structural elements in the matrix with the family relationships Parent/Child, FamRevStru(i2,i1),
!                   is given): (i2=1), 1 and 2 => new reactant/product to a irreversible or reversible reaction (internal connection);
!                   (i2=2), 3 => new reaction; (i2=3), 4 => new reversible reaction; (i2=4), 5 and 6 => new metabolite (reactant or product)
!                   of a irreversible or reversible reaction; (i2=5), 7 => new reaction (irreversible or reversible) and new metabolite;
!                   (i2=6), all the considered five structural elements
!      Hmax(i2,i1) : maximum entropy of structure i1, obtained by -log2[1/Nch(i2,i1)] (results in bit) for each one of the five i2 type
!                    relationship Parent/Child considered. The number or numbers corresponding to the structural elements in the matrix
!                    with the family relationships Parent/Child, FamRevStru(i2,i1), is given: (i2=1), 1 and 2 => new reactant/product to
!                    a irreversible or reversible reaction (internal connection); (i2=2), 3 => new reaction; (i2=3), 4 => new reversible
!                    reaction; (i2=4), 5 and 6 => new metabolite (reactant or product) of a irreversible or reversible reaction; (i2=5),
!                    7 => new reaction (irreversible or reversible) and new metabolite; (i2=6), all the considered five structural elements
!      SumfNch(i2,i1) : sum of the naked frequencies of the Childs of Parent i1 for each one of the type i2 relationships Parent/Child
!                       considered. The number or numbers corresponding to the structural elements in the matrix with the family relationships
!                       Parent/Child, FamRevStru(i2,i1), is given: (i2=1), 1 and 2 => new reactant/product to a irreversible or reversible
!                       reaction (internal connection); (i2=2), 3 => new reaction; (i2=3), 4 => new reversible reaction; (i2=4), 5 and
!                       6 => new metabolite (reactant or product) of a irreversible or reversible reaction; (i2=5), 7 => new reaction
!                       (irreversible or reversible) and new metabolite; (i2=6), all the considered five structural elements
!      RfN(i3,i2,i1) : ratio of the naked frequency of a Child structure i2, fN(i2), divided by sum of the naked frequencies of the Childs
!                      of Parent i1, SumfNch, for each one of the considered type i3 relationships Parent/Child. The number or numbers
!                      corresponding to the structural elements in the matrix with the family relationships Parent/Child, FamRevStru(i2,i1),
!                      is given: (i3=1), 1 and 2 => new reactant/product to a irreversible or reversible reaction (internal connection);
!                      (i3=2), 3 => new reaction; (i3=3), 4 => new reversible reaction; (i3=4), 5 and 6 => new metabolite (reactant or
!                      product) of a irreversible or reversible reaction; (i3=5), 7 => new reaction (irreversible or reversible) and
!                      new metabolite; (i3=6), all the considered five structural elements
!      H(i2,i1) : entropy (results in bit) of structure i1 for each one of the five i2 type relantionship Parent/Child considered. This
!                 property is obtained by H(i1) = - Sum [RfN(j) x log2(RfN(j)) ], with j being a children of structure i1 with a relationship
!                 of type i2 and RfN(j) = fN(j) / SumfNch(i2,i1), where fN(j) is the naked frequency of Child j and SumfNch(i2,i1) is the
!                 sum of the naked frequencies of of the children of Parent i1 for a type i2 relantionship Parent/Child.
!                 The number or numbers corresponding to the structural elements in the matrix with the family relationships
!                 Parent/Child, FamRevStru(i2,i1), is given: (i2=1), 1 and 2 => new reactant/product to a irreversible or reversible reaction
!                 (internal connection); (i2=2), 3 => new reaction; (i2=3), 4 => new reversible reaction; (i2=4), 5 and 6 => new metabolite
!                 (reactant or product) of a irreversible or reversible reaction; (i2=5), 7 => new reaction (irreversible or reversible)
!                 and new metabolite; (i2=6), all the considered five structural elements
!      deltaH(i2,i1) : children information (results in bit) of structure i1 for each one of the five i2 type relationship Parent/Child considered. The
!                      value of deltaH(i1) is null if all the children are equally frequent, and high if the distribution of children frequencies is very
!                      uneven. This property is obtained by deltaH(i1) = Hmax(n) - H(i1) = -log2[1/Nch(i2,i1)] - ( - Sum [RfN(j) x log2(RfN(j)) ], with
!                      j being a children of structure i1 with a relationship of type i2 ). The number or numbers corresponding to the
!                      structural elements in the matrix with the family relationships Parent/Child, FamRevStru(i2,i1), is given:
!                      (i2=1), 1 and 2 => new reactant/product to a irreversible or reversible reaction (internal connection);
!                      (i2=2), 3 => new reaction; (i2=3), 4 => new reversible reaction; (i2=4), 5 and 6 => new metabolite (reactant
!                      or product) of a irreversible or reversible reaction; (i2=5), 7 => new reaction (irreversible or reversible)
!                      and new metabolite; (i2=6), all the considered five structural elements

      implicit none

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      DECLARATION OF THE VARIABLES     !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!      Allocatable arrays: declaration
       ! Integer: 2 bits
      integer*2, DIMENSION(:,:), allocatable :: in_nGen_nStru
       ! Logical
      logical, allocatable :: repeat

!      Static arrays: declaration
      ! Integer: 1 bit
      integer*1 :: FamRevStru(4000,4000), maxNumGen
      ! Integer: 2 bits
      integer*2 :: DataRevStru(12,4000), nrefStruLow(15), nrefStruHigh(15)
      ! Integer: 4 bits
      integer*4 :: ResultsRevStruct(7,4000), const(10), numStructures
      integer*4 :: i, i1, i2, i3
      ! Real: 4 bits
      real :: fN(4000), Nch(6,4000), Hmax(6,4000), SumfNch(6,4000)
      real :: RfN(6,4000,4000), H(6,4000), deltaH(6,4000)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      INPUT FILES                      !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Reads the files regarding the information obtained that is not dependent of the metabolic
!      network used
     !
    !  Read the file "in_nGen_nStru.txt", which stores the information regarding the possible
    !  structures till the number of generations considered (5 generations)
      ! Allocate the variable in_nGen_nStru(i2,i1)
      allocate(in_nGen_nStru(2,15), repeat)
      !
      const(1) = 0
      ! Logical variable. The cycle is performed while the variable is "true"
      repeat = .true.
      open (1,file='in_nGen_nStru.txt')
      do while (repeat)
        const(1) = const(1) + 1
        read(1,*) (in_nGen_nStru(i1,const(1)), i1=1,2)
        if( const(1) .eq. (in_nGen_nStru(1,1)+1) ) then
          repeat= .false.
        end if
      end do
      close(1)
      const(1) = 0
      maxNumGen = in_nGen_nStru(1,1)
      numStructures = in_nGen_nStru(2,1)
      do i1=2,(maxNumGen+1)
        nrefStruLow(i1-1) = in_nGen_nStru(1,i1)
        nrefStruHigh(i1-1) = in_nGen_nStru(2,i1)
      end do
      ! Deallocate the variables in_nGen_nStru(i2,i1) and repeat
      deallocate(in_nGen_nStru, repeat)
     !
    !  Read the file 'Structures_General.txt', which stores the general information
    !  regarding the structures (no restrictions)
      open (1,file='Structures_General.txt')
      do i1=1,numStructures
        read(1,*) (DataRevStru(i2,i1), i2=1,12)
      end do
      close(1)
     !
    !  Read the file 'Family_Structures.txt' which stores the family relationships Parent-Child between
    ! the structures
      open (1,file='Family_Structures.txt')
      do i1=1,numStructures
        read(1,*) (FamRevStru(i2,i1), i2=1,numStructures)
      end do
      close(1)


!      Reads the files regarding the information obtained in the metabolic network
     !
    !  Reads the information regarding the structures in the metabolic network: "naked" frequency,
    !  "embedded" frequency, ...
      open (1,file='ResultsRevStruct.txt')
      do i1=1,numStructures
        read(1,*) (ResultsRevStruct(i2,i1), i2=1,7)
      end do
      close(1)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      N OF CHILDREN / SUM OF FREQ.    !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Obtain the number of children for each type of relationship Parent/Child of a given Parent structure;
!      obtain the sum of naked frequencies of the Childs of a given Parent connected by a given type of
!      relationship Parent/Chlid
      ! Initializes the the variables Nch(i2,i1), Hmax(i2,i1), SumfNch(i2,i1), H(i2,i1) and deltaH(i2,i1).
      ! Obtain the naked frequency of each structure
      do i1=1,numStructures
        !
        ! Obtain the naked frequency of each structure
        fN(i1) = Real(ResultsRevStruct(1,i1))
        !
        do i2=1,6
          Nch(i2,i1) = Real(0)
          Hmax(i2,i1) = Real(0)
          SumfNch(i2,i1) = Real(0)
          H(i2,i1) = Real(0)
          deltaH(i2,i1) = Real(0)
        end do
        !
      end do
      !
      ! Obtains the number of children and the sum of the naked frequencies of the children for each type i2 of
      ! relationship Parent/Child of a given i1 Parent structure. The sum of the naked frequencies is obtained by:
      !   SumfNch(i2,i1) = Sum [fN(j)], with j being a children of structure i1 with a relationship of type i2
      do i=1,(maxNumGen-1)  ! From the first to the generation maxNumGen-1
        !
        do 100 i1=nrefStruLow(i),nrefStruHigh(i)  ! Checks all the structures i1 of generation i
          !
          ! Obtain the children of structure i1
          do 105 i2=nrefStruLow(i+1),nrefStruHigh(i+1)  ! Checks all the structures i2 of generation i+1
            !
            ! Structure i2 is a Child of structure i1?
            if( FamRevStru(i2,i1) .ne. 0 ) then
              !
              ! All five types of relationships PARENT/Child
              !  Obtain the total number of children of a given Parent structure
              Nch(6,i1) = Nch(6,i1) + Real(1)
              !
              !  Adds to the value of the sum of the naked frequencies of the children of Parent i1 the contribution
              !  of Child i2
              SumfNch(6,i1) = SumfNch(6,i1) + fN(i2)
              !
              !
              ! Check the type of relationship Parent/Child
              ! i) new reactant/product to a irreversible or reversible reaction (internal connection)?
              if( ( FamRevStru(i2,i1) .eq. 1 ) .or. ( FamRevStru(i2,i1) .eq. 2 ) ) then
                ! Add a new child
                Nch(1,i1) = Nch(1,i1) + Real(1)
                !
                ! Adds to the value of the sum of the naked frequencies of the children of Parent i1 the contribution
                ! of Child i2
                SumfNch(1,i1) = SumfNch(1,i1) + fN(i2)
                !
                ! All the children of structure i1 were found?
                if( Int(Nch(6,i1)) .eq. DataRevStru(7,i1) ) then
                  goto 100  ! Check the next Parent structure i1
                else
                  goto 105  ! Check the next structure i2
                end if
              end if
              ! ii) new reaction?
              if( FamRevStru(i2,i1) .eq. 3 ) then
                ! Add a new child
                Nch(2,i1) = Nch(2,i1) + Real(1)
                !
                ! Adds to the value of the sum of the naked frequencies of the children of Parent i1 the contribution
                ! of Child i2
                SumfNch(2,i1) = SumfNch(2,i1) + fN(i2)
                !
                ! All the children of structure i1 were found?
                if( Int(Nch(6,i1)) .eq. DataRevStru(7,i1) ) then
                  goto 100  ! Check the next Parent structure i1
                else
                  goto 105  ! Check the next structure i2
                end if
              end if
              ! iii) new reversible reaction?
              if( FamRevStru(i2,i1) .eq. 4 ) then
                ! Add a new child
                Nch(3,i1) = Nch(3,i1) + Real(1)
                !
                ! Adds to the value of the sum of the naked frequencies of the children of Parent i1 the contribution
                ! of Child i2
                SumfNch(3,i1) = SumfNch(3,i1) + fN(i2)
                !
                ! All the children of structure i1 were found?
                if( Int(Nch(6,i1)) .eq. DataRevStru(7,i1) ) then
                  goto 100  ! Check the next Parent structure i1
                else
                  goto 105  ! Check the next structure i2
                end if
              end if
              ! iv) new metabolite (reactant or product) of a irreversible or reversible reaction?
              if( (FamRevStru(i2,i1) .eq. 5) .or. (FamRevStru(i2,i1) .eq. 6) ) then
                ! Add a new child
                Nch(4,i1) = Nch(4,i1) + Real(1)
                !
                ! Adds to the value of the sum of the naked frequencies of the children of Parent i1 the contribution
                ! of Child i2
                SumfNch(4,i1) = SumfNch(4,i1) + fN(i2)
                !
                ! All the children of structure i1 were found?
                if( Int(Nch(6,i1)) .eq. DataRevStru(7,i1) ) then
                  goto 100  ! Check the next Parent structure i1
                else
                  goto 105  ! Check the next structure i2
                end if
              end if
              ! v) new reaction (irreversible or reversible) and new metabolite?
              if( FamRevStru(i2,i1) .eq. 7 ) then
                ! Add a new child
                Nch(5,i1) = Nch(5,i1) + Real(1)
                !
                ! Adds to the value of the sum of the naked frequencies of the children of Parent i1 the contribution
                ! of Child i2
                SumfNch(5,i1) = SumfNch(5,i1) + fN(i2)
                !
                ! All the children of structure i1 were found?
                if( Int(Nch(6,i1)) .eq. DataRevStru(7,i1) ) then
                  goto 100  ! Check the next Parent structure i1
                else
                  goto 105  ! Check the next structure i2
                end if
              end if
              !
            end if
105       continue
100     continue
      end do


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      MAXIMUM ENTROPY / RATIOS OF      !
!      FREQUENCIES / ENTROPY            !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Obtain the maximum entropy, Hmax, for each type of relationship Parent/Child of a given
!      Parent structure. Hmax = -log2(1/Nch), where Nch is the number of children of the Parent.
      !
      ! Determines Hmax for all the structures till the generation maxNumGen - 1
      do i1=1,nrefStruHigh(maxNumGen-1)  ! From 1 to maximum reference number of a structure of
        !                                ! generation maxNumGen - 1
        do i2=1,6  ! Determines Hmax for all the considered relationships Parent/Child: i2=1, internal
          ! connection; i2=2, new reaction; i2=3, new reversible reaction; i2=4, new metabolite;
          ! i2=5, new metabolite and new reaction; i2=6, all the relationshiphs
          !
          ! Structure i1 has at least one Child with the relationship i2?
          if( Nch(i2,i1) .ne. 0) then
            ! Structure i1 has at least one Child with the relationship i2
            !
            Hmax(i2,i1) = - ( alog(Real(1)/Nch(i2,i1))/alog(Real(2)) )  ! NOTE: log2(x) = ln(x)/ln(2)
            !
          end if
        end do
      end do


!      Obtain the ratios of frequencies, RfN, of the Childs of a given Parent for all the considered
!      relationships Parent/Child. The ratios of a given Child i2 are obtained as following:
!         RfN(i3,i2,i1) = fN(i2) / SumfNch(i3,i1), where fN(i2) is the naked frequency of Child i2 and
!          SumfNch(i3,i1) is the sum of the naked frequencies of the Childs that have a relationship
!          Parent/Child of type i3 with the Parent i1
!      Obtain the entropy, H, for a given Parent for all the considered types of relationships Parent/Child.
!      The information is given by the RfNs of the Child:
!         H(i3,i1) = - Sum [RfN(j) x log2(RfN(j)) ], with j being a children of structure i1 with a
!          relationship Parent/Child of type i3
      !
      ! Initializes the variable RfN(i3,i2,i1)
      do i1=1,numStructures
        do i2=1,numStructures
          do i3=1,6
            RfN(i3,i2,i1) = Real(0)
          end do
        end do
      end do
      !
      ! Determines the RfN of each Child i2 of Parent i1 for all the considered relationships Parent/Child i3
      do i=1,(maxNumGen-1)  ! From the first to the generation maxNumGen-1
        !
        do 110 i1=nrefStruLow(i),nrefStruHigh(i)  ! Checks all the Parent structures i1 of generation i
          !
          ! Turns to zero the support variable const(1), used to control the number of children of
          ! structure i1
          const(1) = 0
          !
          ! SumfNch(6,i1) not equal to zero? Is there at least one instance of at least one of the
          ! Childs of structure i1?
          if( Int(SumfNch(6,i1)) .ne. 0 ) then
            !
            ! Obtain the children of structure i1
            do 115 i2=nrefStruLow(i+1),nrefStruHigh(i+1)  ! Checks all the structures i2 of generation i+1
              !
              ! Structure i2 is a Child of structure i1?
              if( FamRevStru(i2,i1) .ne. 0 ) then
                !
                ! All the Relationships Parent/Child considered
                !  Adds 1 to the number of children of structure i1 found
                const(1) = const(1) + 1
                !
                ! Structure i2 has at least one instance?
                if( ResultsRevStruct(1,i2) .ne. 0 ) then
                  ! Determines the RfN by considering all the types of relationships Parent/Child
                  RfN(6,i2,i1) = fN(i2) / SumfNch(6,i1)
                  !
                  ! Adds to the total entropy of Parent i1
                  H(6,i1) = H(6,i1) - ( RfN(6,i2,i1) * ( alog(RfN(6,i2,i1)) / alog(Real(2)) ) )
                  !
                else
                  ! All the children of structure i1 were found?
                  if( const(1) .eq. DataRevStru(7,i1) ) then
                    goto 110  ! Check the next Parent structure i1
                  else
                    goto 115  ! Check the next structure i2
                  end if
                end if
                !
                !
                ! Check the type of relationship Parent/Child
                ! i) new reactant/product to a irreversible or reversible reaction (internal connection)?
                if( ( FamRevStru(i2,i1) .eq. 1 ) .or. ( FamRevStru(i2,i1) .eq. 2 ) ) then
                  !
                  ! Determines the RfN by considering the relationships Parent/Child of type i)
                  RfN(1,i2,i1) = fN(i2) / SumfNch(1,i1)
                  !
                  ! Adds to the total entropy of Parent i1
                  H(1,i1) = H(1,i1) - ( RfN(1,i2,i1) * ( alog(RfN(1,i2,i1)) / alog(Real(2)) ) )
                  !
                  ! All the children of structure i1 were found?
                  if( const(1) .eq. DataRevStru(7,i1) ) then
                    goto 110  ! Check the next Parent structure i1
                  else
                    goto 115  ! Check the next structure i2
                  end if
                end if
                ! ii) new reaction?
                if( FamRevStru(i2,i1) .eq. 3 ) then
                  !
                  ! Determines the RfN by considering the relationships Parent/Child of type ii)
                  RfN(2,i2,i1) = fN(i2) / SumfNch(2,i1)
                  !
                  ! Adds to the total entropy of Parent i1
                  H(2,i1) = H(2,i1) - ( RfN(2,i2,i1) * ( alog(RfN(2,i2,i1)) / alog(Real(2)) ) )
                  !
                  ! All the children of structure i1 were found?
                  if( const(1) .eq. DataRevStru(7,i1) ) then
                    goto 110  ! Check the next Parent structure i1
                  else
                    goto 115  ! Check the next structure i2
                  end if
                end if
                ! iii) new reversible reaction?
                if( FamRevStru(i2,i1) .eq. 4 ) then
                  !
                  ! Determines the RfN by considering the relationships Parent/Child of type iii)
                  RfN(3,i2,i1) = fN(i2) / SumfNch(3,i1)
                  !
                  ! Adds to the total entropy of Parent i1
                  H(3,i1) = H(3,i1) - ( RfN(3,i2,i1) * ( alog(RfN(3,i2,i1)) / alog(Real(2)) ) )
                  !
                  ! All the children of structure i1 were found?
                  if( const(1) .eq. DataRevStru(7,i1) ) then
                    goto 110  ! Check the next Parent structure i1
                  else
                    goto 115  ! Check the next structure i2
                  end if
                end if
                ! iv) new metabolite (reactant or product) of a irreversible or reversible reaction?
                if( (FamRevStru(i2,i1) .eq. 5) .or. (FamRevStru(i2,i1) .eq. 6) ) then
                  !
                  ! Determines the RfN by considering the relationships Parent/Child of type iv)
                  RfN(4,i2,i1) = fN(i2) / SumfNch(4,i1)
                  !
                  ! Adds to the total entropy of Parent i1
                  H(4,i1) = H(4,i1) - ( RfN(4,i2,i1) * ( alog(RfN(4,i2,i1)) / alog(Real(2)) ) )
                  !
                  ! All the children of structure i1 were found?
                  if( const(1) .eq. DataRevStru(7,i1) ) then
                    goto 110  ! Check the next Parent structure i1
                  else
                    goto 115  ! Check the next structure i2
                  end if
                end if
                ! v) new reaction (irreversible or reversible) and new metabolite?
                if( FamRevStru(i2,i1) .eq. 7 ) then
                  !
                  ! Determines the RfN by considering the relationships Parent/Child of type v)
                  RfN(5,i2,i1) = fN(i2) / SumfNch(5,i1)
                  !
                  ! Adds to the total entropy of Parent i1
                  H(5,i1) = H(5,i1) - ( RfN(5,i2,i1) * ( alog(RfN(5,i2,i1)) / alog(Real(2)) ) )
                  !
                  ! All the children of structure i1 were found?
                  if( const(1) .eq. DataRevStru(7,i1) ) then
                    goto 110  ! Check the next Parent structure i1
                  else
                    goto 115  ! Check the next structure i2
                  end if
                end if
              end if
115         continue
          end if
110     continue
      end do


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      CHILDREN INFORMATION             !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Obtain the children information (results in bit) of structure i1 for each one of the five i2 type relationship
!      Parent/Child considered. This property is obtained by:
!         deltaH = Hmax - H = -log2[1/Nch(i2,i1)] - ( - Sum [RfN(j) x log2(RfN(j)) ], with j being a children of
!          structure i1 with a relationship of type i2 )
      !
      ! Determines the children information for each one of the considered relationships Parent/child
      do i1=1,nrefStruHigh(maxNumGen-1)  ! Does the loop till the last structure of generation maxNumGen-1
        !                                ! (structures of the last generation do not have Childs)
        do i2=1,6  ! Does the loop to all the considered relationships Parent/child
          !
          ! SumfNch(i2,i1) not equal to zero? Is there at least one instance of at least one of the
          ! Childs of structure i1?
          if( Int(SumfNch(i2,i1)) .ne. 0 ) then
            ! Determines the cildren information, Hmax - H
            deltaH(i2,i1) = Hmax(i2,i1) - H(i2,i1)
            !
          end if
          !
        end do
      end do


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      OUTPUT FILES                     !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the general information regarding to the children information by
!      considering all the relationships Parent/Child. File:
!      "ChildrenInf_AllRel.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='ChildrenInf_AllRel.txt')
      do i1=1,numStructures
        !             N of Children; Hmax (bit);      SumfNch      ; H (bit); deltaH (bit)
        write(8,1001) Int(Nch(6,i1)), Hmax(6,i1), Int(SumfNch(6,i1)), H(6,i1), deltaH(6,i1)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the general information regarding to the children information by
!      considering relationships Parent/Child of type: i) new reactant/product to a
!      irreversible or reversible reaction (internal connection). File:
!      "ChildrenInf_I.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='ChildrenInf_I.txt')
      do i1=1,numStructures
        !             N of Children; Hmax (bit);      SumfNch      ; H (bit); deltaH (bit)
        write(8,1001) Int(Nch(1,i1)), Hmax(1,i1), Int(SumfNch(1,i1)), H(1,i1), deltaH(1,i1)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the general information regarding to the children information by
!      considering relationships Parent/Child of type: ii) new reaction. File:
!      "ChildrenInf_II.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='ChildrenInf_II.txt')
      do i1=1,numStructures
        !             N of Children; Hmax (bit);      SumfNch      ; H (bit); deltaH (bit)
        write(8,1001) Int(Nch(2,i1)), Hmax(2,i1), Int(SumfNch(2,i1)), H(2,i1), deltaH(2,i1)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the general information regarding to the children information by
!      considering relationships Parent/Child of type: iii) new reversible reaction.
!      File:
!      "ChildrenInf_III.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='ChildrenInf_III.txt')
      do i1=1,numStructures
        !             N of Children; Hmax (bit);      SumfNch      ; H (bit); deltaH (bit)
        write(8,1001) Int(Nch(3,i1)), Hmax(3,i1), Int(SumfNch(3,i1)), H(3,i1), deltaH(3,i1)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the general information regarding to the children information by
!      considering relationships Parent/Child of type: iv) new metabolite. File:
!      "ChildrenInf_IV.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='ChildrenInf_IV.txt')
      do i1=1,numStructures
        !             N of Children; Hmax (bit);      SumfNch      ; H (bit); deltaH (bit)
        write(8,1001) Int(Nch(4,i1)), Hmax(4,i1), Int(SumfNch(4,i1)), H(4,i1), deltaH(4,i1)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the general information regarding to the children information by
!      considering relationships Parent/Child of type: v) new reaction and new
!      metabolite. File:
!      "ChildrenInf_V.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='ChildrenInf_V.txt')
      do i1=1,numStructures
        !             N of Children; Hmax (bit);      SumfNch      ; H (bit); deltaH (bit)
        write(8,1001) Int(Nch(5,i1)), Hmax(5,i1), Int(SumfNch(5,i1)), H(5,i1), deltaH(5,i1)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the RfN of Child i2 of Parent i1 by considering all the relationships
!      Parent/Child. The ratio of frequencies is obtained by: RfN = fN/SumfNch.
!      File:
!      "RfN_AllRel.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='RfN_AllRel.txt')
      do i1=1,numStructures
        write(8,1002) (RfN(6,i2,i1), i2=1,numStructures)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the RfN of Child i2 of Parent i1 by considering the relationship
!      Parent/Child of type: i) new reactant/product to a irreversible or reversible
!      reaction (internal connection). The ratio of frequencies is obtained by:
!      RfN = fN/SumfNch. File:
!      "RfN_I.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='RfN_I.txt')
      do i1=1,numStructures
        write(8,1002) (RfN(1,i2,i1), i2=1,numStructures)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the RfN of Child i2 of Parent i1 by considering the relationship
!      Parent/Child of type: ii) new reaction. The ratio of frequencies is obtained
!      by: RfN = fN/SumfNch. File:
!      "RfN_II.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='RfN_II.txt')
      do i1=1,numStructures
        write(8,1002) (RfN(2,i2,i1), i2=1,numStructures)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the RfN of Child i2 of Parent i1 by considering the relationship
!      Parent/Child of type: iii) new reversible reaction. The ratio of frequencies
!      is obtained by: RfN = fN/SumfNch. File:
!      "RfN_III.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='RfN_III.txt')
      do i1=1,numStructures
        write(8,1002) (RfN(3,i2,i1), i2=1,numStructures)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the RfN of Child i2 of Parent i1 by considering the relationship
!      Parent/Child of type: iv) new metabolite. The ratio of frequencies is
!      obtained by: RfN = fN/SumfNch. File:
!      "RfN_IV.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='RfN_IV.txt')
      do i1=1,numStructures
        write(8,1002) (RfN(4,i2,i1), i2=1,numStructures)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the RfN of Child i2 of Parent i1 by considering the relationship
!      Parent/Child of type: new reaction and new metabolite. The ratio of frequencies
!      is obtained by: RfN = fN/SumfNch. File:
!      "RfN_V.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='RfN_V.txt')
      do i1=1,numStructures
        write(8,1002) (RfN(5,i2,i1), i2=1,numStructures)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Formats
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
1001  format ( i8, ',', f8.5, ',', i10, ',', f8.5, ',', f8.5)
1002  format (4000(f8.5, ','))


end program ShannonEntropy
