!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! TAMES  Topological Analysis of Metabolic Structures                 !
! Copyright (C) 2017                                                   !
! Paulo A. N. Dias, paulo.dias29@gmail.com                             !
! Marco Seabra dos Reis, marco@eq.uc.pt                                !
! Armindo Salvador, armindo.salvador@gmail.com                         !
! Pedro Martins, pmartins@iscac.pt                                     !
!                                                                      !
! This file is part of TAMES.                                          !
!                                                                      !
! TAMES is free software: you can redistribute it and/or modify        !
! it under the terms of the GNU General Public License as published by !
! the Free Software Foundation, either version 3 of the License, or    !
! any later version.                                                   !
!                                                                      !
! You should have received a copy of the GNU General Public License    !
! along with TAMES.  If not, see <http://www.gnu.org/licenses/>.       !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                         !
! Program "Structures Enumeration":                       !
! - Implementation of an algorithm to determine all the   !
!   possible structures within a given topological scope  !
!   by using a bottom-up approach: structural components  !
!   are successively added to a seed structure.           !
!                                                         !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      Program allstructures

! CONSTANTS
!      numGen : number ofthe generation
!      maxNumGen : maximum number of generations considered to generate structures
!      maxNumMet : support constant used to reduce the scope for a considered number of maxNumGen by defining the maximum
!                  number of metabolites of a structure
!      maxNumReact : support constant used to reduce the scope for a considered number of maxNumGen by defining the
!                    maximum number of metabolites of a structure
!      numPermutMet : number of permutations for the given set of metabolites
!      numPermutIrrReact : number of permutations for the given set of irreversible reactions
!      numPermutRevReact : number of permutations for the given set of reversible reactions
!      numPermDirInvReact : number of permutations between the direct and inverse single direction reactions for all
!                           the considered reactions (which can be reversible) and for each permutation of the reactions
!      numStructures : number of structures found for the considered maximum number of generations (maxNumGen)
!      isoStruRef : reference number of a structure in matrix RevStruct(i3,i2,i1) to which the structure in matrix
!                   supRevStruct(i2,i1,1) is a isomorphic representation. We store its reference number to construct the
!                   Parent/Child relationship between the structure that lead to the isomorphic representation shown in
!                   matrix supRevStruct(i2,i1,1) and structure isoStruRef
!      isoStruct : if the structure represented in matrix supRevStruct(i3,i2,1) is a isomorphic representation of a structure
!                  stored in matrix RevStruct(i3,i2,i1), isoStruct is filled with "1"; if it is a new structure, isoStruct
!                  is filled with "0"
!
!
! VECTORS
!      numStruGen(i1) : number of structures of the i1 generation
!      nrefStruLow(i1) : lower reference number of a structure of the i1 generation
!      nrefStruHigh(i1) : higher reference number of a structure of the i1 generation
!      nMet(i1) : number of metabolites involved per i generation
!      nReact(i1) : number of reactions involved per i generation
!      supReactRevStru(i1) : for each reaction i1 of a (possibly) new structure, gives the information regarding its
!                            reversibility. The reactions should be ordered, with the irreversible reactions comming
!                            first. If reaction i1 is reversible, the corresponding cell will be filled with 1; if it
!                            is irreversible, the cell will be filled with 0
!      supMetDegRevStruTemp_(i1) : gives the degree for each metabolite of a (possibly) new structure. i1=[(3 x (nmet - 1)) + 1],
!                                  total n of connections of the metabolite; i1=[(3 x (nmet - 1)) + 2], n of connections
!                                  comming to the metabolite (in); i1=[(3 x (nmet - 1)) + 3], n of connections going out of
!                                  the metabolite (out). The metabolitesmay not be organized according accoridng to their degree
!      supReactDegRevStruTemp_(i1) : gives the degree for each reaction of a possibly new structure. i1=[(3 x (nmet - 1)) + 1],
!                                    connections of the metabolite; i1=[(3 x (nmet - 1)) + 2], n of connections comming to
!                                    the total n ofmetabolite (in); i1=[(3 x (nmet - 1)) + 3], n of connections going out of
!                                    the metabolite (out). The reactions may not be organized according to its reversibility and
!                                    according to its degree
!      supMetDegRevStru(i1) : gives the degree for each metabolite of a (possibly) new structure. i1=[(3 x (nmet - 1)) + 1],
!                             total n of connections of the metabolite; i1=[(3 x (nmet - 1)) + 2], n of connections
!                             comming to the metabolite (in); i1=[(3 x (nmet - 1)) + 3], n of connections going out of
!                             the metabolite (out)
!      supReactDegRevStru(i1) : gives the degree for each reaction of a possibly new structure. The irreversible reactions
!                               are stored first, and then, in case at least one exists, the reversible reactions are stored.
!                               i1=[(3 x (nmet - 1)) + 1], connections of the metabolite; i1=[(3 x (nmet - 1)) + 2],
!                               n of connections comming to the total n ofmetabolite (in); i1=[(3 x (nmet - 1)) + 3],
!                               n of connections going out of the metabolite (out)
!      numNodes(i1) : number of nodes of the (possibly) new struture: (i1=1) number of metabolites; (i1=2) number of reactions;
!                     (i1=3) number of reversible reactions ; (i1=4) number of irreversible reactions
!      supReactRevStruTemp(i1) : information regarding the reversibility of the reactions of the (possibly) new structure. If a
!                                reaction is reversible, the corresponding cell is filled with 1; it it is irreversible, the
!                                corresponding cell is filled with 0. The reactions may not be ordered according to its reversibility
!      supReactRevStru(i1) : information regarding the reversibility of the reactions of the (possibly) new structure. If a
!                            reaction is reversible, the corresponding cell is filled with 1; it it is irreversible, the
!                            corresponding cell is filled with 0. The reactions are ordered according to the reversibility:
!                            irreversible reactions first, with the reversible reactions being stored after
!      const(i1) : support vector with the dimension integer*4
!      testcount(i1) : support vector to test if the code correctly obtained all the structures
!
!
! MATRICES
!      RevStruct(i3,i2,i1) : matrix with the possible structures using a bipartite and directed
!                            graph representation. i1) number of the structure; i2) number of the metabolite;
!                            i3) [4*(i3-1) + 1] reactant of the direct reaction, [4*(i3-1) + 2] product of
!                            the direct reaction, [4*(i3-1) + 3] reactant of the inverse reaction, [4*(i3-1) + 4]
!                            product of the inverse reaction. If a metabolite is connected to a reaction, the
!                            corresponding cell is filled with 1; otherwise, is filled with 0
!      supRevStructTemp(i3,i2,1) : support matrix to store temporarily a new structure. This structure will be reorganized
!                                  into matrix supRevStruct(i3,i2,1)
!      supRevStruct(i3,i2,1) : support matrix to store a new a (possibly) new structure. The pattern of this structure
!                              is ordered according to degree of the nodes (metabolites and reactions), and will be compared
!                              to others already found to determine if it is a isomorphic representation of one of them
!      DataRevStru(i2,i1) : i1) reference number of the structure. i2) addtional data for each structure: (i2=1) number
!                           of generation; (i2=2) number of metabolites; (i2=3) number of reactions; (i2=4) number of
!                           reversible reactions; (i2=5) number of Ancestors; (i2=6) number of Parents; (i2=7) number of
!                           Childs; (i2=8) number of Offspring; (i2=9) number of interconnections; (i2=10) maximum possible
!                           number of interconnections; (i2=11) number of connections metabolite/reaction (reversibility not
!                           considered); (i2=12) maximum possible number of connections metabolite/reaction (reversibility
!                           not considered)
!      ReactRevStru(i2,i1) : i1) reference number of the structure. i2) number of the reaction of the structure: if a
!                            a reaction is reversible, the corresponding cell is filled with 1, and if it is not, with 0
!      MetDegRevStru(i2,i1) : gives the degree for each metabolite of structure i1: i2=[(3 x (nmet - 1)) + 1], total n of
!                             connections of the metabolite; i2=[(3 x (nmet - 1)) + 2], n of connections comming to the
!                             metabolite (in); i2=[(3 x (nmet - 1)) + 3], n of connections going out of the metabolite (out)
!      ReactDegRevStru(i2,i1) : gives the degree for each reaction of structure i1. The irreversible reactions are stored
!                               first and then, in case at least one exists, the reversible reactions are stored.
!                               i2=[(3 x (nmet - 1)) + 1], connections of the reaction; i2=[(3 x (nmet - 1)) + 2],
!                               n of connections comming to the reaction (in); i2=[(3 x (nmet - 1)) + 3],
!                               n of connections going out of the reaction (out)
!      FamRevStru(i2,i1) : matrix with the family relationships Parent/Child. i1) Parent structure; i2) Child structure.
!                          If two structures are not directly related, 0 is introduced in the respective cell; if they are
!                          directly related, a number corresponding to the strutural component in which they differ is
!                          introduced in the respective cell: 1 => new reactant/product to a irreversible reaction (internal
!                          connection), 2 => new reactant/product to a reversible reaction (internal connection),
!                          3 => new reaction, 4 => new reversible reaction, 5 => new metabolite (reactant or product) of a
!                          irreversible reaction, 6 => new metabolite (reactant or product) of a reversible reaction,
!                          7 => new reaction and metabolite
!      InclRevStru(i2,i1) : inclusion matrix. If the structure i1 can be found within structure i2, the corresponding cell
!                           is filled with "1"; if not, the cell is filled with "0"
!      EquiMetRevStru(i2,i1) : information regarding the equivalent metabolites of a given structure i1. If two metabolites are
!                              equivalent, the corresponding cell is filled with "1"; if not, is filled with "0". The first
!                              metabolite is compared with the ones in higher positions, the second metabolite is compared to
!                              metabolites who are also in higher positions, ... till we compare the last two metabolites.
!                              For the metabolite in a general position X (lower than the last position available), it will
!                              be compared to metabolites in increasing higher positions in the cell. Considering the case
!                              where the metabolite in the position X is compared to the metabolite in the position Y, the
!                              information regarding the topological equivalence of the two nodes will be stored in the cell:
!                              i2 = (Somatrio[i=1,(X-1)] n Met. - i ) + ( Y - X )
!      EquiReactRevStru(i2,i1) : information regarding the equivalent reactions of a given structure i1. If two reactions are
!                                equivalent, the corresponding cell is filled with "1"; if not, is filled with "0". The first
!                                reaction is compared with the ones in higher positions, the second reaction is compared to
!                                reactions who are also in higher positions, ... till we compare the last two reactions.
!                                For the reaction in a general position X (lower than the last position available), it will
!                                be compared to reactions in increasing higher positions in the cell. Considering the case
!                                where the reaction in the position X is compared to the reaction in the position Y, the
!                                information regarding the topological equivalence of the two nodes will be stored in the cell:
!                                i2 = (Somatrio[i=1,(X-1)] n Met. - i ) + ( Y - X )
!      DifNodesStru(i2,i1) : stores the information indicating if structure i1 has different nodes by taking into account their
!                            degree (total and IN). If all the nodes of a given type (metabolites, irreversible reactions and
!                            reversible reactions) are all different, the corresponding cell is filled with "1"; if there is
!                            at least two nodes of the same type with the same degree (total and IN), the corresponding cell
!                            is filled with "0". (i2=1) => metabolites; (i2=2) => irreversible reactions; (i2=3) => reversible
!                            reactions
!      PermutMet(i2,i1) : possible permutations for a set of metabolites, obtained by calling the subroutine allPermutations.
!                         i1) number of the permutation; i2) number of the metabolites ordered by this permutation
!      PermutIrrReact(i2,i1) : possible permutations for a set of irreversible reactions, obtained by calling the subroutine
!                              allPermutations. i1) number of the permutation; i2) number of the reactions ordered by this
!                              permutation
!      PermutRevReact(i2,i1) : possible permutations for a set of reversible reactions, obtained by calling the subroutine
!                              allPermutations. i1) number of the permutation; i2) number of the reactions ordered by this
!                              permutation
!      PermutDirInvReact(i2,i1) : possible permutations between the direct and inverse single direction reactions of a reversible
!                                 reaction. The matrix was obtained by calling the subroutine allRevReactPerm. These permutations
!                                 permutations take into account the total number of reversible reactions in the structure.
!                                 i1) number of the permutation; i2) ordering the direct and inverse single direction reactions
!                                 for all the reversible reactions

      implicit none

      ! Integer: 1 bit
      integer*1 :: RevStruct(32,8,6000), supRevStructTemp(32,8,1), supRevStruct(32,8,1)
      integer*1 :: FamRevStru(6000,6000), InclRevStru(6000,6000), ReactRevStru(8,6000)
      integer*1 :: EquiMetRevStru(30,6000), EquiReactRevStru(30,6000), DifNodesStru(3,6000)
      integer*1 :: supReactRevStruTemp(8), supReactRevStru(8), isoStruct
      integer*1 :: maxNumMet, maxNumReact
      ! Integer: 2 bits
      integer*2 :: nMet(7), nReact(7), numGen, PermutMet(8,60000), PermutIrrReact(8,60000)
      integer*2 :: PermutRevReact(8,60000), PermutDirInvReact(16,6000)
      integer*2 :: DataRevStru(12,6000), MetDegRevStru(24,6000), ReactDegRevStru(24,6000)
      integer*2 :: numStruGen(15), nrefStruLow(15), nrefStruHigh(15), supMetDegRevStruTemp(24)
      integer*2 :: supReactDegRevStruTemp(24), supMetDegRevStru(24), supReactDegRevStru(24)
      integer*2 :: numNodes(4), numPermDirInvReact, maxNumGen, isoStruRef, testcount(1)
      ! Integer: 4 bits
      integer*4 :: numPermutMet, numPermutIrrReact, numPermutRevReact, numStructures
      integer*4 :: const(5), i, j, i1, i2, i3
       !
       ! Variables used to obtain structures by adding the structural component i) new reactant/product (internal
       ! connection)
      integer*4 :: iI1, iI2, iI3, iI4, iI_1, iI_2, iI_DM1, iI_DR1, iI_rev1
      integer*4 :: wrI_1, wrI_2, wrI_3, wrI_4, wrI_5, cleanI_1, cleanI_2
      integer*4 :: cleanI_3, cleanI_4, cleanI_5, cleanI_6, cleanI_7
      integer*4 :: cleanI_8, cleanI_9, cleanI_10, cleanI_11
       !
       ! Variables used to obtain structures by adding the structural component ii) new reaction
      integer*4 :: iII1, iII2, iII3, iII4, iII_1, iII_2, iII_DM1, iII_DR1
      integer*4 :: iII_rev1, wrII_1, wrII_2, wrII_3, wrII_4, wrII_5, cleanII_1
      integer*4 :: cleanII_2, cleanII_3, cleanII_4, cleanII_5, cleanII_6
      integer*4 :: cleanII_7, cleanII_8, cleanII_9, cleanII_10, cleanII_11
       !
       ! Variables used to obtain structures by adding the structural component iii) new reversible reaction
      integer*4 :: iIII1, iIII2, iIII3, iIII_1, iIII_2, iIII_DM1, iIII_DR1
      integer*4 :: iIII_rev1, wrIII_1, wrIII_2, wrIII_3, wrIII_4, wrIII_5
      integer*4 :: cleanIII_1, cleanIII_2, cleanIII_3, cleanIII_4
      integer*4 :: cleanIII_5, cleanIII_6, cleanIII_7, cleanIII_8
      integer*4 :: cleanIII_9, cleanIII_10, cleanIII_11
       !
       ! Variables used to obtain structures by adding the structural component iv) new metabolite
      integer*4 :: iIV1, iIV2, iIV3, iIV_1, iIV_2, iIV_DM1, iIV_DR1, iIV_rev1
      integer*4 :: wrIV_1, wrIV_2, wrIV_3, wrIV_4, wrIV_5, cleanIV_1
      integer*4 :: cleanIV_2, cleanIV_3, cleanIV_4, cleanIV_5, cleanIV_6
      integer*4 :: cleanIV_7, cleanIV_8, cleanIV_9, cleanIV_10, cleanIV_11
       !
       ! Variables used to obtain structures by adding the structural component v) new metabolite + new reaction
      integer*4 :: iV1, iV2, iV3, iV_1, iV_2, iV_DM1, iV_DR1, iV_rev1
      integer*4 :: wrV_1, wrV_2, wrV_3, wrV_4, wrV_5, cleanV_1, cleanV_2
      integer*4 :: cleanV_3, cleanV_4, cleanV_5, cleanV_6, cleanV_7
      integer*4 :: cleanV_8, cleanV_9, cleanV_10, cleanV_11


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      INPUT FILES                      !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Reads the variables used to define the topological scope: maximum number
!      of metabolites (maxNumMet), maximum number of reactions (maxNumReact), and
!      maximum number of generations (maxNumGen)
    !
      ! Initiates the variables maxNumMet, maxNumReact and maxNumGen
      maxNumMet = 0
      maxNumReact = 0
      maxNumGen = 0
    !
      ! Reads the file "Topological_Scope.txt"
      open (1,file='Topological_Scope.txt')
      do i1=1,3
        !
        ! Obtains the maximum number of metabolites, maxNumMet
        if( i1 .eq. 1 ) then
          read (1,*) maxNumMet
        end if
        !
        ! Obtains the maximum number of reactions, maxNumReact
        if( i1 .eq. 2 ) then
          read (1,*) maxNumReact
        end if
        !
        ! Obtains the maximum number of generations, maxNumGen
        if( i1 .eq. 3 ) then
          read (1,*) maxNumGen
        end if
        !
      end do
      close (1)
     !
      ! Creating the vector numStruGen(i2,i1), which stores the number of structures per generation
      do i1=1,15
        numStruGen(i1) = 0
      end do


!      Creating the seed structure
      do i1=1,6000
        do i2=1,8
          do i3=1,32
            RevStruct(i3,i2,i1) = 0
          end do
        end do
      end do
      ! Support matrix used to obtain the structures by adding a new structural component to Parent-structures
      do i1=1,8
        supReactRevStru(i1) = 0
        supReactRevStruTemp(i1) = 0
        do i2=1,32
          supRevStruct(i2,i1,1) = 0
          supRevStructTemp(i2,i1,1) = 0
        end do
      end do
      ! Support matrix used to obtain the degree of the metabolites of the new structure by adding a new structural
      ! to Parent-structures
      do i1=1,24
        supMetDegRevStru(i1) = 0
        supMetDegRevStruTemp(i1) = 0
        supReactDegRevStru(i1) = 0
        supReactDegRevStruTemp(i1) = 0
      end do
      !
      ! First structure is created
      RevStruct(2,1,1) = 1  ! Metabolite 1 is a product of reaction 1
      RevStruct(1,2,1) = 1  ! Metabolite 2 is a reactant of reaction 1
      !
      ! Degree of the metabolites
      ! Metabolite 1
      MetDegRevStru(1,1) = 1  ! Total connections
      MetDegRevStru(2,1) = 1  ! N connections IN (product of 1 reaction)
      MetDegRevStru(3,1) = 0  ! N connections OUT
      ! Metabolite 2
      MetDegRevStru(4,1) = 1  ! Total connections
      MetDegRevStru(5,1) = 0  ! N connections IN
      MetDegRevStru(6,1) = 1  ! N connections OUT (reactant of 1 reaction)
      !
      ! Degree of the reaction 1
      ReactDegRevStru(1,1) = 2  ! Total connections
      ReactDegRevStru(2,1) = 1  ! N connections IN (has 1 reactant)
      ReactDegRevStru(3,1) = 1  ! N connections OUT (has 1 product)
      !
      ! Number of the generation
      numGen = 1
      !
      ! Counting the total number of structures
      numStructures = 1
      ! Counting the number of structures of the numGen (in this case, 1st generation)
      numStruGen(numGen) = numStruGen(numGen) + 1
      ! Lower reference number of a structure of the numGen generation (in this case, 1st generation)
      nrefStruLow(numGen) = 1
      ! Higher reference number of a structure of the numGen generation (in this case, 1st generation)
      nrefStruHigh(numGen) = 1
      !
!      Filling the matrix DataRevStru(i2,i1) with the relevant data of the seed structure
      ! Number of the generation
      DataRevStru(1,1) = numGen
      ! Number of metabolites
      DataRevStru(2,1) = 2
      ! Number of reactions
      DataRevStru(3,1) = 1
      ! Number of reversible reactions
      DataRevStru(4,1) = 0
      ! Number of Ancestors
      DataRevStru(5,1) = 0
      ! Number of Parents
      DataRevStru(6,1) = 0
      ! Number of Childs
      DataRevStru(7,1) = 0
      ! Number of Offspring
      DataRevStru(8,1) = 0
      ! Number of interconnections
      DataRevStru(9,1) = 2
      ! Maximum possible number of interconnections
      DataRevStru(10,1) = 2 * DataRevStru(2,1) * DataRevStru(3,1)
      ! Number of connections metabolite/reaction (reversibility not considered)
      DataRevStru(11,1) = 2
      ! Maximum number of connections metabolite/reaction (reversibility not considered)
      DataRevStru(12,1) = DataRevStru(2,1) * DataRevStru(3,1)
      !
!      Creating the matrix ReactRevStru(i2,i1) - information about the reversibility of the reactions
      do i1=1,6000
        do i2=1,8
          ReactRevStru(i2,i1) = 0
        end do
      end do


!      Creating the structures from the seed-structure by successively adding a new structural component
      do i=1,(maxNumGen-1)  ! Performs the loop (maxNumGen-1) times
      ! Introducing a structural component at a time to each one of the structures of the previous generation
        !
        numGen = numGen + 1  ! Controls the number of the generation in which we are currently in the loop
        !
        ! Creates the higher and lower reference numbers of the structures of this new generation
        nrefStruLow(numGen) = nrefStruHigh(numGen-1) + 1
        nrefStruHigh(numGen) = nrefStruHigh(numGen-1)  ! Every time a new structure is found in this generation,
        ! a 1 should be added to nrefStruHigh(numGen)
      !
      !
      ! i) a new reactant/product to a existing reaction (internal connection)
        !
        ! Support constant used to determine if the new structure is a isomorphic representation of another
        ! found before
        const(1) = 0
        !
        do 5 iI1=nrefStruLow(numGen-1),nrefStruHigh(numGen-1)  ! Performs the loop to all the structures of the previous
        ! generation
          !
          ! Checks if it is possible to add the structural component between the pair metabolite/reaction (iI2/iI3).
          ! For that to happen, the number of existing connections metabolite/reaction (reversibility not considered)
          ! must be lower than the maximum possible number of connections metabolite/reaction
          if( DataRevStru(11,iI1) .ne. DataRevStru(12,iI1) ) then
            ! At least one new structure might be obtained by adding a reactant/product to a existing reaction (internal
            ! connection). As such, one can determine the permutations needed to test for isomorphism
            !
            ! Fills the support vector numNodes(i1), which stores the number of metabolites, reactions, reversible
            ! reactions and of irrreversible reactions of the (possibly) new structure
            ! Number of metabolites = number of metabolites of the Parent-Structure
            numNodes(1) = DataRevStru(2,iI1)
            ! Number of reactions = number of reactions of the Parent-Structure
            numNodes(2) = DataRevStru(3,iI1)
            ! Number of reversible reactions = number of reversible reactions of the Parent-Structure
            numNodes(3) = DataRevStru(4,iI1)
            ! Number of irreversible reactions = number of reactions of the Parent-Structure - number of reversible
            !                                    reactions of the Parent-Structure
            numNodes(4) = DataRevStru(3,iI1) - DataRevStru(4,iI1)
            !
            ! Call the subroutine "allPermutations" to obtain the permutations for the metabolites
            call allPermutations(numNodes(1), PermutMet, numPermutMet)
            ! The new structures have at least one irreversible reaction?
            if( numNodes(4) .ne. 0 ) then
              ! Call the subroutine "allPermutations" to obtain the permutations for the irreversible reactions
              call allPermutations(numNodes(4), PermutIrrReact, numPermutIrrReact)
            end if
            ! The new structures have at least one reversible reaction?
            if( numNodes(3) .ne. 0 ) then
              ! Call the subroutine "allPermutations" to obtain the permutations for the reversible reactions
              call allPermutations(numNodes(3), PermutRevReact, numPermutRevReact)
              ! Call the subroutine "allRevReactPerm" to obtain the permutations for the direct and inverse reactions
              ! for all the (possibly) reversible reactions
              call allRevReactPerm(numNodes(3), PermutDirInvReact, numPermDirInvReact)
            end if
            !
            do iI2=1,DataRevStru(2,iI1)  ! Test for each one of the metabolites of the iI1 structure
              do iI3=1,DataRevStru(3,iI1)    ! Test for each one of the reactions of the iI1 structure
                ! Test if the metabolite iI2 is a reactant or a product of reaction iI3
                if( ( RevStruct( (4*(iI3 - 1) + 1), iI2, iI1) .eq. 0  ) .and. &
                ( RevStruct( (4*(iI3 - 1) + 2), iI2, iI1) .eq. 0  ) ) then
                  !
                  ! Adds a connection between the metabolite iI2 and the reaction iI3: first as a reactant (iI4=1), and then
                  ! as a product (iI4=2)
                  do iI4=1,2
                    !
                    ! Creates the vector supRevStructTemp(iI_2,iI_1,1), that stores the information temporarily for a
                    do iI_1=1,DataRevStru(2,iI1)                      ! possible new structure
                      ! The columns corresponding to the reactions (iI_2) are equal to: 4 x number of reactions
                      do iI_2=1,(4*DataRevStru(3,iI1))
                        supRevStructTemp(iI_2,iI_1,1) = RevStruct(iI_2,iI_1,iI1)
                      end do
                    end do
                    ! Creates the vectors supMetDegRevStruTemp(i1) and supReactDegRevStruTemp(i1)
                    do iI_DM1=1,(3*DataRevStru(2,iI1))  ! 3 x number of metabolites of Parent-Structure iI1
                      supMetDegRevStruTemp(iI_DM1) = MetDegRevStru(iI_DM1,iI1)
                    end do
                    do iI_DR1=1,(3*DataRevStru(3,iI1))  ! 3 x number of reactions of Parent-Structure iI1
                      supReactDegRevStruTemp(iI_DR1) = ReactDegRevStru(iI_DR1,iI1)
                    end do
                    !
                    ! Fills the cell for the direct single-direction reaction
                    supRevStructTemp( (4*(iI3 - 1) + iI4), iI2, 1) = 1
                    !
                    ! Adds this connection to the degree of metabolite iI2 and reaction iI3
                    ! Total connections of metabolite iI2
                    supMetDegRevStruTemp(3*(iI2 - 1) + 1) = &
                    supMetDegRevStruTemp(3*(iI2 - 1) + 1) + 1  ! Adds 1 connection
                    ! Total connections of reaction iI3
                    supReactDegRevStruTemp(3*(iI3 - 1) + 1) = &
                    supReactDegRevStruTemp(3*(iI3 - 1) + 1) + 1  ! Adds 1 connection
                    ! Metabolite iI2 is a reactant or a product of reaction iI3?
                    if( iI4 .eq. 1 ) then
                      ! Metabolite iI2 is a reactant of reactions iI3
                      !
                      ! Metabolite iI2: adds 1 connection OUT (reactant of 1 reaction)
                      supMetDegRevStruTemp(3*(iI2 - 1) + 3) = &
                      supMetDegRevStruTemp(3*(iI2 - 1) + 3) + 1
                      !
                      ! Reaction iI3: adds 1 connection IN (1 new reactant)
                      supReactDegRevStruTemp(3*(iI3 - 1) + 2) = &
                      supReactDegRevStruTemp(3*(iI3 - 1) + 2) + 1
                      !
                    else
                      ! Metabolite iI2 is a product of reactions iI3
                      !
                      ! Metabolite iI2: adds 1 connection IN (product of 1 reaction)
                      supMetDegRevStruTemp(3*(iI2 - 1) + 2) = &
                      supMetDegRevStruTemp(3*(iI2 - 1) + 2) + 1
                      !
                      ! Reaction iI3: adds 1 connection OUT (1 new product)
                      supReactDegRevStruTemp(3*(iI3 - 1) + 3) = &
                      supReactDegRevStruTemp(3*(iI3 - 1) + 3) + 1
                      !
                    end if
                    !
                    ! Is the reaction reversible? If it is, fills the cell for the inverse single-direction reaction
                    if( ReactRevStru(iI3,iI1) .eq. 1 ) then
                      !
                      ! Adds the corresponding connection to the degree of metabolite iI2 and reaction iI3
                      ! Total connections of metabolite iI2
                      supMetDegRevStruTemp(3*(iI2 - 1)+1) = &
                      supMetDegRevStruTemp(3*(iI2 - 1) + 1) + 1  ! Adds 1 connection
                      ! Total connections of reaction iI3
                      supReactDegRevStruTemp(3*(iI3 - 1) + 1) = &
                      supReactDegRevStruTemp(3*(iI3 - 1) + 1) + 1  ! Adds 1 connection
                      !
                      ! The metabolite is a reactant or product of the direct reaction?
                      if( iI4 .eq. 1 ) then
                      ! Metabolite iI2 is a reactant of the direct reaction iI3
                        ! So, metabolite iI2 is also a product of reaction iI3
                        !
                        ! Adds 1 to the cell of the metabolite as a product of the inverse reaction
                        supRevStructTemp( (4*(iI3 - 1) + 4), iI2, 1) = 1
                        !
                        ! Metabolite iI2: adds 1 connection IN (product of 1 reaction)
                        supMetDegRevStruTemp(3*(iI2 - 1) + 2) = &
                        supMetDegRevStruTemp(3*(iI2 - 1) + 2) + 1
                        ! Reaction iI3: adds 1 connection OUT (1 new product)
                        supReactDegRevStruTemp(3*(iI3 - 1) + 3) = &
                        supReactDegRevStruTemp(3*(iI3 - 1) + 3) + 1
                        !
                      else
                      ! Metabolite iI2 is a product of the direct reaction iI3
                        ! Metabolite iI2 is also a reactant of reaction iI3
                        !
                        ! Adds 1 to the cell of the metabolite as a reactant of the inverse reaction
                        supRevStructTemp( (4*(iI3 - 1) + 3), iI2, 1) = 1
                        !
                        ! Metabolite iI2: adds 1 connection OUT (reactant of 1 reaction)
                        supMetDegRevStruTemp(3*(iI2 - 1) + 3) = &
                        supMetDegRevStruTemp(3*(iI2 - 1) + 3) + 1
                        !
                        ! Reaction iI3: adds 1 connection IN (1 new reactant)
                        supReactDegRevStruTemp(3*(iI3 - 1) + 2) = &
                        supReactDegRevStruTemp(3*(iI3 - 1) + 2) + 1
                        !
                      end if
                    end if
                    !
                    ! Stores the information of the reversibility of the reactions for this (possibly) new structure
                    do iI_rev1=1,DataRevStru(3,iI1)  ! Number of reactions of the Parent-Structure iI1
                      supReactRevStruTemp(iI_rev1) = ReactRevStru(iI_rev1,iI1)
                    end do
                    !
                    ! Call the subroutine organizingStruct
                    call organizingStruct(supRevStructTemp, supMetDegRevStruTemp, &
                    supReactDegRevStruTemp, numNodes, supReactRevStruTemp, supRevStruct, &
                    supMetDegRevStru, supReactDegRevStru, supReactRevStru)
                    !
                    ! Creates the support constant const(1), which stores the possibly final reference number of
                    ! this new structure (in the case it is not an isomorphic representation of a previous found
                    ! structure)
                    const(1) = numStructures + 1
                    !
                    ! Stores the general information regarding this new structure in matrix DataRevStru(i2,i1) at line
                    ! i1=const(1). If we find that this new structure is a isomorphic representation of a structure
                    ! found before, this line will be filled with zeros again
                    ! Number of the generation
                    DataRevStru(1,const(1)) = numGen
                    ! Number of metabolites
                    DataRevStru(2,const(1)) = DataRevStru(2,iI1)
                    ! Number of reactions
                    DataRevStru(3,const(1)) = DataRevStru(3,iI1)
                    ! Number of reversible reactions
                    DataRevStru(4,const(1)) = DataRevStru(4,iI1)
                    ! Number of Ancestors
                    DataRevStru(5,const(1)) = 0  ! This information will be obtained latter
                    ! Number of Parents
                    DataRevStru(6,const(1)) = 1  ! First Parent found
                    ! Number of Childs
                    DataRevStru(7,const(1)) = 0
                    ! Number of Offspring
                    DataRevStru(8,const(1)) = 0  ! This information will be obtained latter
                    ! Number of interconnections
                    if( ReactRevStru(iI3,iI1) .eq. 1 ) then
                      ! If the reaction iI3 is reversible, adds 2 new interconnections
                      DataRevStru(9,const(1)) = DataRevStru(9,iI1) + 2
                    else
                      ! If the reaction iI3 is irreversible, adds 1 new interconnection
                      DataRevStru(9,const(1)) = DataRevStru(9,iI1) + 1
                    end if
                    ! Maximum possible number of interconnections
                    DataRevStru(10,const(1)) = DataRevStru(10,iI1)
                    ! Number of connections metabolite/reaction (reversibility not considered); adds a new
                    ! connection metabolite/reaction when compared to the Parent
                    DataRevStru(11,const(1)) = DataRevStru(11,iI1) + 1
                    ! Maximum number of connections metabolite/reaction (reversibility not considered)
                    DataRevStru(12,const(1)) = DataRevStru(12,iI1)
                    !
                    ! Is this a isomorphic representation of a structure found before?
                    ! Call the subroutine isomorphismTest
                    call isomorphismTest(supRevStruct, RevStruct, DataRevStru, &
                    MetDegRevStru, ReactDegRevStru, supMetDegRevStru, &
                    supReactDegRevStru, nrefStruLow(numGen), nrefStruHigh(numGen), &
                    const(1), PermutMet, PermutIrrReact, PermutRevReact, &
                    PermutDirInvReact, numPermutMet, numPermutIrrReact, &
                    numPermutRevReact, numPermDirInvReact, isoStruct, isoStruRef)
                    ! If isoStruct = 1, the structure is a isomorphic representation of structure isoStruRef;
                    if(isoStruct .eq. 1) then
                      !
                      ! Is this a new Parent/Child relantionship? If it is, the corresponding cell in matrix FamRevStru(i2,i1)
                      ! must be "0"
                      if(FamRevStru(isoStruRef,iI1) .eq. 0) then
                        ! Structure iI1 is a Parent of the structure isoStruRef, found before
                        ! Reaction iI3 is reversible?
                        if( ReactRevStru(iI3,iI1) .eq. 1 ) then
                          ! i.2) New reactant/product of a reversible reaction
                          FamRevStru(isoStruRef,iI1) = 2
                        else
                          ! i.1) New reactant/product of a irreversible reaction
                          FamRevStru(isoStruRef,iI1) = 1
                        end if
                        !
                        ! Number of Parents of structure isoStruRef: adds structure iI1 as one of its Parents
                        DataRevStru(6,isoStruRef) = DataRevStru(6,isoStruRef) + 1
                        !
                        ! Number of Childs of structure iI1: adds structure isoStruRef as one of its Childs
                        DataRevStru(7,iI1) = DataRevStru(7,iI1) + 1
                      end if
                      !
                      ! The matrix DataRevStru(i2,i1) must be turned to zeros at line i1=const(1)
                      do cleanI_1=1,12
                        DataRevStru(cleanI_1,const(1)) = 0
                      end do
                      !
                    ! If isoStruct = 0, this structure was found for the first time
                    else
                      !
                      ! Adds a new structure to the total number of structures
                      numStructures = numStructures + 1
                      ! Adds a new structure to the total number of structures of this generation (the goal is to store
                      ! the highest reference number of a structure of this generation)
                      nrefStruHigh(numGen) = nrefStruHigh(numGen) + 1
                      ! Adds a new structure to the total number of structures of this generation
                      numStruGen(numGen) = numStruGen(numGen) + 1
                      write(*,2002) numStruGen(numGen), numGen, numStructures
                      !
                      ! Stores the relationship Parent/Child between the Parent iI1 and the new Child numStructures
                      ! Reaction iI3 is reversible?
                      if( ReactRevStru(iI3,iI1) .eq. 1 ) then
                        ! i.2) New reactant/product of a reversible reaction
                        FamRevStru(numStructures,iI1) = 2
                      else
                        ! i.1) New reactant/product of a irreversible reaction
                        FamRevStru(numStructures,iI1) = 1
                      end if
                      !
                      ! Number of Childs of structure iI1: adds structure numStructures as one of its Childs
                      DataRevStru(7,iI1) = DataRevStru(7,iI1) + 1
                      !
                      ! Writes the new structure to the matrix RevStruct(i3,i2,i1) at line i1=numStructures
                      do wrI_1=1,DataRevStru(2,iI1)  ! Number of metabolites of the Parent iI1 (equal to the number of
                      ! metabolites of it's Childs)
                        do wrI_2=1,(4*DataRevStru(3,iI1))  ! 4 x number of reactions of the Parent iI1 (equal to the number of
                        ! reactions of it's Childs)
                          ! Writes the new structure into matrix RevStruct(i3,i2,i1) at line i1=numStructures
                          RevStruct(wrI_2,wrI_1,numStructures) = supRevStruct(wrI_2,wrI_1,1)
                        end do
                      end do
                      !
                      ! Writes the information regarding the reversibility of the reactions of the new found structure
                      ! (it is the same as for the Parent structure) into the matrix ReactRevStru(i2,i1) and of the degree
                      ! of the metabolites into the matrix ReactDegRevStru(i2,i1) at line i1=numStructures
                      do wrI_3=1,numNodes(2)  ! Number of reactions of the structure numStructures
                        ReactRevStru(wrI_3,numStructures) = supReactRevStru(wrI_3)
                      end do
                      do wrI_4=1,(3*numNodes(2))  ! 3 x number of reactions of structure numStructures
                        ReactDegRevStru(wrI_4,numStructures) = supReactDegRevStru(wrI_4)
                      end do
                      !
                      ! Writes the the information regarding the degree of the metabolites into matrix MetDegRevStru(i2,i1)
                      ! at line i1=numStructures
                      do wrI_5=1,(3*NumNodes(1))  ! Number of metabolites of the structure numStructures
                        MetDegRevStru(wrI_5,numStructures) = supMetDegRevStru(wrI_5)
                      end do
                      !
                    end if
                    !
                    ! Cleans the the variables supRevStructTemp(i2,i1), supRevStruct(i2,i1), supMetDegRevStruTemp(i1),
                    ! supMetDegRevStru(i1), supReactDegRevStruTemp(i1), supReactDegRevStru(i1), supReactRevStruTemp(i1) and
                    ! supReactRevStru(i1)
                    do cleanI_2=1,numNodes(1)  ! Number of metabolites
                      supMetDegRevStruTemp(cleanI_2) = 0
                      supMetDegRevStru(cleanI_2) = 0
                      do cleanI_3=1,(4*numNodes(2))  ! 4 x number of reactions
                        supRevStructTemp(cleanI_3,cleanI_2,1) = 0
                        supRevStruct(cleanI_3,cleanI_2,1) = 0
                        if( ( cleanI_2 .eq. 1 ) .and. &
                        ( cleanI_3 .le. numNodes(2) ) ) then
                          supReactDegRevStruTemp(cleanI_3) = 0
                          supReactDegRevStru(cleanI_3) = 0
                          supReactRevStruTemp(cleanI_3) = 0
                          supReactRevStru(cleanI_3) = 0
                        end if
                      end do
                    end do
                    !
                    ! Turns the variables isoStruct and isoStruRef into "0"
                    isoStruct = 0
                    isoStruRef = 0
                    !
                    ! Turns const(1) into zero
                    const(1) = 0
                    !
                  end do
                end if
              end do
            end do
            !
            ! Turns to zero the matrices with the permutations for: i) metabolites - PermutMet(i2,i1); ii) irreversible
            ! reactions PermutIrrReact(i2,i1); iii) reversible reactions PermutRevReact(i2,i1); iv) direct and inverse
            ! single-direction reaction for all the reversible reactions - PermutRevReact(i2,i1)
            do cleanI_4=1,numPermutMet  ! Number of permutations for the metabolites
              do cleanI_5=1,numNodes(1)  ! Number of metabolites of the new found Childs of Parent iI1
                PermutMet(cleanI_5,cleanI_4) = 0
              end do
            end do
            if( numNodes(4) .ne. 0 ) then
              do cleanI_6=1,numPermutIrrReact  ! Number of permutations for the irreversible reactions
                do cleanI_7=1,numNodes(4)  ! Number of irreversible reactions
                  PermutIrrReact(cleanI_7,cleanI_6) = 0
                end do
              end do
            end if
            if( numNodes(3) .ne. 0 ) then
              do cleanI_8=1,numPermutRevReact  ! Number of permutations for the reversible reactions
                do cleanI_9=1,numNodes(3)  ! Number of reversible reactions
                  PermutRevReact(cleanI_9,cleanI_8) = 0
                end do
              end do
              do cleanI_10=1,numPermDirInvReact ! Number of permutations for the direct and inverse reactions of a reversible reaction
                do cleanI_11=1,(2*numNodes(3))  ! 2x number of reversible reactions
                  PermutDirInvReact(cleanI_10,cleanI_11) = 0
                end do
              end do
            end if
            ! Cleans the number of nodes of newly found Childs of Parent iI1
            numNodes(1) = 0
            numNodes(2) = 0
            numNodes(3) = 0
            numNodes(4) = 0
            ! Turns to zero the number of permutations
            numPermutMet = 0
            numPermutIrrReact = 0
            numPermutRevReact = 0
            numPermDirInvReact = 0
          end if
5       continue
      !
      !
      !
      !
      ! ii) a new reaction (connects two existing metabolites as a pair reactant/product)
        !
        ! Support constant used to determine if the new structure is a isomorphic representation of another
        ! found before
        const(1) = 0
        !
        do 10 iII1=nrefStruLow(numGen-1),nrefStruHigh(numGen-1)  ! Performs the loop to all the structures of the previous
        ! generation
          !
          ! The Child-structures of the Parent-structure iII1 have a number of reactions superior to maxNumReact?
          if( (DataRevStru(3,iII1) + 1) .gt. maxNumReact ) goto 10
          !
          !
          ! Fills the support vector numNodes(i1), which stores the number of metabolites, reactions, reversible
          ! reactions and of irrreversible reactions of the (possibly) new structure
          ! Number of metabolites   = number of metabolites of the Parent-Structure
          numNodes(1) = DataRevStru(2,iII1)
          ! Number of reactions = number of reactions of the Parent-Structure + 1 (Adds the new reaction)
          numNodes(2) = DataRevStru(3,iII1) + 1
          ! Number of reversible reactions = number of reversible reactions of the Parent-Structure
          numNodes(3) = DataRevStru(4,iII1)
          ! Number of irreversible reactions = (number of reactions of the Parent-Structure + 1) - number of reversible
          !                                    reactions of the Parent-Structure
          numNodes(4) = (DataRevStru(3,iII1) + 1) - DataRevStru(4,iII1)
          !
          ! Call the subroutine "allPermutations" to obtain the permutations for the metabolites
          call allPermutations(numNodes(1), PermutMet, numPermutMet)
          ! Call the subroutine "allPermutations" to obtain the permutations for the irreversible reactions
          call allPermutations(numNodes(4), PermutIrrReact, numPermutIrrReact)
          ! The new structures have at least one reversible reaction?
          if( numNodes(3) .ne. 0 ) then
            ! Call the subroutine "allPermutations" to obtain the permutations for the reversible reactions
            call allPermutations(numNodes(3), PermutRevReact, numPermutRevReact)
            ! Call the subroutine "allRevReactPerm" to obtain the permutations for the direct and inverse reactions
            ! for all the (possibly) reversible reactions
            call allRevReactPerm(numNodes(3), PermutDirInvReact, numPermDirInvReact)
          end if
          !
          ! Calls all the pairs of metabolites of the Parent structure (Considering the matrix iII2/iII3, it looks
          ! only at the upper part of ths quadratic matrix)
          do iII2=1,(DataRevStru(2,iII1)-1)
            do iII3=(iII2+1),DataRevStru(2,iII1)
              ! Considers both situations: iII4=1 -> iII2 is the reactant and iII3 is the product of the new reaction;
              ! iII4=2 -> iII2 is the product and iII3 is the reactant of the new reaction;
              do iII4=1,2
                !
                ! Creates the vector supRevStructTemp(iII_2,iII_1,1), that stores the information for a possible new structure
                do iII_1=1,DataRevStru(2,iII1)
                  ! The columns corresponding to the reactions (iII_2) are equal to: 4 x number of reactions
                  do iII_2=1,(4*(DataRevStru(3,iII1)+1))
                    supRevStructTemp(iII_2,iII_1,1) = RevStruct(iII_2,iII_1,iII1)
                  end do
                end do
                !
                ! Creates the vectors supMetDegRevStruTemp(i1) and supReactDegRevStruTemp(i1)
                do iII_DM1=1,(3*DataRevStru(2,iII1))  ! 3 x number of metabolites of Parent-Structure iII1
                  supMetDegRevStruTemp(iII_DM1) = MetDegRevStru(iII_DM1,iII1)
                end do
                do iII_DR1=1,(3*DataRevStru(3,iII1))  ! 3 x number of reactions of Parent-Structure iII1
                  supReactDegRevStruTemp(iII_DR1) = ReactDegRevStru(iII_DR1,iII1)
                end do
                !
                ! Introduces the degree of the new reaction numNodes(2) -> equal to: DataRevStru(3,iII1) + 1
                supReactDegRevStruTemp( (3*(numNodes(2) - 1)) + 1 ) = 2  ! Number of connections (1 reactant and 1 product)
                supReactDegRevStruTemp( (3*(numNodes(2) - 1)) + 2 ) = 1  ! Number of connections IN (1 reactant)
                supReactDegRevStruTemp( (3*(numNodes(2) - 1)) + 3 ) = 1  ! Number of connections OUT (1 product)
                !
                ! Introduces the new reaction between the metabolites iII2 and iII3
                if( iII4 .eq. 1 ) then
                  ! If iII4 equal to "1", iII2 = reactant and iII3 = product of the new reaction - equal to DataRevStru(3,iII1) + 1
                  ! Metabolite iII2 -> reactant
                  !  i3 = [(4 x (n reaction - 1)) + 1], where n reaction = (DataRevStru(3,iII1) + 1)
                  supRevStructTemp( ((4*DataRevStru(3,iII1)) + 1), iII2, 1) = 1
                  ! Degree of metabolite iII2: reactant (OUT) of reaction numNodes(2)
                  supMetDegRevStruTemp( (3*(iII2 - 1)) + 1 ) = &
                  supMetDegRevStruTemp( (3*(iII2 - 1)) + 1 ) + 1  ! Total connections
                  supMetDegRevStruTemp( (3*(iII2 - 1)) + 3 ) = &
                  supMetDegRevStruTemp( (3*(iII2 - 1)) + 3 ) + 1  ! OUT connections (reactant)
                  !
                  ! Metabolite iII3 -> product
                  !  i3 = [(4 x (n reaction - 1)) + 2], where n reaction = (DataRevStru(3,iII1) + 1)
                  supRevStructTemp( ((4*DataRevStru(3,iII1)) + 2), iII3, 1) = 1
                  ! Degree of metabolite iII3: product (IN) of reaction numNodes(2)
                  supMetDegRevStruTemp( (3*(iII3 - 1)) + 1 ) = &
                  supMetDegRevStruTemp( (3*(iII3 - 1)) + 1 ) + 1  ! Total connections
                  supMetDegRevStruTemp( (3*(iII3 - 1)) + 2 ) = &
                  supMetDegRevStruTemp( (3*(iII3 - 1)) + 2 ) + 1  ! IN connections (product)
                  !
                else
                  ! If iII4 equal to "2", iII2 = product and iII3 = reactant of the new reaction - equal to DataRevStru(3,iII1) + 1
                  ! Metabolite iII2 -> product
                  !  i3 = [(4 x (n reaction - 1)) + 2], where n reaction = (DataRevStru(3,iII1) + 1)
                  supRevStructTemp( ((4*DataRevStru(3,iII1)) + 2), iII2, 1) = 1
                  ! Degree of metabolite iII2: product (IN) of reaction numNodes(2)
                  supMetDegRevStruTemp( (3*(iII2 - 1)) + 1 ) = &
                  supMetDegRevStruTemp( (3*(iII2 - 1)) + 1 ) + 1  ! Total connections
                  supMetDegRevStruTemp( (3*(iII2 - 1)) + 2 ) = &
                  supMetDegRevStruTemp( (3*(iII2 - 1)) + 2 ) + 1  ! IN connections (product)
                  !
                  ! Metabolite iII3 -> reactant
                  !  i3 = [(4 x (n reaction - 1)) + 1], where n reaction = (DataRevStru(3,iII1) + 1)
                  supRevStructTemp( ((4*DataRevStru(3,iII1)) + 1), iII3, 1) = 1
                  ! Degree of metabolite iII3: reactant (OUT) of reaction numNodes(2)
                  supMetDegRevStruTemp( (3*(iII3 - 1)) + 1 ) = &
                  supMetDegRevStruTemp( (3*(iII3 - 1)) + 1 ) + 1  ! Total connections
                  supMetDegRevStruTemp( (3*(iII3 - 1)) + 3 ) = &
                  supMetDegRevStruTemp( (3*(iII3 - 1)) + 3 ) + 1  ! OUT connections (product)
                  !
                end if
                !
                ! Stores the information of the reversibility of the reactions for this (possibly) new structure
                do iII_rev1=1,DataRevStru(3,iII1)  ! Number of reactions of the Parent-Structure iII1
                  supReactRevStruTemp(iII_rev1) = ReactRevStru(iII_rev1,iII1)
                end do
                ! Stores the reversibility of the new reaction numNodes(2)
                supReactRevStruTemp(numNodes(2)) = 0
                !
                ! Call the subroutine organizingStruct
                call organizingStruct(supRevStructTemp, supMetDegRevStruTemp, &
                supReactDegRevStruTemp, numNodes, supReactRevStruTemp, supRevStruct, &
                supMetDegRevStru, supReactDegRevStru, supReactRevStru)
                !
                !
                ! Creates the support constant const(1), which stores the possibly final reference number of
                ! this new structure (in the case it is not an isomorphic representation of a previous found
                ! structure)
                const(1) = numStructures + 1
                !
                ! Stores the general information regarding this new structure in matrix DataRevStru(i2,i1) at line
                ! i1=const(1). If we find that this new structure is a isomorphic representation of a structure
                ! found before, this line will be filled with zeros again
                ! Number of the generation
                DataRevStru(1,const(1)) = numGen
                ! Number of metabolites
                DataRevStru(2,const(1)) = DataRevStru(2,iII1)
                ! Number of reactions: adds a reaction to the number of reactions of the Parent iII1
                DataRevStru(3,const(1)) = DataRevStru(3,iII1) + 1
                ! Number of reversible reactions
                DataRevStru(4,const(1)) = DataRevStru(4,iII1)
                ! Number of Ancestors
                DataRevStru(5,const(1)) = 0  ! This information will be obtained latter
                ! Number of Parents
                DataRevStru(6,const(1)) = 1  ! First Parent found
                ! Number of Childs
                DataRevStru(7,const(1)) = 0
                ! Number of Offspring
                DataRevStru(8,const(1)) = 0  ! This information will be obtained latter
                ! Number of interconnections: adds 2 interconnections to the ones of Parent iII1 due to the new reaction
                DataRevStru(9,const(1)) = DataRevStru(9,iII1) + 2
                ! Maximum possible number of interconnections: 2 x n metabolite x n of reactions
                DataRevStru(10,const(1)) = 2*DataRevStru(2,const(1))*DataRevStru(3,const(1))
                ! Number of connections metabolite/reaction (reversibility not considered); adds 2 new connections
                ! metabolite/reaction when compared to the Parent iII1 due to the new reaction
                DataRevStru(11,const(1)) = DataRevStru(11,iII1) + 2
                ! Maximum number of connections metabolite/reaction (reversibility not considered)
                DataRevStru(12,const(1)) = DataRevStru(2,const(1)) * DataRevStru(3,const(1))
                !
                ! Is this a isomorphic representation of a structure found before?
                ! Call the subroutine isomorphismTest
                call isomorphismTest(supRevStruct, RevStruct, DataRevStru, &
                MetDegRevStru, ReactDegRevStru, supMetDegRevStru, &
                supReactDegRevStru, nrefStruLow(numGen), nrefStruHigh(numGen), &
                const(1), PermutMet, PermutIrrReact, PermutRevReact, &
                PermutDirInvReact, numPermutMet, numPermutIrrReact, &
                numPermutRevReact, numPermDirInvReact, isoStruct, isoStruRef)
                ! If isoStruct = 1, the structure is a isomorphic representation of structure isoStruRef;
                if(isoStruct .eq. 1) then
                  !
                  ! Is this a new Parent/Child relantionship? If it is, the correspinding cell in matrix FamRevStru(i2,i1)
                  ! must be "0"
                  if(FamRevStru(isoStruRef,iII1) .eq. 0) then
                    !
                    ! Structure iII1 is a Parent of the structure isoStruRef, found before. One adds "3" to the corresponding
                    ! cell in matrix FamRevStru(i2,i1), to show that they only differ of 1 reaction
                    FamRevStru(isoStruRef,iII1) = 3
                    !
                    ! Number of Parents of structure isoStruRef: adds structure iII1 as one of its Parents
                    DataRevStru(6,isoStruRef) = DataRevStru(6,isoStruRef) + 1
                    !
                    ! Number of Childs of structure iII1: adds structure isoStruRef as one of its Childs
                    DataRevStru(7,iII1) = DataRevStru(7,iII1) + 1
                    !
                  end if
                  !
                  ! The matrix DataRevStru(i2,i1) must be turned to zeros at line i1=const(1)
                  do cleanII_1=1,12
                    DataRevStru(cleanII_1,const(1)) = 0
                  end do
                  !
                ! If isoStruct = 0, this structure was found for the first time
                else
                  !
                  ! Adds a new structure to the total number of structures
                  numStructures = numStructures + 1
                  ! Adds a new structure to the total number of structures of this generation (the goal is to store
                  ! the highest reference number of a structure of this generation)
                  nrefStruHigh(numGen) = nrefStruHigh(numGen) + 1
                  ! Adds a new structure to the total number of structures of this generation
                  numStruGen(numGen) = numStruGen(numGen) + 1
                  write(*,2002) numStruGen(numGen), numGen, numStructures
                  !
                  ! Stores the relationship Parent/Child between the Parent iII1 and the new Child numStructures. One adds
                  ! "3" to the corresponding cell in matrix FamRevStru(i2,i1), to show that they only differ of 1 reaction
                  FamRevStru(numStructures,iII1) = 3
                  !
                  ! Number of Childs of structure iII1: adds structure numStructures as one of its Childs
                  DataRevStru(7,iII1) = DataRevStru(7,iII1) + 1
                  !
                  ! Writes the new structure to the matrix RevStruct(i3,i2,i1) at line i1=numStructures
                  do wrII_1=1,DataRevStru(2,iII1)  ! Number of metabolites of the Parent iII1 (equal to the number of
                  ! metabolites of it's Childs)
                    do wrII_2=1,(4*(DataRevStru(3,iII1)+1))  ! 4 x number of reactions of the Parent iII1 + 1 (equal to the number
                    ! of reactions of it's Childs)
                      ! Writes the new structure into matrix RevStruct(i3,i2,i1) at line i1=numStructures
                      RevStruct(wrII_2,wrII_1,numStructures) = &
                      supRevStruct(wrII_2,wrII_1,1)
                      ! Turns to zero the suport matrix supRevStruct(i3,i2,1)
                      supRevStruct(wrII_2,wrII_1,1) = 0
                    end do
                  end do
                  !
                  ! Writes the information regarding the reversibility of the reactions of the new found structure
                  ! (it is the same as for the Parent structure) into the matrix ReactRevStru(i2,i1) and of the degree
                  ! of the reactions into matrix ReactDegRevStru(i2,i1) at line i1=numStructures
                  do wrII_3=1,numNodes(2)  ! Number of reactions of the structure numStructures
                    ReactRevStru(wrII_3,numStructures) = supReactRevStru(wrII_3)
                  end do
                  do wrII_4=1,(3*numNodes(2))  ! Number of reactions of the structure numStructures
                    ReactDegRevStru(wrII_4,numStructures) = supReactDegRevStru(wrII_4)
                  end do
                  !
                  ! Writes the the information regarding the degree of the metabolites into matrix MetDegRevStru(i2,i1)
                  ! at line i1=numStructures
                  do wrII_5=1,(3*numNodes(1))  ! 3 x number of metabolites of structure numStructures
                    MetDegRevStru(wrII_5,numStructures) = supMetDegRevStru(wrII_5)
                  end do
                  !
                end if
                !
                ! Cleans the the variables supRevStructTemp(i2,i1), supRevStruct(i2,i1), supMetDegRevStruTemp(i1),
                ! supMetDegRevStru(i1), supReactDegRevStruTemp(i1), supReactDegRevStru(i1), supReactRevStruTemp(i1) and
                ! supReactRevStru(i1)
                do cleanII_2=1,numNodes(1)  ! Number of metabolites
                  supMetDegRevStruTemp(cleanII_2) = 0
                  supMetDegRevStru(cleanII_2) = 0
                  do cleanII_3=1,(4*numNodes(2))  ! 4 x number of reactions
                    supRevStructTemp(cleanII_3,cleanII_2,1) = 0
                    supRevStruct(cleanII_3,cleanII_2,1) = 0
                    if( ( cleanII_2 .eq. 1 ) .and. &
                    ( cleanII_3 .le. numNodes(2) ) ) then
                      supReactDegRevStruTemp(cleanII_3) = 0
                      supReactDegRevStru(cleanII_3) = 0
                      supReactRevStruTemp(cleanII_3) = 0
                      supReactRevStru(cleanII_3) = 0
                    end if
                  end do
                end do
                !
                ! Turns the variables isoStruct and isoStruRef into "0"
                isoStruct = 0
                isoStruRef = 0
                !
                ! Turns const(1) into zero
                const(1) = 0
                !
              end do
            end do
          end do
          !
          ! Turns to zero the matrices with the permutations for: i) metabolites - PermutMet(i2,i1); ii) irreversible
          ! reactions PermutIrrReact(i2,i1); iii) reversible reactions PermutRevReact(i2,i1); iv) direct and inverse
          ! single-direction reaction for all the reversible reactions - PermutRevReact(i2,i1)
          do cleanII_4=1,numPermutMet   ! Number of permutations for the metabolites
            do cleanII_5=1,numNodes(1)  ! Number of metabolites of the new found Childs of Parent iII1
              PermutMet(cleanII_5,cleanII_4) = 0
            end do
          end do
          if( numNodes(4) .ne. 0 ) then
            do cleanII_6=1,numPermutIrrReact  ! Number of permutations for the irreversible reactions
              do cleanII_7=1,numNodes(4)  ! Number of irreversible reactions
                PermutIrrReact(cleanII_7,cleanII_6) = 0
              end do
            end do
          end if
          if( numNodes(3) .ne. 0 ) then
            do cleanII_8=1,numPermutRevReact  ! Number of permutations for the reversible reactions
              do cleanII_9=1,numNodes(3)  ! Number of reversible reactions
                PermutRevReact(cleanII_9,cleanII_8) = 0
              end do
            end do
            do cleanII_10=1,numPermDirInvReact  ! Number of permutations for the direct and inverse reactions of a reversible reaction
              do cleanII_11=1,(2*numNodes(3))  ! 2x number of reversible reactions
                PermutDirInvReact(cleanII_10,cleanII_11) = 0
              end do
            end do
          end if
          ! Cleans the number of nodes of newly found Childs of Parent iII1
          numNodes(1) = 0
          numNodes(2) = 0
          numNodes(3) = 0
          numNodes(4) = 0
          ! Turns to zero the number of permutations
          numPermutMet = 0
          numPermutIrrReact = 0
          numPermutRevReact = 0
          numPermDirInvReact = 0
10      continue
      !
      !
      !
      !
      ! iii) a new reversible reaction
        !
        ! Support constant used to determine if the new structure is a isomorphic representation of another
        ! found before
        const(1) = 0
        !
        do 15 iIII1=nrefStruLow(numGen-1),nrefStruHigh(numGen-1)    ! Performs the loop to all the structures of the previous
        ! generation
          !
          ! Introduces a new reversible reaction into Parent iIII1 with at least one irreversible reaction. For that
          ! to happen, the number of reversible reactions of the Parent iIII1 must be lower than its total number of
          ! reactions
          if( DataRevStru(4,iIII1) .lt. DataRevStru(3,iIII1) ) then
            !
            ! Fills the support vector numNodes(i1), which stores the number of metabolites, reactions, reversible
            ! reactions and of irrreversible reactions of the (possibly) new structure
            ! Number of metabolites = number of metabolites of the Parent-Structure
            numNodes(1) = DataRevStru(2,iIII1)
            ! Number of reactions = number of reactions of the Parent-Structure
            numNodes(2) = DataRevStru(3,iIII1)
            ! Number of reversible reactions = number of reversible reactions of the Parent-Structure + 1
            numNodes(3) = DataRevStru(4,iIII1) + 1
            ! Number of irreversible reactions = number of reactions of the Parent-Structure - (number of reversible
            !                                  reactions of the Parent-Structure + 1)
            numNodes(4) = DataRevStru(3,iIII1) - (DataRevStru(4,iIII1) + 1)
            !
            ! Call the subroutine "allPermutations" to obtain the permutations for the metabolites
            call allPermutations(numNodes(1), PermutMet, numPermutMet)
            ! The new structures have at least one irreversible reaction?
            if( numNodes(4) .ne. 0 ) then
              ! Call the subroutine "allPermutations" to obtain the permutations for the irreversible reactions
              call allPermutations(numNodes(4), PermutIrrReact, numPermutIrrReact)
            end if
            ! The new structures have at least one reversible reaction
            ! Call the subroutine "allPermutations" to obtain the permutations for the reversible reactions
            call allPermutations(numNodes(3), PermutRevReact, numPermutRevReact)
            ! Call the subroutine "allRevReactPerm" to obtain the permutations for the direct and inverse reactions
            ! for all the (possibly) reversible reactions
            call allRevReactPerm(numNodes(3), PermutDirInvReact, numPermDirInvReact)
            !
            ! Tests all the iIII2 reactions of the Parent iIII1
            do iIII2=1,DataRevStru(3,iIII1)
              ! Is reaction iIII2 irreversible, i.e., ReactRevStru(iIII2,iIII1) equal to "0"?
              if( ReactRevStru(iIII2,iIII1) .eq. 0 ) then
                ! The reaction is irreversible. So, a new structure might be found
                !
                ! Creates the vector supRevStructTemp(iIII_2,iIII_1,1), that stores the information for a possible new structure
                do iIII_1=1,DataRevStru(2,iIII1)  ! Number of metabolites of Parent-Structure iIII1
                  ! The columns corresponding to the reactions (iIII_2) are equal to: 4 x number of reactions
                  do iIII_2=1,(4*DataRevStru(3,iIII1))  ! 4 x number of reactions of Parent-Structure iIII1
                    supRevStructTemp(iIII_2,iIII_1,1) = RevStruct(iIII_2,iIII_1,iIII1)
                  end do
                end do
                !
                ! Creates the vectors supMetDegRevStruTemp(i1) and supReactDegRevStruTemp(i1)
                do iIII_DM1=1,(3*DataRevStru(2,iIII1))  ! 3 x number of metabolites of Parent-Structure iIII1
                  supMetDegRevStruTemp(iIII_DM1) = MetDegRevStru(iIII_DM1,iIII1)
                end do
                do iIII_DR1=1,(3*DataRevStru(3,iIII1))  ! 3 x number of reactions of Parent-Structure iIII1
                  supReactDegRevStruTemp(iIII_DR1) = ReactDegRevStru(iIII_DR1,iIII1)
                end do
                !
                ! Creates the support constant const(1), which stores the possibly final reference number of
                ! this new structure (in the case it is not an isomorphic representation of a previous found
                ! structure)
                const(1) = numStructures + 1
                !
                ! Stores the general information regarding this new structure in matrix DataRevStru(i2,i1) at line
                ! i1=const(1). If we find that this new structure is a isomorphic representation of a structure
                ! found before, this line will be filled with zeros again
                ! Number of the generation
                DataRevStru(1,const(1)) = numGen
                ! Number of metabolites
                DataRevStru(2,const(1)) = DataRevStru(2,iIII1)
                ! Number of reactions
                DataRevStru(3,const(1)) = DataRevStru(3,iIII1)
                ! Number of reversible reactions: adds 1 to the number of reversible reactions of the Parent iIII1
                DataRevStru(4,const(1)) = DataRevStru(4,iIII1) + 1
                ! Number of Ancestors
                DataRevStru(5,const(1)) = 0  ! This information will be obtained latter
                ! Number of Parents
                DataRevStru(6,const(1)) = 1  ! First Parent found
                ! Number of Childs
                DataRevStru(7,const(1)) = 0
                ! Number of Offspring
                DataRevStru(8,const(1)) = 0  ! This information will be obtained latter
                ! Number of interconnections: initially equal to the number of interconnections of the Parent iIII1;
                ! extra connections are added at the same time we introduce the inverse reaction
                DataRevStru(9,const(1)) = DataRevStru(9,iIII1)
                ! Maximum possible number of interconnections: 2 x n metabolite x n of reactions
                DataRevStru(10,const(1)) = 2*DataRevStru(2,const(1))*DataRevStru(3,const(1))
                ! Number of connections metabolite/reaction (reversibility not considered): equal to the ones of the
                ! Parent iIII1
                DataRevStru(11,const(1)) = DataRevStru(11,iIII1)
                ! Maximum number of connections metabolite/reaction (reversibility not considered)
                DataRevStru(12,const(1)) = DataRevStru(2,const(1))*DataRevStru(3,const(1))
                !
                ! Checks if the metabolite iIII3 is connected to reaction iIII2. If it is, adds a connection
                ! between the metabolite iIII3 and the inverse reaction of reaction iIII2
                do iIII3=1,DataRevStru(2,iIII1)  ! Number of metabolites of Parent-Structure iIII1
                  !
                  ! The metabolite iIII3 is a reactant of reaction iIII2, i.e., RevStruct(i3,i2,i1) = 1?
                  !  i3 = [(4x(nreactions - 1)) +        0          +     1     ]
                  !                                direct reaction=0 ; reactant=1
                  if( RevStruct( ((4*(iIII2 - 1)) + 0 + 1), iIII3, iIII1) .eq. 1 ) then
                    ! Adds a connection between the inverse reaction of reaction iIII2 and metabolite iIII3
                    !  i3 = [(4x(nreactions - 1)) +         2         +     2     ]
                    !                               inverse reaction=2 ; product=2
                    supRevStructTemp( ((4*(iIII2 - 1)) + 2 + 2), iIII3, 1) = 1
                    !
                    ! Adds a new interconnection to this potential new structure
                    DataRevStru(9,const(1)) = DataRevStru(9,const(1)) + 1
                    !
                    ! Degree of metabolite iIII3: product (IN) of reaction iIII2
                    supMetDegRevStruTemp( (3*(iIII3 - 1)) + 1 ) = &
                    supMetDegRevStruTemp( (3*(iIII3 - 1)) + 1 ) + 1  ! Total connections
                    supMetDegRevStruTemp( (3*(iIII3 - 1)) + 2 ) = &
                    supMetDegRevStruTemp( (3*(iIII3 - 1)) + 2 ) + 1  ! IN connections:
                    !                                                                                              !    (product)
                    ! Degree of reaction iIII2: 1 product (OUT)
                    supReactDegRevStruTemp( (3*(iIII2 - 1)) + 1 ) = &
                    supReactDegRevStruTemp( (3*(iIII2 - 1)) + 1 ) + 1  ! Total connections
                    supReactDegRevStruTemp( (3*(iIII2 - 1)) + 3 ) = &
                    supReactDegRevStruTemp( (3*(iIII2 - 1)) + 3 ) + 1  ! OUT connections:
                    !                                                                                                  !     (1 product)
                  end if
                  !
                  ! The metabolite iIII3 is a product of reaction iIII2, i.e., RevStruct(i3,i2,i1) = 1?
                  !  i3 = [(4x(nreactions - 1)) +        0          +     2     ]
                  !                                direct reaction=0 ; product=2
                  if( RevStruct( ((4*(iIII2 - 1)) + 0 + 2), iIII3, iIII1) .eq. 1 )  then
                    ! Adds a connection between the inverse reaction of reaction iIII2 and metabolite iIII3
                    !  i3 = [(4x(nreactions - 1)) +         2         +     1     ]
                    !                               inverse reaction=2 ; reactant=1
                    supRevStructTemp( ((4*(iIII2 - 1)) + 2 + 1), iIII3, 1) = 1
                    !
                    ! Adds a new interconnection to this potential new structure
                    DataRevStru(9,const(1)) = DataRevStru(9,const(1)) + 1
                    !
                    ! Degree of metabolite iIII3: reactant (OUT) of reaction iIII2
                    supMetDegRevStruTemp( (3*(iIII3 - 1)) + 1 ) = &
                    supMetDegRevStruTemp( (3*(iIII3 - 1)) + 1 ) + 1  ! Total connections
                    supMetDegRevStruTemp( (3*(iIII3 - 1)) + 3 ) = &
                    supMetDegRevStruTemp( (3*(iIII3 - 1)) + 3 ) + 1  ! OUT connections:
                    !                                                                                              !     (reactant)
                    ! Degree of reaction iIII2: 1 reactant (IN)
                    supReactDegRevStruTemp( (3*(iIII2 - 1)) + 1 ) = &
                    supReactDegRevStruTemp( (3*(iIII2 - 1)) + 1 ) + 1  ! Total connections
                    supReactDegRevStruTemp( (3*(iIII2 - 1)) + 2 ) = &
                    supReactDegRevStruTemp( (3*(iIII2 - 1)) + 2 ) + 1  ! IN connections:
                    !                                                                                                  !    (1 reactant)
                  end if
                end do
                !
                ! Stores the information of the reversibility of the reactions for this (possibly) new structure
                do iIII_rev1=1,DataRevStru(3,iIII1)  ! Number of reactions of the Parent-Structure iIII1
                  supReactRevStruTemp(iIII_rev1) = ReactRevStru(iIII_rev1,iIII1)
                end do
                ! Reaction iIII2 is now reversible
                supReactRevStruTemp(iIII2) = 1
                !
                ! Call the subroutine organizingStruct
                call organizingStruct(supRevStructTemp, supMetDegRevStruTemp, &
                supReactDegRevStruTemp, numNodes, supReactRevStruTemp, supRevStruct, &
                supMetDegRevStru, supReactDegRevStru, supReactRevStru)
                !
                !
                ! Is this a isomorphic representation of a structure found before?
                ! Call the subroutine isomorphismTest
                call isomorphismTest(supRevStruct, RevStruct, DataRevStru, &
                MetDegRevStru, ReactDegRevStru, supMetDegRevStru, &
                supReactDegRevStru, nrefStruLow(numGen), nrefStruHigh(numGen), &
                const(1), PermutMet, PermutIrrReact, PermutRevReact, &
                PermutDirInvReact, numPermutMet, numPermutIrrReact, &
                numPermutRevReact, numPermDirInvReact, isoStruct, isoStruRef)
                ! If isoStruct = 1, the structure is a isomorphic representation of structure isoStruRef;
                if(isoStruct .eq. 1) then
                  !
                  ! Is this a new Parent/Child relantionship? If it is, the correspinding cell in matrix FamRevStru(i2,i1)
                  ! must be "0"
                  if(FamRevStru(isoStruRef,iIII1) .eq. 0) then
                    !
                    ! Structure iIII1 is a Parent of the structure isoStruRef, found before. One adds "4" to the corresponding
                    ! cell in matrix FamRevStru(i2,i1), to show that they only differ of 1 reversible reaction
                    FamRevStru(isoStruRef,iIII1) = 4
                    !
                    ! Number of Parents of structure isoStruRef: adds structure iIII1 as one of its Parents
                    DataRevStru(6,isoStruRef) = DataRevStru(6,isoStruRef) + 1
                    !
                    ! Number of Childs of structure iIII1: adds structure isoStruRef as one of its Childs
                    DataRevStru(7,iIII1) = DataRevStru(7,iIII1) + 1
                    !
                  end if
                  !
                  ! The matrix DataRevStru(i2,i1) must be turned to zeros at line i1=const(1)
                  do cleanIII_1=1,12
                    DataRevStru(cleanIII_1,const(1)) = 0
                  end do
                  !
                ! If isoStruct = 0, this structure was found for the first time
                else
                  !
                  ! Adds a new structure to the total number of structures
                  numStructures = numStructures + 1
                  ! Adds a new structure to the total number of structures of this generation (the goal is to store
                  ! the highest reference number of a structure of this generation)
                  nrefStruHigh(numGen) = nrefStruHigh(numGen) + 1
                  ! Adds a new structure to the total number of structures of this generation
                  numStruGen(numGen) = numStruGen(numGen) + 1
                  write(*,2002) numStruGen(numGen), numGen, numStructures
                  !
                  ! Stores the relationship Parent/Child between the Parent iIII1 and the new Child numStructures. One adds "4"
                  ! to the corresponding cell in matrix FamRevStru(i2,i1), to show that they only differ of 1 reversible reaction
                  FamRevStru(numStructures,iIII1) = 4
                  !
                  ! Number of Childs of structure iIII1: adds structure numStructures as one of its Childs
                  DataRevStru(7,iIII1) = DataRevStru(7,iIII1) + 1
                  !
                  ! Writes the new structure to the matrix RevStruct(i3,i2,i1) at line i1=numStructures
                  do wrIII_1=1,DataRevStru(2,iIII1)  ! Number of metabolites of the Parent iIII1 (equal to the number of
                  ! metabolites of it's Childs)
                    do wrIII_2=1,(4*DataRevStru(3,iIII1))  ! 4 x number of reactions of the Parent iIII1 (equal to the number
                    ! of reactions of it's Childs)
                      ! Writes the new structure into matrix RevStruct(i3,i2,i1) at line i1=numStructures
                      RevStruct(wrIII_2,wrIII_1,numStructures) = &
                      supRevStruct(wrIII_2,wrIII_1,1)
                      ! Turns to zero the suport matrix supRevStruct(i3,i2,1)
                      supRevStruct(wrIII_2,wrIII_1,1) = 0
                    end do
                  end do
                  !
                  ! Writes the information regarding the reversibility of the reactions of the new found structure
                  ! (it is the same as for the Parent structure) into the matrix ReactRevStru(i2,i1) and of the degree
                  ! of the reactions into matrix ReactDegRevStru(i2,i1) at line i1=numStructures
                  do wrIII_3=1,numNodes(2)  ! Number of reactions of the structure numStructures
                    ReactRevStru(wrIII_3,numStructures) = supReactRevStru(wrIII_3)
                  end do
                  do wrIII_4=1,(3*numNodes(2))  ! Number of reactions of the structure numStructures
                    ReactDegRevStru(wrIII_4,numStructures) = supReactDegRevStru(wrIII_4)
                  end do
                  !
                  ! Writes the the information regarding the degree of the metabolites into matrix MetDegRevStru(i2,i1)
                  ! at line i1=numStructures
                  do wrIII_5=1,(3*numNodes(1))  ! 3 x number of metabolites of structure numStructures
                    MetDegRevStru(wrIII_5,numStructures) = supMetDegRevStru(wrIII_5)
                  end do
                  !
                end if
                !
                ! Cleans the the variables supRevStructTemp(i2,i1), supRevStruct(i2,i1), supMetDegRevStruTemp(i1),
                ! supMetDegRevStru(i1), supReactDegRevStruTemp(i1), supReactDegRevStru(i1), supReactRevStruTemp(i1) and
                ! supReactRevStru(i1)
                do cleanIII_2=1,numNodes(1)  ! Number of metabolites
                  supMetDegRevStruTemp(cleanIII_2) = 0
                  supMetDegRevStru(cleanIII_2) = 0
                  do cleanIII_3=1,(4*numNodes(2))  ! 4 x number of reactions
                    supRevStructTemp(cleanIII_3,cleanIII_2,1) = 0
                    supRevStruct(cleanIII_3,cleanIII_2,1) = 0
                    if( ( cleanIII_2 .eq. 1 ) .and. &
                    ( cleanIII_3 .le. numNodes(2) ) ) then
                      supReactDegRevStruTemp(cleanIII_3) = 0
                      supReactDegRevStru(cleanIII_3) = 0
                      supReactRevStruTemp(cleanIII_3) = 0
                      supReactRevStru(cleanIII_3) = 0
                    end if
                  end do
                end do
                !
                ! Turns the variables isoStruct and isoStruRef into "0"
                isoStruct = 0
                isoStruRef = 0
                !
                ! Turns const(1) into zero
                const(1) = 0
                !
              end if
            end do
            !
            ! Turns to zero the matrices with the permutations for: i) metabolites - PermutMet(i2,i1); ii) irreversible
            ! reactions PermutIrrReact(i2,i1); iii) reversible reactions PermutRevReact(i2,i1); iv) direct and inverse
            ! single-direction reaction for all the reversible reactions - PermutRevReact(i2,i1)
            do cleanIII_4=1,numPermutMet    ! Number of permutations for the metabolites
              do cleanIII_5=1,numNodes(1)  ! Number of metabolites of the new found Childs of Parent iIII1
                PermutMet(cleanIII_5,cleanIII_4) = 0
              end do
            end do
            if( numNodes(4) .ne. 0 ) then
              do cleanIII_6=1,numPermutIrrReact  ! Number of permutations for the irreversible reactions
                do cleanIII_7=1,numNodes(4)  ! Number of irreversible reactions
                  PermutIrrReact(cleanIII_7,cleanIII_6) = 0
                end do
              end do
            end if
            do cleanIII_8=1,numPermutRevReact  ! Number of permutations for the reversible reactions
              do cleanIII_9=1,numNodes(3)  ! Number of reversible reactions
                PermutRevReact(cleanIII_9,cleanIII_8) = 0
              end do
            end do
            do cleanIII_10=1,numPermDirInvReact ! Number of permutations for the direct and inverse reactions of a reversible reaction
              do cleanIII_11=1,(2*numNodes(3))  ! 2x number of reversible reactions
                PermutDirInvReact(cleanIII_10,cleanIII_11) = 0
              end do
            end do
            ! Cleans the number of nodes of newly found Childs of Parent iIII1
            numNodes(1) = 0
            numNodes(2) = 0
            numNodes(3) = 0
            numNodes(4) = 0
            ! Turns to zero the number of permutations
            numPermutMet = 0
            numPermutIrrReact = 0
            numPermutRevReact = 0
            numPermDirInvReact = 0
          end if
15      continue
      !
      !
      !
      !
      ! iv) a new metabolite (as a product or as a reactant to a existing irreversible or reversible reaction)
        !
        ! Support constant used to determine if the new structure is a isomorphic representation of another
        ! found before
        const(1) = 0
        !
        do 20 iIV1=nrefStruLow(numGen-1),nrefStruHigh(numGen-1)  ! Performs the loop to all the structures of the previous
        ! generation
          !
          ! The Child-structures of the Parent-structure iIV1 has a number of metabolites superior to maxNumMet?
          if( (DataRevStru(2,iIV1) + 1) .gt. maxNumMet ) goto 20
          !
          !
          ! Fills the support vector numNodes(i1), which stores the number of metabolites, reactions, reversible
          ! reactions and of irrreversible reactions of the (possibly) new structure
          ! Number of metabolites = number of metabolites of the Parent-Structure + 1
          numNodes(1) = DataRevStru(2,iIV1) + 1
          ! Number of reactions = number of reactions of the Parent-Structure
          numNodes(2) = DataRevStru(3,iIV1)
          ! Number of reversible reactions = number of reversible reactions of the Parent-Structure
          numNodes(3) = DataRevStru(4,iIV1)
          ! Number of irreversible reactions = number of reactions of the Parent-Structure - number of reversible
          !                                    reactions of the Parent-Structure
          numNodes(4) = DataRevStru(3,iIV1) - DataRevStru(4,iIV1)
          !
          ! Call the subroutine "allPermutations" to obtain the permutations for the metabolites
          call allPermutations(numNodes(1), PermutMet, numPermutMet)
          ! The new structures have at least one irreversible reaction?
          if( numNodes(4) .ne. 0 ) then
            ! Call the subroutine "allPermutations" to obtain the permutations for the irreversible reactions
            call allPermutations(numNodes(4), PermutIrrReact, numPermutIrrReact)
          end if
          ! The new structures have at least one reversible reaction?
          if( numNodes(3) .ne. 0 ) then
            ! Call the subroutine "allPermutations" to obtain the permutations for the reversible reactions
            call allPermutations(numNodes(3), PermutRevReact, numPermutRevReact)
            ! Call the subroutine "allRevReactPerm" to obtain the permutations for the direct and inverse reactions
            ! for all the (possibly) reversible reactions
            call allRevReactPerm(numNodes(3), PermutDirInvReact, numPermDirInvReact)
          end if
          !
          ! Introduces a connection between one of the existing reactions iIV2 and a new metabolite
          do iIV2=1,DataRevStru(3,iIV1)  ! Number of reactions of the Parent iIV1 (equal to the number reactions
          ! of the Child structures
            !
            ! Introduces the connection between the new metabolite and the reaction iIV2: i) as a reactant, iIV3=1, and then
            ! ii) as a product, iV3=2
            do iIV3=1,2
              !
              ! Creates the vector supRevStructTemp(iIV_2,iIV_1,1), that stores the information for a possible new structure
              do iIV_1=1,(DataRevStru(2,iIV1)+1)  ! Number of metabolites of Parent iIV1 + 1
                ! The columns corresponding to the reactions (iIV_2) are equal to: 4 x number of reactions
                do iIV_2=1,(4*DataRevStru(3,iIV1))  ! 4 x number of reactions of Parent IV1
                  supRevStructTemp(iIV_2,iIV_1,1) = RevStruct(iIV_2,iIV_1,iIV1)
                end do
              end do
              ! Creates the vectors supMetDegRevStruTemp(i1) and supReactDegRevStruTemp(i1)
              do iIV_DM1=1,(3*(DataRevStru(2,iIV1)+1))  ! 3 x (number of metabolites of Parent-Structure iIV1 + 1)
                supMetDegRevStruTemp(iIV_DM1) = MetDegRevStru(iIV_DM1,iIV1)
              end do
              do iIV_DR1=1,(3*DataRevStru(3,iIV1))  ! 3 x number of reactions of Parent-Structure iIV1
                supReactDegRevStruTemp(iIV_DR1) = ReactDegRevStru(iIV_DR1,iIV1)
              end do
              !
              ! Creates the support constant const(1), which stores the possibly final reference number of
              ! this new structure (in the case it is not an isomorphic representation of a previous found
              ! structure)
              const(1) = numStructures + 1
              !
              ! Stores the general information regarding this new structure in matrix DataRevStru(i2,i1) at line
              ! i1=const(1). If we find that this new structure is a isomorphic representation of a structure
              ! found before, this line will be filled with zeros again
              ! Number of the generation
              DataRevStru(1,const(1)) = numGen
              ! Number of metabolites: adds a new metabolite to the total of Parent iIV1
              DataRevStru(2,const(1)) = DataRevStru(2,iIV1) + 1
              ! Number of reactions
              DataRevStru(3,const(1)) = DataRevStru(3,iIV1)
              ! Number of reversible reactions
              DataRevStru(4,const(1)) = DataRevStru(4,iIV1)
              ! Number of Ancestors
              DataRevStru(5,const(1)) = 0  ! This information will be obtained latter
              ! Number of Parents
              DataRevStru(6,const(1)) = 1  ! First Parent found
              ! Number of Childs
              DataRevStru(7,const(1)) = 0
              ! Number of Offspring
              DataRevStru(8,const(1)) = 0  ! This information will be obtained latter
              ! Number of interconnections
              if( ReactRevStru(iIV2,iIV1) .eq. 0 ) then
                ! If the reaction iIV2 is irreversible, i.e., ReactRevStru(iIV2,iIV1) = 0, adds 1 new interconnection
                ! to the total of Parent iIV1
                DataRevStru(9,const(1)) = DataRevStru(9,iIV1) + 1
              else
                ! If the reaction iIV2 is reversible, i.e., ReactRevStru(iIV2,iIV1) = 1, adds 2 new interconnections
                ! to the total of Parent iIV1
                DataRevStru(9,const(1)) = DataRevStru(9,iIV1) + 2
              end if
              ! Maximum possible number of interconnections: 2 x n metabolite x n of reactions
              DataRevStru(10,const(1)) = 2*DataRevStru(2,const(1))*DataRevStru(3,const(1))
              ! Number of connections metabolite/reaction (reversibility not considered): adds 1 to the total number
              ! of Parent iIV1
              DataRevStru(11,const(1)) = DataRevStru(11,iIV1) + 1
              ! Maximum number of connections metabolite/reaction (reversibility not considered)
              DataRevStru(12,const(1)) = DataRevStru(2,const(1)) * DataRevStru(3,const(1))
              !
              ! Adds the new metabolite
              ! If iIV3=1, adds the new metabolite numNodes(1) as a reactant of the direct reaction
              if( iIV3 .eq. 1 ) then
                !
                ! New metabolite DataRevStru(2,const(1)) is a reactant of the direct reaction, i.e., supRevStruct(i3,i2,i1) = 1,
                !    where i3 = [(4x(nreactions - 1)) +        0          +     1     ]
                !                                        direct reaction=0 ; reactant=1
                supRevStructTemp( ((4*(iIV2 - 1)) + 0 + 1) , numNodes(1), 1) = 1
                !
                ! Degree of metabolite numNodes(1): reactant (OUT) of reaction iIV2
                supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 1 ) = &
                supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 1 ) + 1 ! Total connect.
                supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 3 ) = &
                supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 3 ) + 1 ! OUT connections:
                !                                                                                                          (reactant)
                ! Degree of reaction iIV2: 1 reactant (IN)
                supReactDegRevStruTemp( (3*(iIV2 - 1)) + 1 ) = &
                supReactDegRevStruTemp( (3*(iIV2 - 1)) + 1 ) + 1 ! Total connections
                supReactDegRevStruTemp( (3*(iIV2 - 1)) + 2 ) = &
                supReactDegRevStruTemp( (3*(iIV2 - 1)) + 2 ) + 1 ! IN connections: 1 reactant
                !
                ! Is the reaction iIV2 reversible?
                if( ReactRevStru(iIV2,iIV1) .eq. 1 ) then
                  ! The reaction iIV2 is reversible
                  ! New metabolite numNodes(1) is a product of the inverse reaction, i.e., supRevStruct(i3,i2,i1) = 1,
                  !  where i3 = [(4x(nreactions - 1)) +        2          +     2     ]
                  !                                     inverse reaction=2 ; product=2
                  supRevStructTemp( ((4*(iIV2 - 1)) + 2 + 2) , numNodes(1), 1) = 1
                  !
                  ! Degree of metabolite numNodes(1): product (IN) of reaction iIV2
                  supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 1 ) = &
                  supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 1 ) + 1 ! Total connect.
                  supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 2 ) = &
                  supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 2 ) + 1 ! OUT connect.:
                  !                                                                                                          (product)
                  ! Degree of reaction iIV2: 1 product (OUT)
                  supReactDegRevStruTemp( (3*(iIV2 - 1)) + 1 ) = &
                  supReactDegRevStruTemp( (3*(iIV2 - 1)) + 1 ) + 1 ! Total connections
                  supReactDegRevStruTemp( (3*(iIV2 - 1)) + 3 ) = &
                  supReactDegRevStruTemp( (3*(iIV2 - 1)) + 3 ) + 1 ! IN connections:
                  !                                                                                                   (1 product)
                end if
                !
              ! If iIV3=2, adds the new metabolite numNodes(1) as a product of the direct reaction
              else
                !
                ! New metabolite numNodes(1) is a product of the direct reaction, i.e., supRevStruct(i3,i2,i1) = 1,
                !    where i3 = [(4x(nreactions - 1)) +        0          +     2     ]
                !                                        direct reaction=0 ; product=2
                supRevStructTemp( ((4*(iIV2 - 1)) + 0 + 2) , numNodes(1), 1) = 1
                !
                ! Degree of metabolite numNodes(1): product (IN) of reaction iIV2
                supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 1 ) = &
                supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 1 ) + 1 ! Total connect.
                supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 2 ) = &
                supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 2 ) + 1 ! OUT connect.:
                !                                                                                                            (product)
                ! Degree of reaction iIV2: 1 product (OUT)
                supReactDegRevStruTemp( (3*(iIV2 - 1)) + 1 ) = &
                supReactDegRevStruTemp( (3*(iIV2 - 1)) + 1 ) + 1 ! Total connections
                supReactDegRevStruTemp( (3*(iIV2 - 1)) + 3 ) = &
                supReactDegRevStruTemp( (3*(iIV2 - 1)) + 3 ) + 1 ! IN connections:
                !                                                                                                     (1 product)
                ! Is the reaction iIV2 reversible?
                if( ReactRevStru(iIV2,iIV1) .eq. 1 ) then
                  ! The reaction iIV2 is reversible
                  ! New metabolite numNodes(1) is a reactant of the inverse reaction, i.e., supRevStruct(i3,i2,i1) = 1,
                  !  where i3 = [(4x(nreactions - 1)) +        2          +     1     ]
                  !                                     inverse reaction=2 ; reactant=1
                  supRevStructTemp( ((4*(iIV2 - 1)) + 2 + 1) , numNodes(1), 1) = 1
                  !
                  ! Degree of metabolite numNodes(1): reactant (OUT) of reaction iIV2
                  supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 1 ) = &
                  supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 1 ) + 1 ! Total connect.
                  supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 3 ) = &
                  supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 3 ) + 1 ! OUT connect.:
                  !                                                                                                          (reactant)
                  ! Degree of reaction iIV2: 1 reactant (IN)
                  supReactDegRevStruTemp( (3*(iIV2 - 1)) + 1 ) = &
                  supReactDegRevStruTemp( (3*(iIV2 - 1)) + 1 ) + 1 ! Total connections
                  supReactDegRevStruTemp( (3*(iIV2 - 1)) + 2 ) = &
                  supReactDegRevStruTemp( (3*(iIV2 - 1)) + 2 ) + 1 ! IN connections:
                  !                                                                                                   (1 reactant)
                end if
              end if
              !
              ! Stores the information of the reversibility of the reactions for this (possibly) new structure
              do iIV_rev1=1,DataRevStru(3,iIV1)  ! Number of reactions of the Parent-Structure iIV1
                supReactRevStruTemp(iIV_rev1) = ReactRevStru(iIV_rev1,iIV1)
              end do
              !
              ! Call the subroutine organizingStruct
              call organizingStruct(supRevStructTemp, supMetDegRevStruTemp, &
              supReactDegRevStruTemp, numNodes, supReactRevStruTemp, supRevStruct, &
              supMetDegRevStru, supReactDegRevStru, supReactRevStru)
              !
              !
              ! Is this a isomorphic representation of a structure found before?
              ! Call the subroutine isomorphismTest
              call isomorphismTest(supRevStruct, RevStruct, DataRevStru, &
              MetDegRevStru, ReactDegRevStru, supMetDegRevStru, &
              supReactDegRevStru, nrefStruLow(numGen), nrefStruHigh(numGen), &
              const(1), PermutMet, PermutIrrReact, PermutRevReact, &
              PermutDirInvReact, numPermutMet, numPermutIrrReact, &
              numPermutRevReact, numPermDirInvReact, isoStruct, isoStruRef)
              ! If isoStruct = 1, the structure is a isomorphic representation of structure isoStruRef;
              if(isoStruct .eq. 1) then
                !
                ! Is this a new Parent/Child relantionship? If it is, the correspinding cell in matrix FamRevStru(i2,i1)
                ! must be "0"
                if(FamRevStru(isoStruRef,iIV1) .eq. 0) then
                  !
                  ! Number of Parents of structure isoStruRef: adds structure iIV1 as one of its Parents
                  DataRevStru(6,isoStruRef) = DataRevStru(6,isoStruRef) + 1
                  !
                  ! Number of Childs of structure iIV1: adds structure isoStruRef as one of its Childs
                  DataRevStru(7,iIV1) = DataRevStru(7,iIV1) + 1
                  !
                  ! Structure iIV1 is a Parent of the structure isoStruRef, found before
                  ! Is reaction iIV2 reversible or irreversible?
                  if( ReactRevStru(iIV2,iIV1) .eq. 0 ) then
                    !
                    ! Reaction iIV2 is irreversible. One adds "5" to the corresponding cell in matrix FamRevStru(i2,i1),
                    ! to show that they only differ of 1 metabolite connected to a irreversible reaction
                    FamRevStru(isoStruRef,iIV1) = 5
                    !
                  else
                    ! Reaction iIV2 is reversible. One adds "6" to the corresponding cell in matrix FamRevStru(i2,i1),
                    ! to show that they only differ of 1 metabolite connected to a reversible reaction
                    FamRevStru(isoStruRef,iIV1) = 6
                    !
                  end if
                end if
                !
                ! The matrix DataRevStru(i2,i1) must be turned to zeros at line i1=const(1)
                do cleanIV_1=1,12
                  DataRevStru(cleanIV_1,const(1)) = 0
                end do
                !
              ! If isoStruct = 0, this structure was found for the first time
              else
                !
                ! Adds a new structure to the total number of structures
                numStructures = numStructures + 1
                ! Adds a new structure to the total number of structures of this generation (the goal is to store
                ! the highest reference number of a structure of this generation)
                nrefStruHigh(numGen) = nrefStruHigh(numGen) + 1
                ! Adds a new structure to the total number of structures of this generation
                numStruGen(numGen) = numStruGen(numGen) + 1
                write(*,2002) numStruGen(numGen), numGen, numStructures
                !
                ! Structure iIV1 is a Parent of the structure numStructures
                ! Is reaction iIV2 reversible or irreversible?
                if( ReactRevStru(iIV2,iIV1) .eq. 0 ) then
                  !
                  ! Reaction iIV2 is irreversible. One adds "5" to the corresponding cell in matrix FamRevStru(i2,i1),
                  ! to show that they only differ of 1 metabolite connected to a irreversible reaction
                  FamRevStru(numStructures,iIV1) = 5
                  !
                else
                  ! Reaction iIV2 is reversible. One adds "6" to the corresponding cell in matrix FamRevStru(i2,i1),
                  ! to show that they only differ of 1 metabolite connected to a reversible reaction
                  FamRevStru(numStructures,iIV1) = 6
                  !
                end if
                !
                ! Number of Childs of structure iIV1: adds structure numStructures as one of its Childs
                DataRevStru(7,iIV1) = DataRevStru(7,iIV1) + 1
                !
                ! Writes the new structure to the matrix RevStruct(i3,i2,i1) at line i1=numStructures
                do wrIV_1=1,(DataRevStru(2,iIV1)+1)  ! Number of metabolites of the Parent iIV1 + 1 (equal to the number of
                ! metabolites of it's Childs)
                  do wrIV_2=1,(4*DataRevStru(3,iIV1))  ! 4 x number of reactions of the Parent iIV1 (equal to the number
                  ! of reactions of it's Childs)
                    ! Writes the new structure into matrix RevStruct(i3,i2,i1) at line i1=numStructures
                    RevStruct(wrIV_2,wrIV_1,numStructures) = supRevStruct(wrIV_2,wrIV_1,1)
                    ! Turns to zero the suport matrix supRevStruct(i3,i2,1)
                    supRevStruct(wrIV_2,wrIV_1,1) = 0
                  end do
                end do
                !
                ! Writes the information regarding the reversibility of the reactions of the new found structure
                ! (it is the same as for the Parent structure) into the matrix ReactRevStru(i2,i1) and of the degree
                ! of the reactions into matrix ReactDegRevStru(i2,i1) at line i1=numStructures
                do wrIV_3=1,numNodes(2)  ! Number of reactions of the structure numStructures
                  ReactRevStru(wrIV_3,numStructures) = supReactRevStru(wrIV_3)
                end do
                do wrIV_4=1,(3*numNodes(2))  ! Number of reactions of the structure numStructures
                  ReactDegRevStru(wrIV_4,numStructures) = supReactDegRevStru(wrIV_4)
                end do
                !
                ! Writes the the information regarding the degree of the metabolites into matrix MetDegRevStru(i2,i1)
                ! at line i1=numStructures
                do wrIV_5=1,(3*numNodes(1))  ! 3 x number of metabolites of structure numStructures
                  MetDegRevStru(wrIV_5,numStructures) = supMetDegRevStru(wrIV_5)
                end do
                !
              end if
              !
              ! Cleans the the variables supRevStructTemp(i2,i1), supRevStruct(i2,i1), supMetDegRevStruTemp(i1),
              ! supMetDegRevStru(i1), supReactDegRevStruTemp(i1), supReactDegRevStru(i1), supReactRevStruTemp(i1) and
              ! supReactRevStru(i1)
              do cleanIV_2=1,numNodes(1)  ! Number of metabolites
                supMetDegRevStruTemp(cleanIV_2) = 0
                supMetDegRevStru(cleanIV_2) = 0
                do cleanIV_3=1,(4*numNodes(2))  ! 4 x number of reactions
                  supRevStructTemp(cleanIV_3,cleanIV_2,1) = 0
                  supRevStruct(cleanIV_3,cleanIV_2,1) = 0
                  if( ( cleanIV_2 .eq. 1 ) .and. &
                  ( cleanIV_3 .le. numNodes(2) ) ) then
                    supReactDegRevStruTemp(cleanIV_3) = 0
                    supReactDegRevStru(cleanIV_3) = 0
                    supReactRevStruTemp(cleanIV_3) = 0
                    supReactRevStru(cleanIV_3) = 0
                  end if
                end do
              end do
              !
              ! Turns the variables isoStruct and isoStruRef into "0"
              isoStruct = 0
              isoStruRef = 0
              !
              ! Turns const(1) into zero
              const(1) = 0
              !
            end do
          end do
          !
          ! Turns to zero the matrices with the permutations for: i) metabolites - PermutMet(i2,i1); ii) irreversible
          ! reactions PermutIrrReact(i2,i1); iii) reversible reactions PermutRevReact(i2,i1); iv) direct and inverse
          ! single-direction reaction for all the reversible reactions - PermutRevReact(i2,i1)
          do cleanIV_4=1,numPermutMet   ! Number of permutations for the metabolites
            do cleanIV_5=1,numNodes(1)  ! Number of metabolites of the new found Childs of Parent iIV1
              PermutMet(cleanIV_5,cleanIV_4) = 0
            end do
          end do
          if( numNodes(4) .ne. 0 ) then
            do cleanIV_6=1,numPermutIrrReact  ! Number of permutations for the irreversible reactions
              do cleanIV_7=1,numNodes(4)  ! Number of irreversible reactions
                PermutIrrReact(cleanIV_7,cleanIV_6) = 0
              end do
            end do
          end if
          if( numNodes(3) .ne. 0 ) then
            do cleanIV_8=1,numPermutRevReact  ! Number of permutations for the reversible reactions
              do cleanIV_9=1,numNodes(3)  ! Number of reversible reactions
                PermutRevReact(cleanIV_9,cleanIV_8) = 0
              end do
            end do
            do cleanIV_10=1,numPermDirInvReact  ! Number of permutations for the direct and inverse reactions of a reversible reaction
              do cleanIV_11=1,(2*numNodes(3))  ! 2x number of reversible reactions
                PermutDirInvReact(cleanIV_10,cleanIV_11) = 0
              end do
            end do
          end if
          ! Cleans the number of nodes of newly found Childs of Parent iIV1
          numNodes(1) = 0
          numNodes(2) = 0
          numNodes(3) = 0
          numNodes(4) = 0
          ! Turns to zero the number of permutations
          numPermutMet = 0
          numPermutIrrReact = 0
          numPermutRevReact = 0
          numPermDirInvReact = 0
20      continue
      !
      !
      !
      !
      ! v) a new metabolite + a new reaction (the new reaction must connect the new metabolite to a existing
      ! metabolite to create a new pair reactant/product)
        !
        ! Support constant used to determine if the new structure is a isomorphic representation of another
        ! found before
        const(1) = 0
        !
        do 25 iV1=nrefStruLow(numGen-1),nrefStruHigh(numGen-1)  ! Performs the loop to all the structures of the previous
        ! generation
          !
          ! The Child-structures of the Parent-structure iV1 has a number of metabolites superior to maxNumMet?
          if( (DataRevStru(2,iV1) + 1) .gt. maxNumMet ) goto 25
          ! The Child-structures of the Parent-structure iV1 has a number of reactions superior to maxNumReact?
          if( (DataRevStru(3,iV1) + 1) .gt. maxNumReact ) goto 25
          !
          !
          ! Fills the support vector numNodes(i1), which stores the number of metabolites, reactions, reversible
          ! reactions and of irrreversible reactions of the (possibly) new structure
          ! Number of metabolites   = number of metabolites of the Parent-Structure + 1
          numNodes(1) = DataRevStru(2,iV1) + 1
          ! Number of reactions = number of reactions of the Parent-Structure + 1 (Adds the new reaction)
          numNodes(2) = DataRevStru(3,iV1) + 1
          ! Number of reversible reactions = number of reversible reactions of the Parent-Structure
          numNodes(3) = DataRevStru(4,iV1)
          ! Number of irreversible reactions = (number of reactions of the Parent-Structure + 1) - number of reversible
          !                                    reactions of the Parent-Structure
          numNodes(4) = (DataRevStru(3,iV1) + 1) - DataRevStru(4,iV1)
          !
          ! Call the subroutine "allPermutations" to obtain the permutations for the metabolites
          call allPermutations(numNodes(1), PermutMet, numPermutMet)
          ! Call the subroutine "allPermutations" to obtain the permutations for the irreversible reactions
          call allPermutations(numNodes(4), PermutIrrReact, numPermutIrrReact)
          ! The new structures have at least one reversible reaction?
          if( numNodes(3) .ne. 0 ) then
            ! Call the subroutine "allPermutations" to obtain the permutations for the reversible reactions
            call allPermutations(numNodes(3), PermutRevReact, numPermutRevReact)
            ! Call the subroutine "allRevReactPerm" to obtain the permutations for the direct and inverse reactions
            ! for all the (possibly) reversible reactions
            call allRevReactPerm(numNodes(3), PermutDirInvReact, numPermDirInvReact)
          end if
          !
          ! A new reaction is used to connect one of the existing metabolite iV2 and a new metabolite as pairs
          ! reactant/ product and product/reactant
          do iV2=1,DataRevStru(2,iV1)  ! Number of metabolites of Parent iV1
            !
            ! Connecting metabolite iV2 and the new metabolite with the new reaction: iV3=1, reactant/product;
            ! iV3=2, product/reactant
            do iV3=1,2
              !
              ! Creates the vector supRevStructTemp(iV_2,iV_1,1), that stores the information for a possible new structure
              do iV_1=1,(DataRevStru(2,iV1)+1)  ! Number of metabolites of Parent iV1 + 1
                ! The columns corresponding to the reactions (iV_2) are equal to: 4 x number of reactions of the new structures,
                do iV_2=1,(4*(DataRevStru(3,iV1)+1))  ! which are equal to the number of reactions of Parent iV1 + 1
                  supRevStructTemp(iV_2,iV_1,1) = RevStruct(iV_2,iV_1,iV1)
                end do
              end do
              ! Creates the vectors supMetDegRevStruTemp(i1) and supReactDegRevStruTemp(i1)
              do iV_DM1=1,(3*(DataRevStru(2,iV1)+1))  ! 3 x (number of metabolites of Parent-Structure iV1 + 1)
                supMetDegRevStruTemp(iV_DM1) = MetDegRevStru(iV_DM1,iV1)
              end do
              do iV_DR1=1,(3*(DataRevStru(3,iV1)+1))  ! 3 x (number of reactions of Parent-Structure iV1 + 1)
                supReactDegRevStruTemp(iV_DR1) = ReactDegRevStru(iV_DR1,iV1)
              end do
              !
              ! Creates the support constant const(1), which stores the possibly final reference number of
              ! this new structure (in the case it is not an isomorphic representation of a previous found
              ! structure)
              const(1) = numStructures + 1
              !
              ! Stores the general information regarding this new structure in matrix DataRevStru(i2,i1) at line
              ! i1=const(1). If we find that this new structure is a isomorphic representation of a structure
              ! found before, this line will be filled with zeros again
              ! Number of the generation
              DataRevStru(1,const(1)) = numGen
              ! Number of metabolites: adds a new metabolite to the total of Parent iV1
              DataRevStru(2,const(1)) = DataRevStru(2,iV1) + 1
              ! Number of reactions: adds a new reaction to the total of Parent iV1
              DataRevStru(3,const(1)) = DataRevStru(3,iV1) + 1
              ! Number of reversible reactions
              DataRevStru(4,const(1)) = DataRevStru(4,iV1)
              ! Number of Ancestors
              DataRevStru(5,const(1)) = 0  ! This information will be obtained latter
              ! Number of Parents
              DataRevStru(6,const(1)) = 1  ! First Parent found
              ! Number of Childs
              DataRevStru(7,const(1)) = 0
              ! Number of Offspring
              DataRevStru(8,const(1)) = 0  ! This information will be obtained latter
              ! Number of interconnections: adds 2 interconnections to the total of Parent iV1
              DataRevStru(9,const(1)) = DataRevStru(9,iV1) + 2
              ! Maximum possible number of interconnections: 2 x n metabolites x n of reactions
              DataRevStru(10,const(1)) = 2*DataRevStru(2,const(1))*DataRevStru(3,const(1))
              ! Number of connections metabolite/reaction (reversibility not considered): adds 2 to the total number
              ! of Parent iV1
              DataRevStru(11,const(1)) = DataRevStru(11,iV1) + 2
              ! Maximum number of connections metabolite/reaction (reversibility not considered)
              DataRevStru(12,const(1)) = DataRevStru(2,const(1)) * DataRevStru(3,const(1))
              !
              !
              ! Degree of reaction numNodes(2): 1 reactant (IN) and 1 product (OUT)
              supReactDegRevStruTemp( (3*(numNodes(2) - 1)) + 1 ) = &
              supReactDegRevStruTemp( (3*(numNodes(2) - 1)) + 1 ) + 2 ! Total connec.
              supReactDegRevStruTemp( (3*(numNodes(2) - 1)) + 2 ) = &
              supReactDegRevStruTemp( (3*(numNodes(2) - 1)) + 2 ) + 1 ! IN: 1 reacta.
              supReactDegRevStruTemp( (3*(numNodes(2) - 1)) + 3 ) = &
              supReactDegRevStruTemp( (3*(numNodes(2) - 1)) + 3 ) + 1 ! OUT: 1 produ.
              !
              ! Adds the new metabolite and the new reaction
              if( iV3 .eq. 1 ) then
                !
                ! Metabolite iV2: reactant of new reaction numNodes(2), i.e., supRevStructTemp(i3,i2,1) = 1,
                !    where i3 = [(4x(nreactions - 1)) +        0          +     1     ], i2 = iV2
                !                                        direct reaction=0 ; reactant=1
                supRevStructTemp( ((4*(numNodes(2) - 1)) + 0 + 1) , iV2, 1) = 1
                !
                ! Degree of metabolite iV2: reactant (OUT) of reaction numNodes(2)
                supMetDegRevStruTemp( (3*(iV2 - 1)) + 1 ) = &
                supMetDegRevStruTemp( (3*(iV2 - 1)) + 1 ) + 1 ! Total connections
                supMetDegRevStruTemp( (3*(iV2 - 1)) + 3 ) = &
                supMetDegRevStruTemp( (3*(iV2 - 1)) + 3 ) + 1 ! OUT connections: reactant
                !
                !
                ! New metabolite numNodes(1): product of the new reaction numNodes(2), i.e., supRevStructTemp(i3,i2,1) = 1,
                !    where i3 = [(4x(nreactions - 1)) +        0          +     2     ], i2 = numNodes(1)
                !                                        direct reaction=0 ; product=2
                supRevStructTemp( ((4*(numNodes(2) - 1)) + 0 + 2) , numNodes(1), 1) = 1
                !
                ! Degree of metabolite numNodes(1): product (IN) of reaction numNodes(2)
                supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 1 ) = &
                supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 1 ) + 1 ! Total connect.
                supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 2 ) = &
                supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 2 ) + 1 ! IN connect.:
                !                                                                                                            (product)
              else
                !
                ! Metabolite iV2: product of new reaction numNodes(2), i.e., supRevStructTemp(i3,i2,1) = 1,
                !    where i3 = [(4x(nreactions - 1)) +        0          +     2     ], i2 = iV2
                !                                        direct reaction=0 ; product=2
                supRevStructTemp( ((4*(numNodes(2) - 1)) + 0 + 2) , iV2, 1) = 1
                !
                ! Degree of metabolite iV2: product (IN) of reaction numNodes(2)
                supMetDegRevStruTemp( (3*(iV2 - 1)) + 1 ) = &
                supMetDegRevStruTemp( (3*(iV2 - 1)) + 1 ) + 1 ! Total connections
                supMetDegRevStruTemp( (3*(iV2 - 1)) + 2 ) = &
                supMetDegRevStruTemp( (3*(iV2 - 1)) + 2 ) + 1 ! IN connections: product
                !
                !
                ! New metabolite numNodes(1): reactant of the new reaction numNodes(2), i.e., supRevStructTemp(i3,i2,1) = 1,
                !    where i3 = [(4x(nreactions - 1)) +        0          +     1     ], i2 = numNodes(1)
                !                                        direct reaction=0 ; reactant=1
                supRevStructTemp( ((4*(numNodes(2) - 1)) + 0 + 1) , numNodes(1), 1) = 1
                !
                ! Degree of metabolite numNodes(1): reactant (OUT) of reaction numNodes(2)
                supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 1 ) = &
                supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 1 ) + 1 ! Total connect.
                supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 3 ) = &
                supMetDegRevStruTemp( (3*(numNodes(1) - 1)) + 3 ) + 1 ! OUT connect.:
                !                                                                                                            (reactant)
              end if
              !
              ! Stores the information of the reversibility of the reactions for this (possibly) new structure
              do iV_rev1=1,DataRevStru(3,iV1)  ! Number of reactions of the Parent-Structure iV1
                supReactRevStruTemp(iV_rev1) = ReactRevStru(iV_rev1,iV1)
              end do
              ! The new reaction numNodes(2) is irreversible
              supReactRevStruTemp(numNodes(2)) = 0
              !
              ! Call the subroutine organizingStruct
              call organizingStruct(supRevStructTemp, supMetDegRevStruTemp, &
              supReactDegRevStruTemp, numNodes, supReactRevStruTemp, supRevStruct, &
              supMetDegRevStru, supReactDegRevStru, supReactRevStru)
              !
              !
              ! Is this a isomorphic representation of a structure found before?
              ! Call the subroutine isomorphismTest
              call isomorphismTest(supRevStruct, RevStruct, DataRevStru, &
              MetDegRevStru, ReactDegRevStru, supMetDegRevStru, &
              supReactDegRevStru, nrefStruLow(numGen), nrefStruHigh(numGen), &
              const(1), PermutMet, PermutIrrReact, PermutRevReact, &
              PermutDirInvReact, numPermutMet, numPermutIrrReact, &
              numPermutRevReact, numPermDirInvReact, isoStruct, isoStruRef)
              ! If isoStruct = 1, the structure is a isomorphic representation of structure isoStruRef;
              if(isoStruct .eq. 1) then
                !
                ! Is this a new Parent/Child relantionship? If it is, the correspinding cell in matrix FamRevStru(i2,i1)
                ! must be "0"
                if(FamRevStru(isoStruRef,iV1) .eq. 0) then
                  !
                  ! Number of Parents of structure isoStruRef: adds structure iV1 as one of its Parents
                  DataRevStru(6,isoStruRef) = DataRevStru(6,isoStruRef) + 1
                  !
                  ! Number of Childs of structure iV1: adds structure isoStruRef as one of its Childs
                  DataRevStru(7,iV1) = DataRevStru(7,iV1) + 1
                  !
                  ! Structure iV1 is a Parent of the structure isoStruRef, found before. One adds "7" to the corresponding cell
                  ! in matrix FamRevStru(i2,i1) to show that they only differ of 1 reaction and 1 metabolite. The reaction is
                  ! used to connect this metabolite to another metabolite as a pair reactant/product
                  FamRevStru(isoStruRef,iV1) = 7
                  !
                end if
                !
                ! The matrix DataRevStru(i2,i1) must be turned to zeros at line i1=const(1)
                do cleanV_1=1,12
                  DataRevStru(cleanV_1,const(1)) = 0
                end do
                !
              ! If isoStruct = 0, this structure was found for the first time
              else
                !
                ! Adds a new structure to the total number of structures
                numStructures = numStructures + 1
                ! Adds a new structure to the total number of structures of this generation (the goal is to store
                ! the highest reference number of a structure of this generation)
                nrefStruHigh(numGen) = nrefStruHigh(numGen) + 1
                ! Adds a new structure to the total number of structures of this generation
                numStruGen(numGen) = numStruGen(numGen) + 1
                write(*,2002) numStruGen(numGen), numGen, numStructures
                !
                ! Structure iV1 is a Parent of the structure numStructures, found before. One adds "7" to the correspondingl
                ! cell in matrix FamRevStru(i2,i1) to show that they only differ of 1 reaction and 1 metabolite. The reaction
                ! is used to connect this metabolite to another metabolite as a pair reactant/product
                FamRevStru(numStructures,iV1) = 7
                !
                !
                ! Number of Childs of structure iV1: adds structure numStructures as one of its Childs
                DataRevStru(7,iV1) = DataRevStru(7,iV1) + 1
                !
                ! Writes the new structure to the matrix RevStruct(i3,i2,i1) at line i1=numStructures
                do wrV_1=1,(DataRevStru(2,iV1)+1)  ! Number of metabolites of the Parent iV1 + 1 (equal to the number of
                ! metabolites of it's Childs)
                  do wrV_2=1,(4*(DataRevStru(3,iV1)+1))  ! 4 x [number of reactions of the Parent iV1 +1] (equal to the
                  ! number of reactions of it's Childs)
                    ! Writes the new structure into matrix RevStruct(i3,i2,i1) at line i1=numStructures
                    RevStruct(wrV_2,wrV_1,numStructures) = supRevStruct(wrV_2,wrV_1,1)
                    ! Turns to zero the suport matrix supRevStruct(i3,i2,1)
                    supRevStruct(wrV_2,wrV_1,1) = 0
                  end do
                end do
                !
                ! Writes the information regarding the reversibility of the reactions of the new found structure
                ! (it is the same as for the Parent structure) into the matrix ReactRevStru(i2,i1) and of the degree
                ! of the reactions into matrix ReactDegRevStru(i2,i1) at line i1=numStructures
                do wrV_3=1,numNodes(2)  ! Number of reactions of the structure numStructures
                  ReactRevStru(wrV_3,numStructures) = supReactRevStru(wrV_3)
                end do
                do wrV_4=1,(3*numNodes(2))  ! Number of reactions of the structure numStructures
                  ReactDegRevStru(wrV_4,numStructures) = supReactDegRevStru(wrV_4)
                end do
                !
                ! Writes the the information regarding the degree of the metabolites into matrix MetDegRevStru(i2,i1)
                ! at line i1=numStructures
                do wrV_5=1,(3*numNodes(1))  ! 3 x number of metabolites of structure numStructures
                  MetDegRevStru(wrV_5,numStructures) = supMetDegRevStru(wrV_5)
                end do
              end if
              !
              ! Cleans the the variables supRevStructTemp(i2,i1), supRevStruct(i2,i1), supMetDegRevStruTemp(i1),
              ! supMetDegRevStru(i1), supReactDegRevStruTemp(i1), supReactDegRevStru(i1), supReactRevStruTemp(i1) and
              ! supReactRevStru(i1)
              do cleanV_2=1,numNodes(1)  ! Number of metabolites
                supMetDegRevStruTemp(cleanV_2) = 0
                supMetDegRevStru(cleanV_2) = 0
                do cleanV_3=1,(4*numNodes(2))  ! 4 x number of reactions
                  supRevStructTemp(cleanV_3,cleanV_2,1) = 0
                  supRevStruct(cleanV_3,cleanV_2,1) = 0
                  if( ( cleanV_2 .eq. 1 ) .and. &
                  ( cleanV_3 .le. numNodes(2) ) ) then
                    supReactDegRevStruTemp(cleanV_3) = 0
                    supReactDegRevStru(cleanV_3) = 0
                    supReactRevStruTemp(cleanV_3) = 0
                    supReactRevStru(cleanV_3) = 0
                  end if
                end do
              end do
              !
              ! Turns the variables isoStruct and isoStruRef into "0"
              isoStruct = 0
              isoStruRef = 0
              !
              ! Turns const(1) into zero
              const(1) = 0
              !
            end do
          end do
          !
          ! Turns to zero the matrices with the permutations for: i) metabolites - PermutMet(i2,i1); ii) irreversible
          ! reactions PermutIrrReact(i2,i1); iii) reversible reactions PermutRevReact(i2,i1); iv) direct and inverse
          ! single-direction reaction for all the reversible reactions - PermutRevReact(i2,i1)
          do cleanV_4=1,numPermutMet    ! Number of permutations for the metabolites
            do cleanV_5=1,numNodes(1)  ! Number of metabolites of the new found Childs of Parent iV1
              PermutMet(cleanV_5,cleanV_4) = 0
            end do
          end do
          if( numNodes(4) .ne. 0 ) then
            do cleanV_6=1,numPermutIrrReact  ! Number of permutations for the irreversible reactions
              do cleanV_7=1,numNodes(4)  ! Number of irreversible reactions
                PermutIrrReact(cleanV_7,cleanV_6) = 0
              end do
            end do
          end if
          if( numNodes(3) .ne. 0 ) then
            do cleanV_8=1,numPermutRevReact  ! Number of permutations for the reversible reactions
              do cleanV_9=1,numNodes(3)  ! Number of reversible reactions
                PermutRevReact(cleanV_9,cleanV_8) = 0
              end do
            end do
            do cleanV_10=1,numPermDirInvReact   ! Number of permutations for the direct and inverse reactions of a reversible reaction
              do cleanV_11=1,(2*numNodes(3))  ! 2x number of reversible reactions
                PermutDirInvReact(cleanV_10,cleanV_11) = 0
              end do
            end do
          end if
          ! Cleans the number of nodes of newly found Childs of Parent iV1
          numNodes(1) = 0
          numNodes(2) = 0
          numNodes(3) = 0
          numNodes(4) = 0
          ! Turns to zero the number of permutations
          numPermutMet = 0
          numPermutIrrReact = 0
          numPermutRevReact = 0
          numPermDirInvReact = 0
25      continue
        !
        !
      end do


!      Test the code: count all the structures with 3 metabolites and 2 reactions
      testcount(1) = 0
      do i1=1,numStructures
        ! The structure as 3 metabolites and 2 reactions?
        !    number of metabolites = 3    ;   number of reactions = 2
        if( (DataRevStru(2,i1) .eq. 3) .and. (DataRevStru(3,i1) .eq. 2) ) then
          ! Adds a new structure to the total of structures with 3 metabolites and 2 reactions
          testcount(1) = testcount(1) + 1
        end if
      end do


!      Obtaining the inclusion matrix, needed to determine the embedded frequency and to determine the number of
!      Ancesters and of Offspring for each structure
      ! Create the inclusion matriz. Initially, it is equal to the DR-structures network
      do i1=1,numStructures  ! Parents
        do i2=1,numStructures  ! Childs
          if( FamRevStru(i2,i1) .ne. 0 ) then
            InclRevStru(i2,i1) = 1
            DataRevStru(8,i1) = DataRevStru(8,i1) + 1  ! Adds to the number of Offspring of structure i1
          else
            InclRevStru(i2,i1) = 0
          end if
        end do
      end do
      !
      ! Obtains the Offspring of a structure by adding the offspring of its Children
      do i=2,(maxNumGen - 1)
        ! Check the offspring of the children of structures of generation ( maxNumGen - i )
        do i1=nrefStruLow( maxNumGen - i ),nrefStruHigh( maxNumGen - i )
          ! The childrens of structure i1 are of generation ( maxNumGen - i + 1 )
          do i2=nrefStruLow( maxNumGen - i + 1 ),nrefStruHigh( maxNumGen - i + 1 )
            ! Is structure i2 a Child of structure i1?
            if( FamRevStru(i2,i1) .ne. 0 ) then
              ! The Offspring of structure i2 is part of generation ( maxNumGen - i + 2 ) to generation maxNumGen
              do i3=nrefStruLow( maxNumGen - i + 2 ),numStructures
                if( InclRevStru(i3,i2) .ne. 0 ) then
                  if( InclRevStru(i3,i1) .eq. 0 ) then
                    InclRevStru(i3,i1) = InclRevStru(i3,i2)
                    DataRevStru(8,i1) = DataRevStru(8,i1) + 1  ! Adds to the number of Offspring of structure i1
                  end if
                end if
              end do
            end if
          end do
        end do
      end do
      !
      ! Determines the number of Ancesters for each one of the structures
      do i2=1,numStructures
        do i1=1,numStructures
          if( InclRevStru(i2,i1) .ne. 0 ) then
            DataRevStru(5,i2) = DataRevStru(5,i2) + 1  ! Adds to the total number of Ancestors of structure i2
          end if
        end do
      end do


!      Determining the equivalent nodes for each one of the structures; obtaining the matrix
!      DifNodesStru(i2,i1), which stores the information indicating if structure i1 has different
!      nodes by taking into account their degree (total and IN)
      do i=1,numStructures  ! Number of structures
        !
        ! Number of metabolites of structure i
        numNodes(1) = DataRevStru(2,i)
        ! Number of reactions of structure i
        numNodes(2) = DataRevStru(3,i)
        ! Number of reversible reactions of structure i
        numNodes(3) = DataRevStru(4,i)
        ! Number of irreversible reactions of structure i
        numNodes(4) = numNodes(2) - numNodes(3)
        !
        ! Call the subroutine "allPermutations" to obtain the permutations for the metabolites
        call allPermutations(numNodes(1), PermutMet, numPermutMet)
        ! Structure i has at least one irreversible reaction?
        if( numNodes(4) .ne. 0 ) then
          ! Call the subroutine "allPermutations" to obtain the permutations for the irreversible reactions
          call allPermutations(numNodes(4), PermutIrrReact, numPermutIrrReact)
        end if
        ! The new structures have at least one reversible reaction?
        if( numNodes(3) .ne. 0 ) then
          ! Call the subroutine "allPermutations" to obtain the permutations for the reversible reactions
          call allPermutations(numNodes(3), PermutRevReact, numPermutRevReact)
          ! Call the subroutine "allRevReactPerm" to obtain the permutations for the direct and inverse reactions
          ! for all the (possibly) reversible reactions
          call allRevReactPerm(numNodes(3), PermutDirInvReact, numPermDirInvReact)
        end if
        !
        ! Equivalent metabolites
        ! Call subroutine EquiMet
        call EquiMet(i, RevStruct, numNodes, MetDegRevStru, ReactDegRevStru, &
        PermutMet, PermutIrrReact, PermutRevReact, PermutDirInvReact, &
        numPermutMet, numPermutIrrReact, numPermutRevReact, numPermDirInvReact, &
        EquiMetRevStru)
        !
        ! Equivalent reactions
        ! Call subroutine EquiReact
        call EquiReact(i, RevStruct, ReactRevStru, numNodes, MetDegRevStru, &
        ReactDegRevStru, PermutMet, PermutIrrReact, PermutRevReact, &
        PermutDirInvReact, numPermutMet, numPermutIrrReact, numPermutRevReact, &
        numPermDirInvReact, EquiReactRevStru)
        !
        ! Compares the the degree (total and IN) of all the metabolites of structure i
        do iI1=1,(numNodes(1)-1)  ! Performs the cycle from 1 to the number of metabolites - 1
          do iI2=(iI1+1),numNodes(1)  ! Performs the cycle from iI1+1 to the number of metabolites
            !
            ! The metabolites iI1 and iI2 have the same total and IN degree?
            if( ( MetDegRevStru(((3*(iI1-1))+1),i) .eq. &
            MetDegRevStru(((3*(iI2-1))+1),i) ) .and. &
            ( MetDegRevStru(((3*(iI1-1))+2),i) .eq. &
            MetDegRevStru(((3*(iI2-1))+2),i) ) ) then
              ! Structure i has at least two metabolites with the same degree (total and IN)
              DifNodesStru(1,i) = 0
              goto 100
            end if
            !
            ! All pairs of metabolites tested?
            if( (iI1 .eq. (numNodes(1)-1) ) .and. (iI2 .eq. numNodes(1)) ) then
              ! Structure i does not have metabolites with the same degree (total and IN)
              DifNodesStru(1,i) = 1
              goto 100
            end if
          end do
        end do
        goto 100
100     continue
        !
        ! Compares the degree (total and IN) of all the irreversible reactions of structure i
        ! Structure i has more than 1 irreversible reactions?
        if( numNodes(4) .gt. 1 ) then
          ! Structure i has >1 irreversible reactions
          do iII1=1,(numNodes(4)-1)  ! Performs the cycle from 1 to the number of irreversible reactions - 1
            do iII2=(iII1+1),numNodes(4)  ! Performs the cycle from iII1+1 to the number of irreversible reactions
              !
              ! The irreversible reactions iII1 and iII2 have the same total and IN degree?
              if( ( ReactDegRevStru(((3*(iII1-1))+1),i) .eq. &
              ReactDegRevStru(((3*(iII2-1))+1),i) ) .and. &
              ( ReactDegRevStru(((3*(iII1-1))+2),i) .eq. &
              ReactDegRevStru(((3*(iII2-1))+2),i) ) ) then
                ! Structure i has at least two irreversible reactions with the same degree (total and IN)
                DifNodesStru(2,i) = 0
                goto 105
              end if
              !
              ! All pairs of irreversible reactions tested?
              if( (iII1 .eq. (numNodes(4)-1)) .and. (iII2 .eq. numNodes(4)) ) then
                DifNodesStru(2,i) = 1
                goto 105
              end if
            end do
          end do
        else
          ! Structure i has <1 irreversible reactions
          DifNodesStru(2,i) = 0
        end if
        goto 105
105     continue
        !
        ! Compares the degree (total and IN) of all the reversible reactions of structure i
        ! Structure i has more than 1 reversible reactions?
        if( numNodes(3) .gt. 1 ) then
          ! Structure i has >1 reversible reactions
          do iIII1=1,(numNodes(3)-1)  ! Performs the cycle from 1 to the number of reversible reactions - 1
            do iIII2=(iIII1+1),numNodes(3)  ! Performs the cycle from iIII1+1 to the number of reversible reactions
              !
              ! The reversible reactions iIII1 and iIII2 have the same total and IN degree?
              if( ( ReactDegRevStru(((3*(numNodes(4)+iIII1-1))+1),i) .eq. &
              ReactDegRevStru(((3*(numNodes(4)+iIII2-1))+1),i) ) .and. &
              ( ReactDegRevStru(((3*(numNodes(4)+iIII1-1))+2),i) .eq. &
              ReactDegRevStru(((3*(numNodes(4)+iIII2-1))+2),i) ) ) then
                ! Structure i has at least two reversible reactions with the same degree (total and IN)
                DifNodesStru(3,i) = 0
                goto 110
              end if
              !
              ! All pairs of reversible reactions tested?
              if( (iIII1 .eq. (numNodes(3)-1)) .and. (iIII2 .eq. numNodes(3)) ) then
                DifNodesStru(3,i) = 1
                goto 110
              end if
            end do
          end do
        else
          ! Structure i has <1 reversible reactions
          DifNodesStru(3,i) = 0
        end if
        goto 110
110     continue
      end do


!      Creating the matrix with the number of metabolites and reactions per generation
      do i=1,7
        nMet(i) = 2 + (i - 1)
        nReact(i) = 1 + (i - 1)
      end do


!      Creating the permutations for a set of increasing ordered numbers. Application of an algorithm
!      to generate permutations in lexicographic order
      ! Call the subroutine allPermutations
      nMet(7) = maxNumMet
      call allPermutations(nMet(7), PermutMet, numPermutMet)
      nReact(7) = maxNumReact
      call allPermutations(nReact(7), PermutRevReact, numPermutRevReact)
      ! Call the subroutine allRevReactPerm
      call allRevReactPerm(nReact(7), PermutDirInvReact, numPermDirInvReact)



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information regarding the metabolites A of the structures to file:
!      "Structures_Metabolite_A.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      const(1) = 1  ! Metabolite A -> const(1) = 1
      if( maxNumMet .ge. const(1) ) then
        open(8,file='Structures_Metabolite_A.txt')
        do i=1,numStructures
          write(8,2000) (RevStruct(j, const(1), i), j=1,(4*maxNumReact))
        end do
        close(8)
      end if
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information regarding the metabolites B of the structures to file:
!      "Structures_Metabolite_B.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      const(1) = 2  ! Metabolite B -> const(1) = 2
      if( maxNumMet .ge. const(1) ) then
        open(8,file='Structures_Metabolite_B.txt')
        do i=1,numStructures
          write(8,2000) (RevStruct(j, const(1), i), j=1,(4*maxNumReact))
        end do
        close(8)
      end if
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information regarding the metabolites C of the structures to file:
!      "Structures_Metabolite_C.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      const(1) = 3  ! Metabolite C -> const(1) = 3
      if( maxNumMet .ge. const(1) ) then
        open(8,file='Structures_Metabolite_C.txt')
        do i=1,numStructures
          write(8,2000) (RevStruct(j, const(1), i), j=1,(4*maxNumReact))
        end do
        close(8)
      end if
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information regarding the metabolites D of the structures to file:
!      "Structures_Metabolite_D.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      const(1) = 4  ! Metabolite D -> const(1) = 4
      if( maxNumMet .ge. const(1) ) then
        open(8,file='Structures_Metabolite_D.txt')
        do i=1,numStructures
          write(8,2000) (RevStruct(j, const(1), i), j=1,(4*maxNumReact))
        end do
        close(8)
      end if
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information regarding the metabolites E of the structures to file:
!      "Structures_Metabolite_E.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      const(1) = 5  ! Metabolite E -> const(1) = 5
      if( maxNumMet .ge. const(1) ) then
        open(8,file='Structures_Metabolite_E.txt')
        do i=1,numStructures
          write(8,2000) (RevStruct(j, const(1), i), j=1,(4*maxNumReact))
        end do
        close(8)
      end if
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information regarding the metabolites F of the structures to file:
!      "Structures_Metabolite_F.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      const(1) = 6  ! Metabolite F -> const(1) = 6
      if( maxNumMet .ge. const(1) ) then
        open(8,file='Structures_Metabolite_F.txt')
        do i=1,numStructures
          write(8,2000) (RevStruct(j, const(1), i), j=1,(4*maxNumReact))
        end do
        close(8)
      end if
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information regarding the metabolites G of the structures to file:
!      "Structures_Metabolite_G.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      const(1) = 7  ! Metabolite G -> const(1) = 7
      if( maxNumMet .ge. const(1) ) then
        open(8,file='Structures_Metabolite_G.txt')
        do i=1,numStructures
          write(8,2000) (RevStruct(j, const(1), i), j=1,(4*maxNumReact))
        end do
        close(8)
      end if


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the general information regarding the structures to file:
!      "Structures_General.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='Structures_General.txt')
      do i=1,numStructures
        write(8,2001) (DataRevStru(j, i), j=1,12)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information regarding the reversibility of each one of the
!      structures to file:
!      "Structures_RevReact.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='Structures_RevReact.txt')
      do i=1,numStructures
        write(8,2000) (ReactRevStru(j, i), j=1,maxNumReact)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information regarding the relantionships Parent/Child between
!      all the structures found to file:
!      "Family_Structures.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='Family_Structures.txt')
      do i=1,numStructures
        write(8,2000) (FamRevStru(j,i), j=1,numStructures)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information regarding
!      all the structures found to file:
!      "InclusionMatrix_Structures.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='InclusionMatrix_Structures.txt')
      do i=1,numStructures
        write(8,2000) (InclRevStru(j,i), j=1,numStructures)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information regarding the degree of the metabolites into the file:
!      "MetDegreeRevStru.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='MetDegreeRevStru.txt')
      do i=1,numStructures
        write(8,2000) (MetDegRevStru(j,i), j=1,(3*maxNumMet))
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information regarding the degree of the reactions into the file:
!      "ReactDegreeRevStru.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='ReactDegreeRevStru.txt')
      do i=1,numStructures
        write(8,2000) (ReactDegRevStru(j,i), j=1,(3*maxNumReact))
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information regarding the equivalent metabolites of each one
!      of the structures into the file:
!      "EquiMetRevStru.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='EquiMetRevStru.txt')
      const(1) = 0
      do i1=1,(maxNumMet-1)
        const(1) = const(1) + (maxNumMet - i1)
      end do
      do i=1,numStructures
        write(8,2000) (EquiMetRevStru(j,i), j=1,const(1))
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information regarding the equivalent reactions of each one
!      of the structures into the file:
!      "EquiReactRevStru.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='EquiReactRevStru.txt')
      const(1) = 0
      do i1=1,(maxNumReact-1)
        const(1) = const(1) + (maxNumReact - i1)
      end do
      do i=1,numStructures
        write(8,2000) (EquiReactRevStru(j,i), j=1,const(1))
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information indicating if a given structure has different nodes by
!      taking into account their degree (total and IN) into the file:
!      "DifNodesStru.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='DifNodesStru.txt')
      do i1=1,numStructures
        write(8,2000) (DifNodesStru(i2,i1), i2=1,3)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information regarding the number of generations considered, the
!      number of structures obtained, and the minimum and maximum reference numbers
!      of the structures belonging to each one of the considered generations
!      into the file:
!      "in_nGen_nStru.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='in_nGen_nStru.txt')
      do i=1,(maxNumGen + 1)
        if( i .eq. 1 ) then
          write(8,2003) maxNumGen, numStructures
        else
          write(8,2003) nrefStruLow(i-1), nrefStruHigh(i-1)
        end if
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Write the information regarding the maximum number of metabolites and of
!      reactions considered for the structures within the assumed number of
!      generations into the file:
!      "in_nMet_nReact.txt"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='in_nMet_nReact.txt')
      write(8,2003) maxNumMet
      write(8,2003) maxNumReact
      close(8)



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Formats
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
2000  format(6000(i3, ','))
2001  format(4(i3, ','), 4(i8, ','), 3(i5, ','), i5)
2002  format('Structure n', i5, ' of generation n', i2, ';  Ref. of the structure:', i5 )
2003  format(2(i5, ','))


      contains


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!     Subroutine to organize the topological structures according to the degree of the
!!     nodes, metabolites and reactions. The reactions should be ordered first according
!!     to its reversibility: irreversible reactions are stored first, and then, if at
!!     least one exists, reversible reactions are stored.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine organizingStruct(supRevStructTemp_, supMetDegRevStruTemp_, &
      supReactDegRevStruTemp_, numNodes_, supReactRevStruTemp_, supRevStruct_, &
      supMetDegRevStru_, supReactDegRevStru_, supReactRevStru_)

! VECTORS
!      supMetDegRevStruTemp_(i1) : gives the degree for each metabolite of a (possibly) new structure. i1=[(3 x (nmet - 1)) + 1],
!                                  total n of connections of the metabolite; i1=[(3 x (nmet - 1)) + 2], n of connections
!                                  comming to the metabolite (in); i1=[(3 x (nmet - 1)) + 3], n of connections going out of
!                                  the metabolite (out). The metabolites may not be organized according to their degree
!      supReactDegRevStruTemp_(i1) : gives the degree for each reaction of a possibly new structure. The irreversible reactions
!                                    are stored first, and then, in case at least one exists, the reversible reactions are stored.
!                                    i1=[(3 x (nmet - 1)) + 1], connections of the metabolite; i1=[(3 x (nmet - 1)) + 2],
!                                    n of connections comming to the total n ofmetabolite (in); i1=[(3 x (nmet - 1)) + 3],
!                                    n of connections going out of the metabolite (out). The reactions may not be organized
!                                    according to their degree
!      supMetDegRevStru_(i1) : gives the degree for each metabolite of a (possibly) new structure. i1=[(3 x (nmet - 1)) + 1],
!                              total n of connections of the metabolite; i1=[(3 x (nmet - 1)) + 2], n of connections
!                              comming to the metabolite (in); i1=[(3 x (nmet - 1)) + 3], n of connections going out of
!                              the metabolite (out)
!      supReactDegRevStru_(i1) : gives the degree for each reaction of a possibly new structure. The irreversible reactions
!                                are stored first, and then, in case at least one exists, the reversible reactions are stored.
!                                i1=[(3 x (nmet - 1)) + 1], connections of the metabolite; i1=[(3 x (nmet - 1)) + 2],
!                                n of connections comming to the total n ofmetabolite (in); i1=[(3 x (nmet - 1)) + 3],
!                                n of connections going out of the metabolite (out)
!      numNodes_(i1) : number of nodes of the (possibly) new struture: (i1=1) number of metabolites; (i1=2) number of reactions;
!                      (i1=3) number of reversible reactions; (i1=4) number of irreversible reactions
!      supReactRevStruTemp_(i1) : information regarding the reversibility of the reactions of the (possibly) new structure. If a
!                                 reaction is reversible, the corresponding cell is filled with 1; it it is irreversible, the
!                                 corresponding cell is filled with 0. The reactions may not be ordered according to its reversibility
!      supReactRevStru_(i1) : information regarding the reversibility of the reactions of the (possibly) new structure. If a
!                             reaction is reversible, the corresponding cell is filled with 1; it it is irreversible, the
!                             corresponding cell is filled with 0. The reactions are ordered according to the reversibility:
!                             irreversible reactions first, with the reversible reactions being stored after
!      supOrdMet_(i1) : stores the order of the metabolites of the (possibly) new structure according to their degree:
!                       (i1=[(3 x (nmet - 1)) + 1 ]), reference number of the metabolite; (i1=[(3 x (nmet - 1)) + 2 ]),
!                       total number of connections; (i1=[(3 x (nmet - 1)) + 3 ]), number of connections IN
!      supOrdReact_(i1) : stores the order of the reactions of the (possibly) new structure according to their degree. The
!                         reactions are stored according to their reversibility: irreversible reactions first, and
!                         reersible reactions after. (i1=[(3 x (nreact - 1)) + 1 ]), reference number of the reaction;
!                         (i1=[(3 x (nreact - 1)) + 2 ]), total number of connections; (i1=[(3 x (nreact - 1)) + 3 ]),
!                         number of connections IN
!      supMetDegRevStru_b(i1) : the same as supMetDegRevStru_(i1)
!      supReactDegRevStru_b(i1) : the same as supReactDegRevStru_(i1)
!      supReactRevStruTemp_b(i1) : the same as supReactRevStruTemp_(i1)
!      supConst_(i1) : support vector used to organize the (possibly) new structure according to the degree of its nodes
!
!
! MATRICES
!      supRevStructTemp_(i3,i2,1) : support matrix to store temporarily a new structure. This structure will be reorganized
!                                   into matrix supRevStruct_(i3,i2,1)
!      supRevStruct_(i3,i2,1) : support matrix to store the pattern of the (possibly) new structure stored in matrix
!                               supRevStructTemp_(i3,i2,1) ordered according to degree of the nodes (metabolites and reactions).
!                               This representation of the structure will be compared to others already found to determine
!                               if it is a isomorphic representation of one of them

      implicit none

      ! IN variables
       ! Integer: 1 bit
      integer*1, intent(in) :: supRevStructTemp_(:,:,:), supReactRevStruTemp_(:)
       ! Integer: 2 bits
      integer*2, intent(in) :: supMetDegRevStruTemp_(:), supReactDegRevStruTemp_(:), numNodes_(:)

      ! OUT variables
       ! Integer: 1 bit
      integer*1, intent(inout) :: supRevStruct_(:,:,:), supReactRevStru_(:)
       ! Integer: 2 bits
      integer*2, intent(inout) :: supMetDegRevStru_(:), supReactDegRevStru_(:)

      !Internal support variables (2 bits)
      integer*2 :: supOrdMet_(24), supOrdReact_(24), supMetDegRevStruTemp_b(24)
      integer*2 :: supReactDegRevStruTemp_b(24), supReactRevStruTemp_b(8), supConst_(2)

      ! General variables (4 bits)
      integer*4 :: iOrg_1, iOrg_2, iOrg_3

!      Organizes the (possibly) new structure according to the degree of the nodes
      ! Creates the support vectors supMetDegRevStruTemp_b(i1), supReactDegRevStruTemp_b(i1) and supReactRevStruTemp_b(i1)
      do iOrg_1=1,(3*numNodes_(1))  ! 3 x number of metabolites of the (possibly) new structure
        supMetDegRevStruTemp_b(iOrg_1) = supMetDegRevStruTemp_(iOrg_1)
      end do
      do iOrg_1=1,(3*numNodes_(2))  ! 3 x number of reactions of the (possibly) new structure
        supReactDegRevStruTemp_b(iOrg_1) = supReactDegRevStruTemp_(iOrg_1)
      end do
      do iOrg_1=1,numNodes_(2)  ! Number of reactions
        supReactRevStruTemp_b(iOrg_1) = supReactRevStruTemp_(iOrg_1)
      end do
      !
      ! Obtains the ordered list for the metabolites of the (possibly) new structure according to their degree
      do iOrg_1=1,numNodes_(1)  ! Number of metabolites of the (possibly) new structure
        supOrdMet_( (3*(iOrg_1 - 1)) + 1 ) = 0   ! "Reference" number of the metabolite at the ordered position iOrg_1
        supOrdMet_( (3*(iOrg_1 - 1)) + 2 ) = 0   ! Number of connections of the metabolite at the ordered position iOrg_1
        supOrdMet_( (3*(iOrg_1 - 1)) + 3 ) = 0   ! Number of connections IN of the metabolite at the ordered position iOrg_1
        supConst_(1) = 0  ! Support constant: stores the "reference" number of the metabolite for the new ordered position iOrg_1
        do iOrg_2=1,numNodes_(1)  ! Number of metabolites of the (possibly) new structure
          !
          ! Checks the metabolite with higher degree (total and IN)
          !     tests: i) number of connections;
          if( supMetDegRevStruTemp_b( (3*(iOrg_2 - 1)) + 1 ) .ge. &
          supOrdMet_( (3*(iOrg_1 - 1)) + 2 ) )  then !.and. &
            ! Number of connections are the same?
            if( supMetDegRevStruTemp_b( (3*(iOrg_2 - 1)) + 1 ) .eq. &
            supOrdMet_( (3*(iOrg_1 - 1)) + 2 ) ) then
              ! If the number of connections are the same, tests: ii) number of connections IN
              if( supMetDegRevStruTemp_b( (3*(iOrg_2 - 1)) + 2 ) .le. &
              supOrdMet_( (3*(iOrg_1 - 1)) + 3 ) ) goto 2500
            end if
            !
            ! Support constant: stores the "reference" number of the metabolite for the new ordered position iOrg_1
            supConst_(1) = iOrg_2
            !
            ! Previous "reference" number of the metabolite to introduce in the ordered position iOrg_1
            supOrdMet_( (3*(iOrg_1 - 1)) + 1 ) = iOrg_2
            ! Total number of connections
            supOrdMet_( (3*(iOrg_1 - 1)) + 2 ) = &
            supMetDegRevStruTemp_b( (3*(iOrg_2 - 1)) + 1 )
            ! Number of connections IN
            supOrdMet_( (3*(iOrg_1 - 1)) + 3 ) = &
            supMetDegRevStruTemp_b( (3*(iOrg_2 - 1)) + 2 )
            !
          end if
          goto 2500
2500      continue
          !
          ! iOrg_2 equal to numNodes_(1) (number of metabolites)?
          if( iOrg_2 .eq. numNodes_(1) ) then
            ! Turns to zero the cells corresponding to the metabolite supConst_(1) to allow the next metabolite with
            ! higher degree to be found
            supMetDegRevStruTemp_b( (3*(supConst_(1) - 1)) + 1 ) = 0
            supMetDegRevStruTemp_b( (3*(supConst_(1) - 1)) + 2 ) = 0
            supMetDegRevStruTemp_b( (3*(supConst_(1) - 1)) + 3 ) = 0
          end if
        end do
      end do
      !
      ! Obtains the ordered list of the reations of the (possibly) new structure according to their reversibility and
      ! degree
      ! Creates the vector supOrdReact_(i1)
      do iOrg_1=1,(3*numNodes_(2))  ! 3 x number of reactions of the (possibly) new structure
        supOrdReact_(iOrg_1) = 0
      end do
      ! Does the loop till the maximum between the number of irreversible reactions (numNodes_(4)) and
      ! the number of reversible reactions (numNodes_(3))
      do iOrg_1=1,max(numNodes_(4),numNodes_(3))
        supConst_(1) = 0
        supConst_(2) = 0
        do iOrg_2=1,numNodes_(2)  ! Number of reactions of the (possibly) new structure
          !
          ! Checks the reversibility of the reaction
          !
          ! Reaction is irreversible
          if( supReactRevStruTemp_b(iOrg_2) .eq. 0 ) then
            !
            ! Checks the reaction with higher degree (total and IN)
            !       tests: i) number of connections;
            if( supReactDegRevStruTemp_b( (3*(iOrg_2 - 1)) + 1 ) .ge. &
            supOrdReact_( (3*(iOrg_1 - 1)) + 2 ) ) then
              ! The number of connections is the same?
              if( supReactDegRevStruTemp_b( (3*(iOrg_2 - 1)) + 1 ) .eq. &
              supOrdReact_( (3*(iOrg_1 - 1)) + 2 ) ) then
                ! If the number of connections are the same, tests: ii) number of connections IN
                if( supReactDegRevStruTemp_b( (3*(iOrg_2 - 1)) + 2 ) .le. &
                supOrdReact_( (3*(iOrg_1 - 1)) + 3 ) ) goto 2505
              end if
              !
              ! Support constant: stores the "reference" number of the reaction for the new ordered position iOrg_1
              supConst_(1) = iOrg_2
              !
              ! Previous "reference" number of the reaction to introduce in the ordered position iOrg_1
              supOrdReact_( (3*(iOrg_1 - 1)) + 1 ) = iOrg_2
              ! Total number of connections
              supOrdReact_( (3*(iOrg_1 - 1)) + 2 ) = &
              supReactDegRevStruTemp_b( (3*(iOrg_2 - 1)) + 1 )
              ! Number of connections IN
              supOrdReact_( (3*(iOrg_1 - 1)) + 3 ) = &
              supReactDegRevStruTemp_b( (3*(iOrg_2 - 1)) + 2 )
              !
              goto 2505
              !
            end if
          end if
          !
          ! Reaction is reversible
          if( supReactRevStruTemp_b(iOrg_2) .eq. 1 ) then
            !
            ! Checks the reaction with higher degree (total and IN)
            !       tests: i) number of connections;
            if( supReactDegRevStruTemp_b( (3*(iOrg_2 - 1)) + 1 ) .ge. &
            supOrdReact_( (3*(numNodes_(4) + iOrg_1 - 1)) + 2 ) ) then
              ! The number of connections is the same?
              if( supReactDegRevStruTemp_b( (3*(iOrg_2 - 1)) + 1 ) .eq. &
              supOrdReact_( (3*(numNodes_(4) + iOrg_1 - 1)) + 2 ) ) then
                ! If the number of connections are the same, tests: ii) number of connections IN
                if( supReactDegRevStruTemp_b( (3*(iOrg_2 - 1)) + 2 ) .le. &
                supOrdReact_( (3*(numNodes_(4) + iOrg_1 - 1)) + 3 ) ) goto 2505
              end if
              !
              ! Support constant: stores the "reference" number of the reaction for the new ordered position (numNodes_(4) + iOrg_1)
              supConst_(2) = iOrg_2
              !
              ! Previous "reference" number of the reaction to introduce in the ordered position (numNodes_(4) + iOrg_1)
              supOrdReact_( (3*(numNodes_(4) + iOrg_1 - 1)) + 1 ) = iOrg_2
              ! Total number of connections
              supOrdReact_( (3*(numNodes_(4) + iOrg_1 - 1)) + 2 ) = &
              supReactDegRevStruTemp_b( (3*(iOrg_2 - 1)) + 1 )
              ! Number of connections IN
              supOrdReact_( (3*(numNodes_(4) + iOrg_1 - 1)) + 3 ) = &
              supReactDegRevStruTemp_b( (3*(iOrg_2 - 1)) + 2 )
              !
            end if
          end if
          !
          goto 2505
2505      continue
          ! iOrg_2 equal to numNodes_(2) (number of reactions)?
          if( iOrg_2 .eq. numNodes_(2) ) then
            !
            ! Irreversible reactions
            if( (numNodes_(4) .ne. 0) .and. (iOrg_1 .le. numNodes_(4)) ) then
              ! Turns to zero the cells corresponding to the irreversible reaction supConst_(1) to allow the next irreversible
              ! reaction with higher degree to be found
              supReactDegRevStruTemp_b( (3*(supConst_(1) - 1)) + 1 ) = 0
              supReactDegRevStruTemp_b( (3*(supConst_(1) - 1)) + 2 ) = 0
              supReactDegRevStruTemp_b( (3*(supConst_(1) - 1)) + 3 ) = 0
              supReactRevStruTemp_b(supConst_(1)) = 2
            end if
            !
            ! Reversible reactions
            if( (numNodes_(3) .ne. 0) .and. (iOrg_1 .le. numNodes_(3)) ) then
              ! Turns to zero the cells corresponding to the reversible reaction supConst_(2) to allow the next reversible
              ! reaction with higher degree to be found
              supReactDegRevStruTemp_b( (3*(supConst_(2) - 1)) + 1 ) = 0
              supReactDegRevStruTemp_b( (3*(supConst_(2) - 1)) + 2 ) = 0
              supReactDegRevStruTemp_b( (3*(supConst_(2) - 1)) + 3 ) = 0
              supReactRevStruTemp_b(supConst_(2)) = 2
            end if
          end if
        end do
      end do
      !
      !
      ! Organizes the pattern of the (possibly) new structure according to the new order of the metabolites and of
      ! the reactions
      do iOrg_1=1,numNodes_(1)  ! Number of metabolites of the (possibly) new structure
        do iOrg_2=1,numNodes_(2)  ! Number of reactions of the (possibly) new structure
          do iOrg_3=1,4  ! For each reaction, 2 unidirectional reactions with 2 cells for the reactant and product
            supRevStruct_( ((4*(iOrg_2 - 1)) + iOrg_3 ), iOrg_1, 1) = &
            supRevStructTemp_(  ((4*( supOrdReact_((3*(iOrg_2 - 1)) + 1) - 1 )) &
            + iOrg_3 ), supOrdMet_( (3*(iOrg_1 - 1)) + 1 ), 1)
          end do
        end do
      end do
      !
      ! Organizes the information regarding the reversibility of the reactions according to the new order
      do iOrg_1=1,numNodes_(2)  ! Number of reactions of the (possibly) new structure
        supReactRevStru_(iOrg_1) = &
        supReactRevStruTemp_( supOrdReact_((3*(iOrg_1 - 1)) + 1) )
      end do
      !
      ! Organizes the information regarding the degrees of the metabolites according to the new order
      do iOrg_1=1,numNodes_(1)  ! Number of metabolites of the (possibly) new structure
        do iOrg_2=1,3
          supMetDegRevStru_( (3*(iOrg_1 - 1)) + iOrg_2 ) = &
          supMetDegRevStruTemp_( (3*( supOrdMet_((3*(iOrg_1 - 1)) + 1) &
          - 1 )) + iOrg_2 )
        end do
      end do
      !
      ! Organizes the information regarding the degrees of the reactions according to the new order
      do iOrg_1=1,numNodes_(2)  ! Number of reactions of the (possibly) new structure
        do iOrg_2=1,3
          supReactDegRevStru_( (3*(iOrg_1 - 1)) + iOrg_2 ) = &
          supReactDegRevStruTemp_( (3*( supOrdReact_( (3*(iOrg_1 - 1)) &
          + 1) - 1 )) + iOrg_2 )
        end do
      end do

      end subroutine organizingStruct


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!     Subroutine to determine all the possible permutations of a increasing ordered
!!     set of numbers. The set is obtained from the number of the metabolites or
!!     reactions. Application of an algorithm to generate permutations in lexicographic
!!     order.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine allPermutations(inNumber_, Permut_, numPermut_)

! LOGICAL VARIABLES
!      repeat_ : loop continues till repeat is FALSE
!
!
! CONSTANTS
!      numPop_ : number of metabolites or reactions, needed to create the set of ordered
!                numbers. It is equal to inNumber_
!      inNumber_ : number of metabolites or reactions. It is the IN argument of the sub-routine
!
!
! VECTORS
!      supPer_(numPop_) : support vector super_(i1) needed to create the new permutation.
!                         i1) each cell corresponds to a number of the new permutation
!
!
! MATRICES
!      Permut_(numPop_,numPermut_) : matrix Permut_(i2,i1) with the permutations of a set of
!                                    numPop_ numbers. i1) number of the permutation; i2) column
!                                    for each numPop_ number.

      implicit none

      ! IN variables
       ! Integer: 2 bits
      integer*2, intent(in) :: inNumber_

      ! OUT variables
       ! Integer: 2 bits
      integer*2, intent(inout) :: Permut_(:,:)
       ! Integer:4 bits
      integer*4, intent(inout) :: numPermut_

      ! Internal support variables
       ! Integer: 2 bits
      integer*2 :: support_(3), supPer_(8), numPop_
       ! Logical variables
      logical :: repeat_

      ! General variables (4 bits)
      integer*4 :: i_1

!      Creating the permutations for a set of increasing ordered numbers. Application of an algorithm
!      to generate permutations in lexicographic order
      ! Generate the set of numbers from the total number of metabolites or reactions
      numPop_ = inNumber_
      do i_1=1,numPop_
        Permut_(i_1,1) = i_1
      end do
      ! Initializes the counting of the number of permutations
      numPermut_ = 1
      ! Logical variable
      repeat_ = .true.
      ! Finding new permutations occurs till "repeat_" is FALSE
      do while (repeat_)
        support_(1) = 0
        ! Determines the highest index k such that Permt_[k] < Permut[k + 1], and k + 1 =< numPop_
        do i_1=1,(numPop_-1)
          if( (Permut_(i_1,numPermut_) .lt. Permut_((i_1+1),numPermut_)) &
          .and. ( i_1 .gt. support_(1) ) ) then
            support_(1) = i_1
          end if
        end do
        ! An index k was found such that Permt_[k] < Permut[k + 1] (support_(1) is not equal to 0)
        if( support_(1) .ne. 0 ) then
          support_(2) = 0
          ! Creates the support vector supPer_(i_1) to help obtaining the new permutation
          do i_1=1,numPop_
            ! The support vector stores the previous permutation
            supPer_(i_1) = Permut_(i_1,(numPermut_))
          end do
          ! Find the largest index l greater than k such that Permut_[k] < Permut_[l]
          do i_1=(support_(1)+1),numPop_
            if( (Permut_(i_1,numPermut_) .gt. Permut_(support_(1),numPermut_)) &
            .and. (i_1 .gt. support_(2)) ) then
              support_(2) = i_1
            end if
          end do
          ! Swap the cells in position k and l: k=support_(1) and l=support_(2)
          supPer_(support_(1)) = Permut_(support_(2),(numPermut_))
          supPer_(support_(2)) = Permut_(support_(1),(numPermut_))
          ! Adds the new permutation to the total number of permutations
          numPermut_ = numPermut_ + 1
          ! Stores the new permutation in the matrix Permut_(j,i). Some changes still need to be done to this
          do i_1=1,support_(1)  ! new permutation
            Permut_(i_1,numPermut_) = supPer_(i_1)
          end do
          support_(3) = 0
          ! Reverse the sequence from Permut_[k + 1] up to and including the final element Permut_[numPop_]
          do i_1=(support_(1)+1),numPop_
            Permut_(i_1,numPermut_) = supPer_(numPop_-support_(3))
            support_(3) = support_(3) + 1
          end do
          ! Turns to zeros the support matrix supPer_(i_1) to be used in the next iteration
          do i_1=1,numPop_
            supPer_(i_1) = 0
          end do
        ! If no index was found were Permt_[k] < Permut[k + 1], there are no more permutations
        else
          ! Logical variable repeat_ is turn to FALSE to stop the loop
          repeat_ = .false.
        end if
      end do

      end subroutine allPermutations


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!     Subroutine to determine all the possible permutations considering the
!!     two single directions reactions of a reversible reaction. So, the possible
!!     permutations between the direct and the inverse reaction of a reversible
!!     reation are obtained.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine allRevReactPerm(numberIn_, RevPerm_, numRevPerm_)

! CONSTANTS
!      numberIn_ : number of reactions
!      nReact_ : number of reactions (equal to numberIn_)
!
!
! VECTORS
!      rdir(nReact_) : support vector that stores the ordenation of the direct reaction
!                      for all the reactions: 0 if it stays in its position, 2 if it
!                      moves to the position of the inverse reaction
!      rinv(nReact_) : support vector that stores the ordenation of the inverse reaction
!                      for all the reactions: 2 if it stays in its position, 0 if it
!                      moves to the position of the direct reaction
!      suppor_(nReact_) : support vector used to count the consecutive number of times the
!                         positions for the direct and inverse reactions of a i1_ reaction
!                         are keep the same in the permutations. After reaching a determined
!                         number, the positions of the direct and inverse reaction of a i1_
!                         reaction are exchanged and count is turn to zero again
!
!
! MATRICES
!      RevPerm_(nReact_,numRevPerm_) :

      implicit none

      ! IN variables
       ! Integer: 2 bits
      integer*2, intent(in) :: numberIn_

      ! OUT variables
       ! Integer: 2 bits
      integer*2, intent(inout) :: numRevPerm_, RevPerm_(:,:)

      !Internal support variables (2 bits)
      integer*2 :: nReact_, rdir_(8), rinv_(8), suppor_(8)

      ! General variables (4 bits)
      integer*4 :: i1_, i2_

!      Obtains the permutations between the direct and the reverse reaction for all
!      the reactions
      ! number of reactions equal to numberIn_
      nReact_ = numberIn_
      ! Creates the support vectors rdir_(i1_), rinv_(i1_) and suppor(i1_)
      do i1_=1,nReact_
        rdir_(i1_) = 0  ! The direct reactions are initially at there natural position
        rinv_(i1_) = 2  ! The inverse reactions are initially at there natural position
        ! The support vector suppor_(i1_), used to count the number of consecutive
        suppor_(i1_) = 0  ! permutations without changing the position of the direct
      end do              ! and inverse reaction of a i1_ reaction
      !
      ! Number of permutations equal to 2^(nreact_)
      numRevPerm_ = (2**(nReact_))
      !
      do i1_=1,numRevPerm_
        do i2_=1,nreact_
          suppor_(i2_) = suppor_(i2_) + 1
          RevPerm_((2*(i2_ - 1) + 1),i1_) = rdir_(i2_)
          RevPerm_((2*(i2_ - 1) + 2),i1_) = rinv_(i2_)
          ! Test if the pair 0/2 or 2/0 for reaction i2_ was kept constant for the last
          ! 2^(total number of reactions - i2_) permutations
          if( ( i1_ .ne. numRevPerm_ ) .and. ( suppor_(i2_) &
          .eq. (2**(nReact_ - i2_)) ) ) then
            ! If it was, change the order of the pair 0/2
            suppor_(i2_) = 0
            rdir_(i2_) = RevPerm_((2*(i2_ - 1) + 2),i1_)
            rinv_(i2_) = RevPerm_((2*(i2_ - 1) + 1),i1_)
          end if
        end do
      end do


      end subroutine allRevReactPerm


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!     Subroutine to test if two given structures are isomorphic representations
!!     of the same structure
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine isomorphismTest(supRevStruct_, RevStruct_, DataRevStru_, &
      MetDegRevStru_, ReactDegRevStru_, supMetDegRevStr_, supReactDegRevStr_, &
      nrefStruLow_, nrefStruHigh_, numStruct_, PermutMet_ ,PermutIrrReact_, &
      PermutRevReact_, PermutDirInvReact_, numPermutMet_, numPermutIrrReact_, &
      numPermutRevReact_, numPermDirInvReact_, isoStruct_, isoStruRef_)

! CONSTANTS
!      numIrrRea_ : number of irreversible reactions of the (possibly) new structure
!      numPermutMet_ : number of permutations for the given set of metabolites
!      numPermutIrrReact_ : number of permutations for the given set of irreversible reactions
!      numPermutRevReact_ : number of permutations for the given set of reversible reactions
!      numPermDirInvReact_ : number of permutations between the direct and inverse single direction reactions for all
!                            the reversible reactions and for each permutation of the reversible reactions
!      numStruct_ : reference number (temporary) of the new structure, stored in matrix supRevStruct_(i2,i1,1).
!                   This reference number is used to access the information temporarily stored in matrix DataRevStru_(i3,i2,i1)
!                   at line numStruct_. Without this, one cannot test if the structure in matrix supRevStruct_(i3,i2,i1) is
!                   a isomorphic representation of a structure in matrix RevStruct_(i3,i2,i1)
!      isoStruRef_ : reference number of a structure in matrix RevStruct_(i3,i2,i1) to which the structure in matrix
!                    supRevStruct_(i2,i1,1) is a isomorphic representation. We store its reference number to construct the
!                    Parent/Child relationship between the structure that lead to the isomorphic representation shown in
!                    matrix supRevStruct_(i2,i1,1) and structure isoStruRef_
!      nrefStruLow_ : lower reference number of a structure of the considered generation stored in matrix RevStruct_(i3,i2,i1)
!      nrefStruHigh_ : higher reference number of a structure of the considered generation stored in matrix RevStruct_(i3,i2,i1)
!      isoStruct_ : if the structure represented in matrix supRevStruct_(i3,i2,1) is a isomorphic representation of a structure
!                   stored in matrix RevStruct_(i3,i2,i1), isoStruct_ is filled with "1"; if it is a new structure, isoStruct_
!                   is filled with "0"
!
!
! VECTORS
!      supMetDegRevStr_(i1) : gives the degree for each metabolite of a (possibly) new structure. i1=[(3 x (nmet - 1)) + 1],
!                             total n of connections of the metabolite; i1=[(3 x (nmet - 1)) + 2], n of connections
!                             comming to the metabolite (IN); i1=[(3 x (nmet - 1)) + 3], n of connections going out of
!                             the metabolite (OUT)
!      supReactDegRevStr_(i1) : gives the degree for each reaction of a possibly new structure. The irreversible reactions
!                               are stored first, and then, in case at least one exists, the reversible reactions are stored.
!                               i1=[(3 x (nmet - 1)) + 1], connections of the metabolite; i1=[(3 x (nmet - 1)) + 2],
!                               n of connections comming to the total n of metabolite (IN); i1=[(3 x (nmet - 1)) + 3],
!                               n of connections going out of the metabolite (OUT)
!
!
! MATRICES
!      supRevStruct_(i3,i2,1) : support matrix to store a new structure. This structure will be compared to others already
!                                found to determine if it is a isomorphic representation of one of them
!      RevStruct_(i3,i2,i1) : matrix with the possible structures using a bipartite and directed
!                             graph representation. i1) number of the structure; i2) number of the metabolite;
!                             i3) [4*(i3-1) + 1] reactant of the direct reaction, [4*(i3-1) + 2] product of
!                             the direct reaction, [4*(i3-1) + 3] reactant of the inverse reaction, [4*(i3-1) + 4]
!                             product of the inverse reaction. If a metabolite is connected to a reaction, the
!                             corresponding cell is filled with 1; otherwise, is filled with 0
!      DataRevStru_(i2,i1) : i1) reference number of the    structure. i2) addtional data for each structure: (i2=1) number
!                            of generation; (i2=2) number of metabolites; (i2=3) number of reactions; (i2=4) number of
!                            reversible reactions; (i2=5) number of Ancestors; (i2=6) number of Parents; (i2=7) number of
!                            Childs; (i2=8) number of Offspring; (i2=9) number of interconnections; (i2=10) maximum possible
!                            number of interconnections; (i2=11) number of connections metabolite/reaction (reversibility not
!                            considered); (i2=12) maximum possible number of connections metabolite/reaction (reversibility
!                            not considered)
!      MetDegRevStru_(i2,i1) : gives the degree for each metabolite of structure i1: i2=[(3 x (nmet - 1)) + 1], total n of
!                              connections of the metabolite; i2=[(3 x (nmet - 1)) + 2], n of connections comming to the
!                              metabolite (IN); i2=[(3 x (nmet - 1)) + 3], n of connections going out of the metabolite (OUT)
!      ReactDegRevStru_(i2,i1) : gives the degree for each reaction of structure i1. The irreversible reactions are stored
!                                first and then, in case at least one exists, the reversible reactions are stored.
!                                i2=[(3 x (nmet - 1)) + 1], connections of the metabolite; i2=[(3 x (nmet - 1)) + 2],
!                                n of connections comming to the total n ofmetabolite (IN); i2=[(3 x (nmet - 1)) + 3],
!                                n of connections going out of the metabolite (OUT)
!      PermutMet_(i2,i1) : possible permutations for a set of metabolites, obtained by calling in the main program the subroutine
!                          allPermutations. i1) number of the permutation; i2) number of the metabolites ordered by this permutation
!      PermutIrrReact_(i2,i1) : possible permutations for a set of irreversible reactions, obtained by calling in the main program
!                               the subroutine allPermutations. i1) number of the permutation; i2) number of the reactions ordered
!                               by this permutation
!      PermutRevReact_(i2,i1) : possible permutations for a set of reversible reactions, obtained by calling in the main program
!                               the subroutine allPermutations. i1) number of the permutation; i2) number of the reactions ordered
!                               by this permutation
!      PermutDirInvReact_(i2,i1) :  possible permutations between the direct and inverse single direction reactions of a (possibly)
!                                   reversible reaction. The matrix was obtained by calling in the main program the subroutine
!                                   allRevReactPerm. These possible permutations take into account the total number of reactions
!                                   in the structure. i1) number of the permutation; i2) ordering the direct and inverse single
!                                   direction reactions for all the reactions

      implicit none

      ! IN variables
       ! Integer: 1 bit
      integer*1, intent(in) :: supRevStruct_(:,:,:), RevStruct_(:,:,:)
       ! Integer: 2 bits
      integer*2, intent(in) :: nrefStruLow_, nrefStruHigh_, numPermDirInvReact_
      integer*2, intent(in) :: DataRevStru_(:,:), MetDegRevStru_(:,:), ReactDegRevStru_(:,:)
      integer*2, intent(in) :: PermutMet_(:,:), PermutIrrReact_(:,:)
      integer*2, intent(in) :: PermutRevReact_(:,:), PermutDirInvReact_(:,:)
      integer*2, intent(in) :: supMetDegRevStr_(:), supReactDegRevStr_(:)
       ! Integer: 4 bits
      integer*4, intent(in) :: numPermutMet_, numPermutIrrReact_, numPermutRevReact_, numStruct_

      ! OUT variables
       ! Integer: 1 bit
      integer*1, intent(inout) :: isoStruct_
       ! Integer: 2 bits
      integer*2, intent(inout) :: isoStruRef_

      !Internal support variables (2 bits)
      integer*2 :: numIrrRea_

      ! General variables (4 bits)
      integer*4 :: iso1_, iso2_, iso3_, iso4_, iso5_, com1_, com2_, com3_
      integer*4 :: com4_, Nmet1_, Nmet2_, Nreact1_, Nreact2_, Nreact3_

!      Test if structure represented in matrix supRevStruct_(i2,i1,1) is a isomorphic representation
!      of a structure i1 stored in matrix RevStruct_(i3,i2,i1)
      ! Number of irreversible reactions of the (possibly) new structure
      numIrrRea_ = DataRevStru_(3,numStruct_) - DataRevStru_(4,numStruct_)
      ! Reference number of a structure already found (same generation)
      do 3000 iso1_=nrefStruLow_,max(nrefStruLow_,nrefStruHigh_)
        ! The test should be done only for structures with the same structural characteristics:
        ! number of metabolites, number of reactions, number of reversible reactions, number
        ! of interconnections, number of connections metabolite/reaction (reversibility not considered)
        if( ( DataRevStru_(2,numStruct_) .eq. DataRevStru_(2,iso1_) ) .and. &
        ( DataRevStru_(3,numStruct_) .eq. DataRevStru_(3,iso1_) ) .and. &
        ( DataRevStru_(4,numStruct_) .eq. DataRevStru_(4,iso1_) ) .and. &
        ( DataRevStru_(9,numStruct_) .eq. DataRevStru_(9,iso1_) ) .and. &
        ( DataRevStru_(11,numStruct_) .eq. DataRevStru_(11,iso1_) ) ) then
          ! Tests if the nodes (metabolites and reactions) of the possibly new structure have the same degree (total and IN) of the
          ! nodes of the previously found structure iso1_
          ! The metabolites have the same degree?
          do Nmet1_=1,DataRevStru_(2,numStruct_)  ! Number of metabolites of the (possibly) new structure
            if( ( MetDegRevStru_( ((3*(Nmet1_ - 1)) + 1) , iso1_) .ne. &
            supMetDegRevStr_( ((3*(Nmet1_ - 1)) + 1) ) ) .and. &
            ( MetDegRevStru_( ((3*(Nmet1_ - 1)) + 2) , iso1_) .ne. &
            supMetDegRevStr_( ((3*(Nmet1_ - 1)) + 2) ) ) ) goto 3000  ! The (possibly)
          end do  ! new structure is not an isomorphic representation of the previous found structure iso1_
          ! The reactions have the same degree?
          do Nreact1_=1,DataRevStru_(3,numStruct_)  ! Number of reactions of the (possibly) new structure
            if( ( ReactDegRevStru_( ((3*(Nreact1_ - 1)) + 1) , iso1_) .ne. &
            supReactDegRevStr_( ((3*(Nreact1_ - 1)) + 1)) ) .and. &
            ( ReactDegRevStru_( ((3*(Nreact1_ - 1)) + 2) , iso1_) .ne. &
            supReactDegRevStr_( ((3*(Nreact1_ - 1)) + 2)) ) ) goto 3000  ! The
          end do  ! (possibly) new structure is not an isomorphic representation of the previous found structure iso1_
          !
          ! The (possibly) new structure might be an isomorphic representation of the previous found structure iso1_
          !
          ! Reference number of the permutation for the metabolites
          do 3005 iso2_=1,numPermutMet_
            !
            ! Checks if it is the first iteration for the permutations of the metabolites
            if( iso2_ .ne. 1 ) then
              ! If it is not the first iteration, tests if the degree of metabolites are kept equal
              do Nmet2_=1,DataRevStru_(2,numStruct_)  ! Number of metabolites of the (possibly) new structure
                if( ( MetDegRevStru_( ((3*(PermutMet_(Nmet2_,iso2_) - 1)) + 1) , iso1_) &
                .ne. supMetDegRevStr_( ((3*(Nmet2_ - 1)) + 1) ) ) .and. &
                ( MetDegRevStru_( ((3*(PermutMet_(Nmet2_,iso2_) - 1)) + 2) , iso1_) &
                .ne. supMetDegRevStr_( ((3*(Nmet2_ - 1)) + 2) ) ) ) goto 3005  ! Checks the next permutation for the metabolites
              end do
            end if
            !
            ! The degree of the metabolites are kept with this permutation. One can continue to check if the (possibly) new
            ! structure is a isomorphic representation of the previously found structure iso1_
            !
            ! The (possibly) new structure has at least one irreversible reaction?
            if( numIrrRea_ .ne. 0 ) then
              ! Checks the permutations for the irreversible reactions
              do 3010 iso5_=1,numPermutIrrReact_
                !
                ! Checks if it is the first permutation for the irreversible reactions
                if( iso5_ .ne. 1 ) then
                  ! If it is not the first iteration, tests if the degree of the reactions are kept equal
                  do Nreact2_=1,numIrrRea_  ! Number of irreversible reactions of the (possibly) new structure
                    if( ( ReactDegRevStru_( ((3*(PermutIrrReact_(Nreact2_,iso5_) - 1)) &
                    + 1)  ,iso1_) .ne. supReactDegRevStr_( ((3*(Nreact2_ - 1)) + 1)) ) &
                    .and. ( ReactDegRevStru_( ((3*(PermutIrrReact_(Nreact2_,iso5_) &
                    - 1)) + 2)  ,iso1_) .ne. supReactDegRevStr_( ((3*(Nreact2_ - 1)) &
                    + 2)) ) ) goto 3010  ! Checks the next permutation for the irreversible
                  end do                                                         ! reactions
                end if
                !
                ! The part of the patterns with the connections metabolite/irreversible reactions of the two structures
                ! (the (possibly) new structure and structure iso1_) can be compared
                do com3_=1,DataRevStru_(2,numStruct_)  ! Number of metabolites of the structures
                  do com4_=1,numIrrRea_  ! Number of irreversible reactions of the structures
                    ! We only need to compare the direct reactions since there is no inverse reaction
                    !
                    ! How to determine the position? ->
                    !     Matrix([(4x(nreaction - 1)) + direct/inverse + reactant/product],[nmetabolite],[nstructure])
                    !      where,
                    !           direct/inverse: "0", because we only need to compare the direct reactions
                    !           reactant/product: "1" for a reactant, and "2" for a product
                    !     For the matrix supRevStruct_(i3,i2,i1), the positions of the metabolites and of the reactions
                    !     are not changed.
                    !     However, for the matrix RevStruct_(i3,i2,i1) the positions of the metabolites and of the reactions
                    !     change according to the respective permutations:
                    !       n of reaction -> PermutIrrReact_(com4_,iso5_)
                    !       direct/inverse reation position ->
                    !        i3: 1st term + 2nd term + 3rd term
                    !          1st term: [4 x [PermutIrrReact_(com4_,iso5_) - 1]]
                    !          2nd term: 0
                    !          3rd term: "1" for a reactant, and "2" for a product (the same value as the one considered to
                    !                    matrix supRevStruct_(i3,i2,i1))
                    !        i2: PermutMet_(com3_,iso2_)
                    !        i1: iso1_
                    !
                    ! Compares the direct single-direction reaction (direct/inverse term equal to "0"):
                    ! Reactant position (reactant/product term equal to "1")
                    if( supRevStruct_( ((4*(com4_ - 1)) + 0 + 1) ,com3_ ,1) .ne. &
                    RevStruct_( ((4*( PermutIrrReact_(com4_,iso5_) - 1 )) + 0 + 1 ) &
                    , PermutMet_(com3_,iso2_) , iso1_) ) goto 3010
                    !          |            1st term                    | |2nd||3rd |
                    ! Product position (reactant/product term equal to "2")
                    if( supRevStruct_( ((4*(com4_ - 1)) + 0 + 2) ,com3_ ,1) .ne. &
                    RevStruct_( ((4*( PermutIrrReact_(com4_,iso5_) - 1 )) + 0 + 2 ) &
                    , PermutMet_(com3_,iso2_) , iso1_) ) goto 3010
                    !
                    ! Is this part of the pattern, i.e., the connections metabolites/irreversible reactions the same between the
                    ! (possibly) new structure and structure iso1_?
                    if( (com3_ .eq. DataRevStru_(2,numStruct_)) .and. &
                    (com4_ .eq. numIrrRea_) ) then
                      ! We can now test the part of the pattern corresponding to the connections between metabolites and reversible
                      goto 3015  ! reactions
                    end if
                  end do
                end do
3010          continue
              ! This part of the code is not available to the cases were there is a correspondence between the connections
              ! metabolite/irreversible reaction of the (possibly) new structure and structure iso1_. So, it means a new
              ! permutation for the metabolites must be tested
              goto 3005
              !
            end if
            !
            ! The part of the pattern between the metabolites and the reversible reactions can now be tested
            goto 3015
3015        continue
            !
            ! The (possibly) new structure has at least one reversible reaction?
            if( DataRevStru_(4,numStruct_) .ne. 0 ) then
              ! Reference number of the permutation for the reversible reactions
              do 3020 iso3_=1,numPermutRevReact_
                ! Checks if it is the first permutation for the reversible reactions
                if( iso3_ .ne. 1 ) then
                  ! If it is not the first iteration, tests if the degree of the reactions are kept equal (only the total
                  ! degree is needed to test the reversible reactions of the two structures: for reversible reactions, the
                  ! IN degree is equal to the OUT degree
                  do Nreact3_=1,DataRevStru_(4,numStruct_)  ! Number of reversible reactions of the (possibly) new structure
                    if( ReactDegRevStru_( ((3*(numIrrRea_ + PermutRevReact_(Nreact3_,iso3_) &
                    - 1)) + 1)  ,iso1_) .ne. supReactDegRevStr_( ((3*(numIrrRea_ + Nreact3_ &
                    - 1)) + 1)) ) goto 3020  ! Checks the next permutation for the reversible
                  end do                                                         ! reactions
                end if
                !
                ! The degree of the metabolites are kept with this permutation. One can continue to check if the (possibly) new
                ! structure is a isomorphic representation of the previously found structure iso1_
                !
                ! Reference number of the permutation for the direct and inverse single direction reactions for
                ! all the reversible reactions
                do 3025 iso4_=1,numPermDirInvReact_
                  !
                  ! Each possible permutation (perm. Met x perm. RevReact x perm. DirInvReact.) will be used to compare
                  ! the already found structures of this generation with the newly created
                  do com1_=1,DataRevStru_(2,numStruct_)  ! Number of metabolites of the structures
                    do com2_=1,DataRevStru_(4,numStruct_)  ! Number of reversible reactions of the structures
                      !
                      ! How to determine the position? ->
                      !     Matrix([(4x(nreaction - 1)) + direct/inverse + reactant/product],[nmetabolite],[nstructure])
                      !      where,
                      !           direct/inverse: "0" if it is the direct reaction, and "2" if it is the inverse reaction
                      !           reactant/product: "1" for a reactant, and "2" for a product
                      !     For the matrix supRevStruct_(i3,i2,i1), the positions of the metabolites and of the reactions
                      !     are not changed.
                      !     However, for the matrix RevStruct_(i3,i2,i1) the positions of the metabolites and of the reactions
                      !     change according to the respective permutations:
                      !       n of reaction -> PermutReact_(com2_,iso3_)
                      !       direct/inverse reation position ->
                      !        i3: 1st term + 2nd term + 3rd term
                      !          1st term: [4 x [PermutReact_(com2_,iso3_) - 1]]
                      !          2nd term: PermutRevReact_([[2 x [n of reaction - 1]] + a],iso4_)
                      !                    where "a" is according to the position considered to matrix supRevStruct_(i3,i2,i1):
                      !                    "1" for the direct reaction, and "2" for the inverse reaction
                      !          3rd term: "1" for a reactant, and "2" for a product (the same value as the one considered to
                      !                    matrix supRevStruct_(i3,i2,i1))
                      !        i2: PermutMet_(com1_,iso2_)
                      !        i1: iso1_
                      !
                      ! Compares the direct single-direction reaction (direct/inverse term equal to "0"):
                      ! Reactant position (reactant/product term equal to "1")
                      if( supRevStruct_( ((4*(numIrrRea_ + com2_ - 1)) + 0 + 1) ,com1_ ,1) &
                      .ne. RevStruct_( ((4*( numIrrRea_ + PermutRevReact_(com2_,iso3_) - 1 )) &
                      + PermutDirInvReact_( ((2*( PermutRevReact_(com2_,iso3_) - 1)) + 1 ) &
                      , iso4_ ) + 1 ) , PermutMet_(com1_,iso2_) , iso1_) ) goto 3025
                      !          |                      1st term                        |   |                           2nd term                                      ||3rd term|
                      ! Product position (reactant/product term equal to "2")
                      if( supRevStruct_( ((4*(numIrrRea_ + com2_ - 1)) + 0 + 2) ,com1_ ,1) &
                      .ne. RevStruct_( ((4*( numIrrRea_ + PermutRevReact_(com2_,iso3_) - 1 )) &
                      + PermutDirInvReact_( ((2*( PermutRevReact_(com2_,iso3_) - 1)) + 1 ) &
                      , iso4_ ) + 2 ) , PermutMet_(com1_,iso2_) , iso1_) ) goto 3025
                      !
                      ! Compares the inverse single-direction reaction (direct/inverse term equal to "0"):
                      ! Reactant position (reactant/product term equal to "1")
                      if( supRevStruct_( ((4*(numIrrRea_ + com2_ - 1)) + 2 + 1) ,com1_ ,1) &
                      .ne. RevStruct_( ((4*( numIrrRea_ + PermutRevReact_(com2_,iso3_) - 1 )) &
                      + PermutDirInvReact_( ((2*( PermutRevReact_(com2_,iso3_) - 1)) + 2 ) &
                      , iso4_ ) + 1 ) , PermutMet_(com1_,iso2_) , iso1_) ) goto 3025
                      ! Product position (reactant/product term equal to "2")
                      if( supRevStruct_( ((4*(numIrrRea_ + com2_ - 1)) + 2 + 2) ,com1_ ,1) &
                      .ne. RevStruct_( ((4*( numIrrRea_ + PermutRevReact_(com2_,iso3_) - 1 )) &
                      + PermutDirInvReact_( ((2*( PermutRevReact_(com2_,iso3_) - 1)) + 2 ) &
                      , iso4_ ) + 2 ) , PermutMet_(com1_,iso2_) , iso1_) ) goto 3025
                      !
                      ! Is this a isomorphic representation of a structure found before? For that to happen, all the cells
                      ! must be compared: all the metabolites (com1_ equal to number of metabolites) and all reactions
                      ! (com2_ equal to number of reversible reactions)
                      if( (com1_ .eq. DataRevStru_(2,numStruct_)) .and. &
                      (com2_ .eq. DataRevStru_(4,numStruct_)) )   then
                        !
                        ! Fills the variable with "1", indicating the structure represented in matrix supRevStruct_(i3,i2,i1)
                        ! is a isomorphic representation of structure iso1_, stored in matrix RevStruct_(i3,i2,i1)
                        isoStruct_ = 1
                        !
                        ! Stores the reference number i1 of the structure of matrix RevStruct_(i3,i2,i1), of which the structure
                        ! represented in matrix supRevStruct_(i2,i1,1) is one of its isomorphic representations. We store its
                        ! reference number to construct the Parent/Child relationship between the structure that lead to the
                        ! isomorphic representation represented in matrix supRevStruct_(i2,i1,1) and structure iso1_
                        !
                        isoStruRef_ = iso1_
                        !
                        !
                        goto 3030
                      end if
                    end do
                  end do
3025            continue
3020          continue
            else
              ! The structure has no reversible reactions
              !
              ! Fills the variable with "1", indicating the structure represented in matrix supRevStruct_(i3,i2,i1)
              ! is a isomorphic representation of structure iso1_, stored in matrix RevStruct_(i3,i2,i1)
              isoStruct_ = 1
              !
              ! Stores the reference number i1 of the structure of matrix RevStruct_(i3,i2,i1), of which the structure
              ! represented in matrix supRevStruct_(i2,i1,1) is one of its isomorphic representations. We store its
              ! reference number to construct the Parent/Child relationship between the structure that lead to the
              ! isomorphic representation represented in matrix supRevStruct_(i2,i1,1) and structure iso1_
              !
              isoStruRef_ = iso1_
              !
              !
              goto 3030
              !
            end if
3005      continue
        end if
3000  continue
      !
      ! This part of the code is not available to the situations in which the structure represented in matrix supRevStruct_(i2,i1,1)
      ! is a isomorphic representation of a structure in matrix RevStruct_(i3,i2,i1)
      !
      ! The structure in matrix supRevStruct_(i2,i1,1) is a new structure, by not being a isomorphic representation of a structure
      ! in matrix RevStruct_(i3,i2,i1). So, isoStruct equal to "0"
      isoStruct_ = 0
      ! As we have found a new structure, we do not store the reference number of a previous found structure in matrix
      ! RevStruct_(i3,i2,i1). So, isoStruRef_ equal to "0"
      isoStruRef_ = 0
      goto 3030
3030  continue

      end subroutine isomorphismTest


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!     Subroutine to determine the equivalent metabolites of a given structure
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine EquiMet(numRefStructure_, RevStruc_, numNod_, MetDegRevStr_, &
      ReactDegRevStr_, PermutMeta_, PermutIrrReacti_, PermutRevReacti_, &
      PermutDirInvReacti_, numPermutMeta_, numPermutIrrReacti_, &
      numPermutRevReacti_, numPermDirInvReacti_, EquiMetRevStru_)

! CONSTANTS
!      numRefStructure_ : reference number of the structure that will be tested to determine the equivalent metabolites
!      numPermutMeta_ : number of permutations for the given set of metabolites
!      numPermutIrrReacti_ : number of permutations for the given set of irreversible reactions
!      numPermutRevReacti_ : number of permutations for the given set of reversible reactions
!      numPermDirInvReacti_ : number of permutations between the direct and inverse single direction reactions for all
!                             the considered reactions (which can be reversible) and for each permutation of the reactions
!
!
! VECTORS
!      numNod_(i1) : number of nodes of the (possibly) new struture: (i1=1) number of metabolites; (i1=2) number of reactions;
!                    (i1=3) number of reversible reactions  ; (i1=4) number of irreversible reactions
!      aid_(i1) : support constant used to obtain the positions to store the information about the topological equivalence of
!                 two metabolites in the matrix EquiMetRevStru_(i2,i1)
!      supMetDegRevSt_(i1) : support vector with the information regarding the degrees (total, IN and OUT) of the metabolites
!                            of the structure numRefStructure_
!      supReactDegRevSt_(i1) : support vector with the information regarding the degrees (total, IN and OUT) of the reactions
!                              of the structure numRefStructure_
!
!
! MATRICES
!      RevStruc_(i3,i2,i1) : matrix with the possible structures using a bipartite and directed
!                            graph representation. i1) number of the structure; i2) number of the metabolite;
!                            i3) [4*(i3-1) + 1] reactant of the direct reaction, [4*(i3-1) + 2] product of
!                            the direct reaction, [4*(i3-1) + 3] reactant of the inverse reaction, [4*(i3-1) + 4]
!                            product of the inverse reaction. If a metabolite is connected to a reaction, the
!                            corresponding cell is filled with 1; otherwise, is filled with 0
!      supRevStruc_(i2,i1,1) : support matrix with the pattern of structure numRefStructure_
!      MetDegRevStr_(i2,i1) : gives the degree for each metabolite of structure i1: i2=[(3 x (nmet - 1)) + 1], total n of
!                             connections of the metabolite; i2=[(3 x (nmet - 1)) + 2], n of connections comming to the
!                             metabolite (IN); i2=[(3 x (nmet - 1)) + 3], n of connections going out of the metabolite (OUT)
!      ReactDegRevStr_(i2,i1) : gives the degree for each reaction of structure i1. The irreversible reactions are stored
!                               first and then, in case at least one exists, the reversible reactions are stored.
!                               i2=[(3 x (nmet - 1)) + 1], connections of the metabolite; i2=[(3 x (nmet - 1)) + 2],
!                               n of connections comming to the total n ofmetabolite (IN); i2=[(3 x (nmet - 1)) + 3],
!                               n of connections going out of the metabolite (OUT)
!      PermutMeta_(i2,i1) : possible permutations for a set of metabolites, obtained by calling the subroutine allPermutations.
!                           i1) number of the permutation; i2) number of the metabolites ordered by this permutation
!      PermutIrrReacti_(i2,i1) : possible permutations for a set of irreversible reactions, obtained by calling the subroutine
!                                allPermutations. i1) number of the permutation; i2) number of the reactions ordered by this
!                                permutation
!      PermutRevReacti_(i2,i1) : possible permutations for a set of reversible reactions, obtained by calling the subroutine
!                                allPermutations. i1) number of the permutation; i2) number of the reactions ordered by this
!                                permutation
!      PermutDirInvReacti_(i2,i1) : possible permutations between the direct and inverse single direction reactions of a reversible
!                                   reaction. The matrix was obtained by calling the subroutine allRevReactPerm. These permutations
!                                   permutations take into account the total number of reversible reactions in the structure.
!                                   i1) number of the permutation; i2) ordering the direct and inverse single   direction reactions
!                                   for all the reversible reactions
!      EquiMetRevStru_(i2,i1) : information regarding the equivalent metabolites of a given structure i1. If two metabolites are
!                               equivalent, the corresponding cell is filled with "1"; if not, is filled with "0". The first
!                               metabolite is compared with the ones in higher positions, the second metabolite is compared to
!                               metabolites who are also in higher positions, ... till we compare the last two metabolites.
!                               For the metabolite in a general position X (lower than the last position available), it will
!                               be compared to metabolites in increasing higher positions in the cell. Considering the case
!                               where the metabolite in the position X is compared to the metabolite in the position Y, the
!                               information regarding the topological equivalence of the two nodes will be stored in the cell:
!                               i2 = (Somatrio[i=1,(X-1)] n Met. - i ) + ( Y - X )

      implicit none

      ! IN variables
       ! Integer: 1 bit
      integer*1, intent(in) :: RevStruc_(:,:,:)
       ! Integer: 2 bits
      integer*2, intent(in) :: PermutMeta_(:,:), PermutIrrReacti_(:,:), PermutRevReacti_(:,:)
      integer*2, intent(in) :: PermutDirInvReacti_(:,:),MetDegRevStr_(:,:)
      integer*2, intent(in) :: ReactDegRevStr_(:,:), numNod_(:), numPermDirInvReacti_
       ! Integer: 4 bits
      integer*4, intent(in) :: numPermutMeta_, numPermutIrrReacti_, numPermutRevReacti_
      integer*4, intent(in) :: numRefStructure_

      ! IN/OUT variables
       ! Integer: 1 bit
      integer*1, intent(inout) :: EquiMetRevStru_(:,:)

      !Internal support variables
       ! Integer: 1 bit
      integer*1 :: supRevStruc_(32,8,1), aid_(2)
       ! Integer: 2 bits
      integer*2 :: supMetDegRevSt_(24), supReactDegRevSt_(24)

      ! General variables (4 bits)
      integer*4 :: iMet1_, iMet2_, eqM1_, eqM2_, eqM3_, eqM4_, cm1_
      integer*4 :: cm2_, cm3_, cm4_, Nmeta1_, Nreac1_, Nreac2_

!      Determines the topologically equivalent metabolites of the structure numRefStructure_
      ! Creates the support matrix supRevStruc_(i3,i2,i1)
      do iMet1_=1,numNod_(1)  ! Number of metabolites
        do iMet2_=1,(4*numNod_(2))  ! 4 x number of reactions
          supRevStruc_(iMet2_,iMet1_,1) = RevStruc_(iMet2_,iMet1_,numRefStructure_)
        end do
      end do
      ! Creates the support vector supMetDegRevSt_(i1)
      do iMet1_=1,(3*numNod_(1))  ! 3 x number of metabolites
        supMetDegRevSt_(iMet1_) = MetDegRevStr_(iMet1_,numRefStructure_)
      end do
      ! Creates the support vector supReactDegRevSt_(i1)
      do iMet1_=1,(3*numNod_(2))     ! 3 x number of reactions
        supReactDegRevSt_(iMet1_) = ReactDegRevStr_(iMet1_,numRefStructure_)
      end do
      !
      aid_(1) = 0
      aid_(2) = 0
      ! Metabolite iMet1_
      do iMet1_=1,(numNod_(1)-1)  ! Number of metabolites - 1
        ! Metabolite iMet2_
        do 3035 iMet2_=(iMet1_+1),numNod_(1)  ! From (iMet1_+1) to number of metabolites
          !
          ! Adds "1" to the support variable aid_(2), needed to determine a position to store the information of
          ! the topological equivalence of the metabolites iMet1_ and iMet2_ in matrix EquiMetRevStru_(i2,i1)
          aid_(2) = aid_(2) + 1
          !
          ! Metabolite iMet1_ and metabolite iMet2_ are topologically equivalent?
          ! For that to happen, the degree (total and IN) of the two metabolites must be the same
          if( ( MetDegRevStr_( ((3*(iMet1_ - 1)) + 1) , numRefStructure_ ) .eq. &
          MetDegRevStr_( ((3*(iMet2_ - 1)) + 1) , numRefStructure_ ) ) .and. &
          ( MetDegRevStr_( ((3*(iMet1_ - 1)) + 2) , numRefStructure_ ) .eq. &
          MetDegRevStr_( ((3*(iMet2_ - 1)) + 2) , numRefStructure_ ) )  ) then
            ! The metabolites iMet1_ and iMet2_ have the same degree (total and IN)
            !
            ! Reference number of the permutation for the metabolites
            do 3040 eqM1_=1,numPermutMeta_
              !
              ! Checks if the metabolite iMet2_ is in the position of metabolite iMet1_
              if( PermutMeta_(iMet1_,eqM1_) .ne. iMet2_ ) goto 3040
              !
              ! Checks if the degree of the metabolites are the same with this permutation
              do Nmeta1_=1,numNod_(1)  ! Number of metabolites of the structure
                if( MetDegRevStr_( ((3*(PermutMeta_(Nmeta1_,eqM1_) - 1)) + 1) &
                , numRefStructure_) .ne. supMetDegRevSt_( ((3*(Nmeta1_ - 1)) &
                + 1) ) ) goto 3040
                if( MetDegRevStr_( ((3*(PermutMeta_(Nmeta1_,eqM1_) - 1)) + 2) &
                , numRefStructure_) .ne. supMetDegRevSt_( ((3*(Nmeta1_ - 1)) &
                + 2) ) ) goto 3040  ! Checks the next permutation for the metabolites
              end do
              !
              ! The degree of the metabolites are kept with this permutation. One can continue to check if the metabolite
              ! iMet1_ is topologically equivalent to metabolite iMet2_
              !
              ! The structure numRefStructure_ has at least one irreversible reaction?
              if( numNod_(4) .ne. 0 ) then
                ! Checks the permutations for the irreversible reactions
                do 3045 eqM4_=1,numPermutIrrReacti_
                  !
                  ! Checks if it is the first permutation for the irreversible reactions
                  if( eqM4_ .ne. 1 ) then
                    ! If it is not the first iteration, tests if the degree of the reactions are kept equal
                    do Nreac1_=1,numNod_(4)  ! Number of irreversible reactions of the structure
                      if( ReactDegRevStr_( ((3*(PermutIrrReacti_(Nreac1_,eqM4_) - 1)) &
                       + 1) , numRefStructure_) .ne. supReactDegRevSt_( ((3*(Nreac1_ - 1)) &
                      + 1)) ) goto 3045
                      if( ReactDegRevStr_( ((3*(PermutIrrReacti_(Nreac1_,eqM4_) - 1)) &
                      + 2) , numRefStructure_) .ne. supReactDegRevSt_( ((3*(Nreac1_ - 1)) &
                      + 2)) ) goto 3045  ! Checks the next permutation for the irreversible
                    end do                                                       ! reactions
                  end if
                  !
                  ! The part of the patterns with the connections metabolite/irreversible reactions of the two isomorphic
                  ! representations of the structure can be compared
                  do cm3_=1,numNod_(1)  ! Number of metabolites of the structure
                    do cm4_=1,numNod_(4)  ! Number of irreversible reactions of the structure
                      ! We only need to compare the direct reactions since there is no inverse reaction
                      !
                      ! How to determine the position? ->
                      !      Matrix([(4x(nreaction - 1)) + direct/inverse + reactant/product],[nmetabolite],[nstructure])
                      !      where,
                      !           direct/inverse: "0", because we only need to compare the direct reactions
                      !           reactant/product: "1" for a reactant, and "2" for a product
                      !     For the matrix supRevStruc_(i3,i2,i1), the positions of the metabolites and of the reactions
                      !     are not changed.
                      !     However, for the matrix RevStruc_(i3,i2,i1) the positions of the metabolites and of the reactions
                      !     change according to the respective permutations:
                      !       n of reaction -> PermutIrrReacti_(cm4_,eqM4_)
                      !       direct/inverse reation position ->
                      !        i3: 1st term + 2nd term + 3rd term
                      !          1st term: [4 x [PermutIrrReacti_(cm4_,eqM4_) - 1]]
                      !          2nd term: 0
                      !          3rd term: "1" for a reactant, and "2" for a product (the same value as the one considered to
                      !                    matrix supRevStruc_(i3,i2,i1))
                      !        i2: PermutMeta_(cm3_,eqM1_)
                      !        i1: numRefStructure_
                      !
                      ! Compares the direct single-direction reaction (direct/inverse term equal to "0"):
                      ! Reactant position (reactant/product term equal to "1")
                      if( supRevStruc_( ((4*(cm4_ - 1)) + 0 + 1) ,cm3_ ,1) .ne. &
                      RevStruc_( ((4*( PermutIrrReacti_(cm4_,eqM4_) - 1 )) + 0 + 1 ) &
                      , PermutMeta_(cm3_,eqM1_) , numRefStructure_) ) goto 3045
                      !         |           1st term                     | |2nd||3rd |
                      ! Product position (reactant/product term equal to "2")
                      if( supRevStruc_( ((4*(cm4_ - 1)) + 0 + 2) ,cm3_ ,1) .ne. &
                      RevStruc_( ((4*( PermutIrrReacti_(cm4_,eqM4_) - 1 )) + 0 + 2 ) &
                      , PermutMeta_(cm3_,eqM1_) , numRefStructure_) ) goto 3045
                      !
                      ! Is this part of the pattern, i.e., the connections metabolites/irreversible reactions are the same between
                      ! the two isomorphic representations of the structure?
                      if( (cm3_ .eq. numNod_(1)) .and. (cm4_ .eq. numNod_(4)) ) then
                      ! We can now test the part of the pattern corresponding to the connections between metabolites and reversible
                        goto 3050  ! reactions
                      end if
                    end do
                  end do
3045            continue
                ! This part of the code is not available to the cases were there is a correspondence between the connections
                ! metabolite/irreversible reaction of the two isomorphic representations of the structure numRefStructure_. So,
                ! it means a new permutation for the metabolites must be tested
                goto 3040
                !
              end if
              !
              ! The part of the pattern between the metabolites and the reversible reactions can now be tested
              goto 3050
3050          continue
              !
              ! The structure numRefStructure_ has at least one reversible reaction?
              if( numNod_(3) .ne. 0 ) then
                ! Reference number of the permutation for the reversible reactions
                do 3055 eqM2_=1,numPermutRevReacti_
                  ! Checks if it is the first permutation for the reversible reactions
                  if( eqM2_ .ne. 1 ) then
                    ! If it is not the first iteration, tests if the degree of the reactions are kept equal (only the total
                    ! degree is needed to test the reversible reactions of the two structures: for reversible reactions, the
                    ! IN degree is equal to the OUT degree
                    do Nreac2_=1,numNod_(3)  ! Number of reversible reactions of the (possibly) new structure
                      if( ReactDegRevStr_( ((3*(numNod_(4) + PermutRevReacti_(Nreac2_,eqM2_) &
                      - 1)) + 1)  ,numRefStructure_) .ne. supReactDegRevSt_( ((3*(numNod_(4) &
                      + Nreac2_ - 1)) + 1)) ) goto 3055  ! Checks the next permutation for the
                    end do                                                       ! reversible reactions
                  end if
                  !
                  ! The degree of the metabolites are kept with this permutation. One can continue to check if the metablites
                  ! iMet1_ and iMet2_ are topologically equivalents
                  !
                  ! Reference number of the permutation for the direct and inverse single direction reactions for
                  ! all the reversible reactions
                  do 3060 eqM3_=1,numPermDirInvReacti_
                    !
                    ! Each possible permutation (perm. Met x perm. RevReact x perm. DirInvReact.) will be used to compare
                    ! the two isomorphic representations of the structure numRefStructure_
                    do cm1_=1,numNod_(1)  ! Number of metabolites of the structures
                      do cm2_=1,numNod_(3)  ! Number of reversible reactions of the structures
                        !
                        ! How to determine the position? ->
                        !      Matrix([(4x(nreaction - 1)) + direct/inverse + reactant/product],[nmetabolite],[nstructure])
                        !      where,
                        !           direct/inverse: "0" if it is the direct reaction, and "2" if it is the inverse reaction
                        !           reactant/product: "1" for a reactant, and "2" for a product
                        !     For the matrix supRevStruc_(i3,i2,i1), the positions of the metabolites and of the reactions
                        !     are not changed.
                        !     However, for the matrix RevStruc_(i3,i2,i1) the positions of the metabolites and of the reactions
                        !     change according to the respective permutations:
                        !       n of reaction -> PermutReact_(cm2_,eqM2_)
                        !       direct/inverse reation position ->
                        !        i3: 1st term + 2nd term + 3rd term
                        !          1st term: [4 x [PermutReact_(cm2_,eqM2_) - 1]]
                        !          2nd term: PermutRevReacti_([[2 x [n of reaction - 1]] + a],eqM3_)
                        !                    where "a" is according to the position considered to matrix supRevStruc_(i3,i2,i1):
                        !                    "1" for the direct reaction, and "2" for the inverse reaction
                        !          3rd term: "1" for a reactant, and "2" for a product (the same value as the one considered to
                        !                    matrix supRevStruc_(i3,i2,i1))
                        !        i2: PermutMeta_(cm1_,eqM1_)
                        !        i1: numRefStructure_
                        !
                        ! Compares the direct single-direction reaction (direct/inverse term equal to "0"):
                        ! Reactant position (reactant/product term equal to "1")
                        if( supRevStruc_( ((4*(numNod_(4) + cm2_ - 1)) + 0 + 1) ,cm1_ ,1) &
                        .ne. RevStruc_( ((4*( numNod_(4) + PermutRevReacti_(cm2_,eqM2_) &
                        - 1 )) + PermutDirInvReacti_( ((2*( PermutRevReacti_(cm2_,eqM2_) &
                        - 1)) + 1 ) , eqM3_ ) + 1 ) , PermutMeta_(cm1_,eqM1_) , &
                        numRefStructure_) ) goto 3060
                        !            |                      1st term                        |   |                           2nd term                                    ||3rd term|
                        ! Product position (reactant/product term equal to "2")
                        if( supRevStruc_( ((4*(numNod_(4) + cm2_ - 1)) + 0 + 2) ,cm1_ ,1) &
                        .ne. RevStruc_( ((4*( numNod_(4) + PermutRevReacti_(cm2_,eqM2_) &
                        - 1 )) + PermutDirInvReacti_( ((2*( PermutRevReacti_(cm2_,eqM2_) &
                        - 1)) + 1 ) , eqM3_ ) + 2 ) , PermutMeta_(cm1_,eqM1_) &
                        , numRefStructure_) ) goto 3060
                        !
                        ! Compares the inverse single-direction reaction (direct/inverse term equal to "0"):
                        ! Reactant position (reactant/product term equal to "1")
                        if( supRevStruc_( ((4*(numNod_(4) + cm2_ - 1)) + 2 + 1) ,cm1_ ,1) &
                        .ne. RevStruc_( ((4*( numNod_(4) + PermutRevReacti_(cm2_,eqM2_) &
                        - 1 )) + PermutDirInvReacti_( ((2*( PermutRevReacti_(cm2_,eqM2_) &
                        - 1)) + 2 ) , eqM3_ ) + 1 ) , PermutMeta_(cm1_,eqM1_) &
                        , numRefStructure_) ) goto 3060
                        ! Product position (reactant/product term equal to "2")
                        if( supRevStruc_( ((4*(numNod_(4) + cm2_ - 1)) + 2 + 2) ,cm1_ ,1) &
                        .ne. RevStruc_( ((4*( numNod_(4) + PermutRevReacti_(cm2_,eqM2_) &
                        - 1 )) + PermutDirInvReacti_( ((2*( PermutRevReacti_(cm2_,eqM2_) &
                        - 1)) + 2 ) , eqM3_ ) + 2 ) , PermutMeta_(cm1_,eqM1_) &
                        , numRefStructure_) ) goto 3060
                        !
                        ! The metabolites iMet1_ and iMet2_ are topologically equivalents? For that to happen, all the cells
                        ! must be compared: all the metabolites (cm1_ equal to number of metabolites) and all reactions
                        ! (cm2_ equal to number of reversible reactions)
                        if( (cm1_ .eq. numNod_(1)) .and. (cm2_ .eq. numNod_(3)) ) then
                          !
                          ! The metabolites iMet1_ and iMet2_ are topologically equivalents
                          EquiMetRevStru_( (aid_(1) + aid_(2)) , numRefStructure_ ) = 1
                          !
                          goto 3035
                        end if
                      end do
                    end do
3060              continue
3055            continue
              else
                ! The structure has no reversible reactions
                !
                ! The metabolites iMet1_ and iMet2_ are topologically equivalents
                EquiMetRevStru_( (aid_(1) + aid_(2)) , numRefStructure_ ) = 1
                !
                goto 3035
                !
              end if
3040        continue
            !
            ! This part of the code is not available to the situations in which the metabolites iMet1_ and iMet2_ are
            ! topologically equivalent
            !
            ! The metabolites iMet1_ and iMet2_ are not topologically equivalent
            EquiMetRevStru_( (aid_(1) + aid_(2)) , numRefStructure_ ) = 0
            !
            goto 3035
          else
            ! The metabolites iMet1_ and iMet2_ are not topologically equivalent
            EquiMetRevStru_( (aid_(1) + aid_(2)) , numRefStructure_ ) = 0
            !
            goto 3035
          end if
3035    continue
        !
        ! Adds (numNod_(1) - iMet1_) to the support variable aid_(1), needed to the determine a position to store the
        ! information of the topological equivalence of the metabolites iMet1_ and iMet2_ in matrix EquiMetRevStru_(i2,i1)
        aid_(1) = aid_(1) + (numNod_(1) - iMet1_)
        !
        ! Turns to "0" the support variable aid_(2)
        aid_(2) = 0
      end do

      end subroutine EquiMet


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!     Subroutine to determine the equivalent reactions of a given structure
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine EquiReact(numRefStructur_, RevStructu_, ReactRevStr_, &
      numNode_, MetDegRevStruc_, ReactDegRevStruc_, PermutMetab_, &
      PermutIrrReac_, PermutRevReac_, PermutDirInvReac_, &
      numPermutMetab_, numPermutIrrReac_, numPermutRevReac_, &
      numPermDirInvReac_, EquiReactRevStru_)

! CONSTANTS
!      numRefStructur_ : reference number of the structure that will be tested to determine the equivalent metabolites
!      numPermutMetab_ : number of permutations for the given set of metabolites
!      numPermutIrrReac_ : number of permutations for the given set of irreversible reactions
!      numPermutRevReac_ : number of permutations for the given set of reversible reactions
!      numPermDirInvReac_ : number of permutations between the direct and inverse single direction reactions for all
!                             the considered reactions (which can be reversible) and for each permutation of the reactions
!
!
! VECTORS
!      numNode_(i1) : number of nodes of the (possibly) new struture: (i1=1) number of metabolites; (i1=2) number of reactions;
!                    (i1=3) number of reversible reactions  ; (i1=4) number of irreversible reactions
!      aidR_(i1) : support constant used to obtain the positions to store the information about the topological equivalence of
!                  two metabolites in the matrix EquiReactRevStru_(i2,i1)
!      supMetDegRevStruc_(i1) : support vector with the information regarding the degrees (total, IN and OUT) of the metabolites
!                            of the structure numRefStructur_
!      supReactDegRevStruc_(i1) : support vector with the information regarding the degrees (total, IN and OUT) of the reactions
!                              of the structure numRefStructur_
!
!
! MATRICES
!      RevStructu_(i3,i2,i1) : matrix with the possible structures using a bipartite and directed
!                              graph representation. i1) number of the structure; i2) number of the metabolite;
!                              i3) [4*(i3-1) + 1] reactant of the direct reaction, [4*(i3-1) + 2] product of
!                              the direct reaction, [4*(i3-1) + 3] reactant of the inverse reaction, [4*(i3-1) + 4]
!                              product of the inverse reaction. If a metabolite is connected to a reaction, the
!                              corresponding cell is filled with 1; otherwise, is filled with 0
!      ReactRevStr_(i2,i1) : i1) reference number of the structure. i2) number of the reaction of the structure: if a
!                            a reaction is reversible, the corresponding cell is filled with 1, and if it is not, with 0
!      supRevStructu_(i2,i1,1) : support matrix with the pattern of structure numRefStructur_
!      MetDegRevStruc_(i2,i1) : gives the degree for each metabolite of structure i1: i2=[(3 x (nmet - 1)) + 1], total n of
!                             connections of the metabolite; i2=[(3 x (nmet - 1)) + 2], n of connections comming to the
!                             metabolite (IN); i2=[(3 x (nmet - 1)) + 3], n of connections going out of the metabolite (OUT)
!      ReactDegRevStruc_(i2,i1) : gives the degree for each reaction of structure i1. The irreversible reactions are stored
!                               first and then, in case at least one exists, the reversible reactions are stored.
!                               i2=[(3 x (nmet - 1)) + 1], connections of the metabolite; i2=[(3 x (nmet - 1)) + 2],
!                               n of connections comming to the total n ofmetabolite (IN); i2=[(3 x (nmet - 1)) + 3],
!                               n of connections going out of the metabolite (OUT)
!      PermutMetab_(i2,i1) : possible permutations for a set of metabolites, obtained by calling the subroutine allPermutations.
!                           i1) number of the permutation; i2) number of the metabolites ordered by this permutation
!      PermutIrrReac_(i2,i1) : possible permutations for a set of irreversible reactions, obtained by calling the subroutine
!                                allPermutations. i1) number of the permutation; i2) number of the reactions ordered by this
!                                permutation
!      PermutRevReac_(i2,i1) : possible permutations for a set of reversible reactions, obtained by calling the subroutine
!                                allPermutations. i1) number of the permutation; i2) number of the reactions ordered by this
!                                permutation
!      PermutDirInvReac_(i2,i1) : possible permutations between the direct and inverse single direction reactions of a reversible
!                                   reaction. The matrix was obtained by calling the subroutine allRevReactPerm. These permutations
!                                   permutations take into account the total number of reversible reactions in the structure.
!                                   i1) number of the permutation; i2) ordering the direct and inverse single   direction reactions
!                                   for all the reversible reactions
!      EquiReactRevStru_(i2,i1) : information regarding the equivalent reactions of a given structure i1. If two reactions are
!                                 equivalent, the corresponding cell is filled with "1"; if not, is filled with "0". The first
!                                 reaction is compared with the ones in higher positions, the second reaction is compared to
!                                 reactionswho are also in higher positions, ... till we compare the last two reactions.
!                                 For the reaction in a general position X (lower than the last position available), it will
!                                 be compared to reactions in increasing higher positions in the cell. Considering the case
!                                 where the reaction in the position X is compared to the metabolite in the position Y, the
!                                 information regarding the topological equivalence of the two nodes will be stored in the cell:
!                                 i2 = (Somatrio[i=1,(X-1)] n React. - i ) + ( Y - X )

      implicit none

      ! IN variables
       ! Integer: 1 bit
      integer*1, intent(in) :: RevStructu_(:,:,:), ReactRevStr_(:,:)
       ! Integer: 2 bits
      integer*2, intent(in) :: PermutMetab_(:,:), PermutIrrReac_(:,:)
      integer*2, intent(in) :: PermutRevReac_(:,:), PermutDirInvReac_(:,:)
      integer*2, intent(in) :: MetDegRevStruc_(:,:), ReactDegRevStruc_(:,:)
      integer*2, intent(in) :: numNode_(:), numPermDirInvReac_
       ! Integer: 4 bits
      integer*4, intent(in) :: numPermutMetab_, numPermutIrrReac_, numPermutRevReac_
      integer*4, intent(in) :: numRefStructur_

      ! IN/OUT variables
       ! Integer: 1 bit
      integer*1, intent(inout) :: EquiReactRevStru_(:,:)

      !Internal support variables
       ! Integer: 1 bit
      integer*1 :: supRevStructu_(32,8,1), aidR_(2)
       ! Integer: 2 bits
      integer*2 :: supMetDegRevStruc_(24), supReactDegRevStruc_(24)

      ! General variables (4 bits)
      integer*4 :: iReact1_, iReact2_, eqR1_, eqR2_, eqR3_, eqR4_
      integer*4 :: cmR1_, cmR2_, cmR3_, cmR4_, Nmetab1_, Nreacti1_, Nreacti2_

!      Determines the topologically equivalent reactions of the structure numRefStructur_
      ! Creates the support matrix supRevStructu_(i3,i2,i1)
      do iReact1_=1,numNode_(1)  ! Number of metabolites
        do iReact2_=1,(4*numNode_(2))  ! 4 x number of reactions
          supRevStructu_(iReact2_,iReact1_,1) = &
          RevStructu_(iReact2_,iReact1_,numRefStructur_)
        end do
      end do
      ! Creates the support vector supMetDegRevStruc_(i1)
      do iReact1_=1,(3*numNode_(1))  ! 3 x number of metabolites
        supMetDegRevStruc_(iReact1_) = &
        MetDegRevStruc_(iReact1_,numRefStructur_)
      end do
      ! Creates the support vector supReactDegRevStruc_(i1)
      do iReact1_=1,(3*numNode_(2))  ! 3 x number of reactions
        supReactDegRevStruc_(iReact1_) = &
        ReactDegRevStruc_(iReact1_,numRefStructur_)
      end do
      !
      aidR_(1) = 0
      aidR_(2) = 0
      ! Reaction iReact1_
      do iReact1_=1,(numNode_(2)-1)  ! Number of reactions - 1
        ! Reaction iReact2_
        do 3065 iReact2_=(iReact1_+1),numNode_(2)  ! From (iReact1_+1) to number of reactions
          !
          ! Adds "1" to the support variable aidR_(2), needed to determine a position to store the information of
          ! the topological equivalence of the reactions iReact1_ and iReact2_ in matrix EquiReactRevStru_(i2,i1)
          aidR_(2) = aidR_(2) + 1
          !
          ! The reaction iReact1_ and iReact2_ are both reversible or are both irreversible?
          if( ReactRevStr_(iReact1_,numRefStructur_) .ne. &
          ReactRevStr_(iReact2_,numRefStructur_) ) then
            ! The reactions iReact1_ and iReact2_ are not topologically equivalent
            EquiReactRevStru_( (aidR_(1) + aidR_(2)) , numRefStructur_ ) = 0
            !
            goto 3065
          end if
          !
          ! Reaction iReact1_ and reaction iReact2_ are topologically equivalent?
          ! For that to happen, the degree (total and IN) of the two reactions must be the same
          if( ( ReactDegRevStruc_( ((3*(iReact1_ - 1)) + 1) , numRefStructur_ ) .eq. &
          ReactDegRevStruc_( ((3*(iReact2_ - 1)) + 1) , numRefStructur_ ) ) .and. &
          ( ReactDegRevStruc_( ((3*(iReact1_ - 1)) + 2) , numRefStructur_ ) .eq. &
          ReactDegRevStruc_( ((3*(iReact2_ - 1)) + 2) , numRefStructur_ ) )  ) then
            ! The reactions iReact1_ and iReact2_ have the same degree (total and IN)
            !
            ! Reference number of the permutation for the metabolites
            do 3070 eqR1_=1,numPermutMetab_
              !
              ! Checks if it is the first permutation for the reversible reactions
              if( eqR1_ .ne. 1 ) then
                ! Checks if the degree of the metabolites are the same with this permutation
                do Nmetab1_=1,numNode_(1)  ! Number of metabolites of the structure
                  if( MetDegRevStruc_( ((3*(PermutMetab_(Nmetab1_,eqR1_) - 1)) &
                  + 1) , numRefStructur_) .ne. supMetDegRevStruc_( ((3*(Nmetab1_ &
                  - 1)) + 1) ) ) goto 3070
                  if( MetDegRevStruc_( ((3*(PermutMetab_(Nmetab1_,eqR1_) - 1)) &
                  + 2) , numRefStructur_) .ne. supMetDegRevStruc_( ((3*(Nmetab1_ &
                  - 1)) + 2) ) ) goto 3070  ! Checks the next permutation for the metabolites
                end do
              end if
              !
              ! The degree of the metabolites are kept with this permutation. One can continue to check if the reaction
              ! iReact1_ is topologically equivalent to reaction iReact2_
              !
              ! The structure numRefStructur_ has at least one irreversible reaction?
              if( numNode_(4) .ne. 0 ) then
                ! Checks the permutations for the irreversible reactions
                do 3075 eqR4_=1,numPermutIrrReac_
                  !
                  ! The reactions iReact1_ and iReact2_ are irreversible?
                  if( ReactRevStr_(iReact1_,numRefStructur_) .eq. 0 ) then
                    !
                    ! Checks if the reaction iReact2_ is in the position of reaction iReact1_
                    if( PermutIrrReac_(iReact1_,eqR4_) .ne. iReact2_ ) goto 3075
                    !
                  end if
                  !
                  ! Checks if it is the first permutation for the irreversible reactions
                  if( eqR4_ .ne. 1 ) then
                    ! If it is not the first iteration, tests if the degree of the reactions are kept equal
                    do Nreacti1_=1,numNode_(4)  ! Number of irreversible reactions of the structure
                      if( ReactDegRevStruc_( ((3*(PermutIrrReac_(Nreacti1_,eqR4_) - 1)) &
                      + 1) , numRefStructur_) .ne. supReactDegRevStruc_( ((3*(Nreacti1_ &
                      - 1)) + 1)) ) goto 3075
                      if( ReactDegRevStruc_( ((3*(PermutIrrReac_(Nreacti1_,eqR4_) - 1)) &
                      + 2) , numRefStructur_) .ne. supReactDegRevStruc_( ((3*(Nreacti1_ &
                      - 1)) + 2)) ) goto 3075  ! Checks the next permutation for the irreversible
                    end do                                                       ! reactions
                  end if
                  !
                  ! The part of the patterns with the connections metabolite/irreversible reactions of the two isomorphic
                  ! representations of the structure can be compared
                  do cmR3_=1,numNode_(1)  ! Number of metabolites of the structure
                    do cmR4_=1,numNode_(4)  ! Number of irreversible reactions of the structure
                      ! We only need to compare the direct reactions since there is no inverse reaction
                      !
                      ! How to determine the position? ->
                      !      Matrix([(4x(nreaction - 1)) + direct/inverse + reactant/product],[nmetabolite],[nstructure])
                      !      where,
                      !           direct/inverse: "0", because we only need to compare the direct reactions
                      !           reactant/product: "1" for a reactant, and "2" for a product
                      !     For the matrix supRevStructu_(i3,i2,i1), the positions of the metabolites and of the reactions
                      !     are not changed.
                      !     However, for the matrix RevStructu_(i3,i2,i1) the positions of the metabolites and of the reactions
                      !     change according to the respective permutations:
                      !       n of reaction -> PermutIrrReac_(cmR4_,eqR4_)
                      !       direct/inverse reation position ->
                      !        i3: 1st term + 2nd term + 3rd term
                      !          1st term: [4 x [PermutIrrReac_(cmR4_,eqR4_) - 1]]
                      !          2nd term: 0
                      !          3rd term: "1" for a reactant, and "2" for a product (the same value as the one considered to
                      !                    matrix supRevStructu_(i3,i2,i1))
                      !        i2: PermutMetab_(cmR3_,eqR1_)
                      !        i1: numRefStructur_
                      !
                      ! Compares the direct single-direction reaction (direct/inverse term equal to "0"):
                      ! Reactant position (reactant/product term equal to "1")
                      if( supRevStructu_( ((4*(cmR4_ - 1)) + 0 + 1) ,cmR3_ ,1) .ne. &
                      RevStructu_( ((4*( PermutIrrReac_(cmR4_,eqR4_) - 1 )) + 0 + 1 ) &
                      , PermutMetab_(cmR3_,eqR1_) , numRefStructur_) ) goto 3075
                      !         |           1st term                     | |2nd||3rd |
                      ! Product position (reactant/product term equal to "2")
                      if( supRevStructu_( ((4*(cmR4_ - 1)) + 0 + 2) ,cmR3_ ,1) .ne. &
                      RevStructu_( ((4*( PermutIrrReac_(cmR4_,eqR4_) - 1 )) + 0 + 2 ) &
                      , PermutMetab_(cmR3_,eqR1_) , numRefStructur_) ) goto 3075
                      !
                      ! Is this part of the pattern, i.e., the connections metabolites/irreversible reactions are the same between
                      ! the two isomorphic representations of the structure?
                      if( (cmR3_ .eq. numNode_(1)) .and. (cmR4_ .eq. numNode_(4)) ) then
                      ! We can now test the part of the pattern corresponding to the connections between metabolites and reversible
                        goto 3080  ! reactions
                      end if
                    end do
                  end do
3075            continue
                ! This part of the code is not available to the cases were there is a correspondence between the connections
                ! metabolite/irreversible reaction of the two isomorphic representations of the structure numRefStructur_. So,
                ! it means a new permutation for the metabolites must be tested
                goto 3070
                !
              end if
              !
              ! The part of the pattern between the metabolites and the reversible reactions can now be tested
              goto 3080
3080          continue
              !
              ! The structure numRefStructur_ has at least one reversible reaction?
              if( numNode_(3) .ne. 0 ) then
                ! Reference number of the permutation for the reversible reactions
                do 3085 eqR2_=1,numPermutRevReac_
                  !
                  ! The reactions iReact1_ and iReact2_ are reversible?
                  if( ReactRevStr_(iReact1_,numRefStructur_) .eq. 1 ) then
                    !
                    ! Checks if the reaction iReact2_ is in the position of reaction iReact1_
                    if( PermutRevReac_((iReact1_ - numNode_(4)),eqR2_) .ne. &
                    (iReact2_ - numNode_(4)) ) goto 3085
                    !
                  end if
                  !
                  ! Checks if it is the first permutation for the reversible reactions
                  if( eqR2_ .ne. 1 ) then
                    ! If it is not the first iteration, tests if the degree of the reactions are kept equal (only the total
                    ! degree is needed to test the reversible reactions of the two structures: for reversible reactions, the
                    ! IN degree is equal to the OUT degree
                    do Nreacti2_=1,numNode_(3)  ! Number of reversible reactions of the (possibly) new structure
                      if( ReactDegRevStruc_( ((3*(numNode_(4) + PermutRevReac_(Nreacti2_,&
                      eqR2_) - 1)) + 1) , numRefStructur_) .ne. supReactDegRevStruc_( &
                      ((3*(numNode_(4) + Nreacti2_ - 1)) + 1)) ) goto 3085  ! Checks the next permutation for the
                    end do                                                       ! reversible reactions
                  end if
                  !
                  ! The degree of the reactions are kept with this permutation. One can continue to check if the reactions
                  ! iReact1_ and iReact2_ are topologically equivalents
                  !
                  ! Reference number of the permutation for the direct and inverse single direction reactions for
                  ! all the reversible reactions
                  do 3090 eqR3_=1,numPermDirInvReac_
                    !
                    ! Each possible permutation (perm. Met x perm. RevReact x perm. DirInvReact.) will be used to compare
                    ! the two isomorphic representations of the structure numRefStructur_
                    do cmR1_=1,numNode_(1)  ! Number of metabolites of the structures
                      do cmR2_=1,numNode_(3)  ! Number of reversible reactions of the structures
                        !
                        ! How to determine the position? ->
                        !      Matrix([(4x(nreaction - 1)) + direct/inverse + reactant/product],[nmetabolite],[nstructure])
                        !      where,
                        !           direct/inverse: "0" if it is the direct reaction, and "2" if it is the inverse reaction
                        !           reactant/product: "1" for a reactant, and "2" for a product
                        !     For the matrix supRevStructu_(i3,i2,i1), the positions of the metabolites and of the reactions
                        !     are not changed.
                        !     However, for the matrix RevStructu_(i3,i2,i1) the positions of the metabolites and of the reactions
                        !     change according to the respective permutations:
                        !       n of reaction -> PermutReact_(cmR2_,eqR2_)
                        !       direct/inverse reation position ->
                        !        i3: 1st term + 2nd term + 3rd term
                        !          1st term: [4 x [PermutReact_(cmR2_,eqR2_) - 1]]
                        !          2nd term: PermutRevReac_([[2 x [n of reaction - 1]] + a],eqR3_)
                        !                    where "a" is according to the position considered to matrix supRevStructu_(i3,i2,i1):
                        !                    "1" for the direct reaction, and "2" for the inverse reaction
                        !          3rd term: "1" for a reactant, and "2" for a product (the same value as the one considered to
                        !                    matrix supRevStructu_(i3,i2,i1))
                        !        i2: PermutMetab_(cmR1_,eqR1_)
                        !        i1: numRefStructur_
                        !
                        ! Compares the direct single-direction reaction (direct/inverse term equal to "0"):
                        ! Reactant position (reactant/product term equal to "1")
                        if( supRevStructu_( ((4*(numNode_(4) + cmR2_ - 1)) + 0 + 1) &
                        ,cmR1_ ,1) .ne. RevStructu_( ((4*( numNode_(4) + &
                        PermutRevReac_(cmR2_,eqR2_) - 1 )) + PermutDirInvReac_( &
                        ((2*( PermutRevReac_(cmR2_,eqR2_) - 1)) + 1 ) , eqR3_ ) &
                        + 1 ) , PermutMetab_(cmR1_,eqR1_) , numRefStructur_) ) &
                        goto 3090
                        !            |                      1st term                        |   |                           2nd term                                    ||3rd term|
                        ! Product position (reactant/product term equal to "2")
                        if( supRevStructu_( ((4*(numNode_(4) + cmR2_ - 1)) + 0 + 2) &
                        ,cmR1_ ,1) .ne. RevStructu_( ((4*( numNode_(4) + &
                        PermutRevReac_(cmR2_,eqR2_) - 1 )) + PermutDirInvReac_( &
                        ((2*( PermutRevReac_(cmR2_,eqR2_) - 1)) + 1 ) , eqR3_ ) &
                        + 2 ) , PermutMetab_(cmR1_,eqR1_) , numRefStructur_) ) &
                        goto 3090
                        !
                        ! Compares the inverse single-direction reaction (direct/inverse term equal to "0"):
                        ! Reactant position (reactant/product term equal to "1")
                        if( supRevStructu_( ((4*(numNode_(4) + cmR2_ - 1)) + 2 + 1) &
                        ,cmR1_ ,1) .ne. RevStructu_( ((4*( numNode_(4) + &
                        PermutRevReac_(cmR2_,eqR2_) - 1 )) + PermutDirInvReac_( &
                        ((2*( PermutRevReac_(cmR2_,eqR2_) - 1)) + 2 ) , eqR3_ ) &
                        + 1 ) , PermutMetab_(cmR1_,eqR1_) , numRefStructur_) ) &
                        goto 3090
                        ! Product position (reactant/product term equal to "2")
                        if( supRevStructu_( ((4*(numNode_(4) + cmR2_ - 1)) + 2 + 2) &
                        ,cmR1_ ,1) .ne. RevStructu_( ((4*( numNode_(4) + &
                        PermutRevReac_(cmR2_,eqR2_) - 1 )) + PermutDirInvReac_( &
                        ((2*( PermutRevReac_(cmR2_,eqR2_) - 1)) + 2 ) , eqR3_ ) &
                        + 2 ) , PermutMetab_(cmR1_,eqR1_) , numRefStructur_) ) &
                        goto 3090
                        !
                        ! The reactions iReact1_ and iReact2_ are topologically equivalents? For that to happen, all the cells
                        ! must be compared: all the metabolites (cmR1_ equal to number of metabolites) and all reactions
                        ! (cmR2_ equal to number of reversible reactions)
                        if( (cmR1_ .eq. numNode_(1)) .and. (cmR2_ .eq. numNode_(3)) ) then
                          !
                          ! The reactions iReact1_ and iReact2_ are topologically equivalents
                          EquiReactRevStru_( (aidR_(1) + aidR_(2)) , numRefStructur_ ) = 1
                          !
                          goto 3065
                        end if
                      end do
                    end do
3090              continue
3085            continue
              else
                ! The structure has no reversible reactions
                !
                ! The reactions iReact1_ and iReact2_ are topologically equivalents
                EquiReactRevStru_( (aidR_(1) + aidR_(2)) , numRefStructur_ ) = 1
                !
                goto 3065
                !
              end if
3070        continue
            !
            ! This part of the code is not available to the situations in which the reactions iReact1_ and iReact2_ are
            ! topologically equivalent
            !
            ! The reactions iReact1_ and iReact2_ are not topologically equivalent
            EquiReactRevStru_( (aidR_(1) + aidR_(2)) , numRefStructur_ ) = 0
            !
            goto 3065
          else
            ! The reactions iReact1_ and iReact2_ are not topologically equivalent
            EquiReactRevStru_( (aidR_(1) + aidR_(2)) , numRefStructur_ ) = 0
            !
            goto 3065
          end if
3065    continue
        !
        ! Adds (numNode_(1) - iReact1_) to the support variable aidR_(1), needed to the determine a position to store the
        ! information of the topological equivalence of the reactions iReact1_ and iReact2_ in matrix EquiReactRevStru_(i2,i1)
        aidR_(1) = aidR_(1) + (numNode_(2) - iReact1_)
        !
        ! Turns to "0" the support variable aidR_(2)
        aidR_(2) = 0
      end do

      end subroutine EquiReact


      end program allstructures
